/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../lib/config.h"
#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <string>
#include <getopt.h>

#include "../lib/infile.h"
#include "MidarUtils.h"
using namespace std;

namespace MidarUtils {

const char *cvsID = "$Id: MidarUtils.cc,v 1.153 2017/09/01 22:54:05 kkeys Exp $";

Gnuplotter gp;

#if HAVE_PTHREAD
void Mutex::lock() {
    int r = pthread_mutex_lock(&m);
    if (r) throw std::runtime_error(strerror(r));
}

bool Mutex::trylock() {
    int r = pthread_mutex_trylock(&m);
    switch (r) {
	case 0: return true;
	case EBUSY: return false;
	default: throw std::runtime_error(strerror(r));
    }
}

void Mutex::unlock() {
    int r = pthread_mutex_unlock(&m);
    if (r) throw std::runtime_error(strerror(r));
}
#endif

Mutex memlock;

int bits_flag = 1;
int min_bits_wraps = 2; // minimum number of wraps needed to infer limited bits
#if ENABLE_TTL_OPTION
int ttl_flag = TTL_DEFAULT;
#endif
int allow_udp_diff_addr = 0;
#if ENABLE_VSIM_OPTION
int vsim_option = VSIM_DEFAULT;
#endif
#if ENABLE_DIST_OPTION
int dist_option = DIST_DEFAULT;
#endif
int min_overlap_option = 0;
int responsive_count = INT_MAX;
double max_clock_offset = 0.200;

// Max fraction of MBT point failures allowed for which series result may be...
double Comparison::oob_frac_allowed = 0; // ...pass.
double Comparison::oob_converge_allowed = 0; // ...pass, IF first and last points both pass.
double Comparison::oob_inconclusive_allowed = 0; // ...inconclusive.
double Comparison::oob_max = -1; // If MBT point failures exceeds this, series must fail.

uint16_t swap16(uint16_t x) {
    return (x >> 8) | (x << 8);
}

// NB: constructor modifies v
MinMedMax::MinMedMax(vector<double> &v) :
    min(0), med(0), max(0)
{
    int n = v.size();
    if (n < 1) return;
    sort(v.begin(), v.end());
    min = v[0];
    max = v[n-1];
    if (n % 2 == 1) {
	med = v[(n-1)/2];
    } else {
	med = (v[n/2 - 1] + v[n/2]) / 2;
    }
}

// NB: constructor re-orders v
MinQuartilesMax::MinQuartilesMax(vector<double> &v) :
    min(0), q1(0), med(0), q3(0), max(0)
{
    int n = v.size();
    if (n < 4) return;
    sort(v.begin(), v.end());
    min = v[0];
    max = v[n-1];
    int nn;
    if (n % 2 == 1) {
	med = v[(n-1)/2];
	nn = (n-1)/2;
    } else {
	med = (v[n/2 - 1] + v[n/2]) / 2;
	nn = n/2;
    }
    if (nn % 2 == 1) {
	q1 = v[(nn-1)/2];
	q3 = v[n - 1 - (nn-1)/2];
    } else {
	q1 = (v[nn/2 - 1] + v[nn/2]) / 2;
	q3 = (v[n - 1 - nn/2] + v[n - nn/2]) / 2;
    }
}

bool target_velocity_greater(const Target *a, const Target *b)
{
    return a->series_velocity > b->series_velocity;
}

bool target_last_timestamp_less(const Target *a, const Target *b)
{
    return a->time.back().min < b->time.back().min;
}

void parse_meth(const char *method, int &meth, int &ttl_expansion)
{
    for (meth = 0; ; meth++) {
	if (meth >= N_METH) {
            throw std::runtime_error(string("invalid probe method \"") +
		method + "\"");
	}
	if (strncmp(method, methname[meth], strlen(methname[meth])) == 0) {
	    break;
	}
    }
    ttl_expansion = 0;
    for (const char *p = method; *p; p++) {
	if (*p == '-' || *p == '+') {
	    ttl_expansion = atoi(p+1);
	    if (*p == '-') ttl_expansion = -ttl_expansion;
	}
    }
}

// average ids per probe
double ids_per_probe(IpidVector *idr)
{
    int n = 0;
    int sum = 0;
    for (index_t i = 1; i < idr->size(); i++) {
	if (!(*idr)[i] || !(*idr)[i-1]) continue;
	sum += (*idr)[i] - (*idr)[i-1];
        n++;
    }
    return double(sum) / n;
}

// median ids per probe
double median_ids_per_probe(IpidVector *idr)
{
    vector<double> v;
    for (index_t i = 1; i < idr->size(); i++) {
	if (!(*idr)[i] || !(*idr)[i-1]) continue;
	double delta = (*idr)[i] - (*idr)[i-1];
	v.push_back(delta);
    }
    MinMedMax mmm(v);
    return mmm.med;
}

static inline bool bits_can_match(const OrderInfo *oi1, const OrderInfo *oi2)
{
    if (oi1->bits == oi2->bits) {
        return (oi1->prefix == oi2->prefix);
    }

    if (oi1->bits > oi2->bits) {
	swap(oi1, oi2);
    }

    big_ipid_t mask = (0xFFFF << oi1->bits) & 0xFFFF;
    return
	(oi2->min_id & mask) == oi1->prefix &&
	(oi2->max_id & mask) == oi1->prefix;
}

static bool detect_outliers(Target *target, OrderInfo *oi, const char *label,
    ostringstream &desc, bool old)
{
    int xxff = 0;
    int xx00 = 0;
    int other = 0;
    for (unsigned i = 0; i + 1 < oi->id->size(); i++) {
        if ((*oi->id)[i+1] < (*oi->id)[i]) {
            // Delta between points i and i+1 is negative.
            //printf "# i=$i points: i-1=%04x i=%04x i+1=%04x i+2=%04x\n",
                //(*oi->id)[$i-1], (*oi->id)[$i], (*oi->id)[$i+1], (*oi->id)[$i+2];
            //printf "# dist(0,1) = %d\n", iddist((*oi->id)[$i], (*oi->id)[$i+1]);
            if (((*oi->id)[i+1] & 0xFF) == 0x00 &&
                ((*oi->id)[i] & 0xFF00) == ((*oi->id)[i+1] & 0xFF00) &&
		(i+2 == oi->id->size() || ((*oi->id)[i+1] & 0xFF00) + 0x100 == ((*oi->id)[i+2] & 0xFF00)))
            {
                // low byte wrapped to 00, and high byte incremented too late
                // e.g. 11fa, 1100, 1205
                //       i    i+1   i+2
                xx00++;
                oi->outliers.set(i+1, oi->id->size());
                i++;

            } else if (((*oi->id)[i] & 0xFF) == 0xFF &&
                (i == 0 || ((*oi->id)[i-1] & 0xFF00) + 0x100 == ((*oi->id)[i] & 0xFF00)) &&
                ((*oi->id)[i] & 0xFF00) == ((*oi->id)[i+1] & 0xFF00))
            {
                // low byte went to ff, and high byte incremented too early
                // e.g. 11fa, 12ff, 1205
                //       i-1    i   i+1
                xxff++;
                oi->outliers.set(i, oi->id->size());
                i++;
            }
        }
    }

    if (xx00 || xxff || other) {
        desc << label << " outliers:";
	if (xx00) { desc << (old ? " old" : " ") << "XX00:" << xx00; }
	if (xxff) { desc << (old ? " old" : " ") << "XXFF:" << xxff; }
	if (other) { desc << " other:" << other; }
        desc << "; ";
    }

    if (oi->outliers.empty()) {
	return false;
    }

    if (oi->id == &target->id) // shared?
	oi->id = new IpidVector(target->id); // make a copy we can modify
    oi->time = new TimeRangeVector(target->time); // make a copy we can modify

    // delete outlier points
    for (int i = oi->id->size() - 1; i >= 0; i--) {
	if (!oi->outliers.test(i)) continue;
	for (unsigned j = i; j < oi->id->size() - 1; j++)
	    (*oi->id)[j] = (*oi->id)[j+1];
	oi->id->resize(oi->id->size() - 1);
	for (unsigned j = i; j < oi->time->size() - 1; j++)
	    (*oi->time)[j] = (*oi->time)[j+1];
	oi->time->resize(oi->time->size() - 1);
    }

    return true;
}

static void compute_bits(OrderInfo *oi, const char *label, ostringstream &desc)
{
    unsigned bits;
    big_ipid_t min = default_wrapsize;
    big_ipid_t max = 0;
    index_t wraps = 0;
    for (index_t i = 0; i < oi->id->size(); i++) {
	if ((*oi->id)[i] < min) min = (*oi->id)[i];
	if ((*oi->id)[i] > max) max = (*oi->id)[i];
	if (i > 0 && (*oi->id)[i-1] &&
	    (*oi->time)[i].min > (*oi->time)[i-1].max && // don't count overlaping samples
	    (*oi->id)[i] < (*oi->id)[i-1])
		wraps++;
    }
    big_ipid_t diff = min ^ max;
    desc << label << " wraps=" << wraps << " diff=" << fmthex(diff) << "; ";
    if (wraps < min_bits_wraps || (diff & 0x8000)) {
        // not enough negative deltas, or no constant high bits
        return;
    }
    desc << label;
    for (bits = 1; bits < 16; bits++) {
	if (diff >> bits == 0) break;
    }
    big_ipid_t wrapsize = 1 << bits;

    // check that wraps are reasonable with this bit interpretation

    // compute_positive_velocities
    vector<double> velocities;
    vector<double> delta_ts;
    for (index_t i = 1; i < oi->id->size(); i++) {
        if (!(*oi->id)[i-1] || !(*oi->id)[i]) continue;
        int32_t delta_id = (*oi->id)[i] - (*oi->id)[i-1];
	if (delta_id < 0) continue;
        double delta_t = (*oi->time)[i].mid() - (*oi->time)[i-1].mid();
        velocities.push_back(delta_id / delta_t);
        delta_ts.push_back(delta_t);
    }
    MinQuartilesMax pvstats(velocities);
    MinMedMax tstats(delta_ts);
    double allowed_delta_t = 3.5 * tstats.med; // 3 or more skipped probes is a discontinuity

    vector<double> wvelocities; // velocities of wrapped segments
    int32_t allowed_delta_id = int32_t(wraps_per_cycle_thresh * wrapsize);
    for (index_t i = 1; i < oi->id->size(); i++) {
        int32_t delta_id = (*oi->id)[i] - (*oi->id)[i-1];
        if (delta_id < 0) {
	    double delta_t = (*oi->time)[i].mid() - (*oi->time)[i-1].mid();
	    double expected_delta_id = delta_t * pvstats.med;
	    delta_id += wrapsize;
	    if (delta_t > allowed_delta_t ||
		delta_id > allowed_delta_id ||
		expected_delta_id > allowed_delta_id)
	    {
		// Gap is too big, or actual or expected delta id is too large.
		// (This is particularly important during Elimination and
		// Corroboration, when target may be a member of multiple
		// subsets separated by a large gap.)
		if (--wraps < min_bits_wraps) {
		    // Without usable wraps, we can't infer reduced bits.
		    desc << " not enough usable wraps to infer bits; ";
		    return;
		}
	    }
            double v = ((*oi->id)[i] + wrapsize - (*oi->id)[i-1]) / delta_t;
//           printf("# points: %d %.6f %04x, %d %.6f %04x; v=%.3f\n",
//               i-1, (*oi->time)[i-1].mid(), (*oi->id)[i-1].mid(),
//               i, (*oi->time)[i].mid(), (*oi->id)[i].mid(),
//               v);
            wvelocities.push_back(v);
        }
    }
    desc << label << " wraps=" << wraps << "; ";
    MinMedMax wvstats(wvelocities);
//    printf("# median neg velocity: %.3f\n", wvstats.med);

    double max_allowed_v = pvstats.q3 * bits_velocity_thresh + 1;
    double barely_allowed_v = pvstats.q3 * 2.0 + 1;
//    printf("# pos velocities: q0=%.3f q1=%.3f q2=%.3f q3=%.3f q4=%.3f allowed=%.3f\n",
//       pvstats.min, pvstats.q1, pvstats.med, pvstats.q3, pvstats.max, max_allowed_v);
    desc << " wv_med=" << fmtfixed(wvstats.med, 3) <<
	", allowed=" << fmtfixed(max_allowed_v, 3) << ",";
    if (wvstats.med > max_allowed_v) {
        // if the velocity of wrapped segments is much higher than the velocity
        // of positive segments, assume this bit interpretation is wrong
//        printf("# %d bits failed\n", bits);
	desc << " wrap velocity " << fmtfixed(wvstats.med , 3) <<
	    " exceeds " << fmtfixed(max_allowed_v, 3) <<
	    " explainable by " << bits << " bits; ";
	return;
    }
    if (wvstats.med > barely_allowed_v) {
        desc << " " << bits << " bits barely passed,";
        desc << " wrap velocity " << fmtfixed(wvstats.med, 3) <<
	    " barely within " << fmtfixed(max_allowed_v, 3) <<
	    " explainable by " << bits << " bits; ";
    } else {
//        printf("# %d bits passed\n", bits);
    }

    oi->bits = bits;
    oi->prefix = min & (0xFFFF << bits);

    desc << " bits=" << bits << ", prefix=" << fmthex(oi->prefix) <<
	"/" << (16 - bits) << "; ";
}

static bool check_velocity(Target *target, OrderInfo *oi, double &v_iqr, const char *label)
{
    ostringstream &desc = target->desc;
    uint32_t wrapsize = 1 << oi->bits;
    big_ipid_t mask = 0xFFFF >> (16 - oi->bits);
    int32_t sum_delta_id = 0;
    int32_t allowed_delta_id = int32_t(wraps_per_cycle_thresh * wrapsize);
    //int32_t monotonic_thresh = allowed_delta_id * allowed_delta_id / 2.0;

    // compute_unwrapped_velocities
    vector<double> velocities;
    vector<double> delta_ts;
    for (index_t i = 1; i < oi->id->size(); i++) {
        if (!(*oi->id)[i-1] || !(*oi->id)[i]) continue;
        int32_t delta_id = (*oi->id)[i] - (*oi->id)[i-1];
	if (delta_id < 0)
	    delta_id += wrapsize;
        double delta_t = (*oi->time)[i].mid() - (*oi->time)[i-1].mid();
	delta_ts.push_back(delta_t);
        velocities.push_back(delta_id / delta_t);
    }

    MinMedMax tstats(delta_ts);
    double allowed_delta_t = 3.5 * tstats.med; // 3 or more skipped probes is a discontinuity
    MinQuartilesMax vstats(velocities);
    oi->v_med = vstats.med;
    v_iqr = vstats.q3 - vstats.q1;

    // printf("## vq1=%.3f, vmed=%.3f, vq3=%.3f\n", vstats.q1, vstats.med, vstats.q3);
    // printf("## bits=%d, wrapsize=%04x, mask=%04x\n", oi->bits, wrapsize, mask);
    desc << label;
    big_ipid_t additional = 0;
    if (oi->id == &target->id && oi->bits != 16) // id vector is shared, and we need to modify
	oi->id = new IpidVector(target->id); // make a copy we can modify
    (*oi->id)[0] &= mask;
    int n_actual = 0, n_expected = 0, n_time = 0; // discontinuity type counts
    for (index_t i = 1; i < oi->id->size(); i++) {
        int32_t delta_id = ((*oi->id)[i] & mask) - ((*oi->id)[i-1] & mask);
        if (delta_id < 0) {
            delta_id += wrapsize;
            additional += wrapsize;
        }
	// cout << "unwrap " << i << ": " << (*oi->id)[i];
        (*oi->id)[i] = ((*oi->id)[i] & mask) + additional;
	// cout << " -> " << (*oi->id)[i] << "  ";
        // We didn't necessarily add the correct amount for unwrapping, but if we
        // think we didn't, we'll mark this segment as a discontinuity.
        double delta_t = (*oi->time)[i].mid() - (*oi->time)[i-1].mid();
        double expected_delta_id = delta_t * oi->v_med;
// 	printf("## %2d %17.6f %04x -> %2d %17.6f %04x:  dt=%10.6f dy=%d (%d):  ",
// 	    i-1, (*oi->time)[i-1].mid(), (*oi->id)[i-1].mid(),
// 	    i, (*oi->time)[i].mid(), (*oi->id)[i].mid(),
// 	    delta_t, delta_id, unsigned(expected_delta_id));
        if (delta_id > allowed_delta_id) {
            // Actual delta id exceeds allowed fraction of id space.
            oi->discontinuities.set(i, oi->id->size());
	    ++n_actual;
        } else if (expected_delta_id > allowed_delta_id) {
            // Expected delta id exceeds allowed fraction of id space.
            oi->discontinuities.set(i, oi->id->size());
	    ++n_expected;
        } else if (delta_t > allowed_delta_t) {
            // delta t is too large
            oi->discontinuities.set(i, oi->id->size());
	    ++n_time;
//       } else if (abs(delta_id - expected_delta_id) * (v_iqr * delta_t) > monotonic_thresh) {
//           # High deviation from expectation AND overall erraticness
//           push @discontinuities, i;
//           desc .= sprintf " i unpred (%.3f * %.3f > %.3f), ",
//               abs(delta_id - expected_delta_id), (v_iqr * delta_t), monotonic_thresh;
//           #print "nonmonotonic\n"
        } else {
	    // cout << "ok\n";
        }
        sum_delta_id += delta_id;
    }
    if (n_actual || n_expected || n_time) {
	desc << " discon";
	if (n_actual) desc << " a=" << n_actual;
	if (n_expected) desc << " x=" << n_expected;
	if (n_time) desc << " t=" << n_time;
	desc << "/" << oi->id->size()-1;
    }
    if (oi->discontinuities.count() > (oi->id->size()-1) * discontinuity_thresh) {
	desc << " too many discon; ";
	return false;
    }
    // cout << "sum_delta_id=" << sum_delta_id << " ";
    // cout << "allowed=" << wraps_per_cycle_thresh * wrapsize * oi->id->size() << "\n";
    if (sum_delta_id > wraps_per_cycle_thresh * wrapsize * oi->id->size()) {
	desc << " too fast; ";
	return false;
    }
    desc << " v ok; ";
    return true;
}

static int repeated_high_bits(const TimeRangeVector *tr, const IpidVector *idr, ostringstream &desc, int min, double time_fudge)
{
    int fixed[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    int n = idr->size() - 1; // number of deltas
    if (n < 1) return 0;
    for (int i = 0; i < n; i++) {
	if (tr && (*idr)[i] > (*idr)[i+1] && (*tr)[i].max + time_fudge < (*tr)[i+1].min - time_fudge) {
	    // Time ranges don't overlap, and delta id is negative
	    continue; // don't count negative delta
	}
        ipid_t diff = (*idr)[i] ^ (*idr)[i+1];
        // sprintf(printbuf, "%04x ^ %04x = %04x  ", (*idr)[i], (*idr)[i+1], diff);
	// cerr << printbuf;
        for (int j = 0; j < 16; j++) {
            if (diff >> (15 - j) != 0) {
                fixed[j]++;
                // cerr << "fixed=" << j << "\n";
                break;
            }
        }
    }
    for (int j = 15; j > 0; j--) {
        fixed[j-1] += fixed[j];
    }
    desc << "fixed " << fixed[1];
    for (int j = 2; j < 16; j++) {
	desc << ",";
	if (fixed[j] != fixed[j-1]) desc << fixed[j];
	if (j < 14 && fixed[j] == 0) {
	    desc << "...";
	    break;
	}
    }
    desc << "/" << n << " ";
    for (int j = min; j >= 4; j--) {
	if (fixed[j] > n/2.0) {
            desc << "RHB(" << j << ")=" << fixed[j] << "; ";
	    return j;
	}
    }
    desc << "no RHB; ";
    return 0;
}

#if 0
// Note: this includes discontinuity segments
static double weighted_avg_velocity(const TimeRangeVector *tr, const IpidVector *idr)
{
    // cerr << "# weighted_avg_velocity\n";
    int n = tr->size();
    if (n < 2) return 0;
    return ((*idr)[n-1] - (*idr)[0]) / ((*tr)[n-1].mid() - (*tr)[0].mid());
}
#endif

// Note: this excludes discontinuity segments, because we don't know what
// happened during them:  could be a sudden speed-up, a sudden slow-down, or
// a reset.
double OrderInfo::calculate_velocity()
{
    const TimeRangeVector *tr = this->time;
    const IpidVector *idr = this->id;
    if (tr->size() < 2) return 0;

#if 1 // XXX ??? doesn't handle wraps
    if (this->discontinuities.empty()) {
	// optimization
	int n = tr->size();
	return ((*idr)[n-1] - (*idr)[0]) / ((*tr)[n-1].mid() - (*tr)[0].mid());
    }
#endif

    int64_t sum_delta_id = 0;
    double sum_delta_t = 0;
    const big_ipid_t mask = 0xFFFF >> (16 - this->bits);
    big_ipid_t wrapsize = 1 << this->bits;
    for (index_t i = 1; i < idr->size(); i++) {
	if (this->discontinuities.test(i)) continue; // skip discontinuity
        int32_t delta_id = ((*idr)[i] & mask) - ((*idr)[i-1] & mask);
        if (delta_id < 0) delta_id += wrapsize;
        sum_delta_id += delta_id;
        sum_delta_t += (*tr)[i].mid() - (*tr)[i-1].mid();
    }
    double v = (sum_delta_t == 0) ? 0 : sum_delta_id / sum_delta_t;
    return v;
}

bool target_is_sorted(Target *target) {
    // If target was probed from a single monitor, its points should be
    // sorted, and we warn or error if they're not.
    // If target was probed from different monitors, we don't expect
    // its points to be sorted, and just return false if they're not.
    int n_unsorted = 0;
    double diff = 0.0;
    double max = 0.0;
    double cumulative = 0.0;
    for (index_t i = 1; i < target->time.size(); i++) {
	if (!target->time[i].min) continue;
	if (target->time[i].min < target->time[i-1].min) {
	    if (!target->src_addr) return false;
	    cerr << "WARNING: unsorted points for " << target->addr << ": " <<
		fmtfixed(target->time[i-1].min, 6) << ", " <<
		fmtfixed(target->time[i].min, 6) << " (src=" <<
		target->src_addr << ")\n";
	    n_unsorted++;
	    diff = target->time[i-1].min - target->time[i].min;
	    cumulative += diff;
	    if (diff > max) max = diff;
	};
    }
    if (n_unsorted > 0) {
	bool fatal = max > 2.0 || n_unsorted > 10;
	cerr << (fatal ? "ERROR" : "WARNING") << ": " << n_unsorted <<
	    " unsorted points for " << target->addr << " (max=" <<
	    fmtfixed(max, 6) << "s, total=" << fmtfixed(cumulative, 6) <<
	    "s, src=" << target->src_addr << ")\n";
	if (fatal) exit(1);
    }
    return n_unsorted == 0;
}

struct indexes_less {
    const TimeRangeVector *tr;
    indexes_less(const Target *target) : tr(&target->time) {}
    bool operator()(const int &a, const int &b) const {
	return (*tr)[a].min < (*tr)[b].min;
    }
};

static void classifySeries(Target *target, bool optimize)
{
    TimeRangeVector *tr = &target->time;
    IpidVector *idr = &target->id;

    // count wraps, remove outliers, detect byte-swapping
    {
	index_t wraps[2] = {0,0};
	OrderInfo *BE = target->oi[ORDER_BE] = new OrderInfo();
	OrderInfo *LE = target->oi[ORDER_LE] = new OrderInfo();

	BE->time = tr; // share data
	BE->id = idr; // share data
	LE->time = tr; // share data
	LE->id = new IpidVector();
	transform(idr->begin(), idr->end(), back_inserter(*LE->id), swap16); // copy & transform

	// print "### $target_addr BE ids: ", join(', ', @$be_idr), "\n";
	// print "### $target_addr LE ids: ", join(', ', @$le_idr), "\n";

	bool bad[2] = {false, false};
	target->order = ORDER_UNKNOWN;
	BE->bits = 16;
	LE->bits = 16;

	// count wraps and remove outliers
	{
	    for (int o = 0; o < 2; o++) {
		if (!bad[o]) {
		    for (index_t i = 1; i < target->oi[o]->id->size(); i++) {
			if ((*target->oi[o]->id)[i] < (*target->oi[o]->id)[i-1])
			    wraps[o]++;
		    }
		}
	    }

	    // set target->wraps now, in case it turns out to be nonmonotonic
	    target->wraps = wraps[ORDER_BE];

	    for (int o = 0; o < 2; o++) {
		if (!bad[o]) {
		    if ((detect_outliers(target, target->oi[o], order_label[o], target->desc, false))) {
			if (target->oi[o]->outliers.count() >
			    target->oi[o]->id->size() * outlier_thresh + 1)
			{
			    target->desc << order_label[o] << " has too many outliers; ";
			    bad[o] = true;
			}
		    }
		}
	    }
	}

	if (bad[ORDER_BE] && bad[ORDER_LE]) {
	    target->desc << "both orders are nonmonotonic";
	    target->classification = CLASS_NONMONOTONIC;
	    goto end;
	}

	// compute bits (after removing outliers)
	// cerr << "# compute_bits " << target->addr << "\n";
	if (bits_flag) {
	    for (int o = 0; o < 2; o++) {
		if (!bad[o]) {
		    compute_bits(target->oi[o], order_label[o], target->desc);
		}
	    }
	}

	// if both orders are still possible, use RHB test to detect byte order
	if (!bad[ORDER_BE] && !bad[ORDER_LE]) {
	    // If one ordering has semi-fixed high bits, that ordering is
	    // probably correct.  For the other ordering to be correct, it would
	    // have semi-fixed NON-high bits, which is very unlikely.
	    const double time_fudge = target->src_addr ? 0 : max_clock_offset;
	    int repeated[2];
	    if (!bad[ORDER_BE]) {
		target->desc << "BE ";
		repeated[ORDER_BE] = repeated_high_bits(BE->time, BE->id, target->desc, 8, time_fudge);
	    }
	    if (!bad[ORDER_LE]) {
		target->desc << "LE ";
		repeated[ORDER_LE] = repeated_high_bits(LE->time, LE->id, target->desc, 8, time_fudge);
	    }
	    if (repeated[ORDER_BE] > repeated[ORDER_LE]) {
		target->desc << "BE wins RHB; ";
		target->order = ORDER_BE;
	    } else if (repeated[ORDER_LE] > repeated[ORDER_BE]) {
		target->desc << "LE wins RHB; ";
		target->order = ORDER_LE;
#if 1
	    } else if (repeated[ORDER_BE] > 0) {
		// weird
		target->desc << "orders have identical nonzero RHB; ";
#endif
	    }
	}

	double v_iqr[2] = {0,0};
	for (int o = 0; o < 2; o++) {
	    if (!bad[o]) {
		// cerr << "# check_velocity " << target->addr << " " << order_label[o] << "\n";
		if (!check_velocity(target, target->oi[o], v_iqr[o], order_label[o])) {
		    bad[o] = true;
		}
	    }
	}

	if (target->order != ORDER_UNKNOWN) {
	    // Order was already chosen
	    if (bad[target->order]) {
		target->desc << "pref order nonmonotonic; ";
		target->classification = CLASS_NONMONOTONIC;
		goto end;
	    }
	} else if (bad[ORDER_BE] && bad[ORDER_LE]) {
	    target->desc << "both orders nonmonotonic; ";
	    target->classification = CLASS_NONMONOTONIC;
	    goto end;
	} else if (bad[ORDER_BE] || bad[ORDER_LE]) {
	    // Only one order is not bad; choose it.
	    target->order = bad[ORDER_BE] ? ORDER_LE : ORDER_BE;
	    target->desc << order_label[target->order] << " by default; ";
	} else {
	    // Order is still unknown.  Compare v_iqr.
	    target->desc << "BE v_iqr=" << v_iqr[ORDER_BE] << ", LE v_iqr=" << v_iqr[ORDER_LE] << ", ";
	    if (v_iqr[ORDER_BE] * 2 < v_iqr[ORDER_LE]) {
		target->desc << "BE has much better v_iqr; ";
		target->order = ORDER_BE;
	    } else if (v_iqr[ORDER_LE] * 2 < v_iqr[ORDER_BE]) {
		target->desc << "LE has much better v_iqr; ";
		target->order = ORDER_LE;
	    } else {
		target->desc << "orders are equally plausible; ";
		target->classification = CLASS_UNKNOWN_ORDER;
		goto end;
	    }
	}

	if (BE->v_med && LE->v_med) {
	    if ((target->order == ORDER_BE) == (BE->v_med > LE->v_med))
		target->desc << "v_med disagrees with byte order; ";
	}

	target->desc << "order=" << ((target->order == ORDER_BE) ? "BE; " : "LE; ");

	target->wraps = wraps[target->order];

	// The wrapfrac test must be after byteorder test.  We wouldn't want to
	// accept the incorrect byteorder by default because an early wrapfrac
	// test ruled out the correct byte order.
	double wrapfrac = double(target->wraps) / target->poi()->id->size();
	if (wrapfrac > wraps_per_cycle_thresh) {
	    target->desc << order_label[target->order] << " too fast (wrapfrac); ";
	    target->classification = CLASS_NONMONOTONIC;
	    goto end;
	}
    }

//    (target->series_velocity, target->model_t, target->model_id) =
//	best_fit(target->time, target->unwrapped_id);

    // target->avg_velocity = weighted_avg_velocity(target->poi()->time, target->poi()->id);
    target->series_velocity = target->poi()->calculate_velocity();
    if (target->series_velocity == 0) {
	target->desc << order_label[target->order] << " zero velocity";
	// target->classification = CLASS_NONMONOTONIC;
	// goto end;
    }

    target->flags |= MONOTONIC;
    target->classification = CLASS_MONOTONIC;
    target->poi()->set_minmax_id();

end:
    return;
}

#ifdef OLDCLASSIFY
static void oldClassifySeries(Target *target, bool optimize)
{
    ...
}
#endif

void classify(Target *target, bool optimize)
{
    if (target->tries < 3) {
	target->classification = CLASS_NOT_ENOUGH_DATA;
	return;
    }
    // cerr << "################ classify " << target->addr << "\n";
    TimeRangeVector *tr = &target->time;
    IpidVector *idr = &target->id;

#if 0
    cerr << "### points: ";
    for (int i = 0; i < tr->size(); i++) {
	cerr << "[" << fmtfixed((*tr)[i].min, 6) << ", " << fmthex((*idr)[i]) << "] ";
    }
    cerr << "\n";
#endif

    if (idr) {
	if (idr->size() != static_cast<index_t>(idr->size()))
	    throw std::runtime_error("sample count would overflow index_t");
	if (!target_is_sorted(target)) {
	    // sort indexes by time
	    vector<int> indexes(idr->size());
	    for (size_t i = 0; i < indexes.size(); i++)
		indexes[i] = i;
	    sort(indexes.begin(), indexes.end(), indexes_less(target));
	    // reorder time and idr by the sorted indexes
	    TimeRangeVector newtime(tr->size());
	    IpidVector newidr(idr->size());
	    for (size_t i = 0; i < indexes.size(); i++) {
		newtime[i] = (*tr)[indexes[i]];
		newidr[i] = (*idr)[indexes[i]];
	    }
	    swap(target->time, newtime);
	    swap(target->id, newidr);
	} else {
	    // no need to sort, but compacting the vectors is still useful
	    TimeRangeVector newtime(*tr);
	    IpidVector newidr(*idr);
	    swap(target->time, newtime);
	    swap(target->id, newidr);
	}

	// count repeats
	for (index_t i = 1; i < idr->size(); i++) {
	    if ((*idr)[i] != 0 && (*idr)[i] == (*idr)[i-1]) {
		target->repeat++;
		target->nondegenerate--;
	    }
	}
    }

    if (ttl_flag && target->expected) {
	target->init_ttl = guess_initial_ttl(target->max_reply_ttl);
	target->min_rev_path_len = target->init_ttl + 1 - target->max_reply_ttl;
    }

    if ((target->responses < responsive_thresh * target->tries) && (target->responses < responsive_count)) {
	target->classification = CLASS_UNRESPONSIVE;
	return;
    } else if (target->ttl_expansions > 1) {
	target->classification = CLASS_UNSTABLE_TTL;
	return;
    } else if (target->unexpected > mostly_thresh * target->tries) {
	target->classification = CLASS_MOSTLY_UNEXPECTED;
	return;
    } else if ((target->expected < responsive_thresh * target->tries) && (target->expected < responsive_count)) {
	target->classification = CLASS_BAD_RESPONSE_MIXED;
	return;
    }
    target->flags |= GOOD_RESPONSE;

    if (target->nondegenerate < responsive_thresh * target->tries && target->nondegenerate < responsive_count) {
	if (target->zero > target->expected * mostly_thresh) {
	    target->classification = CLASS_MOSTLY_ZERO;
	} else if (target->reflect > target->expected * mostly_thresh) {
	    target->classification = CLASS_MOSTLY_REFLECT;
	} else if (target->repeat > (target->expected - 1) * mostly_thresh) {
	    target->classification = CLASS_MOSTLY_REPEAT;
	} else {
	    target->classification = CLASS_BAD_ID_MIXED;
	}
	return;
    }

    target->flags |= GOOD_ID;

#ifdef OLDCLASSIFY
    target->desc << " OLD: ";
    Target oldtarget(*target); // copy constructor
    oldClassifySeries(&oldtarget, optimize);
    target->desc << oldtarget.desc.str();

    target->desc << " NEW: ";
#endif
    classifySeries(target, optimize);

#ifdef OLDCLASSIFY
    if (oldtarget.order != target->order)
	target->flags |= DISORDER;
    else if (target->order != ORDER_UNKNOWN &&
	oldtarget.order != ORDER_UNKNOWN &&
	oldtarget.poi()->bits != target->poi()->bits)
	    target->flags |= DISBITS;
    if (oldtarget.classification != target->classification)
	target->flags |= DISCLASSIFY;
    oldtarget.freeResponses();
#endif
}

// not threadsafe!
void ResponseData::parse(char *buf)
{
    FieldSplitter<18> fields(buf, " \t");

    if (!fields[5] /*probe_dst_addr*/) {
	throw std::runtime_error("syntax error");
    }
    result = *fields[0];
    // bogus_cycle_id [1]
    target_addr = ip4addr_t(fields[2]);
    method = fields[3];
    src_addr = ip4addr_t(fields[4]);
    // probe_dst_addr [5]
    probe_ttl = fields[6] ? atoi(fields[6]) : 0;
    probe_ipid = fields[7] ? atoi(fields[7]) : 0;
    probe_time = fields[8] ? atof(fields[8]) : 0;
    resp_src_addr = fields[9] ? ip4addr_t(fields[9]) : ip4addr_t(0);
    reply_ttl = fields[10] ? atoi(fields[10]) : 0;
    reply_quoted_ttl = ttl_flag && fields[11] ? atoi(fields[11]) : 0;
    reply_ipid = fields[12] ? atoi(fields[12]) : 0;
    resp_icmp_type = fields[13] && isdigit(*fields[13]) ? atoi(fields[13]) : -1;
    resp_icmp_code = fields[14] && isdigit(*fields[14]) ? atoi(fields[14]) : -1;
    resp_tcp_flags = fields[15] && isdigit(*fields[15]) ? atoi(fields[15]) : -1;
    // resp_warts_id [16]
    resp_time = fields[17] ? atof(fields[17]) : 0;

    if (resp_time > 0) {
	if (prev_resp_time > 0 &&
	    (resp_time - prev_resp_time > 30 || resp_time - prev_resp_time < -1))
	{
	    cerr << "WARNING: response time jumped " <<
		fmtfixed(resp_time - prev_resp_time, 6) << " seconds (" <<
		fmtfixed(prev_resp_time, 6) << " to " <<
		fmtfixed(resp_time, 6) << ")\n";
	}
	prev_resp_time = resp_time;
    }
}

// not threadsafe!
void ResponseData::parseRadarGun(char *buf, const char *meth)
{
    FieldSplitter<6> fields(buf, " \t");

    if (!fields[4] /*reply_ipid*/) {
	throw std::runtime_error("syntax error");
    }
    target_addr = ip4addr_t(fields[0]);
    method = meth;
    if (fields[1][0] == '-') {
	result = 'N';
	return;
    }
    result = 'T';
    src_addr = ip4addr_t(0);
    probe_ttl = 0;
    probe_ipid = fields[4] ? atoi(fields[4]) : 0;
    probe_time = resp_time = fields[2] ? atof(fields[2]) : 0;
    resp_src_addr = fields[1] ? ip4addr_t(fields[1]) : ip4addr_t(0);
    reply_ttl = 0;
    reply_quoted_ttl = 0;
    reply_ipid = fields[3] ? atoi(fields[3]) : 0;
    if (strncmp(fields[5], "Icmp", 4) == 0) {
	int type, code;
	sscanf(fields[5], "Icmp_t=%d_c=%d", &type, &code);
	resp_icmp_type = type;
	resp_icmp_code = code;
    } else {
	resp_icmp_type = -1;
	resp_icmp_code = -1;
    }
    if (strncmp(fields[5], "Tcp", 3) == 0) {
	resp_tcp_flags = 0;
    } else {
	resp_tcp_flags = -1;
    }
}

bool Comparison::parse_oob_allowed(const char *str) {
    char *p;
    oob_frac_allowed = strtod(str, &p);
    if (oob_frac_allowed < 0 || oob_frac_allowed >= 0.5) {
	cerr << "oob-allowed first parameter must be in range [0.0, 0.5)\n";
	return false;
    }
    if (!*p) goto ok;
    if (*p++ != ',') return false;
    oob_converge_allowed = strtod(p, &p);
    if (oob_converge_allowed < 0 || oob_converge_allowed >= 0.5) {
	cerr << "oob-allowed second parameter must be in range [0.0, 0.5)\n";
	return false;
    }
    if (!*p) goto ok;
    if (*p++ != ',') return false;
    oob_inconclusive_allowed = strtod(p, &p);
    if (oob_inconclusive_allowed < 0 || oob_inconclusive_allowed >= 1.0) {
	cerr << "oob-allowed third parameter must be in range [0.0, 1.0)\n";
	return false;
    }
    if (!*p) goto ok;
    return false;
ok:
    if (oob_converge_allowed == 0)
	oob_converge_allowed = oob_frac_allowed;
    oob_max = oob_frac_allowed;
    if (oob_max < oob_converge_allowed)
	oob_max = oob_converge_allowed;
    if (oob_max < oob_inconclusive_allowed)
	oob_max = oob_inconclusive_allowed;
    return true;
}

void Comparison::gap_oob_check(const Target *left, const Target *right,
    const bool &optimize, const double &time_fudge)
{
    // If the two methods are the same, id values must be strictly
    // increasing for them to possibly be from a shared counter.  But if the
    // methods are different, it is possible that one uses preincrement and
    // the other uses postincrement, so id values must only be nondecreasing.
    int min_dist = (left->meth == right->meth) ? 1 : 0;

    const big_ipid_t lmask = 0xFFFF >> (16 - left->poi()->bits);
    const big_ipid_t lprefix = left->poi()->prefix;
    const big_ipid_t rmask = 0xFFFF >> (16 - right->poi()->bits);
    const big_ipid_t rprefix = right->poi()->prefix;

    if (left->time.back().max + time_fudge < right->time.front().min - time_fudge) {
	// no overlap
	if (iddist((left->poi()->id->back() & lmask) | lprefix, (right->poi()->id->front() & rmask) | rprefix) < min_dist) {
	    oob++;
	    if (!result) result = RESULT_NO_OVERLAP_OOB;
	}
    } else {
	// soft overlap
	int left_n = left->time.size();
	if ((!left->poi()->discontinuities.test(left_n-1) &&
	    iddist(((*left->poi()->id)[left_n-1] & lmask) | lprefix, ((*right->poi()->id)[1] & rmask) | rprefix) < min_dist) ||
	    (!right->poi()->discontinuities.test(1) &&
	    iddist(((*left->poi()->id)[left_n-2] & lmask) | lprefix, ((*right->poi()->id)[0] & rmask) | rprefix) < min_dist))
	{
	    oob++;
	    if (!result) result = RESULT_NO_OVERLAP_OOB;
	}
    }
}

// The Monotonic Bounds Test (MBT)
void Comparison::overlap_oob_check(const bool &swapped,
    const bool &optimize, const double &time_fudge)
{
    const Target *p; // series from which we get points
    const Target *s; // series from which we get segments
    if (!swapped) {
	p = r1;
	s = r2;
    } else {
	p = r2;
	s = r1;
    }

    const TimeRangeVector *pt = p->poi()->time;
    const IpidVector *pid = p->poi()->id;
    const big_ipid_t pmask = 0xFFFF >> (16 - p->poi()->bits);
    const big_ipid_t pprefix = p->poi()->prefix;
    const TimeRangeVector *st = s->poi()->time;
    const IpidVector *sid = s->poi()->id;
    const big_ipid_t smask = 0xFFFF >> (16 - s->poi()->bits);
    const big_ipid_t sprefix = s->poi()->prefix;
    const ProbeBitset &sdiscon = s->poi()->discontinuities;
    const ProbeBitset &pdiscon = p->poi()->discontinuities;

    // we can exit early if oob passes this upper bound
    const int oob_allowed = oob_max * (pid->size() + sid->size() - 2);

    unsigned pi = 0; // index of a point in P
    unsigned si_low = 0; // index of a lower bound point in S
    unsigned si_up = 0; // index of an upper bound point in S

    // If the two methods are the same, id values must be strictly
    // increasing for them to possibly be from a shared counter.  But if the
    // methods are different, it is possible that one uses preincrement and
    // the other uses postincrement, so id values must only be nondecreasing.
    const int min_dist = (p->meth == s->meth) ? 1 : 0;

    // Find first point in P that has a lower time bound in S.
    const double lower_time_bound = st->front().max + 2 * time_fudge;
    while (pi < pt->size() && (*pt)[pi].min < lower_time_bound)
	pi++;

    // For each point in P whose time is bounded by a pair of points in S,
    // make sure its id is also bounded by the pair of points in S.
    const double upper_time_bound = st->back().min - 2 * time_fudge;
    while (pi < pt->size() && (*pt)[pi].max < upper_time_bound) {
	bool xdisco = false;
	big_ipid_t id, min_id, max_id;
	// find lower bound of S segment
	while (si_low+1 < st->size() &&
	    (*st)[si_low+1].max + time_fudge < (*pt)[pi].min - time_fudge)
	{
	    si_low++;
	}

	// find upper bound of S segment
	for (si_up = si_low + 1; true; si_up++) {
	    // The outer loop condition guarantees we will find an upper
	    // bound before walking off the end of the array.
	    if (sdiscon.test(si_up))
		goto nextpoint; // discontinuity is unusable
	    if ((*st)[si_up].min - time_fudge > (*pt)[pi].max + time_fudge)
		break; // found the upper bound
	}

	// If P has a discontinuity that overlaps the S segment, AND the S
	// segment width is at least 1/3 of the P discontinuity width, the S
	// segment shouldn't be trusted.  (The width requirement serves two
	// purposes:  it means the S segment is not significantly more
	// trustworthy than the P discontinuity (narrower==better), and it
	// means the P discontinuity is due to a large delta id, not a large
	// sampling gap (delta t).)
	if (!pdiscon.empty()) {
	    const double widthfactor = 0.333333;
	    double s_right = (*st)[si_up].max;
	    double s_left = (*st)[si_low].min;
	    // search forward from pi
	    for (unsigned pdi = pi+1; pdi < pt->size(); pdi++) {
		double p_right = (*pt)[pdi].max;
		double p_left = (*pt)[pdi-1].min;
		if (p_left > s_right)
		    break; // P segment is after the end of S segment
		if (!pdiscon.test(pdi))
		    continue; // no discontinuity in P here
		if (p_left < s_right && p_right > s_left &&
		    s_right - s_left > widthfactor * (p_right - p_left))
		{
		    // S segment overlaps P discontinuity and is wide enough
		    xdisco = true;
		    n_xdisco++;
		    goto done_xdisco;
		}
	    }
	    // search backward from pi
	    for (unsigned pdi = pi; pdi > 0; pdi--) {
		double p_right = (*pt)[pdi].max;
		double p_left = (*pt)[pdi-1].min;
		if (p_right < s_left)
		    break; // P segment is before the beginning of S segment
		if (!pdiscon.test(pdi))
		    continue; // no discontinuity in P here
		if (p_left < s_right && p_right > s_left &&
		    s_right - s_left > widthfactor * (p_right - p_left))
		{
		    // S segment overlaps P discontinuity and is wide enough
		    xdisco = true;
		    n_xdisco++;
		    goto done_xdisco;
		}
	    }
	}
	done_xdisco:

	// test P point against S id bounds
	n_bounded++;
	id = ((*pid)[pi] & pmask) | pprefix;
	min_id = ((*sid)[si_low] & smask) | sprefix;
	max_id = ((*sid)[si_up] & smask) | sprefix;

	if ((min_id <= max_id) ?
	    // bounding box does not wrap
	    ((id < min_id + min_dist) || (id > max_id - min_dist)) :
	    // bounding box wraps
	    ((id < min_id + min_dist) && (id > max_id - min_dist)))
	{
	    // Point is out of bounds.
	    if (xdisco) {
		n_xdisco_oob++;
		n_bounded--; // doesn't count
	    } else {
		oob++;
		if (!swapped) r1_mbt_fail.set(pi, r1->id.size());
		    else r2_mbt_fail.set(pi, r2->id.size());
		if (first_fail > (*pt)[pi].max) first_fail = (*pt)[pi].max;
		if (last_fail < (*pt)[pi].min) last_fail = (*pt)[pi].min;
		if (oob > oob_allowed) {
		    if (!result) result = RESULT_OUT_OF_BOUNDS;
		    if (optimize) return;
		}
	    }
	} else {
	    if (xdisco) {
		n_bounded--; // doesn't count
	    } else {
		if (!swapped) r1_mbt_pass.set(pi, r1->id.size());
		    else r2_mbt_pass.set(pi, r2->id.size());
		if (first_pass > (*pt)[pi].max) first_pass = (*pt)[pi].max;
		if (last_pass < (*pt)[pi].min) last_pass = (*pt)[pi].min;
	    }
	}   

	nextpoint:
	pi++;
    }
}

// returns true if it interpolated and set dist
bool Comparison::interpolation_check(
    // point parameters
    const TimeRangeVector *&pt, const IpidVector *&pid,
    unsigned &pi, const big_ipid_t &pwrapsize,
    // segment parameters
    const TimeRangeVector *&st, const IpidVector *&sid,
    const unsigned &si, const big_ipid_t &swrapsize,
    const double &sv, const ProbeBitset &sdiscon,
    // other parameters
    const bool &optimize, double &dist, const double &time_fudge)
{
    if (sdiscon.test(si)) {
	// don't interpolate in a discontinuity
	// cout << "  discontinuity"; // XXX
	pi++;
	return false;
    }

    big_ipid_t actual_id = (*pid)[pi] % pwrapsize;

    if (0) { // old_oob
	// Points on the other line that are closer than time_fudge may have
	// been time-swapped with this point.  So, to find boundaries, we choose
	// the nearest points on either side of this point that are farther away
	// than time_fudge (but ignoring discontinuities).

	unsigned low = si-1;
	const big_ipid_t *min_id = &(*sid)[low];
	while ((*pt)[pi].mid() - time_fudge < (*st)[low].mid() + time_fudge) {
	    if (low == 0 || sdiscon.test(low)) {
		min_id = 0;
		break;
	    } else {
		min_id = &(*sid)[--low];
	    }
	}

	unsigned high = si;
	const big_ipid_t *max_id = &(*sid)[high];
	while ((*pt)[pi].mid() + time_fudge > (*st)[high].mid() - time_fudge) {
	    if (high == st->size() - 1 || sdiscon.test(high+1)) {
		max_id = 0;
		break;
	    } else {
		max_id = &(*sid)[++high];
	    }
	}

	if ((min_id && iddist(actual_id, *min_id, swrapsize) >= 0) ||
	    (max_id && iddist(actual_id, *max_id, swrapsize) <= 0))
	{
	    // cout << hex;
	    // if (min_id && iddist(actual_id, *min_id, swrapsize))
		// cout << " (id " << actual_id << " < min_id " << *min_id << ")";
	    // if (max_id && iddist(actual_id, *max_id, swrapsize) < 0)
		// cout << " (id " << actual_id << " > max_id " << *max_id << ")";
	    // cout << dec;
	    // cout << " ***OOB*** "; // XXX
	    old_oob++;
	    // if (!result) result = RESULT_OUT_OF_BOUNDS;
	    // if (optimize) return false;
	}
    }

    if (dist_option || !optimize) {
	double expected_id =
	    interpolate((*st)[si-1].mid(), (*sid)[si-1], (*st)[si].mid(), (*sid)[si],
		(*pt)[pi].mid());
	while (expected_id >= swrapsize) {
	    expected_id -= swrapsize;
	}
	// cout << "  actual=" << actual_id; // XXX
	// cout << "  expected=" << expected_id; // XXX
	dist = iddist(actual_id, expected_id);
    }
    pi++;
    return true;
}

void Comparison::compare(Target * const &_r1, Target * const &_r2, bool optimize)
{
    this->init(_r1, _r2);

    // note: src_addr==0 means target was probed from multiple sources
    const bool same_src = (r1->src_addr == r2->src_addr && r1->src_addr != 0);

    const double v1 = r1->series_velocity;
    const double v2 = r2->series_velocity;
    const OrderInfo *oi1 = r1->poi(); // oi for r1
    const OrderInfo *oi2 = r2->poi(); // oi for r2
    const TimeRangeVector *tr1 = oi1->time; // times for r1
    const TimeRangeVector *tr2 = oi2->time; // times for r2
    const IpidVector *idr1 = oi1->id; // ids for r1
    const IpidVector *idr2 = oi2->id; // ids for r2

#ifdef ENABLE_FEATURES
    if (r1->features && r2->features) {
	if (!r1->features->match(r2->features)) {
#if 0
	    if (!result) result = RESULT_FEATUREDIFF;
	    if (optimize) return;
#else
	    this->featurediff = true;
#endif
	}
    }
#endif

    const double time_fudge = same_src ? 0 : max_clock_offset;

    // boundaries of overlap zone
    const double first_t = max(tr1->front().mid(), tr2->front().mid());
    const double last_t = min(tr1->back().mid(), tr2->back().mid());

    // positive overlap indicates time of overlap
    // negative overlap indicates time of gap
    overlap = last_t - first_t;

    if (r1->order != r2->order && r1->meth == r2->meth) {
	// note: different methods are allowed to be in different orders
	if (!result) result = RESULT_SWAPPED;
	if (optimize) return;

    } else if (ttl_flag && r1->init_ttl != r2->init_ttl && r1->meth == r2->meth) {
	// note: different methods are allowed have different initial TTLs
	// cerr << "\nr1:" << int(r1->init_ttl) << " r2:" << int(r2->init_ttl) << "\n";
	if (!result) result = RESULT_DIFF_TTL;
	if (optimize) return;

    } else if (!bits_can_match(oi1, oi2)) {
	if (!result) result = RESULT_BITS;
	if (optimize) return;

    } else if (vsim_option >= VSIM_ALL && vsim_fail(v2, v1)) {
	if (!result) result = RESULT_VSIM;
	if (optimize) return;

    } else if (overlap < 0) {
	if (v1 * -overlap * 4 > default_wrapsize) {
	    // The gap between probe sets is too large to make any meaningful
	    // inferences.
	    if (!result) {
		result = RESULT_BIG_GAP;
		if (vsim_option == VSIM_NEI && vsim_fail(v2, v1)) {
		    result = RESULT_BIG_GAP_VSIM;
		}
	    }
	    // n_extrap = n1 + n2;
	    return;
	}
    }

    if (tr1->back().min - time_fudge < tr2->front().max + time_fudge) {
	gap_oob_check(r1, r2, optimize, time_fudge);
	if (result && optimize) return;
    } else if (tr2->back().min - time_fudge < tr1->front().max + time_fudge) {
	gap_oob_check(r2, r1, optimize, time_fudge);
	if (result && optimize) return;
    } else {
	overlap_oob_check(false, optimize, time_fudge);
	if (result && optimize) return;
	overlap_oob_check(true, optimize, time_fudge);
	if (result && optimize) return;
	if (!result && oob > 0) {
	    if (oob > oob_inconclusive_allowed * n_bounded) {
		result = RESULT_OUT_OF_BOUNDS;
		if (optimize) return;
	    } else {
		double frac = (first_fail > first_pass && last_fail < last_pass) ?
		    oob_converge_allowed : oob_frac_allowed;
		if (oob > frac * n_bounded) {
		    result = RESULT_INCONCLUSIVE;
		    if (optimize) return;
		}
	    }
	}
    }

    if (dist_option) {
	double tot_dist = 0;
	// old method uses fixed clock offset and RTT error
	const double old_time_fudge = (same_src ? 0 : .200) + .500;
	if (overlap >= 0) {
	    // do distance and old_oob tests for interpolated points
	    big_ipid_t wrapsize1 = 1 << oi1->bits; // wrapsize for id1
	    big_ipid_t wrapsize2 = 1 << oi2->bits; // wrapsize for id2
	    unsigned i1 = 0, i2 = 0;
	    while ((*tr1)[i1].mid() < first_t) { i1++; }
	    while ((*tr2)[i2].mid() < first_t) { i2++; }
	    while (i1 < tr1->size() && i2 < tr2->size() &&
		((*tr1)[i1].mid() <= last_t || (*tr2)[i2].mid() <= last_t))
	    {
		// print "interp:  "; # XXX
		// cout << "i1=" << i1 << "  [" << fixed << (*tr1)[i1] << "," << ((*idr1)[i1] & 0xFFFF) << "]    "; // XXX
		// cout << "i2=" << i2 << "  [" << fixed << (*tr2)[i2] << "," << ((*idr2)[i2] & 0xFFFF) << "]    "; // XXX
		double dist = 0;
		if ((*tr1)[i1].mid() < (*tr2)[i2].mid()) {
		    // cout << "point 1";
		    bool interpolated = interpolation_check(tr1, idr1, i1, wrapsize1,
			tr2, idr2, i2, wrapsize2, v2, oi2->discontinuities,
			optimize, dist, old_time_fudge);
		    if (optimize && result) return;
		    if (!interpolated) continue;
		} else if ((*tr2)[i2].mid() < (*tr1)[i1].mid()) {
		    // cout << "point 2";
		    bool interpolated = interpolation_check(tr2, idr2, i2, wrapsize2,
			tr1, idr1, i1, wrapsize1, v1, oi1->discontinuities,
			optimize, dist, old_time_fudge);
		    if (optimize && result) return;
		    if (!interpolated) continue;
		} else {
		    // extremely rare
		    // XXX need to do old_oob check!
		    // cout << "simultaneous"; // XXX
		    if (dist_option || !optimize) {
			big_ipid_t id1 = (*idr1)[i1] % wrapsize1;
			big_ipid_t id2 = (*idr2)[i2] % wrapsize2;
			dist = iddist(id1, id2);
		    }
		    i1++;
		    i2++;
		}
		tot_dist += fabs(dist);
		n_interp++;
		// cout << "  dist=" << dist << "\n"; // XXX
	    }
	}
	if (n_interp > 0) {
	    edist = idist = tot_dist / n_interp;
	    // cout << "idist=" << idist << "\n"; // XXX

	    if (dist_option >= DIST_INTERP && idist >= dist_thresh) {
		if (!result) result = RESULT_IDIST;
		if (optimize) return;
	    }
	}

	if (n_interp < min_point_thresh && (!result || !optimize)) {
	    // Find distances for some extrapolated points, starting with those
	    // closest to the overlap region and working our way outward until
	    // we have enough.

	    struct ExtrapolationData {
		const char * const &label;
		const int &inc; // increment for i
		int i; // index in actual points
		int n; // number of actual points
		const TimeRangeVector *t; // actual points
		const IpidVector *id; // actual points
		double slope; // slope of extrapolation line
		double base_t; // point to extrapolate from
		big_ipid_t base_id; // point to extrapolate from
		ExtrapolationData(const char * const &_label, const int &_inc) :
		    label(_label), inc(_inc) { }
		void init(const Target * const &data,
		    const Target * const &base,
		    const unsigned &index)
		{
		    t = data->poi()->time;
		    id = data->poi()->id;
		    slope = base->series_velocity;
		    base_t = (*base->poi()->time)[index].mid();
		    base_id = (*base->poi()->id)[index];
		    n = t->size();
		}
	    };

	    ExtrapolationData head("head", -1), tail("tail", +1);

	    if (tr1->front().mid() > tr2->front().mid()) {
		// cout << "head: 1\n";
		head.init(r2, r1, 0);
	    } else {
		// cout << "head: 2\n";
		head.init(r1, r2, 0);
	    }
	    // find head point nearest to interpolation zone
	    head.i = 0;
	    while (head.i+1 < head.n && (*head.t)[head.i+1].mid() < first_t)
		head.i++;

	    if (tr1->back().mid() < tr2->back().mid()) {
		// cout << "tail: 1\n";
		tail.init(r2, r1, oi1->time->size() - 1);
	    } else {
		// cout << "tail: 2\n";
		tail.init(r1, r2, oi2->time->size() - 1);
	    }
	    // find tail point nearest to interpolation zone
	    tail.i = tail.n;
	    while (tail.i-1 >= 0 && (*tail.t)[tail.i-1].mid() > last_t)
		tail.i--;

	    if (0) { // old_oob
		if (overlap < -2 * old_time_fudge) {
		    if (iddist(tail.base_id, head.base_id) <= 0) {
			old_oob++;
			// if (!result) result = RESULT_OUT_OF_BOUNDS;
			// if (optimize) return;
		    }
		} else if (overlap < 0) {
		    if (iddist((*head.id)[head.n-2], head.base_id) <= 0 ||
			iddist(tail.base_id, (*tail.id)[1]) <= 0)
		    {
			old_oob++;
			// if (!result) result = RESULT_OUT_OF_BOUNDS;
			// if (optimize) return;
		    }
		}
	    }

	    if (dist_option || !optimize) {
		while (n_interp + n_extrap < min_point_thresh &&
		    (head.i >= 0 || tail.i < tail.n))
		{
		    // print "extrap:  "; # XXX
		    // cout << "head:" << head.i << "/" << head.n << " tail:" << tail.i << "/" << tail.n << "\n";
		    ExtrapolationData *ed;
		    if (head.i < 0) {
			ed = &tail;
		    } else if (tail.i >= tail.n) {
			ed = &head;
		    } else {
			ed = (head.base_t - (*head.t)[head.i].mid() < (*tail.t)[tail.i].mid() - tail.base_t) ?
			    &head : &tail;
		    }
		    // cout << ed->label << ": " << ed->i << "  [" << setprecision(15) << (*ed->t)[ed->i].mid() << "," << (*ed->id)[ed->i] << "]    "; // XXX
		    // cout << "base [" << setprecision(15) << ed->base_t << "," << ed->base_id << "]    "; // XXX
		    // cout << "slope " << ed->slope << "    "; // XXX
		    double expected_id = extrapolate(ed->base_t, ed->base_id, ed->slope, (*ed->t)[ed->i].mid());
		    big_ipid_t actual_id = (*ed->id)[ed->i];
		    ed->i += ed->inc;
		    while (expected_id >= default_wrapsize) {
			expected_id -= default_wrapsize;
		    }
		    while (expected_id < 0) {
			expected_id += default_wrapsize;
		    }
		    actual_id %= default_wrapsize;
		    double dist = iddist(actual_id, expected_id);
		    tot_dist += fabs(dist);
		    n_extrap++;
		    // cout << "actual=" << actual_id << "   "; // XXX
		    // cout << "expected=" << expected_id << "   "; // XXX
		    // cout << "dist=" << dist << "\n"; // XXX
		}
		if ((n_extrap + n_interp) > 0)
		    edist = tot_dist / (n_extrap + n_interp);
		if (dist_option >= DIST_ALL && edist >= dist_thresh) {
		    if (!result) result = RESULT_EDIST;
		}
	    }
	}
    } // dist_option

    if (!result) {
	if (overlap < 0) {
	    result = RESULT_NO_OVERLAP;
	} else if (n_bounded < min_overlap_option) {
	    result = RESULT_POOR_OVERLAP;
	}
    }

    if (vsim_option == VSIM_NEI &&
	result_types[result] == RESTYPE_NOT_ENOUGH_INFO &&
	vsim_fail(v2, v1))
    {
	switch (result) {
	    case RESULT_BIG_GAP:      result = RESULT_BIG_GAP_VSIM;      break;
	    case RESULT_POOR_OVERLAP: result = RESULT_POOR_OVERLAP_VSIM; break;
	    case RESULT_NO_OVERLAP:   result = RESULT_NO_OVERLAP_VSIM;   break;
	    default: break;
	}
    }

    if (!result) {
	_r1->flags |= SHARED;
	_r2->flags |= SHARED;
	result = RESULT_SHARED;
    }
}

void TargetFilterVector::append(const std::string &label, int status,
    TargetFilter::Func *f1,
    TargetFilter::Func *f2,
    TargetFilter::Func *f3)
{
    push_back(TargetFilter(label, status, f1, f2, f3));
}

void TargetFilterVector::init()
{
    using namespace TargetFilters;

    append("insufficient good responses", STATUS_BAD,
	Not(FlagF(GOOD_RESPONSE)));
    append(   "NOT_ENOUGH_DATA", STATUS_BAD_SUB,
	new ClassF(CLASS_NOT_ENOUGH_DATA));
    append(   "UNRESPONSIVE", STATUS_BAD_SUB,
	new ClassF(CLASS_UNRESPONSIVE));
    append(   "UNSTABLE_TTL", STATUS_BAD_SUB,
	new ClassF(CLASS_UNSTABLE_TTL));
    append(   "MOSTLY_UNEXPECTED", STATUS_BAD_SUB,
	new ClassF(CLASS_MOSTLY_UNEXPECTED));
    append(   "BAD_RESPONSE_MIXED", STATUS_BAD_SUB,
	new ClassF(CLASS_BAD_RESPONSE_MIXED));
    append("degenerate", STATUS_MEDIUMBAD,
	new FlagF(GOOD_RESPONSE), Not(FlagF(GOOD_ID)));
    append(   "MOSTLY_ZERO", STATUS_MEDIUMBAD_SUB,
	new ClassF(CLASS_MOSTLY_ZERO));
    append(   "MOSTLY_REFLECT", STATUS_MEDIUMBAD_SUB,
	new ClassF(CLASS_MOSTLY_REFLECT));
    append(   "MOSTLY_REPEAT", STATUS_MEDIUMBAD_SUB,
	new ClassF(CLASS_MOSTLY_REPEAT));
    append(   "BAD_ID_MIXED", STATUS_MEDIUMBAD_SUB,
	new ClassF(CLASS_BAD_ID_MIXED));
    append("not monotonic", STATUS_MEDIUM,
	new FlagF(GOOD_ID), Not(FlagF(MONOTONIC)));
    append(   "UNKNOWN_ORDER", STATUS_MEDIUM_SUB,
	new ClassF(CLASS_UNKNOWN_ORDER));
    append(   "NONMONOTONIC", STATUS_MEDIUM_SUB,
	new ClassF(CLASS_NONMONOTONIC));
    append("MONOTONIC", STATUS_GOOD,
	new FlagF(MONOTONIC));

    append("monotonic, BE", 0,
	new ClassF(CLASS_MONOTONIC), new OrderF(ORDER_BE));
    append("monotonic, LE", 0,
	new ClassF(CLASS_MONOTONIC), new OrderF(ORDER_LE));
#ifdef OLDCLASSIFY
    append("new and old classification disagree", 0,
	new FlagF(DISCLASSIFY));
    append("new and old bits disagree", 0,
	new FlagF(DISBITS));
    append("mono, new and old order disagree", 0,
	new ClassF(CLASS_MONOTONIC), new FlagF(DISORDER));
    append("nomono, new and old order disagree", 0,
	Not(ClassF(CLASS_MONOTONIC)), new FlagF(DISORDER));
#endif
    append("monotonic, byte order by RHBs", 0,
	new FlagF(MONOTONIC), new DescF("wins RHBs"));
    append("monotonic, byte order by default", 0,
	new FlagF(MONOTONIC), new DescF("wins by default"));
    append("monotonic, byte order by v_iqr", 0,
	new FlagF(MONOTONIC), new DescF("better v_iqr"));
    append("suspicious byte order", 0,
	new FlagF(GOOD_ID), Not(OrderF(ORDER_UNKNOWN)), new DescF("v_med disagrees with byte order"));

    append("XX00 outlier", 0,
	new FlagF(GOOD_ID), new DescF(" XX00:"));
    append("XXFF outlier", 0,
	new FlagF(GOOD_ID), new DescF(" XXFF:"));

    if (bits_flag) {
	append("BE, limited precision", 0,
	    new ClassF(CLASS_MONOTONIC), new OrderF(ORDER_BE), new LimBitsF(true));
	append("LE, limited precision", 0,
	    new ClassF(CLASS_MONOTONIC), new OrderF(ORDER_LE), new LimBitsF(true));
    }

    if (allow_udp_diff_addr) {
	append("Responses from diff addr", 0,
	    new FlagF(GOOD_ID), new FlagF(DIFFADDR));
	append("Responses from same and diff addr", 0,
	    new FlagF(GOOD_ID), new FlagF(SAMEADDR), new FlagF(DIFFADDR));
    }

    for (int rt = 0; rt < N_RESPONSETYPE; rt++) {
	append(std::string(responsetype_label[rt]) + " responses", 0,
	    new FlagF(GOOD_ID), new RespTypeF(rt));
    }

    if (ttl_flag) {
	append("fwd < rev path len", 0,
	    new FlagF(GOOD_ID), new PathLenF(-1));
	append("fwd = rev path len", 0,
	    new FlagF(GOOD_ID), new PathLenF(0));
	append("fwd > rev path len", 0,
	    new FlagF(GOOD_ID), new PathLenF(1));
	append("large path len diff", 0,
	    new FlagF(GOOD_ID), new PathLenDiffF(12));
    }
}

void CompFilterVector::operator()(const std::string &label, int status,
    ComparisonFilter::Func *f1,
    ComparisonFilter::Func *f2,
    ComparisonFilter::Func *f3)
{
    push_back(ComparisonFilter(label, status, f1, f2, f3));
}

void CompFilterVector::operator()(const ResResult &result)
{
    operator()(result_label[result], restype_status[result_types[result]],
	new ComparisonFilters::ResultF(result));
}

TargetFilterVector targetfilters;
unsigned TargetStat::max_examples = 50;
unsigned ComparisonStat::max_examples = 50;

set<string> plotted;

void print_bits_cell(ofstream &out, const Target *const r)
{
    out << "<td>" << int(r->poi()->bits);
    if (r->poi()->bits)
	out << " (" << fmthex(r->poi()->prefix) <<
	    "/" << dec << 16 - r->poi()->bits << ")";
}

#ifdef ENABLE_FEATURES
void print_features_cell(ofstream &out, const Target *const r)
{
    out << "<td>";
    if (r->features) {
	for (int i = 0; i < N_FEATURES; i++) {
	    if ((*r->features)[i] >= 0)
		out << (*r->features)[i] << " ";
	    else
		out << "? ";
	}
    }
}
#endif

string URLencode(const string &in)
{
    string out;
    for (size_t i = 0; i < in.length(); i++) {
	if (isalnum(in[i]) || strchr(".-_", in[i])) {
	    out += in[i];
	} else {
	    out += '%';
	    out += "0123456789ABCDEF"[in[i] >> 4];
	    out += "0123456789ABCDEF"[in[i] % 16];
	}
    }
    return out;
}

void print_classification_cell(ofstream &html1, const TargetStat *targetstat,
    uint64_t total_targets, string label, const char *gptype, bool gpscript)
{
    uint64_t n_matches = targetstat->count;
    string filename(label + ".html");
    ofstream html2;

    if (n_matches) {
	html2.open(filename.c_str());
	if (!html2)
	    cerr << "can't open " << filename << ": " << strerror(errno) << endl;
    }
    if (!n_matches || !html2) {
	html1 << "    <td>" << htmlnum(n_matches);
    } else {
	html1 << "    <td><a href='" << URLencode(filename) << "'>" << htmlnum(n_matches) << "</a>";
    }
    html1 << "<td class='percent'>" << fmtpercent(n_matches, total_targets) << "\n";
    if (!n_matches || !html2) return;

    html2 << "<html>\n";
    html2 << "<head>\n";
    html2 << "<style>\n";
    html2 << "td { text-align: right; }\n";
    html2 << "td, th { border: 1px solid black; padding: 0.5ex; }\n";
    html2 << ".idtable { border-collapse: collapse; font-family: monospace; font-size: 70%; background: white; }\n";
    html2 << ".idtable th, .idtable td { border: 0px transparent; padding: 0px 1ex; }\n";
    // Insert a thin space. This does not work in all browsers (notably IE),
    // but when it fails, it's harmless.
    html2 << ".numeric { white-space: nowrap; }\n";
    html2 << ".numeric span:after { content: \" \"; font-size: 50%; }\n";
    html2 << "</style>\n";
    html2 << "<title>" << label << "</title>\n";
    html2 << "</head>\n";
    html2 << "<body>\n";
    html2 << "<h2>" << label << "</h2>\n";

    // cerr << "# plotting (" << n_matches << ", " << targetstat->examples.size() << ")\n";
    for (vector<const Target*>::const_iterator it = targetstat->examples.begin();
	it != targetstat->examples.end(); it++)
    {
	const Target *r = (*it);

	//print
	    //"meth: ", $ref->[$i_meth], " ", $methname[$ref->[$i_meth]], "; ",
	    //"usable: ", $ref->[$i_usable_flag], "; ",
	    //"swapped: ", $ref->[$i_swapped_flag], "; ",
	    //"outliers: ", ($ref->[$i_outliers] && scalar @{$ref->[$i_outliers]}), "; ",
	    //"classification: ", $class_label[$ref->[$i_classification]], "\n";
	//$| = 1;

	if (!(r->flags & GOOD_ID)) continue;

	// print "## found: $filename $target_addr ", $ref->[$i_desc], "\n";
	html2 << "<div style='border: 1px solid black; padding: 1ex;'>\n";
	html2 << "<h3>" << r->addr << " " << methname[r->meth] << "</h3>\n";
	html2 << "<table style='border-collapse: collapse'>\n";
	html2 << "<tr><td>nondegenerate/responses/tries <td>" << r->nondegenerate << "/" << r->responses << "/" << r->tries << "\n";
	html2 << "<tr><td>source <td>" << r->src_addr << "\n";

        html2 << "<tr><td>response types";
	html2 << "<td>";
	for (int i = 0; i < N_RESPONSETYPE; i++) {
	    if (r->responsetypes & (1 << i))
		html2 << responsetype_label[i] << " ";
	}
	html2 << "\n";

	html2 << "<tr><td>classification <td><b>" << class_label[r->classification] << "</b>\n";

        html2 << "<tr><td>fwd path length<td>";
	if (r->min_fwd_path_len) html2 << int(r->min_fwd_path_len); else html2 << "-";
        html2 << "\n";

	if (ttl_flag) {
	    html2 << "<tr><td>rev init ttl, final ttl, path len<td>";
	    if (r->init_ttl) html2 << int(r->init_ttl); else html2 << "-";
	    html2 << ", ";
	    if (r->max_reply_ttl) html2 << int(r->max_reply_ttl); else html2 << "-";
	    html2 << ", ";
	    if (r->min_rev_path_len) html2 << int(r->min_rev_path_len); else html2 << "-";
	    html2 << "\n";
	}

	html2 << "</table>\n";

	html2 << "<table style='border-collapse: collapse;'>\n";
	html2 << "<tr><th> <th>id<sub>min</sub> <th>id<sub>max</sub> <th>v<sub>med</sub> <th>bits <th>prefix <th>outliers <th>discontinuities\n";
	for (int order = 0; order <= 1; order++) {
	    if (!r->oi[order]) continue;
	    html2 << "<tr";
	    if (r->order == order) html2 << " style='font-weight: bold'";
	    html2 << "><td>" << order_label[order];
	    html2 << "<td>" << fmthex(r->oi[order]->min_id);
	    html2 << "<td>" << fmthex(r->oi[order]->max_id);
	    html2 << "<td>" << r->oi[order]->v_med;
	    html2 << "<td>" << int(r->oi[order]->bits);
	    html2 << "<td>" << fmthex(r->oi[order]->prefix) << "/" << (16 - r->oi[order]->bits);
	    html2 << "<td>" << r->oi[order]->outliers.count();
	    html2 << "<td>" << r->oi[order]->discontinuities.count();
	    html2 << "\n";
	}
	html2 << "</table>\n";

	html2 << "<br>desc: " << r->desc.str() << "\n<br>\n";

        string plotname = string("id-t-") + string(r->addr) + '-' + methname[r->meth] + '@' + string(r->src_addr);
        string title = string(r->addr) + " " + methname[r->meth] + " @" + string(r->src_addr);
        if (plotted.find(plotname) != plotted.end()) {
            // cerr << "# cached plot: " << plotname << "\n";
	} else {
	    const Target *data[] = {r};
            gp.open(plotname.c_str(), title.c_str(), gptype, gpscript);
	    gp.plot_id_vs_time_compact(1, data, 0, true);
	    gp.close();
	    plotted.insert(plotname);
	}

	html2 << "<table style='border-collapse: collapse;'>\n";
	html2 << "<tr>\n";
	if (r->flags & GOOD_ID) {
	    html2 << "<td><img src='" << plotname << ".png'>\n";
	    // html2 << "<td><img src='" << plotname << "-unwrapped.png'>\n";
	}

        html2 << "<td style='vertical-align:top'>\n<table class='idtable'>\n";
        html2 << "<tr><th><th>BE<th>LE\n";
        big_ipid_t mask1 = 0xFFFF;
        big_ipid_t mask2 = 0xFFFF;
        big_ipid_t prefix1 = 0;
        big_ipid_t prefix2 = 0;
        big_ipid_t prev_id1 = 0;
        big_ipid_t prev_id2 = 0;

	if (r->oi[ORDER_BE]) {
	    mask1 = 0xFFFF >> (16 - r->oi[ORDER_BE]->bits);
	    prefix1 = r->oi[ORDER_BE]->prefix;
	}
	if (r->oi[ORDER_LE]) {
	    mask2 = 0xFFFF >> (16 - r->oi[ORDER_LE]->bits);
	    prefix2 = r->oi[ORDER_LE]->prefix;
	}

        double t;
        for (unsigned i = 0; i < r->time.size(); i++) {
            if (i > 0 && i % 34 == 0) { // column is getting long, start a new one
                html2 << "</table>\n";
		html2 << "<td style='vertical-align:top'>\n<table class='idtable'>\n";
                html2 << "<tr><th><th>BE<th>LE\n";
            }
	    html2 << "<tr>";
	    t = r->time[i].mid();
            time_t t_int = int(t + 0.0005);
            char buf[128], tbuf[16];
            strftime(tbuf, sizeof(tbuf), "%H:%M:%S", gmtime(&t_int));
            sprintf(buf, "%s.%03d", tbuf, int((t - t_int) * 1000 + 0.5));
            html2 << "<td>" << buf;

	    big_ipid_t id = r->id[i] & 0xFFFF;
	    html2 << "<td";
	    if (r->oi[ORDER_BE] && r->oi[ORDER_BE]->outliers.test(i))
		html2 << " style='color:#FF0000'";
	    if (id < prev_id1)
		html2 << " style='border-top: 1px solid #aaa'";
	    html2 << ">";
	    html2 << fmthex((id & mask1) | prefix1);
	    prev_id1 = id;

	    id = swap16(id);
	    html2 << "<td";
	    if (r->oi[ORDER_LE] && r->oi[ORDER_LE]->outliers.test(i))
		html2 << " style='color:#FF0000'";
	    if (id < prev_id2)
		html2 << " style='border-top: 1px solid #aaa'";
	    html2 << ">";
	    html2 << fmthex((id & mask2) | prefix2);
	    prev_id2 = id;

	    html2 << "\n";
        }

	html2 << "</table>\n";
	html2 << "</table>\n";
	html2 << "</div>\n";
    }

    html2 << "<br>";
    if (n_matches > targetstat->max_examples) {
	html2 << "stopped after " << targetstat->max_examples;
    } else {
	html2 << "done: " << n_matches;
    }
    html2 << " of " << n_matches << "\n";
    html2 << "</body>\n";
    html2 << "</html>\n";
    html2.close();
}

void print_resolution_cell(ofstream &html, const ComparisonStat *compstat,
    uint64_t n_pairs, string label, const char *gptype, bool gpscript)
{
    uint64_t n_matches = compstat->count;
    string filename(label + ".html");
    ofstream html2;

    if (n_matches) {
	html2.open(filename.c_str());
	if (!html2)
	    cerr << "can't open " << filename << ": " << strerror(errno) << endl;
    }
    if (!n_matches || !html2) {
	html << "<td>" << htmlnum(n_matches);
    } else {
	html << "<td><a href='" << URLencode(filename) << "'>" << htmlnum(n_matches) << "</a>";
    }
    html << "<td class='percent'>" << fmtpercent(n_matches, n_pairs) << "\n";
    if (!n_matches || !html2) return;

    html2.setf(ios_base::fixed, ios_base::floatfield);
    html2.precision(4);

    html2 << "<html>\n";
    html2 << "<head>\n";
    html2 << "<style>\n";
    html2 << "td { text-align: right; vertical-align: top; }\n";
    // html2 << "td, th { border: 1px solid black; padding: 0.5ex; }\n";
    html2 << "td, th { border: 1px solid black; padding: 0.1ex 0.5ex; }\n";
    html2 << ".idtable { border-collapse: collapse; font-family: monospace; font-size: 70%; background: white; }\n";
    html2 << ".idtable th, .idtable td { border: 0px transparent; padding: 0px 1ex; }\n";
    // Insert a thin space. This does not work in all browsers (notably IE),
    // but when it fails, it's harmless.
    html2 << ".numeric { white-space: nowrap; }\n";
    html2 << ".numeric span:after { content: \" \"; font-size: 50%; }\n";
    html2 << ".idtable tr td { border-top: 1px solid #ccc; }\n";
    html2 << ".idtable tr.nobar td { border-top: 1px solid #fff; }\n";
    html2 << ".idtable td.pass { color:green }\n";
    html2 << ".idtable td.fail { color:red }\n";
    html2 << "</style>\n";
    html2 << "<title>" << label << "</title>\n";
    html2 << "</head>\n";
    html2 << "<body>\n";
    html2 << "<h2>" << label << "</h2>\n";

    // cerr << "# plotting (" << n_matches << ", " << compstat->examples.size() << ")\n";
    for (vector<const Comparison*>::const_iterator it = compstat->examples.begin();
	it != compstat->examples.end(); it++)
    {
	const Comparison *comparison = (*it);
        const Target *r1 = comparison->r1;
        const Target *r2 = comparison->r2;
        html2 << "<div style='border: 1px solid black; padding: 1ex;'>\n";
        html2 << "<table style='border-collapse: collapse'>\n";
        html2 << "<tr><th><th>target 1<th>target 2\n";
        html2 << "<tr><td>target" <<
	    "<td>" << r1->addr << " " << methname[r1->meth] <<
	    "<td>" << r2->addr << " " << methname[r2->meth] << "\n";
        html2 << "<tr><td>monitor" <<
	    "<td>" << r1->src_addr <<
	    "<td>" << r2->src_addr << "\n";
        html2 << "<tr><td>byte order" <<
	    "<td>" << r1->order <<
	    "<td>" << r2->order << "\n";
        html2 << "<tr><td>responses of expected type" <<
	    "<td>" << htmlnum(r1->expected) <<
	    "<td>" << htmlnum(r2->expected) << "\n";

        html2 << "<tr><td>response types";
	html2 << "<td>";
	for (int i = 0; i < N_RESPONSETYPE; i++) {
	    if (r1->responsetypes & (1 << i))
		html2 << responsetype_label[i] << " ";
	}
	html2 << "<td>";
	for (int i = 0; i < N_RESPONSETYPE; i++) {
	    if (r2->responsetypes & (1 << i))
		html2 << responsetype_label[i] << " ";
	}
	html2 << "\n";

        html2 << "<tr><td>discontinuities" <<
	    "<td>" << r1->poi()->discontinuities.count() <<
	    "<td>" << r2->poi()->discontinuities.count() << "\n";

        html2 << "<tr><td>velocity" <<
	    "<td>" << r1->series_velocity <<
	    "<td>" << r2->series_velocity << "\n";

        html2 << "<tr><td>fwd path len<td>";
	if (r1->min_fwd_path_len) html2 << int(r1->min_fwd_path_len); else html2 << "-";
	html2 << "<td>";
	if (r2->min_fwd_path_len) html2 << int(r2->min_fwd_path_len); else html2 << "-";
	html2 << "\n";

	if (ttl_flag) {
	    html2 << "<tr><td>rev init ttl, final ttl, path len<td>";
	    if (r1->init_ttl) html2 << int(r1->init_ttl); else html2 << "-";
	    html2 << ", ";
	    if (r1->max_reply_ttl) html2 << int(r1->max_reply_ttl); else html2 << "-";
	    html2 << ", ";
	    if (r1->min_rev_path_len) html2 << int(r1->min_rev_path_len); else html2 << "-";
	    html2 << "<td>";
	    if (r2->init_ttl) html2 << int(r2->init_ttl); else html2 << "-";
	    html2 << ", ";
	    if (r2->max_reply_ttl) html2 << int(r2->max_reply_ttl); else html2 << "-";
	    html2 << ", ";
	    if (r2->min_rev_path_len) html2 << int(r2->min_rev_path_len); else html2 << "-";
	    html2 << "\n";
	}

        html2 << "<tr><td>bits";
	print_bits_cell(html2, r1);
	print_bits_cell(html2, r2);
	html2 << "\n";

#ifdef ENABLE_FEATURES
	if (r1->features || r2->features) {
	    html2 << "<tr><td>features";
	    print_features_cell(html2, r1);
	    print_features_cell(html2, r2);
	    html2 << "\n";
	}
#endif

        html2 << "</table>\n";
        html2 << "<br>v2/v1: " << r2->series_velocity / r1->series_velocity << "\n";
	if (dist_option) {
	    html2 << "<br>interpolated: " << comparison->n_interp << "\n";
	    html2 << "<br>extrapolated: " << comparison->n_extrap << "\n";
	}
        html2 << "<br>overlap: " << setprecision(3) << comparison->overlap << " s\n";
        html2 << "<br>bounds checked: " << comparison->n_bounded << "\n";
        html2 << "<br>out-of-bounds: " << comparison->oob << " (" <<
	    comparison->old_oob << ")\n";
        html2 << "<br>xdisco_oob: " << comparison->n_xdisco_oob << "\n";
	if (dist_option) {
	    html2 << "<br>idist: " << setprecision(4) << comparison->idist << "\n";
	    html2 << "<br>edist: " << setprecision(4) << comparison->edist << "\n";
	}
        html2 << "<br>result: <b>" << result_label[comparison->result] << "</b>\n";
        if (r1->known_nodeid || r2->known_nodeid) {
            html2 << "<br>known aliases: <b>" <<
                (r1->known_nodeid == r2->known_nodeid ? "yes" : "no") << "</b>\n";
        }

        string plotname = string("id-t-") +
	    string(r1->addr) + '-' + methname[r1->meth] + '@' + string(r1->src_addr) + '-' +
	    string(r2->addr) + '-' + methname[r2->meth] + '@' + string(r2->src_addr);
        string title =
	    string(r1->addr) + " " + methname[r1->meth] + " vs " +
	    string(r2->addr) + " " + methname[r2->meth];
        if (plotted.find(plotname) != plotted.end()) {
            // cerr << "# cached plot: " << plotname << "\n";
        } else {
	    const Target *data[] = {r1, r2};
	    const char *labels[] = {"target 1", "target 2"};
            gp.open(plotname.c_str(), title.c_str(), Gnuplotter::PNG, gpscript);
            gp.plot_id_vs_time_compact(2, data, labels);
	    gp.close();
            plotted.insert(plotname);
        }

        html2 << "<table><tr>\n";
        html2 << "<td>\n<img src='" << plotname << ".png'>\n";

	html2 << "<td style='vertical-align:top'>\n<table class='idtable'>\n";
	html2 << "<tr><th>send<th>recv<th>i<th style='color:blue'>tgt1<th style='color:red'>tgt2\n";
	unsigned i1 = 0, i2 = 0, indx;
	TimeRangeVector *time1 = r1->poi()->time, *time2 = r2->poi()->time;
	IpidVector *id1 = r1->poi()->id, *id2 = r2->poi()->id;
	big_ipid_t mask1 = 0xFFFF >> (16 - r1->poi()->bits);
	big_ipid_t mask2 = 0xFFFF >> (16 - r2->poi()->bits);
	big_ipid_t prefix1 = r1->poi()->prefix;
	big_ipid_t prefix2 = r2->poi()->prefix;
	const TimeRange *tr = 0, *prev_tr = 0;
	big_ipid_t id;
	const char *fmt, *rowstyle, *idstyle;
	const bool same_src = (r1->src_addr != 0 && r1->src_addr == r2->src_addr);
	const double time_fudge = same_src ? 0 : max_clock_offset;
	while (i1 < time1->size() || i2 < time2->size()) {
	    if (i1 + i2 > 0 && (i1 + i2) % 34 == 0) { // column is getting long, start a new one
		html2 << "</table>\n";
		html2 << "<td style='vertical-align:top'>\n<table class='idtable'>\n";
		html2 << "<tr><th>send<th>recv<th>i<th style='color:blue'>tgt1<th style='color:red'>tgt2\n";
	    }
	    if (i1 == time1->size() ||
		(i2 != time2->size() && (*time2)[i2].mid() < (*time1)[i1].mid()))
	    {
		indx = i2;
		idstyle = comparison->r2_mbt_fail.test(i2) ? " class='fail'" :
		    comparison->r2_mbt_pass.test(i2) ? " class='pass'" : "";
		tr = &(*time2)[i2];
		id = ((*id2)[i2] & mask2) | prefix2;
		fmt = "<tr%s><td>%s.%03d<td>%s.%03d<td>%d<td>-<td%s>%04x\n";
		i2++;
	    } else {
		indx = i1;
		idstyle = comparison->r1_mbt_fail.test(i1) ? " class='fail'" :
		    comparison->r1_mbt_pass.test(i1) ? " class='pass'" : "";
		tr = &(*time1)[i1];
		id = ((*id1)[i1] & mask1) | prefix1;
		fmt = "<tr%s><td>%s.%03d<td>%s.%03d<td>%d<td%s>%04x<td>-\n";
		i1++;
	    }
	    rowstyle = (prev_tr && tr->min - prev_tr->max > 2 * time_fudge) ?
		"" : " class='nobar'";
	    time_t tmin_int = int(tr->min + 0.0005);
	    time_t tmax_int = int(tr->max + 0.0005);
	    char buf[128], tmin_buf[16], tmax_buf[16];
	    strftime(tmin_buf, sizeof(tmin_buf), "%H:%M:%S", gmtime(&tmin_int));
	    strftime(tmax_buf, sizeof(tmax_buf), ":%S", gmtime(&tmax_int));
	    sprintf(buf, fmt, rowstyle,
		tmin_buf, int((tr->min - tmin_int) * 1000 + 0.5),
		tmax_buf, int((tr->max - tmax_int) * 1000 + 0.5),
		indx, idstyle, id);
	    html2 << buf;
	    prev_tr = tr;
	}
	html2 << "</table>\n";
        html2 << "</table>\n";

        html2 << "</div>\n";
    }
    html2 << "<br>";
    if (n_matches > compstat->max_examples) {
	html2 << "stopped after " << compstat->max_examples;
    } else {
	html2 << "done: " << n_matches;
    }
    html2 << " of " << n_matches << "\n";
    html2 << "</body>\n";
    html2 << "</html>\n";
}

// ttl comparisons
void init_ttl_comparison(ofstream &html, const MultiMethTargetSet &targets,
    const int *n_targets, const char *gptype, bool gpscript)
{
    if (!ttl_flag) return;
    cerr << "# TTLs\n";
    html << "\n<h2>TTLs</h2>\n";
    html << "\n";
    html << "<p>\n";
    int init_ttl_data[N_METH][256];
    memset(init_ttl_data, 0, sizeof(init_ttl_data));
    for (MultiMethTargetSet::const_iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!(*it)->target[meth]) continue;
	    if (!((*it)->target[meth]->flags & GOOD_ID)) continue;
	    if ((*it)->target[meth]->init_ttl == 0) continue;
	    init_ttl_data[meth][(*it)->target[meth]->init_ttl]++;
	}
    }

    html << "<table style='border-collapse: collapse'>\n";
    html << "<tr><th>inferred initial ttl ";
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	html << "<th colspan=2>" << methname[meth];
    }
    html << "<th colspan=2>" << "total";
    html << "</tr>\n";
    for (unsigned i = 0; i < N_TTL; i++) {
	html << "<tr><td>" << init_ttls[i] << "  ";
	int total = 0;
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!n_targets[meth]) continue;
	    html << "<td>" << htmlnum(init_ttl_data[meth][init_ttls[i]]) <<
		" <td class='percent'>" <<
		fmtpercent(init_ttl_data[meth][init_ttls[i]], n_targets[meth]) << "";
	    total += init_ttl_data[meth][init_ttls[i]];
	}
	html << "<td>" << htmlnum(total) <<
	    " <td class='percent'>" << fmtpercent(total, targets.size());
	html << "</tr>\n";
    }

    html << "</table>\n";

    uint64_t *data[N_METH];
    for (int i = 0; i < N_METH; i++) {
	data[i] = newZeroedArray(uint64_t, 256);
    }
    for (MultiMethTargetSet::const_iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!(*it)->target[meth]) continue;
	    if (!((*it)->target[meth]->flags & GOOD_ID)) continue;
	    if ((*it)->target[meth]->max_reply_ttl == 0) continue;
	    data[meth][(*it)->target[meth]->max_reply_ttl]++;
	}
    }

    gp.open("ttl-distribution", "TTL distribution", gptype, gpscript);
    fprintf(gp.f, "set xlabel 'received TTL'\n");
    fprintf(gp.f, "set ylabel 'fraction'\n");
    fprintf(gp.f, "set xrange [0:288]\n");
    // fprintf(gp.f, "set yrange [0:?]\n");
    fprintf(gp.f, "set key top left\n");
    fprintf(gp.f, "set xtics 16\n");
    fprintf(gp.f, "set mxtics 4\n");
    //fprintf(gp.f, "set logscale y\nset mytics 10\n");
    gp.plot_binned_histogram(N_METH, 256, data, 0, 1, methname);
    html << "<br><img src='ttl-distribution.png'>\n";

    gp.open("ttl-log-distribution", "TTL distribution", gptype, gpscript);
    fprintf(gp.f, "set xlabel 'received TTL'\n");
    fprintf(gp.f, "set ylabel 'fraction'\n");
    fprintf(gp.f, "set xrange [0:288]\n");
    // fprintf(gp.f, "set yrange [0:?]\n");
    fprintf(gp.f, "set key top left\n");
    fprintf(gp.f, "set xtics 16\n");
    fprintf(gp.f, "set mxtics 4\n");
    fprintf(gp.f, "set logscale y\nset mytics 10\n");
    gp.plot_binned_histogram(N_METH, 256, data, 0, 1, methname);
    html << "<br><img src='ttl-log-distribution.png'>\n";

    for (int i = 0; i < N_METH; i++) {
	delete[] data[i];
    }

    for (int i = 0; i < N_METH; i++) {
	data[i] = newZeroedArray(uint64_t, 512);
    }
    for (MultiMethTargetSet::const_iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!(*it)->target[meth]) continue;
	    if (!((*it)->target[meth]->flags & GOOD_ID)) continue;
	    if ((*it)->target[meth]->max_reply_ttl == 0) continue;
	    if ((*it)->target[meth]->min_fwd_path_len == 0) continue;
	    int init_ttl = (*it)->target[meth]->min_fwd_path_len - 1 +
		(*it)->target[meth]->max_reply_ttl;
	    data[meth][init_ttl]++;
	}
    }

    gp.open("init-ttl-distribution", "Initial TTL distribution", gptype, gpscript);
    fprintf(gp.f, "set xlabel 'received TTL + fwd path len - 1'\n");
    fprintf(gp.f, "set ylabel 'fraction'\n");
    fprintf(gp.f, "set xrange [0:288]\n");
    // fprintf(gp.f, "set yrange [0:?]\n");
    fprintf(gp.f, "set key top left\n");
    fprintf(gp.f, "set xtics 16\n");
    fprintf(gp.f, "set mxtics 4\n");
    //fprintf(gp.f, "set logscale y\nset mytics 10\n");
    gp.plot_binned_histogram(N_METH, 512, data, 0, 1, methname);
    html << "<br><img src='init-ttl-distribution.png'>\n";

    gp.open("init-ttl-log-distribution", "Initial TTL distribution", gptype, gpscript);
    fprintf(gp.f, "set xlabel 'received TTL + fwd path len - 1'\n");
    fprintf(gp.f, "set ylabel 'fraction'\n");
    fprintf(gp.f, "set xrange [0:288]\n");
    // fprintf(gp.f, "set yrange [0:?]\n");
    fprintf(gp.f, "set key top left\n");
    fprintf(gp.f, "set xtics 16\n");
    fprintf(gp.f, "set mxtics 4\n");
    fprintf(gp.f, "set logscale y\nset mytics 10\n");
    gp.plot_binned_histogram(N_METH, 512, data, 0, 1, methname);
    html << "<br><img src='init-ttl-log-distribution.png'>\n";

    html << "<br>Peaks:\n";
    html << "<ul>\n";
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	html << "<li>" << methname[meth] << ": ";
	for (int i = 1; i < 512 - 1; i++) {
	    if (data[meth][i] > data[meth][i-1] + 1 &&
		data[meth][i] > data[meth][i+1] + 1)
	    {
		if (data[meth][i] > data[meth][i-2] + 2 &&
		    data[meth][i] > data[meth][i+2] + 2)
		{
		    html << "<b>" << i << "</b>" << " "; // strong peak
		} else {
		    html << i << " "; // weak peak
		}
	    }
	}
	html << "\n";
    }
    html << "</ul>\n";

    // density map of fwd path len vs. received ttl
    if (0) { // XXX broken - density array too small?
	int density[288][65];
	for (int i = 0; i < N_METH; i++) {
	    if (!n_targets[i]) continue;
	    memset(density, 0, sizeof(density));
	    int total_targets = 0;
	    for (MultiMethTargetSet::const_iterator it = targets.begin(); it != targets.end(); it++) {
		if (!(*it)->target[i]) continue;
		if (!((*it)->target[i]->flags & GOOD_ID)) continue;
		if (!((*it)->target[i]->max_reply_ttl)) continue;
		if (!((*it)->target[i]->min_fwd_path_len)) continue;
		total_targets++;
		int x = (*it)->target[i]->max_reply_ttl + (*it)->target[i]->min_fwd_path_len - 1;
		int y = (*it)->target[i]->min_fwd_path_len;
		density[x][y]++;
	    }
	    if (!total_targets) continue;

	    char filename[64];
	    char title[64];

	    sprintf(filename, "reply-ttl-log-density-%s", methname[i]);
	    sprintf(title, "Reply TTL density: %s (%d addrs)",
		methname[i], total_targets);
	    gp.open(filename, title, gptype, gpscript);
	    html << "<br><img src='" << filename << ".png'>\n\n";
	    // fprintf(gp.f, "set grid xtics ytics\n");
	    fprintf(gp.f, "set xlabel 'received TTL + fwd path len - 1'\n");
	    fprintf(gp.f, "set ylabel 'forward path length'\n");
	    fprintf(gp.f, "set xtics 16\nset mxtics 4\n");
	    fprintf(gp.f, "set ytics 4\n");
	    fprintf(gp.f, "set logscale cb\n");
	    fprintf(gp.f, "set pm3d map\n");
	    fprintf(gp.f, "set palette rgbformulae -7,-5,-15\n");
	    fprintf(gp.f, "set pm3d corners2color c1\n");
	    fprintf(gp.f, "splot '-' title ''\n");
	    fprintf(gp.f, "set xrange [0:256]\n");
	    fprintf(gp.f, "set yrange [0:64]\n");
	    fprintf(gp.f, "set cbrange [:0.2]\n");
	    for (int x = 0; x <= 288; x++) {
		for (int y = 0; y <= 64; y++) {
		    fprintf(gp.f, "%d %d %f\n", x, y, density[x][y] / double(total_targets));
		}
		fprintf(gp.f, "\n");
	    }
	    fprintf(gp.f, "e\n");
	    gp.close();

	}
    }

    for (int i = 0; i < N_METH; i++) {
	delete[] data[i];
    }
}

// fwd/reverse path length comparison
void path_length_comparison(ofstream &html, const MultiMethTargetSet &targets,
    const char *gptype, bool gpscript)
{
    if (!ttl_flag) return;
    cerr << "# path lengths\n";
    html << "\n<h2>Path lengths</h2>\n";
    html << "\n";
    html << "<p>\n";

    uint64_t overflow = 0;
    int offset = -40;
    int nbins = -offset * 2 + 1;
    uint64_t *data[N_METH];
    for (int i = 0; i < N_METH; i++) {
	data[i] = newZeroedArray(uint64_t, nbins);
    }
    for (MultiMethTargetSet::const_iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!(*it)->target[meth]) continue;
	    if (!((*it)->target[meth]->flags & GOOD_ID)) continue;
	    if ((*it)->target[meth]->min_fwd_path_len == 0) continue;
	    if ((*it)->target[meth]->min_rev_path_len == 0) continue;
	    int diff = (*it)->target[meth]->min_fwd_path_len -
		(*it)->target[meth]->min_rev_path_len;
	    if (diff < offset) {
		diff = offset;
		overflow++;
	    } else if (diff > -offset) {
		diff = -offset;
		overflow++;
	    }
	    data[meth][diff-offset]++;
	}
    }

    gp.open("path-len-diff", "Forward and reverse path length difference", gptype, gpscript);
    fprintf(gp.f, "set xlabel '(fwd path len) - (rev path len)'\n");
    fprintf(gp.f, "set ylabel 'fraction'\n");
    // fprintf(gp.f, "set xrange [-30:30]\n");
    // fprintf(gp.f, "set yrange [0:?]\n");
    fprintf(gp.f, "set key top left\n");
    fprintf(gp.f, "set xtics 5\n");
    fprintf(gp.f, "set mxtics 5\n");
    //fprintf(gp.f, "set logscale y\nset mytics 10\n");
    gp.plot_binned_histogram(N_METH, nbins, data, offset, 1, methname, "linespoints");
    html << "<br><img src='path-len-diff.png'>\n";
    if (overflow) {
	html << "<br>warning: " << overflow << " differences overflowed.\n";
    }

    for (int i = 0; i < N_METH; i++) {
	delete[] data[i];
    }
}

void print_rtt_distribution(ofstream &html, const MultiMethTargetSet &targets,
    const char *gptype, bool gpscript)
{
    cerr << "# RTTs\n";
    html << "\n<h2>RTTs</h2>\n";
    html << "\n";
    html << "<p>\n";
    int rtt_data_nbins = 200;
    double rtt_data_binsize = 2.0 / rtt_data_nbins;
    uint64_t *data[N_METH];
    for (int i = 0; i < N_METH; i++) {
	data[i] = newZeroedArray(uint64_t, rtt_data_nbins);
    }
    for (MultiMethTargetSet::const_iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!(*it)->target[meth]) continue;
	    if (!((*it)->target[meth]->flags & MONOTONIC)) continue;
	    data[meth][int((*it)->target[meth]->max_rtt / rtt_data_binsize)]++;
	}
    }

    gp.open("rtt-distribution", "distribution of RTTs for monotonic targets",
	gptype, gpscript);
    fprintf(gp.f, "set xlabel 'target''s maximum RTT (seconds)'\n");
    fprintf(gp.f, "set ylabel 'fraction'\n");
    // fprintf(gp.f, "set xrange [0:2]\n");
    // fprintf(gp.f, "set yrange [0:?]\n");
    fprintf(gp.f, "set xtics 0.1\n");
    fprintf(gp.f, "set mxtics 5\n");
    fprintf(gp.f, "set grid xtics ytics\n");
    fprintf(gp.f, "set key bottom right\n");
    gp.plot_binned_cdf(false, N_METH, rtt_data_nbins, data, 0, rtt_data_binsize, methname);
    html << "<br><img src='rtt-distribution.png'>\n";

    for (int i = 0; i < N_METH; i++) {
	delete[] data[i];
    }
    gp.close();
}

void OobDistribution::plot(std::ofstream &html, const char *gptype, bool gpscript)
{
    cerr << "# MBT point failure count distribution\n";
    html << "\n<h2>MBT point failure count distribution</h2>\n";

    long sum = 0;
    int max = 0;
    for (int i = 1; i < SIZE; i++) {
	sum += data[i];
	if (data[i]) max = i;
    }

    if (sum == 0) {
	html << "<br>no MBT point failures\n\n";
	html << flush;
	return;
    }

    gp.open("oob-cdf", "MBT point failure count distribution", gptype, gpscript);
    html << "<br><img src='oob-cdf.png'>\n\n";
    fprintf(gp.f, "set title ''\n");
    fprintf(gp.f, "set grid xtics ytics\n");
    fprintf(gp.f, "set xlabel 'MBT point failures'\n");
    fprintf(gp.f, "set ylabel 'fraction of pairs where (1 <= point failures <= x)'\n");
    fprintf(gp.f, "set key top left\n");
    fprintf(gp.f, "set logscale y\n");

    fprintf(gp.f, "plot '-' using ($0+1):($1/%ld) title '' w steps\n", sum);

    long running = 0;
    for (int i = 1; i <= max; i++) {
	running += data[i];
	fprintf(gp.f, "%ld\n", running);
    }
    fprintf(gp.f, "e\n");

    gp.close();
    html << flush;
}

void print_monitor_stats(std::ofstream &html, MonitorSet &monitors,
    const int *n_targets)
{
    html << "\n<h2>Per-monitor statistics</h2>\n";
    html << "<p>A \"monitor series\" is an IPID time series constructed " <<
	"from the first response from each target.\n";
    html << "If the targets are on different routers scattered around the " <<
	"Internet, a monitor series <i>should</i> appear random.\n";
    html << "Values that deviate from expected values are highlighted.\n";
    html << "This is useful as a sanity check on the monitor setup.\n";
    html << "<ul>\n";
    html << "<li>probes - Total number of probes sent.\n";
    html << "<li>resp. - Total number of responses received.\n";
    html << "<li>TTL &sigma; - Standard deviation of reply TTL of " <<
	"monitor series.  A low stddev indicates a near-constant TTL, " <<
	"which suggests most responses may be coming from the same " <<
	"router or network.\n";
    html << "<li>ID-t correlation - Correlation coefficient of ID " <<
	"vs. time for monitor series.  Should be near 0.  A value near 1 " <<
	"suggests most responses may be coming from the same router.\n";
    html << "<li>ID wrap - Fraction of &Delta;ID that are negative for " <<
	"monitor series.  Should be near 0.5.  A value near 0 suggests " <<
	"most responses may be coming from the same router.\n";
    html << "</ul>\n";
    html << "<table style='border-collapse: collapse; font-size: smaller;'>\n";
    int n_meth = 0;
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	n_meth++;
    }
    html << "<tr><th rowspan=2>monitor";
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	html << "<th colspan=5 style='border-width: 1 1 1 2'>" << methname[meth];
    }
    html << "\n<tr>";
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	html <<
	    "<th style='border-width: 1 1 1 2'>probes" << 
	    "<th>resp." <<
	    "<th>TTL<br>&sigma;" <<
	    "<th>ID-t<br>corr." <<
	    "<th>ID<br>wrap";
    }
    html << "\n";
    for (vector<Monitor*>::iterator it = monitors.beginByName();
	it != monitors.endByName(); ++it)
    {
	long mon_probes = 0;
	for (int meth = 0; meth < N_METH; meth++) {
	    mon_probes += (*it)->n_probes[meth];
	}
	if (mon_probes == 0) continue;
	html << "<tr><td>";
	if ((*it)->name != "")
	    html << "<span title='" << (*it)->addr << "'>" << (*it)->name << "</span>";
	else
	    html << (*it)->addr;
	html << "\n";
	html << flush;
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!n_targets[meth]) continue;
	    int gb;
	    // probes
	    html << "  <td style='border-width: 1 1 1 2'>" << (*it)->n_probes[meth] << "\n";
	    // responses
	    if ((*it)->n_probes[meth] > 0) {
		gb = 255 * (1 - pow(double((*it)->n_probes[meth] -
		    (*it)->n_responses[meth]) / (*it)->n_probes[meth], 2));
		html << "  <td style='background: rgb(255," << gb << "," << gb << ")'>" <<
		    (*it)->n_responses[meth] << "\n";
	    } else {
		html << "  <td style='background: #DDDDDD;'>--\n";
	    }
	    // TTL stddev
	    {
		double sum = 0;
		double n = 0;
		for (int i = 0; i < 256; i++) {
		    n += (*it)->reply_ttl_histogram[meth][i];
		    sum += i * (*it)->reply_ttl_histogram[meth][i];
		}
		double mean = sum / n;
		sum = 0;
		for (int i = 0; i < 256; i++) {
		    sum += (*it)->reply_ttl_histogram[meth][i] * pow(mean - i, 2);
		}
		if (n > 1) {
		    double ttl_stddev = sqrt(sum / (n - 1));
		    gb = 255 * (ttl_stddev / 16);
		    if (gb > 255) gb = 255;
		    html << "  <td style='background: rgb(255," << gb << "," << gb << ")'>" <<
			fmtfixed(ttl_stddev, 2) << "\n";
		} else {
		    html << "  <td style='background: #DDDDDD;'>--\n";
		}
	    }
	    // correlation
	    if ((*it)->id[meth].size() > 1) {
		double corr = correlation((*it)->time[meth], (*it)->id[meth]);
		gb = 255 * (1 - corr);
		html << "  <td style='background: rgb(255," << gb << "," << gb << ")'>" <<
		    fmtfixed(corr, 3) << "\n";
	    } else {
		html << "  <td style='background: #DDDDDD;'>--\n";
	    }
	    // wraps
	    double wraps = 0;
	    long n = 0;
	    for (size_t i = 0; i < (*it)->id[meth].size(); i++) {
		if (i > 0 && (*it)->id[meth][i-1]) {
		    n++;
		    // (*(*it)->time[meth])[i].min > (*(*it)->time[meth])[i-1].max && // don't count overlaping samples
		    if ((*it)->id[meth][i] < (*it)->id[meth][i-1])
			wraps++;
		}
	    }
	    if (n > 0) {
		wraps /= n;
		gb = 255 * (1 - pow(fabs(wraps - 0.5) / 0.5, 2));
		html << "  <td style='background: rgb(255," << gb << "," << gb << ")'>" <<
		    fmtfixed(wraps, 3) << "\n";
	    } else {
		html << "  <td style='background: #DDDDDD;'>--\n";
	    }
	    // free unnecessary space
	    vector<double> newtimevec;
	    vector<ipid_t> newidvec;
	    swap((*it)->time[meth], newtimevec);
	    swap((*it)->id[meth], newidvec);
	}
	html << "\n";
    }

    html << "</table>\n";
}

void Gnuplotter::open(const string &filename, const string &title,
    const char *gptype, bool gpscript)
{
    close();
    char buf[1024] = "gnuplot";
    // cerr << "# plot: " << title << "\n";
    if (!f) {
	if ((ispipe = !gpscript)) {
	    f = popen("gnuplot", "w");
	} else {
	    sprintf(buf, "%.*s.gp", int(sizeof(buf)-4), filename.c_str());
	    f = fopen(buf, "w");
	}
    }
    if (!f) {
	cerr << "can't open " << buf << ": " << strerror(errno) << endl;
	exit(1);
    }
    fprintf(f, "reset\n");
    fprintf(f, "%sset terminal png small\n",
	((strcmp(gptype, PNG) != 0) ? "#" : ""));
    fprintf(f, "%sset terminal postscript eps color 14\n",
	((strcmp(gptype, EPS) != 0) ? "#" : ""));
    fprintf(f, "%sset size 0.7,0.7\n", ((strcmp(gptype, EPS) != 0) ? "#" : ""));
    fprintf(f, "set output '%s.%s'\n", filename.c_str(), gptype);
    fprintf(f, "set title '%s'\n", title.c_str());
}

void Gnuplotter::close()
{
    if (!f) return;
    if (ispipe) {
	fflush(f);
    } else {
	fclose(f);
	f = 0;
    }
}

void Gnuplotter::plot_cdf(int nsets, vector<double> *data,
    const char * const *labels, const char *with)
{
    if (!with) with = "linespoints";
    fprintf(f, "set ylabel 'CDF'\n");
    fprintf(f, "set yrange [0:1]\n");
    fprintf(f, "set ytics 0.1\n");
    fprintf(f, "set mytics 5\n");
    const char *prefix = "plot";
    for (int i = 0; i < nsets; i++) {
	if (data[i].empty()) continue;
	int n = data[i].size();
	fprintf(f, "%s '-' using ($1):(($0+1)/%d) title '%s (%d)' with %s",
	    prefix, n, labels[i], n, with);
	if (strchr(with, 'p'))
	    fprintf(f, " ps %f", pointsize(i, nsets));
	prefix = ",";
    }
    fprintf(f, "\n");
    for (int i = 0; i < nsets; i++) {
	if (data[i].empty()) continue;
	sort(data[i].begin(), data[i].end());
	for (vector<double>::iterator j = data[i].begin(); j != data[i].end(); j++) {
	    fprintf(f, "%f\n", *j);
	}
	fprintf(f, "e\n");
    }
}

void Gnuplotter::plot_binned_cdf(bool logbins, int nsets, int nbins, uint64_t **data,
    double offset, double binsize, const char * const *labels, bool logy, bool absolute)
{
    const char *with = "lines";
    if (logbins)
	fprintf(f, "set logscale x\nset mxtics 10\n");
    if (absolute)
	fprintf(f, "set ylabel 'count where X <= x'\n");
    else
	fprintf(f, "set ylabel 'fraction where X <= x'\n");
    if (logy) {
	fprintf(f, "set logscale y\nset mytics 10\n");
	if (!absolute)
	    fprintf(f, "set yrange [:1]\n");
    } else {
	if (!absolute) {
	    fprintf(f, "set yrange [0:1]\n");
	    fprintf(f, "set ytics 0.1\n");
	    fprintf(f, "set mytics 5\n");
	}
    }
    const char *prefix = "plot";
    vector<uint64_t> total(nsets);
    for (int i = 0; i < nsets; i++) {
	total[i] = 0;
	for (int j = 0; j < nbins; j++) {
	    total[i] += data[i][j];
	}
	if (!total[i]) continue;
	// append ".0" to potentially large number so gnuplot reads it as a float
	fprintf(f, "%s '-' using ($1):(($2)", prefix);
	if (!absolute)
	    fprintf(f, "/%" PRIu64 ".0", total[i]);
	fprintf(f, ") title '%s (%" PRIu64 ")' with %s", labels[i], total[i], with);
	if (strchr(with, 'p'))
	    fprintf(f, " ps %f", pointsize(i, nsets));
	prefix = ",";
    }
    fprintf(f, "\n");
    for (int i = 0; i < nsets; i++) {
	if (!total[i]) continue;
	uint64_t sum = 0;
	for (int j = 0; j < nbins; j++) {
	    //if (data[i][j] == 0) continue;
	    sum += data[i][j];
	    double x = j * binsize + binsize/2 + offset; 
	    if (logbins) x = exp(x);
	    // append ".0" to potentially large number so gnuplot reads it as a float
	    fprintf(f, "%f %" PRIu64 ".0\n", x, sum);
	    if (sum == total[i]) break;
	}
	fprintf(f, "e\n");
    }
}

void Gnuplotter::plot_binned_histogram(int nsets, int nbins, uint64_t **data,
    int offset, int binsize, const char * const *labels, const char *with)
{
    vector<uint64_t> sum(nsets);
    bool normalized = true;
    bool cumulative = false;
    //if (logbins) fprintf(f, "set logscale x\nset mxtics 10\n");
    if (cumulative) {
	fprintf(f, "set ylabel '%s'\n", normalized ? "CDF" : "cumulative count");
    } else {
	fprintf(f, "set ylabel '%s'\n", normalized ? "fraction" : "count");
    }
    if (cumulative && normalized) fprintf(f, "set yrange [0:1]\n");
    const char *prefix = "plot";
    for (int i = 0; i < nsets; i++) {
	sum[i] = 0;
	for (int j = 0; j < nbins; j++) {
	    sum[i] += data[i][j];
	}
	if (!sum[i]) continue;
	fprintf(f, "%s '-' using ($1):", prefix);
	if (normalized)
	    fprintf(f, "(($2)/%" PRIu64 ")", sum[i]);
	else
	    fprintf(f, "($2)");
	fprintf(f, " title '%s (%" PRIu64 ")' with %s", labels[i], sum[i], with);
	if (strchr(with, 'p')) fprintf(f, " ps %f", pointsize(i, nsets));
	prefix = ",";
    }
    fprintf(f, "\n");
    for (int i = 0; i < nsets; i++) {
	if (!sum[i]) continue;
	uint64_t val = 0;
	// for (int j = 0; j < nbins; j++) { sum += data[i][j]; }
	for (int j = 0; j < nbins; j++) {
	    if (cumulative) val += data[i][j]; else val = data[i][j];
	    // if (logbins) fprintf(f, "%f %" PRIu64 "\n", exp(j * binsize + offset), val) else
	    fprintf(f, "%d %" PRIu64 "\n", j * binsize + offset, val);
	}
	fprintf(f, "e\n");
    }
}

void Gnuplotter::plot_id_vs_time_compact(unsigned n, const Target **data,
    const char * const *labels, bool both_orders)
{
    // print "# plot_id_vs_time meths: ", join(' ', @meths), "\n";

    double xorigin = 0.10;
    double yorigin = 0.1;
    double xsize = 0.87;
    double ysize = 0.83;
    // unsigned gpoffset = 946684800; // gnuplot stores time as seconds since year 2000

    vector<int> first_point(n);
    vector<int> last_point(n);

    // Pick a plot range that excludes periods when series do not overlap
    if (n == 1) {
	first_point[0] = 0;
	last_point[0] = data[0]->time.size() - 1;
    } else {
	// find median delta_t
	vector<double> delta_ts;
	for (unsigned d = 0; d < n; d++) {
	    first_point[0] = 0;
	    last_point[0] = data[0]->time.size() - 1;
	    for (unsigned i = 1; i < data[d]->time.size(); i++) {
		double delta_t = data[d]->time[i].mid() - data[d]->time[i-1].mid();
		delta_ts.push_back(delta_t);
	    }
	}
	MinMedMax tstats(delta_ts);
	double window = 3.5 * tstats.med;
	// find first and last point in each series that is within time window
	// of a point in another series
	for (unsigned d1 = 0; d1 < n; d1++) {
	    for (unsigned d2 = 0; d2 < n; d2++) {
		if (d1 == d2) continue;
		for (unsigned i = 0; i < data[d1]->time.size(); i++) {
		    for (unsigned j = 0; j < data[d2]->time.size(); j++) {
			if (data[d1]->time[i].mid() > data[d2]->time[j].mid() - window &&
			    data[d1]->time[i].mid() < data[d2]->time[j].mid() + window)
			{
			    first_point[d1] = i;
			    i = data[d1]->time.size(); // break outer loop
			    break;
			}
		    }
		}
		for (int i = data[d1]->time.size(); i > 0; i--) {
		    for (unsigned j = 0; j < data[d2]->time.size(); j++) {
			if (data[d1]->time[i-1].mid() > data[d2]->time[j].mid() - window &&
			    data[d1]->time[i-1].mid() < data[d2]->time[j].mid() + window)
			{
			    last_point[d1] = i-1;
			    i = 0; // break outer loop
			    break;
			}
		    }
		}
	    }
	}
    }

    double max_t = 0;
    double min_t = DBL_MAX;
    unsigned ok = 0;
    for (unsigned i = 0; i < n; i++) {
	if (!(data[i]->flags & GOOD_ID)) continue;
	if (last_point[i] < first_point[i]) continue;
	ok++;
	double t = data[i]->time[first_point[i]].min;
	if (t < min_t) min_t = t;
	t = data[i]->time[last_point[i]].max;
	if (t > max_t) max_t = t;
    }
    if (!ok) return;

    // gnuplot can't handle sub-second precision in time input, so we have to
    // handle all the time formatting ourselves.
    bool minutes = false;
    int xtics = int(ceil((max_t - min_t) / 8));
    if (xtics > 600) {
	xtics = xtics / 60;
	minutes = true;
    }
    if (xtics < 2) xtics = 2;
    else if (xtics > 30) xtics = int(ceil(xtics / 60.0)) * 60;
    else if (xtics > 10) xtics = int(ceil(xtics / 30.0)) * 30;

    int mxtics = 2;
    if (minutes) {
	mxtics =
	    (xtics <= 3) ? 6 :
	    (xtics <= 10) ? xtics / 1 :
	    (xtics <= 30) ? xtics / 5 :
	    xtics / 10;
    } else {
	mxtics =
	    (xtics <= 1) ? 5 :
	    (xtics <= 10) ? xtics / 1 :
	    (xtics <= 30) ? xtics / 5 :
	    xtics / 10;
    }

    if (minutes) xtics *= 60;

    ostringstream xtics_str;
    bool first_xtics = true;
    time_t start = floor((min_t - 1) / xtics) * xtics;
    // Subtracting start from all times in gp file reduces loss of precision
    // and reduces file size.
    for (time_t t = start; t <= max_t + 1; t += xtics) {
	char buf[10];
	struct tm* tm = gmtime(&t);
	strftime(buf, sizeof(buf), minutes ? "%H:%M" : "%H:%M:%S", tm);
	if (!first_xtics) xtics_str << ", ";
	xtics_str << "\"" << buf << "\" " << (t - start);
	first_xtics = false;
	for (int m = 1; m < mxtics; m++) {
	    xtics_str << ", \"\" " << setprecision(16) << (t - start + m * double(xtics)/mxtics) << " 1";
	}
    }

    fprintf(f, "# start time: 0 = %ld\n", start);
    //fprintf(f, "set grid xtics ytics\n");
    //fprintf(f, "set grid xtics\n");
    //fprintf(f, "set grid xtics ytics\n");
    fprintf(f, "set xlabel 'time'\n");
    // fprintf(f, "set xdata time\n");
    // fprintf(f, "set timefmt '%%s'\n");
    fprintf(f, "set format x '%s'\n", minutes ? "%H:%M" : "%H:%M:%S");
    fprintf(f, "set xtics (%s)\n", xtics_str.str().c_str());
    // fprintf(f, "set mxtics %u\n", mxtics);
    //fprintf(f, "set ytics nomirror\n");
    // fprintf(f, "set y2tics 4096\n");
    fprintf(f, "set ytics ('0000' 0, '1000' 4096, '2000' 8192, '3000' 12288, '4000' 16384, '5000' 20480, '6000' 24576, '7000' 28672, '8000' 32768, '9000' 36864, 'a000' 40960, 'b000' 45056, 'c000' 49152, 'd000' 53248, 'e000' 57344, 'f000' 61440, '0000' 65536)\n");
    //fprintf(f, "set format y2 '%%04x'\n");
    fprintf(f, "set ylabel 'ID (hex)'\n");
    fprintf(f, "set key top left\n");
    fprintf(f, "set xrange [%d:%d]\n",
	int(min_t - 1 - start), int(max_t + 1 - start));
    fprintf(f, "set yrange [0:%d]\n", default_wrapsize);
    //fprintf(f, "set y2range [0:%d]\n", default_wrapsize);

    fprintf(f, "set multiplot\n");
    fprintf(f, "set lmargin 0\n");
    fprintf(f, "set rmargin 0\n");
    fprintf(f, "set tmargin 0\n");
    fprintf(f, "set bmargin 0\n");
    fprintf(f, "set size %f, %f\n", xsize, ysize);
    fprintf(f, "set origin %f, %f\n", xorigin, yorigin);

    // create axes and legend.
    const char *gpprefix = "plot";
    for (unsigned i = 0; i < n; i++) {
	if (!(data[i]->flags & GOOD_ID)) continue;
	if (both_orders) {
	    fprintf(f, "%s -1 title '%s' with lp lt %d pt %d",
		gpprefix, order_label[0], cooked_linetype[2*i+0], cooked_pointtype[2*i+0]);
	    fprintf(f, "%s -1 title '%s' with lp lt %d pt %d",
		",", order_label[1], cooked_linetype[2*i+1], cooked_pointtype[2*i+1]);
	} else {
	    fprintf(f, "%s -1 title '%s' with lp lt %d pt %d",
		gpprefix, labels[i], cooked_linetype[i], cooked_pointtype[i]);
	}
	gpprefix = ",";
    }
    fprintf(f, "\n");

    fprintf(f, "set key off\n");
    fprintf(f, "set xlabel ''\n");
    fprintf(f, "set ylabel ''\n");
    fprintf(f, "set noxtics\n");
    fprintf(f, "set noytics\n");
    fprintf(f, "set noy2tics\n");
    fprintf(f, "unset title\n");

//    my $prefix = 'plot';
//    for (my $i = 0; $i < scalar @$refs; $i++) {
//       next unless $refs->[$i][$i_usable_flag];
//       my $my_wrapsize = 1 << $refs->[$i][$i_bits];
//       fprintf(f, $prefix, " '-' using 1:2 title '", $labels->[$i],
//           "' with lp ps ", gp_pointsize($i, scalar @$labels));
//       $prefix = ',';
//    }

//    for (my $i = 0; $i < scalar @$refs; $i++) {
//       next unless $refs->[$i][$i_usable_flag];
//       my $my_wrapsize = 1 << $refs->[$i][$i_bits];
//       fprintf(f, $prefix, "$my_wrapsize title '' ls 0");
//    }

    // individual plots have no axes or legend
    for (int i = n - 1; i >= 0; i--) {
	if (!(data[i]->flags & GOOD_ID)) continue;
	for (int order = 0; order < 2; order++) {
	    if (data[i]->order != ORDER_UNKNOWN) {
		if (!both_orders && order != data[i]->order)
		    continue;
	    }
	    int index = both_orders ? 2*i+order : i;

	    // raw
	    fprintf(f, "# %s %s raw\n", labels ? labels[i] : "", order_label[order]);
	    fprintf(f, "set size %f, %f\n", xsize, ysize);
	    fprintf(f, "set origin %f, %f\n", xorigin, yorigin);
	    fprintf(f, "set yrange [%d:%d]\n", 0, default_wrapsize);
	    fprintf(f, "set border 15 lt %d\n", -1);
	    fprintf(f, "plot '-' using 1:2:3:4 title '' with xerrorlines lt %d pt %d\n",
		raw_linetype[index], raw_pointtype[index]);
	    double prev_t_min = 0;
	    double prev_t_max = 0;
	    double prev_t = 0;
	    big_ipid_t prev_id = 0;

	    // plot every point in [min_t,max_t] plus 1 extra off each side
	    int first = first_point[i];
	    int last = last_point[i];
	    while (first > 0 &&
		data[i]->time[first].max > min_t)
		    first--;
	    while (last < (int)data[i]->time.size() - 1 &&
		data[i]->time[last].min < max_t)
		    last++;

	    for (int j = first; j <= last; j++) {
		double t_min = data[i]->time[j].min - start;
		double t_max = data[i]->time[j].max - start;
		double t = data[i]->time[j].mid() - start;
		big_ipid_t id = data[i]->id[j] & 0xFFFF;
		if (both_orders && order == ORDER_LE)
		    id = swap16(id);
		if (id < prev_id) {
		    // wrap between prev_id and id
		    fprintf(f, "%f %d %f %f\n\n", t, id + default_wrapsize,
			t_min, t_max);
		    fprintf(f, "%f %d %f %f\n", prev_t,
			prev_id - default_wrapsize, prev_t_min, prev_t_max);
		}
		fprintf(f, "%f %d %f %f\n", t, id, t_min, t_max);
		prev_t_min = t_min;
		prev_t_max = t_max;
		prev_t = t;
		prev_id = id;
	    }
	    fprintf(f, "e\n");
	    if (!both_orders && data[i]->order == ORDER_UNKNOWN) break;

	    // cooked
	    OrderInfo *oi = data[i]->oi[order];
	    if (!oi) break;
	    unsigned bits = data[i]->oi[order]->bits;
	    big_ipid_t prefix = data[i]->oi[order]->prefix;
	    big_ipid_t mask = 0xFFFF >> (16 - bits);
	    big_ipid_t wrapsize = 1 << bits;
	    double scale = 1 << (16 - bits);
	    fprintf(f, "# %s %s cooked\n", labels ? labels[i] : "", order_label[order]);
	    fprintf(f, "set size %f, %f\n", xsize, ysize / scale);
	    fprintf(f, "set origin %f, %f\n", xorigin, yorigin + ysize * prefix / default_wrapsize);
	    fprintf(f, "set yrange [%d:%d]\n", prefix, prefix + wrapsize);
	    fprintf(f, "set border 15 lt %d\n", bits == 16 ? -1 : 0);
	    fprintf(f, "plot '-' using 1:2:3:4 title '' with xerrorlines lt %d pt %d\n",
		cooked_linetype[index], cooked_pointtype[index]);
	    prev_t_min = 0;
	    prev_t_max = 0;
	    prev_t = 0;
	    prev_id = 0;
	    for (int j = first; j <= last; j++) {
		double t_min = data[i]->time[j].min - start;
		double t_max = data[i]->time[j].max - start;
		double t = (*oi->time)[j].mid() - start;
		big_ipid_t id = ((*oi->id)[j] & mask) | oi->prefix;
		if (oi->discontinuities.test(j)) {
		    // discontinuity between j-1 and j
		    fprintf(f, "\n");
		} else if (id < prev_id) {
		    // wrap between prev_id and id
		    fprintf(f, "%f %d %f %f\n\n", t, id + wrapsize, t_min, t_max);
		    fprintf(f, "%f %d %f %f\n", prev_t,
			prev_id - wrapsize, prev_t_min, prev_t_max);
		}
		fprintf(f, "%f %d %f %f\n", t, id, t_min, t_max);
		prev_t = t;
		prev_id = id;
	    }
	    fprintf(f, "e\n");
	}
    }
    fprintf(f, "unset multiplot\n");
}

const char *Gnuplotter::PNG = "png";
const char *Gnuplotter::EPS = "eps";

int Gnuplotter::raw_pointtype[]    = {8, 10, 4, 12};
int Gnuplotter::cooked_pointtype[] = {9, 11, 5, 13};
int Gnuplotter::raw_linetype[]     = {5, 9, 19, 13};
int Gnuplotter::cooked_linetype[]  = {3, 1, 2, 4};

} // namespace MidarUtils
