/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// #define OLDCLASSIFY

#include "../lib/config.h"
#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <string>
#include <exception>
#include <stdexcept>
#include <algorithm>
#include <bitset>
#include <math.h>
#include <limits.h>
#include <float.h>
#ifdef HAVE_PTHREAD
# include <pthread.h>
#endif
#include "../lib/ip4addr.h"

namespace MidarUtils {

extern const char *cvsID;

const uint32_t dist_thresh = 500;
const double mostly_thresh = .75;
const double responsive_thresh = .25;
const double outlier_thresh = .05;
const double discontinuity_thresh = .5;
const uint32_t default_wrapsize = 0x10000;
const double wraps_per_cycle_thresh = .3;

enum ProbeMethod { METH_TCP, METH_UDP, METH_ICMP, METH_INDIR, N_METH }; // in descending order of reliability
const char * const methname[N_METH] = { "tcp", "udp", "icmp", "indir" };

enum ResponseType {
    RT_TCP_RST, RT_TCP_OTHER,
    RT_ICMP_PORT_UNREACH, RT_ICMP_ECHO_REPLY, RT_ICMP_TIME_EXCEED, RT_ICMP_OTHER,
    RT_OTHER,
    N_RESPONSETYPE
};
const char * const responsetype_label[N_RESPONSETYPE] = {
    "TCP RST", "TCP other",
    "ICMP port unreach", "ICMP echo reply", "ICMP time exceed", "ICMP other",
    "other"
};

enum Classification {
    CLASS_UNKNOWN = 0,
    CLASS_NOT_ENOUGH_DATA, CLASS_UNRESPONSIVE, CLASS_UNSTABLE_TTL,
    CLASS_MOSTLY_UNEXPECTED, CLASS_BAD_RESPONSE_MIXED,
    CLASS_MOSTLY_ZERO, CLASS_MOSTLY_REFLECT,
    CLASS_MOSTLY_REPEAT, CLASS_BAD_ID_MIXED, CLASS_UNKNOWN_ORDER,
    CLASS_NONMONOTONIC, CLASS_MONOTONIC, N_CLASS
};
const char * const class_label[N_CLASS] = {
    "unknown",
    "not enough data", "unresponsive", "unstable TTL",
    "mostly unexpected", "bad response mixed",
    "mostly zero", "mostly reflect",
    "mostly repeat", "bad id mixed", "unknown order",
    "nonmonotonic", "monotonic"
};

enum Status { STATUS_BAD=1, STATUS_MEDIUMBAD, STATUS_MEDIUM, STATUS_GOOD,
    STATUS_BAD_SUB, STATUS_MEDIUMBAD_SUB, STATUS_MEDIUM_SUB, STATUS_GOOD_SUB };
const char *const status_style[] = { "", "class='bad'", "class='mediumbad'", "class='medium'", "class='good'",
    "class='bad subgroup'", "class='mediumbad subgroup'", "class='medium subgroup'", "class='good subgroup'" };

// restype_label and restype_status must line up with ResultType
enum ResultType { RESTYPE_NONE=0, RESTYPE_UNSHARED=1, RESTYPE_NOT_ENOUGH_INFO, RESTYPE_SHARED };
const std::string restype_label[] = { "?", "UNSHARED", "NOT_ENOUGH_INFO", "SHARED" };
const int restype_status[] = { 0, STATUS_BAD, STATUS_MEDIUM, STATUS_GOOD };

enum ResResult {
#   define defresult(label, conclusive)  RESULT_##label,
#   include "results.h"
    N_RESULT
};
const char * const result_label[N_RESULT] = {
#   define defresult(label, conclusive)  #label,
#   include "results.h"
};
const ResultType result_types[N_RESULT] = {
#   define defresult(label, type)  RESTYPE_##type,
#   include "results.h"
};

enum Order { ORDER_UNKNOWN = -1, ORDER_BE = 0 /* big endian */, ORDER_LE = 1 /* little endian */ };
const char * const order_label[] = { "BE", "LE", "??" };

enum vsim_options { VSIM_NONE, VSIM_NEI, VSIM_ALL };
enum dist_options { DIST_NONE, DIST_CALC, DIST_INTERP, DIST_ALL };

extern int bits_flag;
extern int allow_udp_diff_addr;

// initializer for variable option that allows all values on the command line
#define ENABLED_LONGOPT(opt, var, val)      \
    { opt, no_argument, var, val }
// initializer for constant option that allows only the default on the cmd line
#define DISABLED_LONGOPT(opt, val, defalt)  \
    { ((val == defalt) ? opt : ""), no_argument, 0, 0 }

#define TTL_DEFAULT 0
#if ENABLE_TTL_OPTION
extern int ttl_flag;
# define TTL_LONGOPT(opt, var, val)  ENABLED_LONGOPT(opt, var, val)
#else
const int ttl_flag = TTL_DEFAULT;
# define TTL_LONGOPT(opt, var, val)  DISABLED_LONGOPT(opt, val, TTL_DEFAULT)
#endif

#define VSIM_DEFAULT VSIM_NONE
#if ENABLE_VSIM_OPTION
extern int vsim_option;
# define VSIM_LONGOPT(opt, var, val)  ENABLED_LONGOPT(opt, var, val)
#else
const int vsim_option = VSIM_DEFAULT;
# define VSIM_LONGOPT(opt, var, val)  DISABLED_LONGOPT(opt, val, VSIM_DEFAULT)
#endif

#define DIST_DEFAULT DIST_NONE
#if ENABLE_DIST_OPTION
extern int dist_option;
# define DIST_LONGOPT(opt, var, val)  ENABLED_LONGOPT(opt, var, val)
#else
const int dist_option = DIST_DEFAULT;
# define DIST_LONGOPT(opt, var, val)  DISABLED_LONGOPT(opt, val, DIST_DEFAULT)
#endif

extern int min_overlap_option;
extern int responsive_count;
extern double max_clock_offset; // clock offset error


const long crude_dist_thresh = 500;
const long min_point_thresh = 30; // # of points wanted for avg dist calc
const double vsim_thresh = 2.0/3.0;
const double vsim_fudge = 2;
const double bits_velocity_thresh = 3.0;
const double repeat_high_thresh = .25;

typedef uint32_t ipid_t;
typedef uint32_t big_ipid_t;
typedef uint16_t index_t;

// target flags
enum TargetFlags {
    GOOD_RESPONSE = 0x001,	// has sufficient good responses
    GOOD_ID       = 0x002,	// has sufficient good ID samples
    MONOTONIC     = 0x004,	// has monotonic ID counter
    SAMEADDR      = 0x008,	// response from same addr
    DIFFADDR      = 0x010,	// response from different addr
    SHARED        = 0x020,	// shares counter with another target
    DISORDER      = 0x040,	// new and old byte order disagree
    DISCLASSIFY   = 0x080,	// new and old classification disagree
    DISBITS       = 0x100	// new and old bits disagree
};

#define newZeroedArray(T, n)  static_cast<T*>(memset(new T[n], 0, sizeof(T)*n))

// htmlnum(n) is a convenient way to output a large number formatted for html
template<class T> struct HtmlNum {
    T n;
    HtmlNum(const T &_n) : n(_n) {}
};
template<class T> inline HtmlNum<T> htmlnum(const T &n) { return HtmlNum<T>(n); }
template<class T>
inline std::ostream& operator<< (std::ostream& out, const HtmlNum<T> &num)
{
    std::ostringstream ost;
    ost << num.n;
    std::string str(ost.str());
    for (int i = str.length() - 3; i > 0; i -= 3) {
        str.insert(i, "<span></span>");
    }
    return out << "<span class='numeric'>" << str << "</span>";
}

// fmtnum(n) is a convenient way to output a large number
template<class T> struct FmtNum {
    std::string str;
    FmtNum(const T &_n) {
	std::ostringstream ost;
	ost << _n;
	str = ost.str();
	for (int i = str.length() - 3; i > 0; i -= 3) {
	    str.insert(i, ",");
	}
    }
};
template<class T> inline FmtNum<T> fmtnum(const T &n) { return FmtNum<T>(n); }
template<class T>
inline std::ostream& operator<< (std::ostream& out, const FmtNum<T> &num)
{
    return out << num.str;
}

// fmtpercent(n,d) is convenient way to output a ratio as a percentage value
struct fmtpercent {
    double val;
    fmtpercent(const double &n) : val(n) {}
    fmtpercent(const double &n, const double &d) : val(d ? n/d : 0) {}
};
inline std::ostream& operator<< (std::ostream& out, const fmtpercent& fp)
{
    std::ios_base::fmtflags old_flags = out.flags();
    out << std::fixed << std::setprecision(2) << (100.0 * fp.val) << "%";
    out.flags(old_flags);
    return out;
}

// fmtfixed(d,p) is a convenient way to output a double as a fixed-precision value
struct fmtfixed {
    double val;
    int prec;
    fmtfixed(const double &n, const int &precision = 3) : val(n), prec(precision) {}
};
inline std::ostream& operator<< (std::ostream& out, const fmtfixed& ff)
{
    std::ios_base::fmtflags old_flags = out.flags();
    out << std::fixed << std::setprecision(ff.prec) << (ff.val);
    out.flags(old_flags);
    return out;
}

// fmthex(n) is a convenient way to output an integer as 4-digit hex
template<class T> struct FmtHex {
    T n;
    FmtHex(const T &_n) : n(_n) {}
};
template<class T> inline FmtHex<T> fmthex(const T &n) { return FmtHex<T>(n); }
template<class T>
inline std::ostream& operator<< (std::ostream& out, const FmtHex<T> &num)
{
    std::ios_base::fmtflags old_flags = out.flags();
    out << std::hex << std::setw(4) << std::setfill('0') << num.n;
    out.flags(old_flags);
    return out;
}


uint16_t swap16(uint16_t x);

// number of ways to choose 2 unordered items out of N
#define choose2count(N) (N * (N - 1) / 2)

// index of choice of ith and jth item, out of a list of
// all possible choices of 2 unordered items out of N
inline int choose2index(int N, int i, int j) {
    if (i > j) std::swap(i, j);
    return (i * (2 * N - i - 1) / 2 + (j-i-1));
}

template<class T> class Array2d {
    size_t rows, cols;
    T *data;
public:
    Array2d(size_t _rows, size_t _cols) : rows(_rows), cols(_cols) {
	data = new T[rows*cols];
    }
    ~Array2d() { delete[] data; }
    T &operator()(size_t i, size_t j) { return data[i*cols + j]; }
};

class MinMedMax {
public:
    double min, med, max;
    explicit MinMedMax(std::vector<double> &v); // NB: constructor modifies v
};

class MinQuartilesMax {
public:
    double min, q1, med, q3, max;
    explicit MinQuartilesMax(std::vector<double> &v); // NB: constructor modifies v
};

template<int N>
struct FieldSplitter {
    const char *word[N];
    FieldSplitter(char *str, const char *sep) {
	word[0] = strtok(str, sep);
	for (int i = 1; i < N; i++) {
	    word[i] = strtok(0, sep);
	}
    }
    const char *&operator[](int i) { return word[i]; }
};

class ProbeBitset {
    std::vector<bool> vec;
public:
    ProbeBitset() : vec() {}
    void set(size_t i, size_t size) {
	if (vec.empty())
	    vec.resize(size);
	vec[i] = true;
    }
    bool test(const size_t &i) const {
	return !vec.empty() && vec[i];
    }
    size_t count() const {
	size_t n = 0;
	for (std::vector<bool>::const_iterator it = vec.begin(); it != vec.end(); ++it)
	    if (*it) n++;
	return n;
    }
    void clear() { vec.clear(); }
    bool empty() const { return vec.empty(); }
};

#if 0
class MethPairBitset {
    std::bitset<choose2count(N_METH)> inner;
public:
    bool test(int t1, int t2) const { return inner.test(choose2index(N_METH, t1, t2)); }
    void set(int t1, int t2) { inner.set(choose2index(N_METH, t1, t2)); }
};
#else
typedef std::bitset<choose2count(N_METH)> bitset_meth_pairs;
struct MethPairBitset : public bitset_meth_pairs {
    MethPairBitset(unsigned long val = 0) : bitset_meth_pairs(val) {}
    bool test(int t1, int t2) const
	{ return (static_cast<const bitset_meth_pairs*>(this))->test(choose2index(N_METH, t1, t2)); }
    void set(int t1, int t2)
	{ (static_cast<bitset_meth_pairs*>(this))->set(choose2index(N_METH, t1, t2)); }
};
#endif

class TimeRange {
public:
    double min;
    double max;
private:
    double mid_value;
public:
    TimeRange() : min(), max(), mid_value() {}
    TimeRange(double _min, double _max) : min(_min), max(_max), mid_value((_min+_max)/2) {}
    double mid() const { return mid_value; }
};

typedef std::vector<TimeRange> TimeRangeVector;
typedef std::vector<big_ipid_t> IpidVector;

// Info for interpreting data with a particular byte order
class OrderInfo {
public:
    TimeRangeVector *time;
    IpidVector *id;
    uint8_t bits;
    ipid_t prefix;
    double v_med;
    ProbeBitset outliers;
    ProbeBitset discontinuities; // discon at i means between points i-1 and i
    // uint32_t wrapsize;
    ipid_t min_id;
    ipid_t max_id;

    OrderInfo() : time(0), id(0),
	bits(16), prefix(0),
	v_med(0), outliers(),
	discontinuities(), /* wrapsize(default_wrapsize), */ min_id(0), max_id(0)
	{ }

    void set_minmax_id()
    {
	if (min_id) return;
	min_id = 0xFFFF;
	max_id = 0;
	ipid_t mask = 0xFFFF >> (16 - bits);
	for (unsigned i = 0; i < id->size(); i++) {
	    ipid_t val = (*id)[i] & mask;
	    if (val < min_id) min_id = val;
	    if (val > max_id) max_id = val;
	}
	min_id += prefix;
	max_id += prefix;
    }

    double calculate_velocity();
};

struct BadclockSet : public std::set<ip4addr_t> {
    bool contains(ip4addr_t addr) { return this->find(addr) != this->end(); }
    void parse(char *str) {
	// syntax: comma-separated list of ip addresses
	for (char *p = str; true; str = ++p) {
	    while (*p && *p != ',') p++;
	    this->insert(ip4addr_t(std::string(str, p-str)));
	    if (!*p) break;
	}
    }
};

struct ResponseData {
    char result;
    // int bogus_cycle_id
    ip4addr_t target_addr;
    const char *method;
    ip4addr_t src_addr; // probe source
    // ip4addr_t probe_dst_addr;
    int probe_ttl;
    ipid_t probe_ipid;
    double probe_time;
    ip4addr_t resp_src_addr; // response source
    int reply_ttl;
    int reply_quoted_ttl;
    ipid_t reply_ipid;
    int16_t resp_icmp_type; // note: -1 indicates non-icmp
    int16_t resp_icmp_code; // note: -1 indicates non-icmp
    int16_t resp_tcp_flags; // note: -1 indicates non-tcp
    // resp_warts_id
    double resp_time;
    double prev_resp_time;
    double rtt;
    int ttl_expansion;

    void parse(char *buf);
    void parseRadarGun(char *buf, const char *meth);
    ResponseData() : prev_resp_time(0) {}
};

#ifdef ENABLE_FEATURES
// Feature tests
enum FeatureID {
    TCP_INIT_TTL,
    ICMP_INIT_TTL,
    INDIR_INIT_TTL,
    TCP_IPID_CLASS,
    ICMP_IPID_CLASS,
    INDIR_IPID_CLASS,
    TCP_IPID_ORDER,
    ICMP_IPID_ORDER,
    INDIR_IPID_ORDER,
    ICMP_ECHO_CODE,
    N_FEATURES
};
const char * const feature_label[N_FEATURES] = {
    "TCP initial TTL",
    "ICMP initial TTL",
    "INDIR initial TTL",
    "TCP IPID class",
    "ICMP IPID class",
    "INDIR IPID class",
    "TCP IPID order",
    "ICMP IPID order",
    "INDIR IPID order",
    "ICMP ECHO REPLY code",
};
// other possible feature tests
//   tcp & icmp share ID counter:  yes, no
//   tcp & indir share ID counter:  yes, no
//   icmp & indir share ID counter:  yes, no
//   icmp response DF bit: always, never, same as request, opposite of request
//   indir response DF bit: yes, no
//   port unreachable response IP length:  number
// these depend on tcp port number:
//   tcp response type:  none, TCP RST, ICMP PORT UNREACH
//   tcp RST seq num:  zero, probe ack num, probe ack num + 1, other
//   tcp RST ack num:  zero, probe seq num, probe seq num + 1, other
//   tcp RST flags:  ...
//   tcp RST response DF bit: yes, no
//   tcp PORT_UNREACH response DF bit: yes, no

enum IPID_CLASS { IPID_ZERO, IPID_REFLECT, IPID_REPEAT, IPID_NONDEGENERATE, N_IPID_CLASS };
const char * const IPID_CLASS_label[N_IPID_CLASS] =
    { "zero", "reflect", "repeat", "nondegen" };

class Feature {
    int16_t v[N_FEATURES];
public:
    Feature() {
	for (int i = 0; i < N_FEATURES; i++) { v[i] = -1; }
    }
    int16_t &operator[](int i) { return v[i]; }

    bool match(Feature *f) {
	for (int i = 0; i < N_FEATURES; i++) {
	    if (v[i] != -1 && f->v[i] != -1 && v[i] != f->v[i]) {
		return false;
	    }
	}
	return true;
    }
};
#endif

// 30 and 32 are both popular, but impossible for us to distinguish, so we
// lump them both into the 32 category.
const int init_ttls[] = { 0, 32, 64, 128, 155, 192, 255 };
#define N_TTL (sizeof(init_ttls)/sizeof(init_ttls[0]))

inline uint8_t guess_initial_ttl(uint8_t ttl)
{
    for (unsigned i = 0; i < N_TTL; i++) {
	if (ttl <= init_ttls[i])
	    return init_ttls[i];
    }
    return 255;
}

#ifdef ENABLE_FEATURES
inline uint16_t class_to_feature(Classification c)
{
    switch (c) {
	case CLASS_UNRESPONSIVE:     return -1; // unknown
	case CLASS_MOSTLY_ZERO:      return IPID_ZERO;
	case CLASS_MOSTLY_REFLECT:   return IPID_REFLECT;
	case CLASS_MOSTLY_REPEAT:    return IPID_REPEAT;
	case CLASS_UNKNOWN_ORDER:    return IPID_NONDEGENERATE;
	case CLASS_NONMONOTONIC:     return IPID_NONDEGENERATE;
	case CLASS_MONOTONIC:        return IPID_NONDEGENERATE;
	default: return -1; // unknown
    }
}
#endif


struct AddrPair {
    ip4addr_t addr[2];
public:
    const ip4addr_t& operator[](size_t i) const { return addr[i]; }
    ip4addr_t& operator[](size_t i) { return addr[i]; }
};

struct addrpair_less {
    bool operator()(const AddrPair &a, const AddrPair &b) const {
	return a.addr[0] != b.addr[0] ? a.addr[0] < b.addr[0] :
	    a.addr[1] < b.addr[1];
    }
};

typedef std::set<AddrPair, addrpair_less> AddrPairSet;


struct Monitor {
    ip4addr_t addr;
    std::string name;
    int n_probes[N_METH];
    int n_responses[N_METH];
    uint8_t reply_ttl_histogram[N_METH][256];
    std::vector<double> time[N_METH];
    std::vector<ipid_t> id[N_METH];
private:
    void init()
    {
	for (int i = 0; i < N_METH; i++) {
	    n_probes[i] = 0;
	    n_responses[i] = 0;
	    memset(&reply_ttl_histogram[i][0], 0, 256);
	}
    }
public:
    Monitor(const ip4addr_t &_a = ip4addr_t(0)) :
	addr(_a) { init(); }
    Monitor(const ip4addr_t &_a, const std::string &_name) :
	addr(_a), name(_name) { init(); }
};

struct monitor_less {
    bool operator()(const Monitor * const &a, const Monitor * const &b) const {
        return (a->addr < b->addr);
    }
};

struct monitor_name_less {
    bool operator()(const Monitor * const &a, const Monitor * const &b) const {
        return (a->name != b->name) ? (a->name < b->name) : (a->addr < b->addr);
    }
};

class MonitorSet : public std::set<Monitor*, monitor_less> {
    std::vector<Monitor*> sorted;
public:
    Monitor *countProbeFrom(ip4addr_t addr, int meth)
    {
	static Monitor key; // static to avoid repeated ctor cost
	key.addr = addr;
	Monitor *monitor;
	MonitorSet::const_iterator it = this->find(&key);
	if (it == this->end()) {
	    monitor = new Monitor(addr);
	    this->insert(monitor);
	} else {
	    monitor = *it; // known monitor
	}
	monitor->n_probes[meth]++;
	return monitor;
    }

    void load(const char *filename)
    {
	char buf[8196];
	int linenum = 0;
	InFile in(filename);
	while (in.gets(buf, sizeof(buf))) {
	    linenum++;
	    if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	    MidarUtils::FieldSplitter<2> fields(buf, " \t\n");
	    Monitor *monitor = new Monitor(ip4addr_t(fields[1]), std::string(fields[0]));
	    std::pair<MonitorSet::iterator, bool> p = this->insert(monitor);
	    if (!p.second) {
		std::cerr << "# WARNING: monitor " << fields[0] <<
		    " has duplicate address " << fields[1] <<"\n";
		delete monitor;
	    }
	}
	in.close();
    }

    void sortByName()
    {
	for (MonitorSet::iterator it = this->begin();
	    it != this->end(); ++it)
	{
	    long mon_probes = 0;
	    for (int meth = 0; meth < N_METH; meth++) {
		mon_probes += (*it)->n_probes[meth];
	    }
	    if (mon_probes == 0) continue;
	    sorted.push_back(*it);
	}
	sort(sorted.begin(), sorted.end(), monitor_name_less());
    }

    std::vector<Monitor*>::iterator beginByName() { return sorted.begin(); }
    std::vector<Monitor*>::iterator endByName() { return sorted.end(); }
    std::vector<Monitor*>::size_type sizeByName() { return sorted.size(); }
};

class Target {
public:
    ip4addr_t addr;
    ip4addr_t src_addr;
    int8_t meth;
    uint8_t responsetypes;
    uint16_t flags;
    index_t tries;
    index_t probes;
    index_t responses;
    index_t expected;
    index_t unexpected;
    index_t zero;
    index_t reflect;
    index_t repeat;
    index_t nondegenerate;
    index_t wraps;
    double wrapfrac;
    TimeRangeVector time;
    IpidVector id;
    double series_velocity;
    double max_rtt;
    Classification classification;
    Order order;
    OrderInfo *oi[2];
    uint8_t ttl_expansions;
    uint8_t ttl;
    uint8_t max_reply_ttl;
    uint8_t min_fwd_path_len;
    uint8_t min_rev_path_len;
    uint8_t init_ttl;
    std::ostringstream desc;
    unsigned known_nodeid;
    mutable unsigned refcount;
#ifdef ENABLE_FEATURES
    Feature *features;
#endif

    explicit Target(ip4addr_t _a = ip4addr_t(0)) : addr(_a), meth(-1),
	responsetypes(), flags(),
	tries(), probes(), responses(), expected(), unexpected(), zero(), 
	reflect(), repeat(), nondegenerate(), wraps(), wrapfrac(), time(), id(),
	series_velocity(), max_rtt(),
	classification(CLASS_UNKNOWN), order(ORDER_UNKNOWN), 
	ttl_expansions(), ttl(), max_reply_ttl(),
	min_fwd_path_len(), min_rev_path_len(),
	init_ttl(), desc(), known_nodeid(), refcount(0)
#ifdef ENABLE_FEATURES
	, features(0)
#endif
	{ oi[0] = oi[1] = 0; }

#ifdef OLDCLASSIFY
    Target(const Target &t) { // copy constructor
	if (t.oi[0] || t.oi[1]) {
	    throw std::runtime_error("Target copy constructor is not defined when oi is set");
	}
	addr             = t.addr;
	src_addr         = t.src_addr;
	meth           = t.meth;
	flags            = t.flags;
	tries            = t.tries;
	probes           = t.probes;
	responses        = t.responses;
	expected         = t.expected;
	unexpected       = t.unexpected;
	zero             = t.zero;
	reflect          = t.reflect;
	repeat           = t.repeat;
	nondegenerate    = t.nondegenerate;
	wraps            = t.wraps;
	wrapfrac         = t.wrapfrac;
	time             = t.time;
	id               = t.id;
	series_velocity  = t.series_velocity;
	max_rtt          = t.max_rtt;
	responsetypes    = t.responsetypes;
	classification   = t.classification;
	order            = t.order;
	oi[0]            = 0;
	oi[1]            = 0;
	ttl_expansions   = t.ttl_expansions;
	ttl              = t.ttl;
	max_reply_ttl    = t.max_reply_ttl;
	min_fwd_path_len = t.min_fwd_path_len;
	min_rev_path_len = t.min_rev_path_len;
	init_ttl         = t.init_ttl;
	// desc             = t.desc;
	known_nodeid     = t.known_nodeid;
	refcount         = 0;
#ifdef ENABLE_FEATURES
	# error "features not implemented"
#endif
    }
#endif

    OrderInfo *poi() const { return oi[order]; } // prefered OrderInfo

    void processResponse(ResponseData *rd)
    {
	using namespace std;
	this->responses++;

	// Older data was post-processed with fix-probe-method-results, which
	// changed some 'T' to 'U'.  We treat them both as 'T', and apply our
	// own updated logic below to decide if the result is unexpected.
	if (rd->result != 'T' && rd->result != 'U') {
	    this->unexpected++;
	    return;
	}

	switch (this->meth) {
	case METH_TCP:
	    // expect any TCP response or ICMP PORT UNREACHABLE, from target
	    if (!(
		rd->resp_src_addr == rd->target_addr &&
		(rd->resp_tcp_flags >= 0 ||
		(rd->resp_icmp_type == 3 && rd->resp_icmp_code == 3))))
	    {
		this->unexpected++;
		return;
	    }
	    break;
	case METH_UDP:
	    // expect ICMP PORT UNREACHABLE (maybe from different addr)
	    if (!((allow_udp_diff_addr || rd->resp_src_addr == rd->target_addr) &&
		rd->resp_icmp_type == 3 && rd->resp_icmp_code == 3))
	    {
		this->unexpected++;
		return;
	    }
	    break;
	case METH_ICMP:
	    // expect ICMP ECHO REPLY from target
	    if (!(rd->resp_src_addr == rd->target_addr &&
		(rd->resp_icmp_type == 0)))
	    {
		this->unexpected++;
		return;
	    }
	    break;
	case METH_INDIR:
	    // expect ICMP TIME EXCEEDED from target (which was not the
	    // probe's destination)
	    if (!(rd->resp_src_addr == rd->target_addr &&
		rd->resp_icmp_type == 11))
	    {
		this->unexpected++;
		return;
	    }
	    break;
	}

	this->expected++;

	rd->rtt = rd->resp_time - rd->probe_time;
	if (rd->rtt < 0 || rd->rtt > 2) {
	    cerr << "WARNING: ignoring response from " << this->addr <<
		" with rtt=" << fmtfixed(rd->rtt, 6) << endl;
	    return;
	}

	if (rd->reply_ipid == rd->probe_ipid) {
	    this->reflect++;
	    // This might be just coincidence, but it's safest to throw it out.
	} else if (rd->reply_ipid == 0) {
	    this->zero++;
	    // This might be just coincidence, but it's safest to throw it out.
	} else {
	    this->nondegenerate++;
	    this->time.push_back(TimeRange(rd->probe_time, rd->resp_time));
	    this->id.push_back(rd->reply_ipid);
	}

	ResponseType responsetype = RT_OTHER;
	if (rd->resp_tcp_flags >= 0) {
	    if (rd->resp_tcp_flags & 0x4)
		responsetype = RT_TCP_RST;
	    else
		responsetype = RT_TCP_OTHER;
	} else if (rd->resp_icmp_type >= 0) {
	    if (rd->resp_icmp_type == 3 && rd->resp_icmp_code == 3)
		responsetype = RT_ICMP_PORT_UNREACH;
	    else if (rd->resp_icmp_type == 0)
		responsetype = RT_ICMP_ECHO_REPLY;
	    else if (rd->resp_icmp_type == 11)
		responsetype = RT_ICMP_TIME_EXCEED;
	    else
		responsetype = RT_ICMP_OTHER;
	}
	this->responsetypes |= (1 << responsetype);

	this->flags |= (rd->resp_src_addr == rd->target_addr) ? SAMEADDR : DIFFADDR;

	if (this->max_reply_ttl < rd->reply_ttl)
	    this->max_reply_ttl = rd->reply_ttl;

	if (this->max_rtt < rd->rtt)
	    this->max_rtt = rd->rtt;

	switch (this->meth) {
	case METH_TCP:
	case METH_UDP:
	case METH_INDIR:
	    if (rd->reply_quoted_ttl) {
		uint8_t fwd_path_len = rd->probe_ttl + 1 - rd->reply_quoted_ttl;
		if (this->min_fwd_path_len == 0 || this->min_fwd_path_len > fwd_path_len)
		    this->min_fwd_path_len = fwd_path_len;
	    }
	}

	this->ttl = rd->probe_ttl;

	if (rd->ttl_expansion != 0)
	    this->ttl_expansions++;
    }

    void freeOrder(const int &o)
    {
	// deallocate space for OrderInfo
	if (this->oi[o]) {
	    if (this->oi[o]->id && this->oi[o]->id != &this->id)
		delete this->oi[o]->id;
	    if (this->oi[o]->time && this->oi[o]->time != &this->time)
		delete this->oi[o]->time;
	    delete this->oi[o];
	    this->oi[o] = 0;
	}
    }

    void freeUnusedOrder()
    {
	// deallocate space for unused OrderInfo
	for (int o = 0; o < 2; o++) {
	    if (this->order != o) {
		freeOrder(o);
	    }
	}
    }

    void freeResponses()
    {
	this->freeOrder(0);
	this->freeOrder(1);
	IpidVector newidvec;
	TimeRangeVector newtimevec;
	swap(this->id, newidvec);
	swap(this->time, newtimevec);
    }
};


struct MultiMethTarget {
    ip4addr_t addr;
    int meth;
    MethPairBitset shared;
#ifdef ENABLE_FEATURES
    Feature feature;
#endif
    Target *target[N_METH];

    MultiMethTarget(ip4addr_t a) : addr(a), meth(-1), shared(), target() { }

    Target *&ptarget() { return target[meth]; }
};

struct mttarget_less {
    bool operator()(const MultiMethTarget * const &a, const MultiMethTarget * const &b) const {
        return (a->addr < b->addr);
    }
};

class MultiMethTargetSet : public std::set<MultiMethTarget*, mttarget_less> {
public:
    MultiMethTarget *findOrInsert(ip4addr_t addr)
    {
	MultiMethTarget *mttarget;
	MultiMethTarget key(addr);
	MultiMethTargetSet::const_iterator it = this->find(&key);
	if (it != this->end()) {
	    mttarget = *it; // known target
	} else {
	    mttarget = new MultiMethTarget(addr);
	    this->insert(mttarget);
	}
	return mttarget;
    }
};


bool target_velocity_greater(const Target *a, const Target *b);
bool target_last_timestamp_less(const Target *a, const Target *b);

// throws std::runtime_error for invalid method name
void parse_meth(const char *method, int &meth, int &ttl_expansion);

double ids_per_probe(IpidVector *idr);
double median_ids_per_probe(IpidVector *idr);

// Simple correlation algorithm
// Calculates square of correlation coefficient of a set of points,
// given as a list of x values and a list of y values.
// Returns 0 if there are not enough points.
// (Must be defined in .h file, not a .c file, because it's a template)
template<class TX, class TY> static double
    correlation(const std::vector<TX> &xr, const std::vector<TY> &yr)
{
    unsigned n = xr.size();
    if (n <= 1) return 0;
    double xoff = xr[0]; // hack to reduce numerical instability
    double yoff = yr[0]; // hack to reduce numerical instability
    double sumx = 0, sumy = 0, sumxy = 0, sumx2 = 0, sumy2 = 0;
    for (unsigned i = 0; i < n; i++) {
        double x = double(xr[i]) - xoff;
        double y = double(yr[i]) - yoff;
        sumx += x;
        sumy += y;
        sumxy += x * y;
        sumx2 += x * x;
        sumy2 += y * y;
    }
    double num = (n * sumxy - sumx * sumy);
    return num * num / ((n * sumx2 - sumx*sumx) * (n * sumy2 - sumy*sumy));
}

// distance from a to b, i.e. (b - a), allowing for wrapping
inline int32_t iddist(big_ipid_t a, big_ipid_t b, big_ipid_t wrapsize = default_wrapsize)
{
    if (a >= wrapsize) a %= wrapsize;
    if (b >= wrapsize) b %= wrapsize;
    int32_t dist = (b > a) ? b - a : wrapsize - a + b;
    if (dist > int32_t(wrapsize) / 2) { dist -= wrapsize; }
    return dist;
}

// distance from a to b, i.e. (b - a), allowing for wrapping
inline double iddist(big_ipid_t a, double b, big_ipid_t wrapsize = default_wrapsize)
{
    if (a >= wrapsize) a %= wrapsize;
    if (b >= wrapsize) b = fmod(b, wrapsize);
    double dist = (b > a) ? b - a : wrapsize - a + b;
    if (dist > wrapsize / 2) { dist -= wrapsize; }
    return dist;
}

// interpolate the value at time t on the line between points t0,id0 and t1,id1
inline double interpolate(double t0, double id0, double t1, double id1, double t,
    big_ipid_t wrapsize = default_wrapsize)
{
    if (!t0 || !t1) return 0; // shouldn't happen
    if (id1 < id0) { id1 += wrapsize; }
    if (t1 == t0) { // very unlikely
        return (id0 + id1) / 2;
    } else {
        return id0 + (id1 - id0) / (t1 - t0) * (t - t0);
    }
}

// extrapolate the value at time t on the line passing through t0,id0 with slope
inline double extrapolate(double t0, double id0, double slope, double t)
{
    return id0 + slope * (t - t0);
}

inline double vsim(double v2, double v1)
{
    return ((v2 + vsim_fudge) / (v1 + vsim_fudge));
}

inline bool vsim_fail(double v2, double v1)
{
    return (vsim(v2, v1) < vsim_thresh);
}

void classify(Target *target, bool optimize);

class Mutex {
#if HAVE_PTHREAD
    pthread_mutex_t m;
public:
    Mutex(pthread_mutexattr_t *attr = 0) {
	int r = pthread_mutex_init(&m, attr);
	if (r) throw std::runtime_error(strerror(r));
    }
    ~Mutex() {
	int r = pthread_mutex_destroy(&m);
	if (r) throw std::runtime_error(strerror(r));
    }
    void lock();
    bool trylock();
    void unlock();
#else
public:
    void lock() { }
    bool trylock() { return true; }
    void unlock() { }
#endif
};

extern Mutex memlock;

struct Callable {
    virtual void operator()(void) = 0;
    virtual ~Callable() {}
};

#if HAVE_PTHREAD
class Thread {
    pthread_t pthread;
public:
    Thread() {}
    ~Thread() {}
    static void *func(void *arg) {
	Callable *callable = static_cast<Callable*>(arg);
	(*callable)();
	return 0;
    }
    void start(Callable *callable, pthread_attr_t *attr = 0) {
	int r = pthread_create(&pthread, attr, func, callable);
	if (r) throw std::runtime_error(strerror(r));
    }
    void join() {
	int r = pthread_join(pthread, 0);
	if (r) throw std::runtime_error(strerror(r));
    }
};
#endif

class TargetFilter {
public:
    struct Func : public std::unary_function<Target*, bool> {
	virtual bool operator()(const Target * const &r) const = 0;
    };
    std::string label;
    int status;
    const Func *f1, *f2, *f3, *f4;

    struct True : public Func {
	bool operator()(const Target * const &r) const { return true; }
    };

    explicit TargetFilter(const std::string _label, int _status,
	const Func *_f1 = new True(), const Func *_f2 = new True(),
	const Func *_f3 = new True(), const Func *_f4 = new True()) :
	label(_label), status(_status), f1(_f1), f2(_f2), f3(_f3), f4(_f4)
	{ }

    bool operator()(const Target * const &r) const {
	return (*f1)(r) && (*f2)(r) && (*f3)(r) && (*f4)(r);
    }
};

namespace TargetFilters {
    class ClassF : public TargetFilter::Func {
	int clas;
    public:
	ClassF(int _clas) : clas(_clas) {}
	bool operator()(const Target * const &r) const
	    { return r->classification == clas; }
    };

    class FlagF : public TargetFilter::Func {
	int flag;
    public:
	FlagF(int _flag) : flag(_flag) {}
	bool operator()(const Target * const &r) const
	    { return r->flags & flag; }
    };

    class RespTypeF : public TargetFilter::Func {
	int rtmask;
    public:
	RespTypeF(int _rt) : rtmask(1 << _rt) {}
	bool operator()(const Target * const &r) const
	    { return r->responsetypes & rtmask; }
    };

    class OrderF : public TargetFilter::Func {
	int order;
    public:
	OrderF(int _order) : order(_order) {}
	bool operator()(const Target * const &r) const
	    { return r->order == order; }
    };

    class LimBitsF : public TargetFilter::Func {
	bool flag;
    public:
	LimBitsF(int _flag) : flag(_flag) {}
	bool operator()(const Target * const &r) const
	    { return (r->poi()->bits == 16) != flag; }
    };

    class PathLenF : public TargetFilter::Func {
	int sign;
    public:
	PathLenF(int _sign) : sign(_sign) {}
	bool operator()(const Target * const &r) const {
	    if (r->min_fwd_path_len == 0 || r->min_rev_path_len == 0) return false;
	    switch (sign) {
		case -1: return (r->min_fwd_path_len < r->min_rev_path_len);
		case  0: return (r->min_fwd_path_len == r->min_rev_path_len);
		case  1: return (r->min_fwd_path_len > r->min_rev_path_len);
	    }
	    return false;
	}
    };

    class PathLenDiffF : public TargetFilter::Func {
	int bigdiff;
    public:
	PathLenDiffF(int _bigdiff) : bigdiff(_bigdiff) {}
	bool operator()(const Target * const &r) const {
	    if (r->min_fwd_path_len == 0 || r->min_rev_path_len == 0) return false;
	    return abs(r->min_fwd_path_len - r->min_rev_path_len) >= bigdiff;
	}
    };

    class DescF : public TargetFilter::Func {
	std::string desc;
    public:
	DescF(std::string _desc) : desc(_desc) {}
	bool operator()(const Target * const &r) const
	    { return r->desc.str().find(desc) != std::string::npos; }
    };

    // Negate any filter function
    template<class TF> class NotF : public TargetFilter::Func {
	const TF tf;
    public:
	explicit NotF(const TF &_tf) : tf(_tf) {}
	bool operator()(const Target * const &r) const {
	    return !tf(r);
	}
    };

    // convenience function lets us write "Not(Foo(x))" instead of "new NotF<Foo>(Foo(x))"
    template<class TF> static inline NotF<TF> *Not(const TF &_tf) {
	return new NotF<TF>(_tf);
    }
}

struct TargetFilterVector : public std::vector<TargetFilter> {
    void append(const std::string &label, int status,
	TargetFilter::Func *f1 = new TargetFilter::True(),
	TargetFilter::Func *f2 = new TargetFilter::True(),
	TargetFilter::Func *f3 = new TargetFilter::True());
    void init();
};

extern TargetFilterVector targetfilters;

class TargetStat {
public:
    static unsigned max_examples;
    uint64_t count;
    std::vector<const Target*> examples;

    TargetStat() : count(0), examples() { }
    
    void insert(const Target * const r)
    {
	// insert r into examples
	if (examples.size() < max_examples) {
	    examples.push_back(r);
	    r->refcount++;
	}
    }
};

class Comparison {
public:

    static double oob_frac_allowed;
    static double oob_converge_allowed;
    static double oob_inconclusive_allowed;
    static double oob_max;

    const Target *r1;
    const Target *r2;
    mutable unsigned refcount;
    ResResult result;
    ProbeBitset r1_mbt_fail;
    ProbeBitset r2_mbt_fail;
    ProbeBitset r1_mbt_pass;
    ProbeBitset r2_mbt_pass;
    double overlap;
    int n_interp;
    int n_extrap;
    int n_bounded; // number of points that passed upper & lower bounds checks
#ifdef ENABLE_FEATURES
    bool featurediff;
#endif
    int oob;
    int n_xdisco;
    int n_xdisco_oob;
    int old_oob;
    double first_pass, last_pass;
    double first_fail, last_fail;
    double idist; // avg distance using interpolation only
    double edist; // avg distance using interpolation and extrapolation

    Comparison() : refcount(0), r1_mbt_fail(), r2_mbt_fail(), r1_mbt_pass(), r2_mbt_pass() {}

    static bool parse_oob_allowed(const char *str);

    void init(const Target *_r1, const Target *_r2) {
	r1 = _r1;
	r2 = _r2;
	result = RESULT_UNKNOWN;
	r1_mbt_fail.clear();
	r2_mbt_fail.clear();
	r1_mbt_pass.clear();
	r2_mbt_pass.clear();
	n_interp = 0;
	n_extrap = 0;
	n_bounded = 0;
	overlap = 0;
#ifdef ENABLE_FEATURES
	featurediff = false;
#endif
	oob = 0;
	n_xdisco = 0;
	n_xdisco_oob = 0;
	old_oob = 0;
	first_pass = first_fail = DBL_MAX;
	last_pass = last_fail = 0;
	idist = 32767;
	edist = 32767;
    }
    void compare(Target * const &_r1, Target * const &_r2, bool optimize = true);

private:
    bool interpolation_check(
	const TimeRangeVector *&pt, const IpidVector *&pid,
	unsigned &pi, const big_ipid_t &pwrapsize,
	const TimeRangeVector *&st, const IpidVector *&sid,
	const unsigned &si, const big_ipid_t &swrapsize,
	const double &sv, const ProbeBitset &sdiscon,
	const bool &optimize, double &dist, const double &time_fudge);
    void gap_oob_check(const Target *left, const Target *right,
	const bool &optimize, const double &time_fudge);
    void overlap_oob_check(const bool &swapped,
	const bool &optimize, const double &time_fudge);
};

class ComparisonFilter {
public:
    struct Func : public std::unary_function<Comparison*, bool> {
	virtual bool operator()(const Comparison * const &comp) const = 0;
    };
    std::string label;
    int status;
    const Func *func1;
    const Func *func2;
    const Func *func3;

    struct True : public Func {
	bool operator()(const Comparison * const &comp) const { return true; }
    };

    explicit ComparisonFilter(const std::string _label, int _status,
	const Func *_func1 = new True(), const Func *_func2 = new True(),
	const Func *_func3 = new True()) :
	label(_label), status(_status), func1(_func1), func2(_func2), func3(_func3)
	{ }

    bool operator()(const Comparison * const &comp) const {
	return (*func1)(comp) && (*func2)(comp) && (*func3)(comp);
    }
};

namespace ComparisonFilters {

    class ResultF : public ComparisonFilter::Func {
	const ResResult result;
    public:
	explicit ResultF(ResResult _result) : result(_result) {}
	bool operator()(const Comparison * const &comp) const {
	    return (comp->result == result);
	}
    };

    class ResultTypeF : public ComparisonFilter::Func {
	const ResultType restype;
    public:
	explicit ResultTypeF(ResultType _restype) : restype(_restype) {}
	bool operator()(const Comparison * const &comp) const {
	    return (restype == result_types[comp->result]);
	}
    };

#ifdef ENABLE_FEATURES
    struct FeaturediffF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->featurediff);
	}
    };
#endif

    struct OverlapF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->overlap > 0);
	}
    };

    class Velocity2F : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit Velocity2F(double _min, double _max = DBL_MAX) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->r2->series_velocity > min && comp->r2->series_velocity <= max;
	}
    };

    class IDistF : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit IDistF(double _min, double _max = DBL_MAX) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->n_interp > 0 && comp->idist >= min && comp->idist <= max;
	}
    };

    class EDistF : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit EDistF(double _min, double _max = DBL_MAX) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->n_interp + comp->n_extrap > 0 && comp->edist >= min && comp->edist <= max;
	}
    };

    class OOBfracF : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit OOBfracF(double _min, double _max = 1.0) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->n_bounded > 0 &&
		comp->oob >= min * comp->n_bounded &&
		comp->oob <= max * comp->n_bounded;
	}
    };

    class OOBcountF : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit OOBcountF(int _min, int _max = INT_MAX) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->n_bounded > 0 && comp->oob >= min && comp->oob <= max;
	}
    };

    class XDiscoOOBcountF : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit XDiscoOOBcountF(int _min, int _max = INT_MAX) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->n_bounded > 0 && comp->n_xdisco_oob >= min && comp->n_xdisco_oob <= max;
	}
    };

    class MBTcountF : public ComparisonFilter::Func {
	const double min, max;
    public:
	explicit MBTcountF(int _min, int _max = INT_MAX) : min(_min), max(_max) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->n_bounded >= min && comp->n_bounded <= max;
	}
    };

    class MethF : public ComparisonFilter::Func {
	const int meth;
    public:
	explicit MethF(int _meth) : meth(_meth) {}
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->meth == meth || comp->r2->meth == meth);
	}
    };

    class Meth2F : public ComparisonFilter::Func {
	const int meth1, meth2;
    public:
	explicit Meth2F(int _meth1, int _meth2) : meth1(_meth1), meth2(_meth2) {}
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->meth == meth1 && comp->r2->meth == meth2) ||
		(comp->r1->meth == meth2 && comp->r2->meth == meth1);
	}
    };

    class SameSourceF : public ComparisonFilter::Func {
    public:
	bool operator()(const Comparison * const &comp) const {
	    // note: src_addr==0 means target was probed from multiple sources
	    return comp->r1->src_addr != 0 &&
		comp->r1->src_addr == comp->r2->src_addr;
	}
    };

    struct InitTTLF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->init_ttl == comp->r2->init_ttl);
	}
    };

    struct FwdPathLenOkF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->min_fwd_path_len > 0 && comp->r2->min_fwd_path_len > 0);
	}
    };

    struct FwdPathLenEqualF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->min_fwd_path_len == comp->r2->min_fwd_path_len);
	}
    };

    struct RevPathLenOkF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->min_rev_path_len > 0 && comp->r2->min_rev_path_len > 0);
	}
    };

    struct RevPathLenEqualF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->min_rev_path_len == comp->r2->min_rev_path_len);
	}
    };

    struct OrderF : public ComparisonFilter::Func {
	bool operator()(const Comparison * const &comp) const {
	    return (comp->r1->order == comp->r2->order);
	}
    };

    class FailOobF : public ComparisonFilter::Func {
	const bool oldflag, newflag;
    public:
	explicit FailOobF(const bool &_old, const bool _new) : oldflag(_old), newflag(_new) {}
	bool operator()(const Comparison * const &comp) const {
	    return comp->result >= RESULT_OUT_OF_BOUNDS &&
		((comp->old_oob > 0) == oldflag) && ((comp->oob > 0) == newflag);
	}
    };

    // Negate any filter function
    template<class CF> class NotF : public ComparisonFilter::Func {
	const CF cf;
    public:
	explicit NotF(const CF &_cf) : cf(_cf) {}
	bool operator()(const Comparison * const &comp) const {
	    return !cf(comp);
	}
    };

    // convenience function lets us write "Not(Foo(x))" instead of "new NotF<Foo>(Foo(x))"
    template<class CF> static inline NotF<CF> *Not(const CF &_cf) {
	return new NotF<CF>(_cf);
    }
}

struct CompFilterVector : public std::vector<ComparisonFilter> {
    void operator()(const std::string &label, int status,
	ComparisonFilter::Func *f1 = new ComparisonFilter::True(),
	ComparisonFilter::Func *f2 = new ComparisonFilter::True(),
	ComparisonFilter::Func *f3 = new ComparisonFilter::True());

    void operator()(const ResResult &result);
};

class ComparisonStat {
public:
    static unsigned max_examples;
    uint64_t count;
    std::vector<const Comparison*> examples;

    ComparisonStat() : count(0), examples() { }
    
    void insert(const Comparison * const comp)
    {
	// insert comp into examples, sorted by descending v2
	if (examples.size() < max_examples ||
	    comp->r2->series_velocity > examples.back()->r2->series_velocity)
	{
	    if (examples.size() == max_examples) {
		const Comparison *worst = examples.back();
		examples.pop_back();
		--worst->refcount;
		--worst->r1->refcount;
		--worst->r2->refcount;
		if (!worst->refcount) {
		    memlock.lock();
		    delete worst;
		    memlock.unlock();
		}
	    }
	    std::vector<const Comparison*>::iterator it;
	    for (it = examples.begin(); it != examples.end(); it++) {
		if (comp->r2->series_velocity > (*it)->r2->series_velocity)
		    break;
	    }
	    examples.insert(it, comp);
	    comp->refcount++;
	    comp->r1->refcount++;
	    comp->r2->refcount++;
	}
    }
};

inline void target_summary_header(std::ofstream &summary, const std::string &cmdline)
{
    using namespace std;
    summary << "# " << cmdline << "\n";
    summary << "# " <<
	left << setw(13) << "target_addr" << right <<
	setw(7) << "method" <<
	setw(6) << "tries" <<
	setw(6) << "probe" <<
	setw(6) << "expct" <<
	setw(6) << "unexp" <<
	setw(6) << "zero" <<
	setw(6) << "rflct" <<
	setw(6) << "rpeat" <<
	setw(6) << "ndgen" <<
	setw(12) << "velocity" <<
	setw(4) << "ttl" <<
	setw(4) << "exp";
    if (ttl_flag) summary << setw(5) << "ittl";
    summary <<
	setw(3) << "LE" <<
	" classification" <<
	"\n";
}

inline void target_summary_row(std::ofstream &summary, const Target *target,
    bool preferred = false)
{
    using namespace std;
    summary <<
	left << setw(15) << target->addr << right <<
	' ' << setw(5) << methname[target->meth] << (preferred ? '*' : ' ') <<
	' ' << setw(5) << target->tries <<
	' ' << setw(5) << target->probes <<
	' ' << setw(5) << target->expected <<
	' ' << setw(5) << target->unexpected <<
	' ' << setw(5) << target->zero <<
	' ' << setw(5) << target->reflect <<
	' ' << setw(5) << target->repeat <<
	' ' << setw(5) << target->nondegenerate <<
	' ' << setw(11) << fmtfixed(target->series_velocity, 6) <<
	' ' << setw(3) << int(target->ttl) <<
	' ' << setw(3) << int(target->ttl_expansions);
    if (ttl_flag) summary << ' ' << setw(4) << int(target->init_ttl);
    summary <<
	' ' << setw(2) << int(target->order) <<
	" " << class_label[target->classification] <<
	"; " << target->desc.str() <<
	"\n";
}

void print_bits_cell(std::ofstream &out, const Target *const r);

void print_classification_cell(std::ofstream &html, const TargetStat *targetstat,
    uint64_t total_targets, std::string label, const char *gptype, bool gpscript = false);

void print_resolution_cell(std::ofstream &html, const ComparisonStat *compstat,
    uint64_t n_pairs, std::string label, const char *gptype, bool gpscript = false);

void init_ttl_comparison(std::ofstream &html, const MultiMethTargetSet &targets,
    const int *n_targets, const char *gptype, bool gpscript);

void path_length_comparison(std::ofstream &html, const MultiMethTargetSet &targets,
    const char *gptype, bool gpscript);

void print_rtt_distribution(std::ofstream &html, const MultiMethTargetSet &targets,
    const char *gptype, bool gpscript);

class OobDistribution {
    static const long SIZE = 200;
    long data[SIZE];
public:
    OobDistribution() { memset(data, 0, sizeof(data)); }
    void count(const Comparison *comp) {
	if (comp->oob > 0) {
	    long i = (comp->oob < SIZE) ? comp->oob : (SIZE-1);
	    data[i]++;
	}
    }
    void plot(std::ofstream &html, const char *gptype, bool gpscript);
};

void print_monitor_stats(std::ofstream &html, MonitorSet &monitors,
    const int *n_targets);

extern std::set<std::string> plotted;

class Gnuplotter {
    static int raw_pointtype[];
    static int cooked_pointtype[];
    static int raw_linetype[];
    static int cooked_linetype[];

    bool ispipe;
public:
    FILE *f;
    static const char *PNG;
    static const char *EPS;

    Gnuplotter() : ispipe(false), f(0) { }
    ~Gnuplotter() { if (f) { close(); if (ispipe) { pclose(f); f = 0; } } }

    void open(const std::string &filename, const std::string &title,
	const char *gptype, bool gpscript);

    void close();

    static double pointsize(int i, int n) { return pow(1.23, n-i-2); }

    void plot_cdf(int nsets, std::vector<double> *data,
	const char * const *labels, const char *with=0);

    void plot_binned_cdf(bool logbins, int nsets, int nbins, uint64_t **data,
	double offset, double binsize, const char * const *labels,
	bool logy = false, bool absolute = false);

    void plot_binned_histogram(int nsets, int nbins, uint64_t **data,
	int offset, int binsize, const char * const *labels, const char *with = "lines");

    void plot_id_vs_time_compact(unsigned n, const Target **data,
	const char * const *labels, bool both_orders = false);
};

extern Gnuplotter gp;

} // namespace MidarUtils
