/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <string>
#include <getopt.h>

using namespace std;

#include "../lib/config.h"
#include "../lib/infile.h"
#include "../lib/MemoryInfo.h"
#define NO_DEBUG_MEMORY 0
MemoryInfo memoryInfo;

#include "MidarUtils.h"
using namespace MidarUtils;

static const char *cvsID = "$Id: estimation-analyze.cc,v 1.107 2016/09/22 19:11:38 kkeys Exp $";
static string cmdline = "";
static const char *argv0;

static const char *htmldir = 0;
static const char *monitorfilename = 0;
static const char *includefilename = 0;
#ifdef ENABLE_FEATURES
static const char *featurefilename = 0;
#endif
static const char *gptype = Gnuplotter::PNG;
static int gpscript = 1;

static ofstream html;
#ifdef ENABLE_FEATURES
static ofstream featurefile;
#endif

static int roundnum = -1;
static int min_round = -1;
static int max_round = -1;
static MultiMethTargetSet unresolvedTargets; // results of estimation probe run
static MultiMethTargetSet targets; // processed results of estimation probe run
static set<ip4addr_t> includeaddrs;
static int n_targets[N_METH];
static int n_probes[N_METH];
static int n_responses[N_METH];
// static int n_extra_responses = 0;
static double min_round_time = DBL_MAX;
static double max_round_time = 0;
static double tot_round_time = 0;
static int n_window_rounds = 0;
static OobDistribution oob_distribution;
static int tries_per_target = -1;
static uint64_t *wrapdata[N_METH+1];

static MonitorSet monitors;
static vector<std::string> errormsgs;

static struct option longopts[] = {
    { "htmldir",       required_argument, 0,         'H' },
    { "include",       required_argument, 0,         'I' },
    { "monitors",      required_argument, 0,         'M' },
#ifdef ENABLE_FEATURES
    { "features",      required_argument, 0,         'f' },
#endif
    TTL_LONGOPT("no-ttl",     &ttl_flag,    0),
    TTL_LONGOPT("ttl",        &ttl_flag,    1),
    DIST_LONGOPT("dist-none", &dist_option, DIST_NONE),
    DIST_LONGOPT("dist-all",  &dist_option, DIST_ALL),
    { "udp-diff",      no_argument,       &allow_udp_diff_addr, 1 },
    { "no-udp-diff",   no_argument,       &allow_udp_diff_addr, 0 },
    { "oob-allowed",   required_argument, 0,         'a' },
    { "overlap",       required_argument, 0,         'o' },
    { "gptype",        required_argument, 0,         't' },
    { "gpscript",      no_argument,       &gpscript, 1 },
    { "no-gpscript",   no_argument,       &gpscript, 0 },
    { 0, 0, 0, 0 }
};

void usageExit() {
    cerr << "Usage: " << argv0 << " [<options>] <estimation_probe_result_file>\n";
    cerr << "--htmldir <dir>         write analysis graphs and webpage to directory <dir>\n";
    cerr << "--gptype {png|eps}      format of graphs (default: png)\n";
    cerr << "--no-gpscript           write gnuplot graphs instead of scripts\n";
    cerr << "--include <file>        include only addresses listed in <file>\n";
    cerr << "--monitors <file>       load monitor names and addrs from <file>\n";
#ifdef ENABLE_FEATURES
    cerr << "--features <file>       write target features to <file>\n";
#endif
#if !ENABLE_TTL_OPTION
    cerr << "--ttl option requires recompiling with ENABLE_TTL_OPTION\n";
#endif
    cerr << "--(no-)ttl              (do not) use the DIFF_TTL test (default: off)\n";
    cerr << "--(no-)udp-diff         (do not) allow UDP response from different address\n";
    cerr << "                        (default: off)\n";
#if !ENABLE_DIST_OPTION
    cerr << "--dist- options require recompiling with ENABLE_DIST_OPTION\n";
#endif
    cerr << "--dist-none             do not use DIST test (default)\n";
    cerr << "--dist-all              use DIST test on all points\n";
    cerr << "--overlap <n>           require at least <n> overlapping points (default: 0)\n";
    cerr << "--oob-allowed <a>[,<b>[,<c>]]\n";
    cerr << "                        fraction of MBT point failures allowed for series to...\n";
    cerr << "                        <a>:  pass (default: 0.0)\n";
    cerr << "                        <b>:  pass, IF first & last points pass (default: <a>)\n";
    cerr << "                        <c>:  be inconclusive (default: <b>)\n";
    cerr << "Note: --htmldir is required.\n";
    exit(1);
}

void loadIncludes(const char *filename)
{
    char buf[8192];
    InFile in(filename);
    cerr << "# loading include addresses from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	char *p = strchr(buf, '\n');
	if (p) *p = '\0';
	ip4addr_t addr(buf);
	includeaddrs.insert(addr);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << includeaddrs.size() << " addresses\n";
    memoryInfo.print("loaded includes");
}

void loadResponses(const char *filename)
{
    ResponseData rd;
    char buf[8192];
    InFile in(filename);
    cerr << "# loading data from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	double t;
	
	if (sscanf(buf, "## started window %*d round %d:", &roundnum) == 1) {
	    if (min_round < 0) min_round = roundnum;
	    if (roundnum > max_round) max_round = roundnum;
	    continue;
	} else if (sscanf(buf, "## finished window %*d round %*d in %lf secs", &t) == 1) {
	    if (t < min_round_time) min_round_time = t;
	    if (t > max_round_time) max_round_time = t;
	    tot_round_time += t;
	    n_window_rounds++;
	    continue;
	} else if (buf[0] == '#' || buf[0] == '\n') {
	    continue; // comment or empty
	}

	if (!strchr("TURHN", buf[0])) continue;

	rd.parse(buf);
	if (rd.result == 'S') continue; // 'S' record is informational, not a real probe

	if (includefilename) {
	    if (includeaddrs.find(rd.target_addr) == includeaddrs.end())
		continue; // skip
	}

	int meth;
	parse_meth(rd.method, meth, rd.ttl_expansion);

	Monitor *monitor = monitors.countProbeFrom(rd.src_addr, meth);

	MultiMethTarget *mttarget = unresolvedTargets.findOrInsert(rd.target_addr);

	if (!mttarget->target[meth]) {
	    mttarget->target[meth] = new Target(rd.target_addr);
	    mttarget->target[meth]->meth = meth;
	    n_targets[meth]++;
	}

	if (rd.ttl_expansion == 0)
	    mttarget->target[meth]->tries++;
	mttarget->target[meth]->probes++;
	n_probes[meth]++;
	if (!strchr("TUH", rd.result)) continue;
	n_responses[meth]++;
	monitor->n_responses[meth]++;
	monitor->reply_ttl_histogram[meth][rd.reply_ttl]++;
	mttarget->target[meth]->processResponse(&rd);
	if (mttarget->target[meth]->probes == 1) {
	    monitor->time[meth].push_back((rd.resp_time + rd.probe_time) / 2);
	    monitor->id[meth].push_back(rd.reply_ipid);
	}

#ifdef ENABLE_FEATURES
	if (meth == METH_ICMP && rd.result == 'T') {
	    mttarget->feature[ICMP_ECHO_CODE] = rd.resp_icmp_code;
	}
#endif
      } catch (const std::runtime_error &e) {
	if (!in.gets(buf, sizeof(buf))) { // error was in last line
	    std::string msg(InFile::Error(in, e).what());
	    msg += " (file appears to be truncated; ignoring error)";
	    errormsgs.push_back(msg);
	    cerr << msg << endl;
	    break;
	} else {
	    throw InFile::Error(in, e);
	}
      }
    }
    in.close();

    memoryInfo.print("loaded responses");
}

#ifdef ENABLE_FEATURES
void feature_table(int fpid, int n, const char * const *labels)
{
    int unknown = 0, total = 0;
    int rows = 1;
    int counter[n];
    memset(counter, 0, n * sizeof(*counter));

    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	if ((*it)->feature[fpid] == -1) {
	    unknown++;
	} else {
	    if (!counter[(*it)->feature[fpid]])
		rows++;
	    counter[(*it)->feature[fpid]]++;
	}
	total++;
    }

    html << "<tr><td rowspan=" << rows << ">" << feature_label[fpid] << "\n";
    html << "    <td>unknown <td>" << htmlnum(unknown) << "<td>" <<
	fmtpercent(unknown, total) << "</tr>\n";
    for (int i = 0; i < n; i++) {
	if (!counter[i]) continue;
	html << "<tr><td>";
	if (labels) {
	    html << labels[i];
	} else {
	    html << i;
	}
	html << "<td>" << htmlnum(counter[i]) << "<td>" <<
	    fmtpercent(counter[i], total) << "</tr>\n";
    }
    html << "\n";
}

void feature_table(int fpid1, int fpid2)
{
    static const char * const labels[2] = { "same", "different" };
    static const int n = 2;
    int unknown = 0, total = 0;
    int rows = 1;
    int counter[n];
    memset(counter, 0, n * sizeof(*counter));

    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	if ((*it)->feature[fpid1] == -1 || (*it)->feature[fpid2] == -1) {
	    unknown++;
	} else {
	    int val = int((*it)->feature[fpid1] != (*it)->feature[fpid2]);
	    if (!counter[val])
		rows++;
	    counter[val]++;
	}
	total++;
    }

    html << "<tr><td rowspan=" << rows << ">" << feature_label[fpid1] << " vs. " << feature_label[fpid2] << "\n";
    html << "    <td>unknown <td>" << htmlnum(unknown) << "<td>" <<
	fmtpercent(unknown, total) << "</tr>\n";
    for (int i = 0; i < n; i++) {
	if (!counter[i]) continue;
	html << "<tr><td>";
	html << labels[i];
	html << "<td>" << htmlnum(counter[i]) << "<td>" <<
	    fmtpercent(counter[i], total) << "</tr>\n";
    }
    html << "\n";
}
#endif

void print_classification_row(ofstream &out, const char *label, int *counter)
{
    int total = 0;
    out << "<tr><td>" << label;
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	out << "<td colspan=2>" << htmlnum(counter[meth]) << " ";
	total += counter[meth];
    }
    out << "<td colspan=2>" << htmlnum(total) << "\n";
}

CompFilterVector compfilters;
Array2d<ComparisonStat> *compstats = 0;
uint64_t n_pairs[choose2count(N_METH)];

void initResolveStats()
{
#if 0
    char printbuf[1024];
    const double distdata_min = 0.1;
    const double distdata_offset = log(distdata_min);
    const double distdata_binsize = 0.02;
    const int nbins = int((log(32768) - distdata_offset) / distdata_binsize);
    int *distdata[4];
    for (int i = 0; i < 4; i++) {
	distdata[i] = new int[nbins];
	memset(distdata[i], 0, sizeof(int)*nbins);
    }
#endif

    using namespace ComparisonFilters;

    compfilters(RESULT_SWAPPED);
    // DIFF_TTL is impossible across methods
    if (bits_flag) {
	compfilters(RESULT_BITS);
    }
    if (vsim_option > VSIM_NEI) {
	compfilters(RESULT_VSIM);
    }
#if 0
    compfilters(RESULT_DIVERGE);
#endif
    compfilters(RESULT_OUT_OF_BOUNDS);
    compfilters("total UNSHARED", STATUS_BAD,
	new ResultTypeF(RESTYPE_UNSHARED));

    compfilters(RESULT_BIG_GAP);
    if (vsim_option == VSIM_NEI)
	compfilters(RESULT_BIG_GAP_VSIM);
    compfilters(RESULT_NO_OVERLAP_OOB);
    if (dist_option >= DIST_INTERP) {
	compfilters(RESULT_IDIST);
    }
    if (dist_option >= DIST_ALL) {
	compfilters(RESULT_EDIST);
    }
    compfilters(RESULT_NO_OVERLAP);
    if (vsim_option == VSIM_NEI)
	compfilters(RESULT_NO_OVERLAP_VSIM);
    if (min_overlap_option > 0) {
	compfilters(RESULT_POOR_OVERLAP);
	if (vsim_option == VSIM_NEI)
	    compfilters(RESULT_POOR_OVERLAP_VSIM);
    }
    compfilters(RESULT_INCONCLUSIVE);
    compfilters("total NOT_ENOUGH_INFO", STATUS_MEDIUM,
	new ResultTypeF(RESTYPE_NOT_ENOUGH_INFO));

    compfilters(RESULT_SHARED);
    compfilters("total pairs compared", 0);

    if (dist_option) {
	compfilters("SHARED-dist-0-500", 0,
	    new ResultTypeF(RESTYPE_SHARED), new EDistF(0, 500));
	compfilters("SHARED-dist-gt-500", 0,
	    new ResultTypeF(RESTYPE_SHARED), new EDistF(500));
	compfilters("UNSHARED-dist-0-500", 0,
	    new ResultTypeF(RESTYPE_UNSHARED), new EDistF(0, 500));
	compfilters("UNSHARED-dist-gt-500", 0,
	    new ResultTypeF(RESTYPE_UNSHARED), new EDistF(500));
    }

    compfilters("OUT_OF_BOUNDS with < 20% OOB", 0,
	new ResultF(RESULT_OUT_OF_BOUNDS), new OOBfracF(0,0.2));

    compfilters("SHARED with xdisco_oob", 0,
	new ResultF(RESULT_SHARED), new XDiscoOOBcountF(1));

    compfilters("SHARED with OOB >= 1", 0,
	new ResultF(RESULT_SHARED), new OOBcountF(1));

    if (ttl_flag) {
	compfilters("same init TTL, shared", 0,
	    new ResultF(RESULT_SHARED), new InitTTLF());
	compfilters("different init TTL, shared", 0,
	    new ResultF(RESULT_SHARED), Not(InitTTLF()));
    }

    compfilters("same fwd path length, shared", 0,
	new ResultF(RESULT_SHARED), new FwdPathLenOkF(), new FwdPathLenEqualF());
    compfilters("different fwd path length, shared", 0,
	new ResultF(RESULT_SHARED), new FwdPathLenOkF(), Not(FwdPathLenEqualF()));
    compfilters("unknown fwd path length, shared", 0,
	new ResultF(RESULT_SHARED), Not(FwdPathLenOkF()));

    if (ttl_flag) {
	compfilters("same rev path length, shared", 0,
	    new ResultF(RESULT_SHARED), new RevPathLenOkF(), new RevPathLenEqualF());
	compfilters("different rev path length, shared", 0,
	    new ResultF(RESULT_SHARED), new RevPathLenOkF(), Not(RevPathLenEqualF()));
	compfilters("unknown rev path length, shared", 0,
	    new ResultF(RESULT_SHARED), Not(RevPathLenOkF()));
    }

    compfilters("same byte order, shared", 0,
	new ResultF(RESULT_SHARED), new OrderF());
    compfilters("different byte order, shared", 0,
	new ResultF(RESULT_SHARED), Not(OrderF()));

#if 0
    compfilters("oob old=fail new=fail", 0,
	new FailOobF(true, true));
    compfilters("oob old=pass new=pass", 0,
	new FailOobF(false, false));
    compfilters("oob old=pass new=fail", 0,
	new FailOobF(false, true));
    compfilters("oob old=fail new=pass", 0,
	new FailOobF(true, false));
#endif

    const int rows = compfilters.size();
    const int cols = choose2count(N_METH);
    compstats = new Array2d<ComparisonStat>(rows, cols);
    memset(n_pairs, 0, sizeof(n_pairs));
}

void resolveTargets()
{
    if (htmldir) {
	Comparison *comp = new Comparison();
	for (MultiMethTargetSet::iterator it = unresolvedTargets.begin(); it != unresolvedTargets.end(); it++) {
	    for (int t1 = 0; t1 < N_METH - 1; t1++) {
		Target *r1 = (*it)->target[t1];
		for (int t2 = t1+1; t2 < N_METH; t2++) {
		    Target *r2 = (*it)->target[t2];
#ifdef ENABLE_FEATURES
		    int result = RESULT_UNKNOWN;
#endif
		    if (r1 && (r1->flags & MONOTONIC) &&
			r2 && (r2->flags & MONOTONIC))
		    {
			if (comp->refcount > 0) // old comp is still in use
			    comp = new Comparison();
			    // else, we can recycle the old comp
			comp->compare(r1, r2, false);

			if (comp->result == RESULT_SHARED)
			    (*it)->shared.set(t1, t2);

			oob_distribution.count(comp);

#if 0
			double v1 = r1->series_velocity;
			double v2 = r2->series_velocity;
			outfile << "# i=" << i << ", " << r1->addr << " v1=" << setprecision(15) << v1 << "\n";
			sprintf(printbuf, "%-15s %-15s %-6s %9.4f %-15s %-15s %-6s %9.4f %7.5f ",
			    string(r1->addr).c_str(), string(r1->src_addr).c_str(),
			    methname[r1->meth], v1,
			    string(r2->addr).c_str(), string(r2->src_addr).c_str(),
			    methname[r2->meth], v2,
			    v2 / v1);
			outfile << printbuf;

			sprintf(printbuf, "%7s %3d %3d %10.3f %3d %12.5f %12.5f\n",
			    result_label[comp->result],
			    comp->n_interp,
			    comp->n_extrap,
			    comp->overlap,
			    comp->oob,
			    comp->idist,
			    comp->edist);
			outfile << printbuf;
#endif

			int col = choose2index(N_METH, r1->meth, r2->meth);

			for (size_t row = 0; row < compfilters.size(); row++) {
			    if (compfilters[row](comp)) {
				(*compstats)(row, col).insert(comp);
				(*compstats)(row, col).count++;
			    }
			}
			n_pairs[col]++;

#if 0
			int bin = comp->edist < distdata_min ? 0 :
			    int((log(comp->edist) - distdata_offset) / distdata_binsize);
			if (bin < 0) bin = 0;
			distdata[int(comp->result == RESULT_SHARED) + 2 * col][bin]++;
#endif
#ifdef ENABLE_FEATURES
			result = comp->result;
#endif
		    }

#ifdef ENABLE_FEATURES
		    // If the higher-reliability method has a monotonic
		    // counter but the lower-reliability method does not
		    // share it, we shouldn't trust the features of the
		    // lower-reliablity method.
		    if (r1 && (r1->flags & MONOTONIC) && result != RESULT_SHARED) {
			if (t2 == METH_ICMP) {
			    (*it)->feature[ICMP_INIT_TTL] = -1;
			    (*it)->feature[ICMP_IPID_CLASS] = -1;
			    (*it)->feature[ICMP_IPID_ORDER] = -1;
			    (*it)->feature[ICMP_ECHO_CODE] = -1;
			} else if (t2 == METH_INDIR) {
			    (*it)->feature[INDIR_INIT_TTL] = -1;
			    (*it)->feature[INDIR_IPID_CLASS] = -1;
			    (*it)->feature[INDIR_IPID_ORDER] = -1;
			}
		    }
#endif
		}
	    }
	}
	if (comp->refcount == 0) delete comp;
    }
    memoryInfo.print("resolved");
}

void printResolveStats()
{
#if 0
    // plot dist
    if (htmldir && dist_option) {
	html << "\n<h2>Average Distance</h2>\n\n";

	gp.open("dist-cdf", "distribution of average distances", gptype, gpscript);
	const char *known_labels[] = {"rejected non known alias", "possible non known alias", "rejected known alias", "possible known alias"};
	const char *unknown_labels[] = {"rejected", "possible"};
	const char **labels = knownfilename ? known_labels : unknown_labels;
	html << "<br><img src='dist-cdf.png'>\n\n";
	fprintf(gp.f, "set logscale x\nset mxtics 10\n");
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'distance'\n");
	fprintf(gp.f, "set key top left\n");
	gp.plot_logbinned_cdf(4, nbins, distdata, distdata_offset, distdata_binsize, labels);
	for (int i = 0; i < 4; i++) {
	    delete[] distdata[i];
	}
	gp.close();
    }
#endif

#ifdef ENABLE_FEATURES
    if (htmldir) {
	html << "\n<h2>Features</h2>\n\n";
	html << "<table style='border-collapse: collapse'>\n";
	html << "<tr><th>feature <th>value <th colspan=2>count </tr>\n";
        feature_table(TCP_INIT_TTL, 256, 0);
	feature_table(ICMP_INIT_TTL, 256, 0);
	feature_table(INDIR_INIT_TTL, 256, 0);
        feature_table(TCP_INIT_TTL, ICMP_INIT_TTL);
	feature_table(TCP_INIT_TTL, INDIR_INIT_TTL);
	feature_table(ICMP_INIT_TTL, INDIR_INIT_TTL);
	feature_table(TCP_IPID_CLASS, N_IPID_CLASS, IPID_CLASS_label);
	feature_table(ICMP_IPID_CLASS, N_IPID_CLASS, IPID_CLASS_label);
	feature_table(INDIR_IPID_CLASS, N_IPID_CLASS, IPID_CLASS_label);
	feature_table(TCP_IPID_ORDER, 2, order_label);
	feature_table(ICMP_IPID_ORDER, 2, order_label);
	feature_table(INDIR_IPID_ORDER, 2, order_label);
	feature_table(ICMP_ECHO_CODE, 256, 0);
	html << "</table>\n";
    }
#endif

    html << flush;
    if (htmldir) {
	html << "\n<h2>Resolution comparisons</h2>\n\n";
	html << "<p>Comparing IP ID time series collected with different " <<
	    "probing methods from the same address identifies addresses " <<
	    "that share an ID counter across methods.</p>\n";
	html << "<table style='border-collapse: collapse'>\n";

	html << "<ul>\n";
	if (vsim_option) {
	    html << "<li>VSIM: (v<sub>2</sub> + e) / (v<sub>1</sub> + e) < r, where " <<
		"e = " << vsim_fudge << " and " <<
		"r = " << setprecision(6) << vsim_thresh << "\n";
	}
	if (min_overlap_option) {
	    html << "<li>POOR_OVERLAP: fewer than " << min_overlap_option <<
		" points overlapped\n";
	}
	if (Comparison::oob_frac_allowed) {
	    html << "<li>oob-allowed: " <<
		Comparison::oob_frac_allowed << "," <<
		Comparison::oob_converge_allowed << "," <<
		Comparison::oob_inconclusive_allowed << "\n";
	}
	html << "</ul>\n";

	html << "<tr><th>category";
	for (int t1 = 0; t1 < N_METH - 1; t1++) {
	    if (!n_targets[t1]) continue;
	    for (int t2 = t1+1; t2 < N_METH; t2++) {
		if (!n_targets[t2]) continue;
		html << "<th colspan=2>" << methname[t1] << ":" << methname[t2];
	    }
	}
	html << "\n";

	for (size_t row = 0; row < compfilters.size(); row++) {
	    // cerr << "# print_row: " << compfilters[row].label << "\n";
	    html << "<tr " <<
		status_style[compfilters[row].status] <<
		"><td class='label'>" << compfilters[row].label << "\n";

	    for (int t1 = 0; t1 < N_METH - 1; t1++) {
		if (!n_targets[t1]) continue;
		for (int t2 = t1+1; t2 < N_METH; t2++) {
		    if (!n_targets[t2]) continue;
		    int col = choose2index(N_METH, t1, t2);
		    string label(compfilters[row].label);
		    label += "-";
		    label += methname[t1];
		    label += "-";
		    label += methname[t2];
		    print_resolution_cell(html, &(*compstats)(row, col),
			n_pairs[col], label, gptype, gpscript);
		}
	    }
	}

	html << "</table>\n";
    }
    memoryInfo.print("resolve stats");
}

struct MethShareData {
    // vector<MultiMethTarget*> examples;
    uint64_t count;
    MethShareData() : /* examples(), */ count() {}
};

static void print_methshare_table(
    MethShareData table[1 << N_METH][1 << choose2count(N_METH)],
    int i, int meth, const string &label)
{
    bitset<N_METH> meths(i);
    MethShareData subtable[1 << choose2count(N_METH)];

    uint64_t total = 0;
    int jmask = 0;
    for (int t1 = 0; t1 < N_METH - 1; t1++) {
	for (int t2 = t1+1; t2 < N_METH; t2++) {
	    if (!meths.test(t1) || !meths.test(t2)) continue;
	    jmask |= (1 << choose2index(N_METH, t1, t2));
	}
    }
    for (int j = 0; j < (1 << choose2count(N_METH)); j++) {
	for (int ii = 0; ii < (1 << N_METH); ii++) {
	    if ((ii & i) != i) continue;
	    subtable[j & jmask].count += table[ii][j].count;
	    total += table[ii][j].count;
	}
    }

    html << "<td>\n";
    html << label << "<br>\n";

    html << "<table style='border-collapse: collapse'>\n";
    html << "<tr>\n";
    int columns = 0;
    for (int t1 = 0; t1 < N_METH - 1; t1++) {
	for (int t2 = t1+1; t2 < N_METH; t2++) {
	    if (!meths.test(t1) || !meths.test(t2)) continue;
	    if (meth < 0) {
		html << "<td>" << methname[t1] << "=" << methname[t2] << "\n";
	    } else if (t1 == meth) {
		html << "<td>" << methname[t2] << "\n";
	    } else if (t2 == meth) {
		html << "<td>" << methname[t1] << "\n";
	    } else {
		continue;
	    }
	    columns++;
	}
    }
    html << "<td colspan=2>count\n";
    for (int j = 0; j < (1 << choose2count(N_METH)); j++) {
	if (subtable[j].count == 0) continue;
	MethPairBitset pairs(j);
	html << "<tr>";
	for (int t1 = 0; t1 < N_METH - 1; t1++) {
	    for (int t2 = t1+1; t2 < N_METH; t2++) {
		if (!meths.test(t1) || !meths.test(t2)) continue;
		if (!(meth < 0 || t1 == meth || t2 == meth)) continue;
		html << "<td>" << (pairs.test(t1, t2) ? "shared" : "-");
	    }
	}
	html << "\n";
	html << "<td>" << subtable[j].count <<
	    "<td class='percent'>" << fmtpercent(subtable[j].count, total) << "\n";
    }
    html << "<tr><td colspan=" << columns <<
	">total <td>" << total << "<td class='percent'>\n";
    html << "</table>\n";
}

static void print_methshare_tables()
{
    cerr << "# counter sharing across probe methods\n";
    html << "\n<h2>Counter sharing across probe methods</h2>\n";

    size_t n_targeted_meths = 0;
    for (int t = 0; t < N_METH; t++) {
	if (n_targets[t] == 0) continue;
	n_targeted_meths++;
    }

    if (n_targeted_meths > 2) {
	html << "<table style='border-collapse: collapse'>\n";
	int n_tables = 0;
	for (int t1 = 0; t1 < N_METH - 1; t1++) {
	    if (n_targets[t1] == 0) continue;
	    for (int t2 = t1+1; t2 < N_METH; t2++) {
		if (n_targets[t2] == 0) continue;
		MethShareData table[2][5][N_METH];
		uint64_t total[2][N_METH];
		memset(total, 0, sizeof(total));

		for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
		    if (!(*it)->target[t1] || !((*it)->target[t1]->flags & MONOTONIC)) continue;
		    if (!(*it)->target[t2] || !((*it)->target[t2]->flags & MONOTONIC)) continue;
		    for (int t3 = 0; t3 < N_METH; t3++) {
			if (t3 == t1 || t3 == t2) continue;
			int shared = (*it)->shared.test(t1, t2);
			int type;
			if (!(*it)->target[t3] || !((*it)->target[t3]->flags & MONOTONIC)) {
			    type = 0; // unavailable
			} else if (!(*it)->shared.test(t1, t3) && !(*it)->shared.test(t2, t3)) {
			    type = 1; // t3 shares with neither
			} else if ((*it)->shared.test(t1, t3) && !(*it)->shared.test(t2, t3)) {
			    type = 2; // t3 shares with t1
			} else if (!(*it)->shared.test(t1, t3) && (*it)->shared.test(t2, t3)) {
			    type = 3; // t3 shares with t2
			} else {
			    type = 4; // t3 shares with both
			}
			table[shared][type][t3].count++;
			total[shared][t3]++;
		    }
		}

		if (n_tables++ % 3 == 0) {
		    html << "<tr>";
		}
		html << "<td style='border: 2px solid black'>";

		html << "<table style='border-collapse: collapse'>\n";
		html << "<tr><th>" << methname[t1] << ":" << methname[t2] << "<th>";
		for (int t3 = 0; t3 < N_METH; t3++) {
		    if (t3 == t1 || t3 == t2) continue;
		    if (!n_targets[t3]) continue;
		    html << "<th colspan=2>" << methname[t3];
		}
		html << "\n";

		for (int shared = 0; shared <= 1; shared++) {
		    html << "<tr><th rowspan=6>" << (shared ? "shared" : "not shared") << "\n";
		    for (int type = 0; type < 5; type++) {
			html << (type==0 ? "    " : "<tr>") << "<td>";
			switch (type) {
			case 0: html << "unavailable\n";  break;
			case 1: html << "shares with neither\n";  break;
			case 2: html << "shares with " << methname[t1] << "\n";  break;
			case 3: html << "shares with " << methname[t2] << "\n";  break;
			case 4: html << "shares with both\n";  break;
			}
			for (int t3 = 0; t3 < N_METH; t3++) {
			    if (t3 == t1 || t3 == t2) continue;
			    if (!n_targets[t3]) continue;
			    html << "    <td>" << htmlnum(table[shared][type][t3].count) <<
				"<td class='percent'>" <<
				fmtpercent(table[shared][type][t3].count, total[shared][t3]) <<
				"\n";
			}
		    }
		    html << "<tr><td>total\n";
		    for (int t3 = 0; t3 < N_METH; t3++) {
			if (t3 == t1 || t3 == t2) continue;
			if (!n_targets[t3]) continue;
			html << "    <td>" << htmlnum(total[shared][t3]) <<
			    "<td class='percent'>\n";
		    }
		}
		html << "</table>\n";
	    }
	}
	html << "</table>\n";
    }

    if (n_targeted_meths > 1) {
	// dimension 1:  each bit is a monotonic method
	// dimension 2:  each bit is a pair of methods with shared counters
	MethShareData table[1 << N_METH][1 << choose2count(N_METH)];
	MethShareData table_alt_udp[1 << N_METH][1 << choose2count(N_METH)];
	MethShareData table_alt_tcp[1 << N_METH][1 << choose2count(N_METH)];

	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    bitset<N_METH> meths(0);
	    int n_meths = 0;
	    for (int t = 0; t < N_METH; t++) {
		if (!(*it)->target[t] || !((*it)->target[t]->flags & MONOTONIC)) continue;
		meths.set(t);
		n_meths++;
	    }
	    if (n_meths < 2) continue;
	    table[meths.to_ulong()][(*it)->shared.to_ulong()].count++;
	    if (n_targets[METH_UDP]) {
		if (((*it)->target[METH_UDP]) &&
		    ((*it)->target[METH_UDP]->flags & MONOTONIC) &&
		    (((*it)->target[METH_UDP]->flags & (DIFFADDR|SAMEADDR)) == DIFFADDR))
		{
		    table_alt_udp[meths.to_ulong()][(*it)->shared.to_ulong()].count++;
		}
	    }
	    if (n_targets[METH_TCP]) {
		if (((*it)->target[METH_TCP]) &&
		    ((*it)->target[METH_TCP]->flags & MONOTONIC) &&
		    ((*it)->target[METH_TCP]->responsetypes & (1 << RT_ICMP_PORT_UNREACH)) &&
		    !((*it)->target[METH_TCP]->responsetypes & (1 << RT_TCP_RST)))
		{
		    table_alt_tcp[meths.to_ulong()][(*it)->shared.to_ulong()].count++;
		}
	    }
	}

	for (size_t n_meth = 2; n_meth <= n_targeted_meths; n_meth++) {
	    html << "<h3>Targets that are monotonic for at least " <<
		n_meth << " probe methods</h3>\n";
	    for (int i = 0; i < (1 << N_METH); i++) {
		bitset<N_METH> meths(i);
		if (meths.count() != n_meth) continue;

		for (int t = 0; t < N_METH; t++) {
		    if (meths.test(t) && !n_targets[t])
			goto next_i;
		}

		html << "<h4>Targets that are monotonic for at least: ";
		for (int t = 0; t < N_METH; t++) {
		    if (meths.test(t))
			html << methname[t] << " ";
		}
		html << "</h4>\n";

		html << "<table class='invisible'>\n";

		html << "<tr>\n";
		print_methshare_table(table, i, -1, "all");

		if (meths.test(METH_TCP)) {
		    print_methshare_table(table_alt_tcp, i, METH_TCP,
			"TCP with port unreachable response");
		}

		if (meths.test(METH_UDP)) {
		    print_methshare_table(table_alt_udp, i, METH_UDP,
			"UDP with response from different address");
		}

		html << "</table>\n";
		next_i:;
	    }
	}
    }
}

Array2d<TargetStat> *targetstats = 0;

void initTargetStats()
{
    targetfilters.init();
    const int rows = targetfilters.size();
    const int cols = N_METH;
    targetstats = htmldir ? new Array2d<TargetStat>(rows, cols) : 0;
}

void classifyTargets()
{
    // calculate distribution of responses, unwrap, correlation...
    for (MultiMethTargetSet::iterator it = unresolvedTargets.begin(); it != unresolvedTargets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!(*it)->target[meth]) continue;
	    classify((*it)->target[meth], false);
	    if ((*it)->target[meth]->nondegenerate == tries_per_target) {
		wrapdata[meth][(*it)->target[meth]->wraps]++;
	    }
	    // cerr << "# classify " << (*it)->addr << ": " << (*it)->target[meth]->desc.str() << "\n";
	}
#ifdef ENABLE_FEATURES
	Target *t;
	t = (*it)->target[METH_TCP];
	if (t && t->expected) {
	    (*it)->feature[TCP_INIT_TTL] = t->init_ttl;
	    (*it)->feature[TCP_IPID_CLASS] = class_to_feature(t->classification);
	    (*it)->feature[TCP_IPID_ORDER] = t->order;
	}
	t = (*it)->target[METH_ICMP];
	if (t && t->expected) {
	    (*it)->feature[ICMP_INIT_TTL] = t->init_ttl;
	    (*it)->feature[ICMP_IPID_CLASS] = class_to_feature(t->classification);
	    (*it)->feature[ICMP_IPID_ORDER] = t->order;
	}
	t = (*it)->target[METH_INDIR];
	if (t && t->expected) {
	    (*it)->feature[INDIR_INIT_TTL] = t->init_ttl;
	    (*it)->feature[INDIR_IPID_CLASS] = class_to_feature(t->classification);
	    (*it)->feature[INDIR_IPID_ORDER] = t->order;
	}
#endif
	if (htmldir) {
	    for (size_t row = 0; row < targetfilters.size(); row++) {
		for (int meth = 0; meth < N_METH; meth++) {
		    if ((*it)->target[meth] &&
			targetfilters[row]((*it)->target[meth]))
		    {
			(*targetstats)(row, meth).insert((*it)->target[meth]);
			(*targetstats)(row, meth).count++;
		    }
		}
	    }
	}
    }
    memoryInfo.print("classified");
}

static uint64_t choose(int n, int k)
{
    uint64_t num = 1;
    uint64_t den = 1;
    if (k <= n/2) k = n-k;
    for (int i = n; i > k; i--) { num *= i; }  //  n! / k!
    for (int i = n-k; i > 1; i--) { den *= i; }  //  (n-k)!
    return num/den;
}

int main(int argc, char *argv[])
{
  try {
    argv0 = argv[0];
    for (int i = 0; i < argc; i++) {
	cmdline += argv[i];
	cmdline += ' ';
    }

    InFile::fork = false;
    allow_udp_diff_addr = false;
    min_overlap_option = 0;

    int opt;
    int longidx;
    while ((opt = getopt_long(argc, argv, "", longopts, &longidx)) != -1) {
	switch (opt) {
	case 'H':
	    htmldir = optarg;
	    break;
	case 'I':
	    includefilename = strdup(optarg);
	    break;
	case 'M':
	    monitorfilename = strdup(optarg);
	    break;
#ifdef ENABLE_FEATURES
	case 'f':
	    featurefilename = optarg;
	    break;
#endif
	case 'o':
	    min_overlap_option = atoi(optarg);
	    break;
	case 'a':
	    if (!Comparison::parse_oob_allowed(optarg))
		usageExit();
	    break;
	case 't':
	    gptype = optarg;
	    break;
	case 0:
	    // long flag
	    break;
	default:
	    usageExit();
	}
    }
    if (optind == argc || !htmldir) { usageExit(); }

    if (htmldir) {
	if (mkdir(htmldir, 0755) < 0 && errno != EEXIST) {
	    cerr << htmldir << ": " << strerror(errno) << endl;
	    exit(1);
	}
    }

#ifdef ENABLE_FEATURES
    if (featurefilename) {
	featurefile.open(featurefilename);
	if (!featurefile) {
	    cerr << "can't open " << featurefilename << ": " << strerror(errno) << endl;
	    exit(1);
	}
    }
#endif

    initTargetStats();
    initResolveStats();
    if (monitorfilename) monitors.load(monitorfilename);
    memoryInfo.print("startup");
    if (includefilename) loadIncludes(includefilename);

    for (int i = optind; i < argc; i++) {
	loadResponses(argv[i]);
	cerr << "# loaded " << unresolvedTargets.size() << " targets" << endl;
	if (tries_per_target < 0) {
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!n_targets[meth]) continue;
		tries_per_target = (*unresolvedTargets.begin())->target[meth]->tries;
		break;
	    }
	    for (int meth = 0; meth < N_METH+1; meth++) {
		wrapdata[meth] = newZeroedArray(uint64_t, tries_per_target);
	    }
	}
	classifyTargets();
	resolveTargets();
#if 1
	for (MultiMethTargetSet::iterator it = unresolvedTargets.begin(); it != unresolvedTargets.end(); it++) {
	    for (int meth = 0; meth < N_METH; meth++) {
		if ((*it)->target[meth]) {
		    if ((*it)->target[meth]->refcount == 0) {
			(*it)->target[meth]->freeResponses();
		    }
		}
	    }
	}
	memoryInfo.print("freeResponses");
#endif
	targets.insert(unresolvedTargets.begin(), unresolvedTargets.end());
	unresolvedTargets.clear();
    }

    if (htmldir) {
	if (chdir(htmldir) < 0) {
	    cerr << htmldir << ": " << strerror(errno) << endl;
	    exit(1);
	}
	html.open("index.html");
	if (!html) {
	    cerr << "can't open " << "index.html" << ": " << strerror(errno) << endl;
	    exit(1);
	}
	html << "<html>\n";
	html << "<head>\n";
	html << "<style>\n";
	html << ".bad { background:#FFBBBB; }\n";       // bad = red
	html << ".mediumbad { background:#FFDDBB; }\n"; // mediumbad = orange
	html << ".medium { background:#FFFFBB; }\n";    // medium = yellow
	html << ".good { background:#BBFFBB; }\n";      // good = green
	html << "td { text-align: right; }\n";
	html << "td.label { text-align: left; }\n";
	html << ".subgroup { font-size: small; }\n";
	html << ".subgroup .label { padding-left: 4ex; }\n";
	html << ".subgroup .label:before { content: \"-- \"; }\n";
	html << "td, th { border: 1px solid black; padding: 0.5ex; }\n";
	html << "td.percent { border-left-style: hidden; }\n";
	// Insert a thin space. This does not work in all browsers (notably IE),
	// but when it fails, it's harmless.
	html << ".numeric { white-space: nowrap; }\n";
	html << ".numeric span:after { content: \" \"; font-size: 50%; }\n";
	html << "table.invisible { border-collapse: collapse; }\n";
	html << "table.invisible > * > * > td { border-style: hidden; text-align: left; vertical-align: top; }\n";
	html << "</style>\n";
	html << "</head>\n";
	html << "<body>\n";
	html << "<h1>Estimation Analysis</h1>\n";
    }

    double avg_round_time = tot_round_time / n_window_rounds;
    if (htmldir) {
	monitors.sortByName();
	html << "\n<h2>The setup</h2>\n";
	html << "<p>Version: " << ::cvsID << "\n";
	html << "<p>Version: " << MidarUtils::cvsID << "\n";
	html << "<p>Command line: " << cmdline << "\n";
	if (!errormsgs.empty()) {
	    html << "<p style='color: red;'>Errors:\n";
	    html << "<ul style='color: red;'>\n";
	    for (unsigned i = 0; i < errormsgs.size(); i++)
		html << "<li>" << errormsgs[i] << "\n";
	    html << "</ul>\n";
	}
	html << "<p>Probed " << targets.size() << " router addresses from " <<
	    monitors.sizeByName() << " monitors.\n";
	html << "<p>round times: min=" << min_round_time <<
	    " avg=" << avg_round_time << " max=" << max_round_time << "\n";
	html << "<p>total time: " << fmtfixed(tot_round_time, 6) << " s (" <<
	    (tot_round_time / 3600) << " hours)\n";
	html << "<p>Probe methods:\n";
	html << "<dl>\n";
	html << "<dt>tcp<dd>TCP SYN ACK sent to unopen port on target\n";
	html << "<dt>udp<dd>UDP packet sent to unopen port on target\n";
	html << "<dt>icmp<dd>ICMP Echo Request sent to target\n";
	html << "<dt>indir<dd>TTL-limited ICMP Echo Request sent to a destination <i>past</i> the target\n";
	html << "</dl>\n";
	memoryInfo.print("html header");
    }

    html << flush;

    // print per-monitor statistics
    if (htmldir) {
	cerr << "# monitor stats\n";
	print_monitor_stats(html, monitors, n_targets);
	memoryInfo.print("per-monitor stats");
    }

    // print distribution of responses
    if (htmldir) {
	cerr << "# response distribution\n";
	html << "\n<h2>Response distribution</h2>\n";
	html << "Definitions:\n";
	html << "<ul>\n";
	html << "<li>An \"unresponsive\" target is one that responded to less " <<
	    "than " << (100*responsive_thresh) << "% of probes" << 
	    " and less than " << responsive_count << " probes total.\n";
	html << "<li>A \"nondegenerate\" response is one where the IP ID was not zero, " <<
	    "reflecting the probe, or a repeat of the previous response.\n";
	html << "<li>\"Mostly\" means more than " << (100*mostly_thresh) <<
	    "% of the responses for the target had the same classificiation.\n";
	html << "<li>An \"nonmonotonic\" target is one whose ID, on " <<
	    "average, increases by more than " <<
	    (100*wraps_per_cycle_thresh) << "% of the ID space (i.e., " <<
	    int(default_wrapsize * wraps_per_cycle_thresh) <<
	    " for 16-bit counters) between probes, " <<
	    "which at our probing rate is roughly " <<
	    int((default_wrapsize * wraps_per_cycle_thresh)/avg_round_time) <<
	    " IDs per second.  Such targets can not reliably be " <<
	    "distinguished from targets with random ID counters.\n";
	html << "</ul>\n";

	html << "\n";
	html << "<p>\n";
	html << "<table style='border-collapse: collapse'>\n";
	html << "<tr><td>                 ";
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!n_targets[meth]) continue;
	    html << "<th colspan=2>" << methname[meth];
	}
	html << "\n    <th colspan=2>" << "total";
	html << "</tr>\n";

	print_classification_row(html, "total methods-targets", n_targets);
	print_classification_row(html, "probes",        n_probes);
	print_classification_row(html, "responses",     n_responses);

	for (size_t row = 0; row < targetfilters.size(); row++) {
	    // cerr << "# print_row: " << targetfilters[row].label << "\n";
	    html << "<tr " <<
		status_style[targetfilters[row].status] <<
		"><td class='label'>" << targetfilters[row].label << "\n";

	    for (int col = 0; col < N_METH; col++) {
		if (!n_targets[col]) continue;
		string label(targetfilters[row].label);
		label += "-";
		label += methname[col];
		print_classification_cell(html, &(*targetstats)(row, col),
		    n_targets[col], label, gptype, gpscript);
	    }
	    html << "<td><td class='percent'>\n";
	}
	delete targetstats; targetstats = 0; // no longer needed

	html << "</table>\n";

	vector<double> data[N_METH*2];
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!(*it)->target[meth]) continue;
		data[2*meth].push_back((*it)->target[meth]->expected /
		    double((*it)->target[meth]->tries));
		data[2*meth+1].push_back((*it)->target[meth]->nondegenerate /
		    double((*it)->target[meth]->tries));
	    }
	}

	html << "<p>\n";
	html << "<table style='border-collapse: collapse'>\n";
	int used_meths = 0;
	for (int meth = 0; meth < N_METH; meth++) {
	    if (n_targets[meth]) used_meths++;
	}
	html << "<tr><th rowspan=2 colspan=" << used_meths << ">probe<br/>methods " << 
	    "<th colspan=4>responsive targets " <<
	    "<th colspan=4>nondegenerate targets " <<
	    "<th colspan=4>monotonic targets\n";
	html << "<tr> " <<
	    "<th colspan=2>union " <<
	    "<th colspan=2>intersection " <<
	    "<th colspan=2>union " <<
	    "<th colspan=2>intersection " <<
	    "<th colspan=2>union " <<
	    "<th colspan=2>intersection\n";
	for (int bits = 1; bits < (1 << N_METH); bits++) {
	    int total_targets = 0;
	    int total_resp_union = 0;
	    int total_resp_intersection = 0;
	    int total_use_union = 0;
	    int total_use_intersection = 0;
	    int total_pred_union = 0;
	    int total_pred_intersection = 0;
	    for (int meth = 0; meth < N_METH; meth++) {
		if (((1 << meth) & bits) && !n_targets[meth]) goto next;
	    }
	    html << "<tr>";
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!n_targets[meth]) continue;
		html << "<td>";
		if (!((1 << meth) & bits)) continue;
		html << methname[meth];
	    }
	    html << "\n";
	    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
		bool resp_union_ok = false;
		bool resp_intersection_ok = true;
		bool use_union_ok = false;
		bool use_intersection_ok = true;
		bool pred_union_ok = false;
		bool pred_intersection_ok = true;
		for (int meth = 0; meth < N_METH; meth++) {
		    if (!((1 << meth) & bits)) continue;
		    if (!(*it)->target[meth]) continue;
		    if ((*it)->target[meth]->classification != CLASS_UNRESPONSIVE) {
			resp_union_ok = true;
		    } else {
			resp_intersection_ok = false;
		    }
		    if ((*it)->target[meth]->flags & GOOD_ID) {
			use_union_ok = true;
		    } else {
			use_intersection_ok = false;
		    }
		    if ((*it)->target[meth]->flags & MONOTONIC) {
			pred_union_ok = true;
		    } else {
			pred_intersection_ok = false;
		    }
		}
		total_targets++;
		if (resp_union_ok) total_resp_union++;
		if (resp_intersection_ok) total_resp_intersection++;
		if (use_union_ok) total_use_union++;
		if (use_intersection_ok) total_use_intersection++;
		if (pred_union_ok) total_pred_union++;
		if (pred_intersection_ok) total_pred_intersection++;
	    }
	    html << "    <td>" << htmlnum(total_resp_union) << "<td class='percent'>" <<
		fmtpercent(total_resp_union, total_targets) << "\n";
	    html << "    <td>" << htmlnum(total_resp_intersection) << "<td class='percent'>" <<
		fmtpercent(total_resp_intersection, total_targets) << "\n";
	    html << "    <td>" << htmlnum(total_use_union) << "<td class='percent'>" <<
		fmtpercent(total_use_union, total_targets) << "\n";
	    html << "    <td>" << htmlnum(total_use_intersection) << "<td class='percent'>" <<
		fmtpercent(total_use_intersection, total_targets) << "\n";
	    html << "    <td>" << htmlnum(total_pred_union) << "<td class='percent'>" <<
		fmtpercent(total_pred_union, total_targets) << "\n";
	    html << "    <td>" << htmlnum(total_pred_intersection) << "<td class='percent'>" <<
		fmtpercent(total_pred_intersection, total_targets) << "\n";
	    next: ;
	}
	html << "</table>\n";

	if (n_targets[METH_TCP]) {
	    uint64_t n_rst = 0;
	    uint64_t n_pu = 0;
	    uint64_t n_mixed = 0;
	    uint64_t n_other = 0;
	    uint64_t n_total = 0;
	    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
		if (!(*it)->target[METH_TCP] || !((*it)->target[METH_TCP]->flags & GOOD_ID))
		    continue;
		switch ((*it)->target[METH_TCP]->responsetypes) {
		    case (1 << RT_TCP_RST):
			n_rst++; break;
		    case (1 << RT_ICMP_PORT_UNREACH):
			n_pu++; break;
		    case ((1 << RT_TCP_RST) | (1 << RT_ICMP_PORT_UNREACH)):
			n_mixed++; break;
		    default:
			n_other++; break;
		}
		n_total++;
	    }
	    html << "<h3>Response types for nondegenerate TCP targets</h3>\n";
	    html << "<table style='border-collapse: collapse'>\n";
	    html << "<tr><td>always TCP RST <td>" << htmlnum(n_rst) <<
		"<td class='percent'>" << fmtpercent(n_rst, n_total) << "\n";
	    html << "<tr><td>always ICMP PORT UNREACHABLE <td>" << htmlnum(n_pu) <<
		"<td class='percent'>" << fmtpercent(n_pu, n_total) << "\n";
	    html << "<tr><td>mixed <td>" << htmlnum(n_mixed) <<
		"<td class='percent'>" << fmtpercent(n_mixed, n_total) << "\n";
	    html << "<tr><td>other <td>" << htmlnum(n_other) <<
		"<td class='percent'>" << fmtpercent(n_other, n_total) << "\n";
	    html << "</table>\n";
	}

	if (n_targets[METH_UDP]) {
	    uint64_t n_same = 0;
	    uint64_t n_diff = 0;
	    uint64_t n_mixed = 0;
	    uint64_t n_total = 0;
	    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
		if (!(*it)->target[METH_UDP] || !((*it)->target[METH_UDP]->flags & GOOD_ID))
		    continue;
		if (((*it)->target[METH_UDP]->flags & (SAMEADDR|DIFFADDR)) == (SAMEADDR|DIFFADDR)) {
		    n_mixed++;
		} else if ((*it)->target[METH_UDP]->flags & SAMEADDR) {
		    n_same++;
		} else if ((*it)->target[METH_UDP]->flags & DIFFADDR) {
		    n_diff++;
		}
		n_total++;
	    }
	    html << "<h3>Response types for nondegenerate UDP targets</h3>\n";
	    html << "<table style='border-collapse: collapse'>\n";
	    html << "<tr><td>always same address <td>" << htmlnum(n_same) <<
		"<td class='percent'>" << fmtpercent(n_same, n_total) << "\n";
	    html << "<tr><td>always different address <td>" << htmlnum(n_diff) <<
		"<td class='percent'>" << fmtpercent(n_diff, n_total) << "\n";
	    html << "<tr><td>mixed <td>" << htmlnum(n_mixed) <<
		"<td class='percent'>" << fmtpercent(n_mixed, n_total) << "\n";
	    html << "</table>\n";
	}

	{
	    html << "\n<h3>Targets with N useful TTL expansions</h3>\n";
	    const int max_allowed_n = 60;
	    int histogram[N_METH][max_allowed_n + 1];
	    memset(histogram, 0, sizeof(histogram));
	    int max_n = 0;
	    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
		for (int meth = 0; meth < N_METH; meth++) {
		    if (!(*it)->target[meth]) continue;
		    int n = int((*it)->target[meth]->ttl_expansions);
		    if (n > max_allowed_n) n = max_allowed_n;
		    if (n > max_n) max_n = n;
		    histogram[meth][n]++;
		}
	    }
	    if (max_n == 0) {
		html << "<p>No useful TTL expansions were seen.</p>";
	    } else {
		html << "<table style='border-collapse: collapse'>\n";
		html << "<tr><th>N";
		for (int meth = 0; meth < N_METH; meth++) {
		    if (!n_targets[meth]) continue;
		    html << "<td>" << methname[meth];
		}
		html << "\n";
		for (int n = 1; n < max_n; n++) {
		    html << "<tr><td>" << n;
		    for (int meth = 0; meth < N_METH; meth++) {
			if (!n_targets[meth]) continue;
			html << "<td>" << histogram[meth][n];
		    }
		    html << "</tr>\n";
		}
	    }
	    html << "</table>\n";
	}

	gp.open("responsiveness", "responsiveness", gptype, gpscript);
	html << "<br><img src='responsiveness.png'>\n\n";
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'Fraction of attempts'\n");
	fprintf(gp.f, "set key top left\n");

	string strlabels[N_METH*2];
	const char *labels[N_METH*2];
	for (int i = 0; i < N_METH; i++) {
	    strlabels[2*i] = string(methname[i]) + ": responses of expected type";
	    labels[2*i] = strlabels[2*i].c_str();
	    strlabels[2*i+1] = string(methname[i]) + ": nondegenerate";
	    labels[2*i+1] = strlabels[2*i+1].c_str();
	}
	gp.plot_cdf(N_METH * 2, data, labels);
	gp.close();
	html << flush;
	memoryInfo.print("classif. table");
    }

    // wrap distribution
    if (htmldir) {
	const char *label[N_METH+1];
	for (int k = 0; k <= tries_per_target-1; k++) {
	    wrapdata[N_METH][k] = choose(tries_per_target-1, k);
	}
	for (int i = 0; i < N_METH; i++) {
	    label[i] = methname[i];
	}
	label[N_METH] = "binomial";
	html << "\n<h2>Wrap distribution</h2>\n\n";

	gp.close();
	gp.open("wraps", "wrap distribution for targets that respond to every probe", gptype, gpscript);
	html << "<br><img src='wraps.png'>\n\n";

	fprintf(gp.f, "set logscale y\nset mytics 10\n");
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'number of wraps\n");
	fprintf(gp.f, "set key top right\n");
	gp.plot_binned_histogram(N_METH+1, tries_per_target, wrapdata, 0, 1, label, "linespoints");
	gp.close();
	html << flush;
	memoryInfo.print("plot wraps");
    }

    // plot velocity
    if (htmldir) {
	html << "\n<h2>Velocity</h2>\n\n";

	gp.close();
	gp.open("velocity", "velocity distribution for monotonic targets", gptype, gpscript);
	html << "<br><img src='velocity.png'>\n\n";

	vector<double> data[N_METH];
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!(*it)->target[meth] || !((*it)->target[meth]->flags & MONOTONIC))
		    continue;
		data[meth].push_back((*it)->target[meth]->series_velocity);
	    }
	}

	fprintf(gp.f, "set logscale x\nset mxtics 10\n");
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'Velocity (ID/s)'\n");
	fprintf(gp.f, "set key bottom right\n");
	gp.plot_cdf(N_METH, data, methname);
	gp.close();
	html << flush;
	memoryInfo.print("plot velocity");
    }

    //## HOWTO: Heat map
    // define 2d grid with points at offset -0.5 * rectangle_size
    // accumulate points into 2d grid
    // GP: set pm3d corners2color c1
    // GP: splot '-'
    // print the grid, plus an extra row and column with 0 z-values

    // velocity comparison
    if (htmldir) {
	html << "\n<h2>Velocity comparison</h2>\n\n";
	html << "<p>Comparison of velocities of targets that had monotonic ID values for two probe methods.\n\n";

	{
	    double size = .05;
	    int offset = -int(log(0.005)/log(10) / size + 0.5);
	    int max = int(log(65536/10.0)/log(10) / size + 0.5);

	    // essentially, a 2D (max+offset) x (max+offset) array of ints
	    vector<vector<int> > data(max+offset);

	    for (int i = 0; i < N_METH; i++) {
		if (!n_targets[i]) continue;
		for (int j = i + 1; j < N_METH; j++) {
		    if (!n_targets[j]) continue;
		    for (int x = 0; x < max+offset; x++) {
			data[x].clear();
			data[x].resize(max+offset, 0);
		    }
		    // cerr << "### comparing " << methname[i] << " and " << methname[j] << "\n";
		    int min_x = INT_MAX;
		    int min_y = INT_MAX;
		    int max_x = 0;
		    int max_y = 0;
		    int total_targets = 0;
		    for (MultiMethTargetSet::iterator it = targets.begin();
			it != targets.end(); it++)
		    {
			if (!(*it)->target[i] || !((*it)->target[i]->flags & MONOTONIC))
			    continue;
			if (!(*it)->target[j] || !((*it)->target[j]->flags & MONOTONIC))
			    continue;
			total_targets++;
			double v;
			v = (*it)->target[i]->series_velocity;
			int x = int(log(v)/log(10) / size + 0.5);
			v = (*it)->target[j]->series_velocity;
			int y = int(log(v)/log(10) / size + 0.5);
			data[x+offset][y+offset]++;
			if (x > max_x) max_x = x;
			if (y > max_y) max_y = y;
			if (x < min_x) min_x = x;
			if (y < min_y) min_y = y;
		    }
		    if (total_targets == 0) continue;

		    char filename[64];
		    sprintf(filename, "velocity-comparison-%s-%s", methname[i], methname[j]);
		    char title[64];
		    sprintf(title, "Velocity comparison: %s vs. %s (%d addrs)",
			methname[i], methname[j], total_targets);
		    gp.open(filename, title, gptype, gpscript);
		    html << "<br><img src='" << filename << ".png'>\n\n";
		    // fprintf(gp.f, "set grid xtics ytics\n");
		    fprintf(gp.f, "set xlabel 'Average velocity of %s'\n", methname[i]);
		    fprintf(gp.f, "set ylabel 'Average velocity of %s'\n", methname[j]);
		    fprintf(gp.f, "set logscale x\nset mxtics 10\n");
		    fprintf(gp.f, "set logscale y\nset mytics 10\n");
		    // fprintf(gp.f, "set logscale z\n");
		    fprintf(gp.f, "set pm3d map\n");
		    fprintf(gp.f, "set palette rgbformulae -7,-5,-15\n");
		    fprintf(gp.f, "set pm3d corners2color c1\n");
		    fprintf(gp.f, "splot '-' title ''\n");
		    fprintf(gp.f, "set xrange [%f:%f]\n",
			pow(10, min_x*size), pow(10, max_x*size));
		    fprintf(gp.f, "set yrange [%f:%f]\n",
			pow(10, min_y*size), pow(10, max_y*size));
		    for (int x = min_x; x <= max_x + 1; x++) {
			for (int y = min_y; y <= max_y + 1; y++) {
			    fprintf(gp.f, "%f %f %d\n",
				pow(10, x*size-size/2), pow(10, y*size-size/2),
				data[x+offset][y+offset]);
			}
			fprintf(gp.f, "\n");
		    }
		    fprintf(gp.f, "e\n");
		    gp.close();
		}
	    }
	}
	html << flush;
	memoryInfo.print("velocity cmp");
    }

    // initial ttl comparison
    if (htmldir) {
	init_ttl_comparison(html, targets, n_targets, gptype, gpscript);
	html << flush;
	memoryInfo.print("init ttl cmp");
    }

    // fwd/reverse path length comparison
    if (htmldir) {
	path_length_comparison(html, targets, gptype, gpscript);
	html << flush;
	memoryInfo.print("length cmp");
    }

    //  print distribution of RTTs
    if (htmldir) {
	print_rtt_distribution(html, targets, gptype, gpscript);
	memoryInfo.print("rtt distr");
    }

    // free unneeded data
    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if ((*it)->target[meth])
		(*it)->target[meth]->freeUnusedOrder();
	}
    }
    memoryInfo.print("freeUnusedOrder");

    printResolveStats();

    // target summary must be generated AFTER shared counter test so it can
    // print the preferred method
    if (htmldir) {
	//html << "<p><a href='target-summary.txt'>target summary</a></p>\n";
	static ofstream summary;
	summary.open("target-summary.txt");
	if (!summary) {
	    cerr << "can't open " << "target-summary.txt" << ": " << strerror(errno) << endl;
	    exit(1);
	}
	target_summary_header(summary, cmdline);
	for (MultiMethTargetSet::iterator it = targets.begin();
	    it != targets.end(); it++)
	{
	    Target **tgt = (*it)->target;
	    int8_t preferred_meth = -1;
	    index_t max_nondegenerate = 0;
	    // find maximum responsiveness
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!tgt[meth] || !(tgt[meth]->flags & MONOTONIC)) continue;
		if (tgt[meth]->nondegenerate > max_nondegenerate) {
		    max_nondegenerate = tgt[meth]->nondegenerate;
		}
	    }
	    // We prefer the first (lowest) qualified method
	    for (int meth = 0; meth < N_METH; meth++) {
		// disqualify method that wasn't used or wasn't monotonic
		if (!tgt[meth] || !(tgt[meth]->flags & MONOTONIC)) continue;
		// disqualify method with less than half the max responsiveness
		if (tgt[meth]->nondegenerate * 2 < max_nondegenerate) continue;
		// found a qualified method
		preferred_meth = meth;
		break;
	    }
	    // But ICMP is preferred over UDP if they share a counter, because
	    // ICMP is less likely to be rate limited in later runs.
	    if ((preferred_meth == METH_UDP) &&
		(*it)->shared.test(METH_UDP, METH_ICMP) &&
		(tgt[METH_ICMP]->nondegenerate * 2 >= max_nondegenerate))
	    {
		preferred_meth = METH_ICMP;
	    }
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!(*it)->target[meth]) continue;
		target_summary_row(summary, (*it)->target[meth], meth == preferred_meth);
	    }
	}
	summary << "# end\n";
	summary.close();
	memoryInfo.print("target summary");
    }

    if (htmldir) {
	oob_distribution.plot(html, gptype, gpscript);
	memoryInfo.print("oob distr");
    }

    if (htmldir) {
	print_methshare_tables();
	memoryInfo.print("methshare tables");
    }

#ifdef ENABLE_FEATURES
    // After resolve(), because resolve() may have cleared features from
    // methnqiues that don't share counters.
    if (featurefilename) {
	featurefile << "# command line: " << cmdline << "\n";
	featurefile << "# features:\n";
	for (int i = 0; i < N_FEATURES; i++) {
	    featurefile << "# " << i << ": " << feature_label[i] << "\n";
	}
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    featurefile << (*it)->addr;
	    for (int i = 0; i < N_FEATURES; i++) {
		featurefile << "\t" << (*it)->feature[i];
	    }
	    featurefile << "\n";
	}
	featurefile << "# end\n";
	featurefile.close();
	memoryInfo.print("features");
    }
#endif

    if (htmldir) {
	html << "</body>\n";
	html << "</html>\n";
	html.close();
    }
    cerr << "# " << argv0 << " done\n";
    memoryInfo.print("done");

  } catch (const std::exception &e) {
    cerr << e.what() << endl;
    exit(1);
  }
}
