/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <assert.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <map>
#include <string>
#include <getopt.h>

static const char *cvsID = "$Id: generate-subsets.cc,v 1.33 2015/09/18 19:25:05 kkeys Exp $";

using namespace std;

#include "../lib/config.h"
#include "../lib/infile.h"

#include "../lib/MemoryInfo.h"
#define NO_DEBUG_MEMORY 0
MemoryInfo memoryInfo;

#include "MidarUtils.h"
using MidarUtils::AddrPair;
using MidarUtils::AddrPairSet;

static string cmdline = "";
static const char *argv0;
static double spacing = 0.600;
static double defaultVelocity = 0.600;
static uint32_t max_max_delta_id = 65536 * MidarUtils::wraps_per_cycle_thresh;
static uint32_t max_delta_id = 65536 * 0.05;
static size_t max_subset_size = 100;
static size_t min_subset_size = 3;
static bool debug = false;
static ofstream *dotfile = 0;
static ofstream *largefile = 0;
static ofstream *smallfile = 0;
static bool split = false;
static long total_small_cost = 0, total_large_cost = 0, total_large_rotcost = 0;

typedef set<ip4addr_t> AddrSet;

struct AddrData {
    double velocity;
    bool partial;
    AddrSet neighbors;
    AddrData() : velocity(0), partial(false), neighbors() {}
};

typedef map<ip4addr_t, AddrData> AddrMap;
typedef map<ip4addr_t, int> AddrCount;
typedef set<AddrMap::iterator> IteratorSet;

namespace std {
    template<>
    struct less<AddrMap::iterator> {
	bool operator()(const AddrMap::iterator &a, const AddrMap::iterator &b) {
	    return a->first < b->first;
	}
    };
}

AddrMap addrs;
IteratorSet partials;

static void usageExit()
{
    cerr << "Usage: " << argv0 << " [<options>] <pairfile>...\n";
    cerr << "Generate elimination subsets from pairs listed in <pairfile>.\n";
    cerr << "Options:\n";
    cerr << "  -o <file>   write subsets to <file>\n";
    cerr << "  -O <file>   write subsets of large sets to <file>.large and small sets to <file>.small\n";
    cerr << "  -v <file>   load velocity information from target summary <file>\n";
    cerr << "  -V <n>      use <n> as the default velocity for targets not listed in -v file\n";
    cerr << "  -d <n>      delay between probes is <n> seconds [" << spacing << "]\n";
    cerr << "  -s <n>      maximum subset size is <n> [" << max_subset_size << "]\n";
    cerr << "  -S <n>      minimum subset size is <n> (supercedes -a) [" << min_subset_size << "]\n";
    cerr << "  -a <n>      maximum ID advancement per round is <n> [" << max_delta_id << "]\n";
    cerr << "  -g <file>   write DOT graph of subsets to <file>\n";
    cerr << "  -D          enable debugging\n";
    exit(1);
}

static void loadPairs(const char *filename)
{
    char buf[8192];
    InFile in(filename);
    AddrPair ap;
    long n_loaded = 0;
    cerr << "# loading pairs from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	MidarUtils::FieldSplitter<3> fields(buf, " \t\n");
	ap[0] = ap[1] = ip4addr_t(0);

	if (!fields[1] || fields[2])
	    throw std::runtime_error("bad format");
	ap[0] = ip4addr_t(fields[0]);
	ap[1] = ip4addr_t(fields[1]);

#if 0
	if (includeFile) {
	    if (include.find(ap[0]) == include.end() ||
		include.find(ap[1]) == include.end())
		    continue;
	}
#endif

	n_loaded++;
	addrs[ap[0]].neighbors.insert(ap[1]);
	addrs[ap[1]].neighbors.insert(ap[0]);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    cerr << "# loaded " << n_loaded << " pairs\n";
    memoryInfo.print("loaded pairs");
}

static void loadVelocities(const char *filename)
{
    char buf[8192];
    InFile in(filename);
    cerr << "# loading velocities from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {

// #    target_addr method  tries probes  expct unexpc   zero reflct repeat ndegen     velocity ttl exp ittl LE classification
//          1.0.0.6  indir     30     30     30      0      0      0      0     30     1.698768  12   0  255  0 monotonic; BE RHB(8)=

	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	// note: "ittl" field is optional, so classification could be in
	// field 14 or 15.
	if (!strstr(buf, " monotonic;")) continue;
	MidarUtils::FieldSplitter<16> fields(buf, " \t\n");
	if (!fields[15]) continue;
	ip4addr_t addr = ip4addr_t(fields[0]);

	AddrMap::iterator amit = addrs.find(addr);
	if (amit == addrs.end()) continue;
	char *endp;
	double velocity = strtod(fields[10], &endp);
	if (*endp) continue;

	// If we're loading an estimation run, there may be multiple lines
	// for the same target, but the preferred one will have "*" in the
	// method.  For other types of runs, there's only one line, and
	// it won't have a "*", so we accept the first line we find.
	if (amit->second.velocity == 0 || strchr(fields[1], '*'))
	    amit->second.velocity = velocity;
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    AddrMap::iterator amit;
    int missing = 0;
    for (amit = addrs.begin(); amit != addrs.end(); ++amit) {
	if (amit->second.velocity == 0) {
	    if (defaultVelocity > 0) {
		amit->second.velocity = defaultVelocity;
		if (missing < 10)
		    cerr << "warning: no velocity for " << amit->first << endl;
	    } else {
		if (missing < 10)
		    cerr << "error: no velocity for " << amit->first << endl;
	    }
	    missing++;
	}
    }
    if (missing) {
	cerr << (defaultVelocity > 0 ? "warning" : "error") <<
	    ": no velocity for " << missing << " addresses" << endl;
	if (!defaultVelocity) exit(1);
    }

    memoryInfo.print("loaded velocities");
}

static long primes[] = {
    2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61,
    67, 71, 73, 79, 83, 89, 97
};

static long prime_ceil(long n)
{
    for (size_t i = 0; i < sizeof(primes)/sizeof(int); i++) {
	if (n <= primes[i])
	    return primes[i];
    }
    return LONG_MAX - 1;
}

class WorkingSet
{
public:
    int setid, subsetid, setsize;
    long cost;
    int pairs;
    double max_velocity, min_velocity, max_set_velocity;
    double H, S, V;
    IteratorSet contents;
    AddrCount neighbors;
    WorkingSet() : setid(0), subsetid(0), setsize(0), cost(0), pairs(0),
	max_velocity(0), min_velocity(DBL_MAX), max_set_velocity(0),
	H(0), S(1), V(1), contents(), neighbors() {}
    ~WorkingSet() { }
    void insert(AddrMap::iterator amit) {
	AddrSet::iterator nit;
	IteratorSet::iterator cit;

	if (dotfile) {
	    if (contents.empty() && subsetid == 0) {
		*dotfile << "subgraph \"set" << setid << "\" {\n";
	    }
	}
	if (!amit->second.partial) setsize++;
	cost++;
	for (cit = contents.begin(); cit != contents.end(); ++cit) {
	    (*cit)->second.neighbors.erase(amit->first);
	    int n = amit->second.neighbors.erase((*cit)->first);
	    pairs += n;
	    if (dotfile && n) {
		*dotfile << "\"" << (*cit)->first << "\" -- \"" << amit->first <<
		    "\" [ color=\"" << H << "," << S << "," << V << "\"";
		if (contents.size() == 1)
		    *dotfile << ", label=\"" << subsetid << "\"";
		*dotfile << " ];\n";
	    }
	}
	contents.insert(amit); // addr referred to by amit is now a member...
	neighbors.erase(amit->first); // ...not a neighbor.
	for (nit = amit->second.neighbors.begin(); nit != amit->second.neighbors.end(); ++nit) {
	    neighbors[*nit]++;
	}
	if (dotfile) {
	    if (V > 0.3) V *= 0.99;
	}
	if (max_velocity < amit->second.velocity)
	    max_velocity = amit->second.velocity;
	if (min_velocity > amit->second.velocity)
	    min_velocity = amit->second.velocity;
	if (max_set_velocity < max_velocity)
	    max_set_velocity = max_velocity;
    }
    void dump() {
	IteratorSet::iterator cit;
	for (cit = contents.begin(); cit != contents.end(); ++cit) {
	    if ((*cit)->second.neighbors.empty()) {
		(*cit)->second.partial = false;
		partials.erase(*cit);
	    } else {
		(*cit)->second.partial = true;
		partials.insert(*cit);
	    }
	}
	bool small = partials.empty() && subsetid == 0;
	ofstream *outfile = small ? smallfile : largefile;
	*outfile << "# set " << setid << "." << subsetid <<
	    " (addrs " << contents.size() <<
	    ", pairs " << pairs <<
	    ", subset_neighbors " << neighbors.size() <<
	    ", partials " << partials.size() <<
	    ", velocity " << min_velocity << " - " << max_velocity <<
	    ", delta_id " << contents.size() * spacing * max_velocity <<
	    ")" << endl;
	for (cit = contents.begin(); cit != contents.end(); ++cit) {
	    *outfile << (*cit)->first << endl;
	    if ((*cit)->second.neighbors.empty())
		addrs.erase(*cit);
	}

	if (partials.empty()) {
	    if (small && split)
		total_small_cost += cost;
	    else
		total_large_cost += cost;
	    cerr << "# full set " << setid <<
		": subsets " << subsetid + 1 <<
		", velocity " << max_set_velocity <<
		", size " << setsize <<
		", cost " << cost;
	    if (!split || !small) {
		long rotations = 1;
		long rotgroups = 1;
		double delta_id = setsize * spacing * max_set_velocity;
		rotgroups = ceil(delta_id / max_delta_id);
		if (rotgroups < ceil(double(setsize) / max_subset_size))
		    rotgroups = ceil(double(setsize) / max_subset_size);
		if (rotgroups > ceil(double(setsize) / min_subset_size))
		    rotgroups = ceil(double(setsize) / min_subset_size);
		if (rotgroups > 1)
		    rotations = prime_ceil(rotgroups) + 1;
		if (rotations > 100) {
		    cerr << ", rotgroups " << rotgroups <<
			", rotations HUGE" <<
			", rotcost HUGE";
		    total_large_rotcost = LONG_MAX;
		} else {
		    // cost to probe this set using a Rotating Groups schedule
		    long rotcost = rotations * setsize;
		    cerr << ", rotgroups " << rotgroups <<
			", rotations " << rotations <<
			", rotcost " << rotcost;
		    if (total_large_rotcost != LONG_MAX)
			total_large_rotcost += rotcost;
		}
	    }
	    cerr << "\n";
	    setsize = 0;
	    setid++;
	    subsetid = 0;
	    if (dotfile) {
		*dotfile << "} // size " << setsize << "\n"; // close graph
		H = 0;
		S = 1;
		V = 1;
	    }
	    cost = 0;
	    max_set_velocity = 0;
	} else {
	    subsetid++;
	    if (dotfile) {
		do {
		    H += 0.180556;
		    if (H > 1) H -= 1;
		} while (H > 0.138 && H < 0.200); // skip yellow
		if (S > 0.3) S *= 0.99;
		V = 1;
	    }
	}
	contents.clear();
	neighbors.clear();
	pairs = 0;
	max_velocity = 0;
	min_velocity = DBL_MAX;
    }
};

void generate()
{
    WorkingSet workingset;

    while (!addrs.empty()) {
	AddrMap::iterator seed;
	seed = addrs.end();

	if (!partials.empty()) {
	    // Start with a low-degree partially solved addr
	    IteratorSet::iterator pit;
	    seed = *partials.begin();
	    for (pit = partials.begin(); pit != partials.end(); ++pit) {
		AddrMap::iterator amit = *pit;
		if (amit->second.neighbors.size() == 1) {
		    seed = amit;
		    break;
		}
		if (amit->second.neighbors.size() < seed->second.neighbors.size())
		    seed = amit;
	    }
	    if (debug) cerr << "# seed: " << seed->first << " " << seed->second.neighbors.size() << " (partial)" << endl;
	} else {
	    seed = addrs.begin();
#if 0 // expensive, and doesn't help
	    // Start with a low-degree addr
	    AddrMap::iterator amit;
	    for (amit = addrs.begin(); amit != addrs.end(); ++amit) {
		if (amit->second.neighbors.size() <= lowest_degree) {
		    seed = amit;
		    break;
		}
		if (amit->second.neighbors.size() < seed->second.neighbors.size())
		    seed = amit;
	    }
	    lowest_degree = seed->second.neighbors.size();
#endif
	    if (debug) cerr << "# seed: " << seed->first << " " << seed->second.neighbors.size() << " (new)" << endl;
	}

	workingset.insert(seed);

	double delta_id;   
	int allowed_delta_id = max_max_delta_id;
	do {
	    // Find the best neighbor to add to the workingset.
	    AddrMap::iterator best_amit = addrs.end();
	    double best_score = -DBL_MAX;
	    double best_delta_id = 0;
	    size_t best_in_links = 0;
	    size_t best_out_links = 0;
	    bool best_partial = false;
	    AddrCount::iterator nit;
	    double delta_t = (workingset.contents.size() + 1) * spacing;
	    // cerr << "# searching for next candidate (set size = " << workingset.neighbors.size() << ")\n";
	    for (nit = workingset.neighbors.begin(); nit != workingset.neighbors.end(); ++nit) {
		AddrMap::iterator amit = addrs.find(nit->first);
		delta_id = delta_t * max(workingset.max_velocity, amit->second.velocity);
		// Prefer a candidate that's linked to the most workingset members.
		size_t in_links = nit->second;
		double score = in_links;
		// If the candidate is already partially complete, raise its
		// score (by a factor < 2, so a non-partial with in_links==2
		// is still better than a partial with in_links==1)
		if (amit->second.partial) score *= 1.9;
		// If adding candidate would complete it, raise its score.
		if (amit->second.neighbors.size() == in_links) score *= 3;
#if 1
		// tiebreaker: prefer node that adds the fewest new neighbors.
		size_t out_links = amit->second.neighbors.size() - in_links;
#if 0 // expensive to calculate, and doesn't help much
		{
		    // don't count amit neighbors that are already workingset neighbors
		    AddrSet::iterator cn = amit->second.neighbors.begin();
		    AddrCount::iterator wn = workingset.neighbors.begin();
		    while (cn != amit->second.neighbors.end() && wn != workingset.neighbors.end()) {
			if (*cn == wn->first) {
			    out_links--;
			    ++cn;
			    ++wn;
			} else if (*cn < wn->first) {
			    ++cn;
			} else {
			    ++wn;
			}
		    }
		}
#endif
#endif
		//cerr << "# candidate " << nit->first <<
		    //"(in_links " << in_links <<
		    //", partial " << amit->second.partial <<
		    //", score " << score <<
		    //")";
		if (score != best_score ? score > best_score :
		    out_links < best_out_links)
		{
		    if (workingset.contents.size() >= 2 && delta_id > allowed_delta_id) {
			// this addr would make the set exceed allowed_delta_id
			//cerr << " too fast\n";
			continue;
		    }
		    best_amit = amit;
		    best_score = score;
		    best_partial = amit->second.partial;
		    best_delta_id = delta_id;
		    best_in_links = in_links;
		    best_out_links = out_links;
		    //cerr << " best so far\n";
		} else {
		    //cerr << " not best so far\n";
		}
	    }
	    if (best_amit == addrs.end()) {
		break;
	    }
	    if (workingset.contents.size() > 10 && // set has reached reasonable size
		best_in_links == 1 && // best has only one link into set
		best_out_links > best_in_links) // neighbor count would grow 
	    {
		// stopping here, where all incomplete nodes are "necks", will
		// keep delta_id from growing while forcing only a minimal
		// number of nodes to be repeated in multiple subsets
		if (debug) cerr << "# stopping at " <<
		    best_amit->first <<
		    " (score " << best_score <<
		    ", velocity " << best_amit->second.velocity <<
		    ", delta_id " << best_delta_id <<
		    ", neighbors " << best_amit->second.neighbors.size() <<
		    ")\n";
		break;
	    }
	    workingset.insert(best_amit);
	    if (debug) cerr << "# add " << (best_partial ? " partial" : "neighbor") << ": " <<
		best_amit->first <<
		" (score " << best_score <<
		", velocity " << best_amit->second.velocity <<
		", delta_id " << best_delta_id <<
		")\n";

	    allowed_delta_id = (workingset.contents.size() < min_subset_size) ?
		max_max_delta_id : max_delta_id;
	    if (debug) cerr << "# allowed_delta_id = " << allowed_delta_id << "\n";

	} while (!workingset.neighbors.empty() &&
	    workingset.contents.size() < max_subset_size &&
	    (workingset.contents.size() + 1) * spacing * workingset.max_velocity < allowed_delta_id);

	workingset.dump();
    }
    if (split) {
	*smallfile << "# total cost: " << total_small_cost << "\n";
	*smallfile << "# end" << endl;
    }
    *largefile << "# total cost: " << total_large_cost << "\n";
    if (total_large_rotcost == LONG_MAX)
	*largefile << "# total rotcost: HUGE\n";
    else
	*largefile << "# total rotcost: " << total_large_rotcost << "\n";
    *largefile << "# end" << endl;
}

int main(int argc, char *const *argv)
{
  try {
    argv0 = argv[0];
    int ch;
    const char *outfilename = 0;
    const char *velocityfilename = 0;
    const char *dotfilename = 0;

    for (int i = 0; i < argc; i++) {
	cmdline += argv[i];
	cmdline += ' ';
    }

    while ((ch = getopt(argc, argv, "o:O:v:g:d:s:S:a:DV:")) != -1) {
	switch (ch) {
	    case 'o':
		outfilename = strdup(optarg);
		split = false;
		break;
	    case 'O':
		outfilename = strdup(optarg);
		split = true;
		break;
	    case 'v':
		velocityfilename = strdup(optarg);
		break;
	    case 'V':
		defaultVelocity = atof(optarg);
		break;
	    case 'g':
		dotfilename = strdup(optarg);
		break;
	    case 'd':
		spacing = atof(optarg);
		break;
	    case 's':
		max_subset_size = atoi(optarg);
		break;
	    case 'S':
		min_subset_size = atoi(optarg);
		break;
	    case 'a':
		max_delta_id = atoi(optarg);
		if (max_delta_id > max_max_delta_id) {
		    cerr << "error: maximum ID advancement must be <= " << max_max_delta_id << endl;
		    exit(1);
		}
		break;
	    case 'D':
		debug = true;
		break;
	    default:
		usageExit();
	}
    }

    argc -= optind;
    argv += optind;

    if (split) {
	largefile = new ofstream((string(outfilename) + ".large").c_str());
	smallfile = new ofstream((string(outfilename) + ".small").c_str());
    } else {
	largefile = smallfile = new ofstream(outfilename);
    }

    *largefile << "# command line: " << cmdline << endl;
    *largefile << "# version: " << ::cvsID  << endl;
    *largefile << "# version: " << MidarUtils::cvsID << endl;
    if (split) {
	*smallfile << "# command line: " << cmdline << endl;
	*smallfile << "# version: " << ::cvsID  << endl;
	*smallfile << "# version: " << MidarUtils::cvsID << endl;
    }
    memoryInfo.print("startup");

#if 0
    if (includeFile) {
	loadInclude(includeFile);
    }
#endif

    for (int i = 0; i < argc; i++) {
	loadPairs(argv[i]);
    }

    if (velocityfilename)
	loadVelocities(velocityfilename);

    if (dotfilename) {
	dotfile = new ofstream(dotfilename);
	*dotfile << "graph \"" << dotfilename << "\" {\n";
	*dotfile << "edge [ color=black, style=solid, penwidth=.1, fontsize=28 ];\n";
	*dotfile << "node [ color=black, shape=point, label=\"\" ];\n";
    }

    generate();
    memoryInfo.print("done");
    if (dotfile) {
	*dotfile << "}\n";
	delete dotfile;
    }

  } catch (const std::exception &e) {
    cerr << e.what() << endl;
    exit(1);
  }
}
