#!/usr/bin/env perl
## 
## Copyright (C) 2011-2018 The Regents of the University of California.
## 
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
## 

# Generate target list for MIDAR corroboration run.

use strict;
use Getopt::Long;
use List::Util 'shuffle';

my $cvsid = '$Id: midar-cor-targets.pl,v 1.22 2012/06/04 18:51:10 kkeys Exp $';
my $cmdline = join(' ', $0, @ARGV);

#### parameters
my $outdir = undef;
my $default_meth = undef;
my $shuffle_flag = 0;
my $minoverlap_threads = 0;
my $same_mon = 1;
my $separate = 0;
my $maxsmall = 0;
my $delay = .600;
my $monfile = undef;
my $allowed_ids_per_round = .3 * 65536;

my $usage =
    "Usage: $0 [<options>] <sets_file> <res_target_files>...\n" .
    "Options:\n" .
    "--outdir <dir>    write results to <dir>\n" .
    "--method <meth>   default probe method for any target not listed in target files\n" .
    "--shuffle         randomly shuffle sets (recommended)\n" .
    "--minoverlap <n>  minimize overlap between sets, using <n> probe threads (experimental)\n" .
    "--nosame          do not try to reassign targets of a set to the same monitor\n" .
    "--separate        write homogeneous and heterogeneous sets to separate files\n" .
    "--maxsmall <n>    write sets of size <=n and >n to separate files (default: $maxsmall)\n" .
    "--delay <n>       delay between probes (seconds) (default: $delay)\n" .
    "--monitors <file> load list of usable monitors from <file>\n";

sub usageExit() {
    print STDERR $usage;
    exit 1;
}

GetOptions(
    "outdir=s" => \$outdir,
    "method=s" => \$default_meth,
    "shuffle!" => \$shuffle_flag,
    "minoverlap=i" => \$minoverlap_threads,
    "same!" => \$same_mon,
    "separate" => \$separate,
    "maxsmall=i" => \$maxsmall,
    "delay=f" => \$delay,
    "monitors|M=s" => \$monfile
    ) or usageExit();

usageExit() if @ARGV < 2;

#### state
my @sets = (); # sets
my @diff_sets = (); # sets where targets have different sources
my @same_sets = (); # sets where all targets have the same source
my %targets; # solution for each target
my %sources; # list of groups for each source
my $heterogeneous = 0;
my $maxgroup = -1;
my $ids_per_round;
my $max_ids_per_round = 0;
my @max_ids_per_round_offenders = ();
my $n_overmax_ids_per_round = 0;
my $n_unassignable = 0;
my $n_random = 0;

sub load_sets() {
    my $file = $ARGV[0];
    print STDERR "# loading sets from $file\n";
    open SETS, "<", $file || die "$file: $!\n";
    my ($setid, $set, $size, $vhi);
    $setid = -1;
    while (<SETS>) {
	# print STDERR "### line: $_";
	if (/^# set /) {
	    # for Elimination run
	    # e.g.: # set 0.0 (addrs 32, ... velocity 6.12345 - 7.11311, delta_id 136.572)
	    if (!/^# set ((\d+)\.(\d+)) \(addrs (\d+),.* velocity (?:[\d.]+ - )?([\d.]+),/) {
		print STDERR "ERROR: syntax error in set line:\n";
		print STDERR $_;
		exit 1;
	    }
	    $setid++;
	    $set = {setid => $setid};
	    $set = {setid => $1, parentid => $2, subsetid => $3, size => $4, vhi => $5};
	    push @sets, $set;
	    # print STDERR "# new set: $_";
	} elsif (/^#|^\s*$/) {
	    next;
	} else {
	    chomp;
	    push @{$set->{targets}}, {addr=>$_};
	    # print STDERR "# added to set: $_\n";
	}
    }
    close SETS;
    shift @ARGV;
    print STDERR "# loaded ", (scalar @sets), " sets\n";
}

sub load_targets() {
    while (@ARGV) {
	my $file = $ARGV[0];
	print STDERR "# loading targets from $file\n";
	open DATA, "<", $file || die "$file: $!\n";
	my $n_targets = 0;
	while (<DATA>) {
	    next if (/^#/ || /^$/);
	    my ($addr, $src, $dst, $hop, $terminal_hop, $cksum, $method, $velocity) = split;
	    if (!defined $velocity) {
		die "$file:$.: syntax error\n";
	    }
	    if (!defined $monfile) {
		$sources{$src}++;
	    } elsif (!$sources{$src}) {
		if ($method =~ /indir/) {
		    $n_unassignable++;
		    next;
		}
		$src = undef;
	    }
	    $n_targets++;
	    $targets{$addr} =
		{src=>$src, dst=>$dst, hop=>$hop,
		terminal_hop=>$terminal_hop, cksum=>$cksum, method=>$method,
		velocity=>$velocity+0.0};
	}
	print STDERR "## loaded $n_targets targets\n";
	close DATA;
	shift @ARGV;
    }
}

sub load_monitors() {
    print STDERR "# loading monitors from $monfile\n";
    open MONITORS, "<", $monfile || die "$monfile: $!\n";
    while (<MONITORS>) {
	next if (/^#/ || /^$/);
	my ($name, $addr) = split;
	$sources{$addr} = 1;
    }
    close MONITORS;
    print STDERR "# loaded ", scalar keys %sources, " monitors\n";
}

#### main
if (defined $outdir) {
    -d $outdir || mkdir $outdir || die "$outdir: $!\n";
}
load_monitors() if defined $monfile;
load_sets();
exit if scalar @sets == 0;
load_targets();
chdir $outdir if defined $outdir;
my @sources = keys %sources;

# select sources for each set
if ($same_mon) {
    {
	for (my $s = 0; $s < scalar @sets; $s++) {
	    print STDERR "# selecting sources: set $s of ", scalar @sets, "\n[A";
	    #print STDERR "\n"; # XXX
	    my $set = $sets[$s];
	    my $n_indir = 0;
	    my %indir_src = {};
	    for my $target (@{$set->{targets}}) {
		if (defined $targets{$target->{addr}}) {
		    if ($targets{$target->{addr}}{method} =~ /^indir/) {
			$indir_src{$targets{$target->{addr}}{src}}++;
			$n_indir++;
		    }
		} elsif ($default_meth) {
		} else {
		    next;
		}
	    }
	    #print STDERR "n_indir=$n_indir\n"; # XXX

	    my $bestsrc = undef;
	    my $besthop = 256;
	    for my $target (@{$set->{targets}}) {
		#print STDERR "    checking",
		#    " src=", $targets{$target}{src},
		#    " hop=", $targets{$target}{hop},
		#    "\n"; # XXX
		my $targetdata = $targets{$target->{addr}};
		if (defined $targetdata &&
		    defined $targetdata->{src} &&
		    $targetdata->{hop} < $besthop)
		{
		    next if $n_indir && !$indir_src{$targetdata->{src}};
		    $bestsrc = $targetdata->{src};
		    $besthop = $targetdata->{hop};
		}
	    }
	    if (!defined $bestsrc) {
		$n_random++;
		$bestsrc = $sources[rand(scalar @sources)];
	    }
	    #print STDERR "bestsrc=$bestsrc; besthop=$besthop;\n"; # XXX

	    my $different = 0;
	    for my $target (@{$set->{targets}}) {
		my $targetdata = $targets{$target->{addr}};
		if (!defined $targetdata) {
		    $target->{src} = (defined $default_meth) ? $bestsrc : undef;
		    next;
		}
		if ($targetdata->{src} eq $bestsrc) {
		    $target->{src} = $bestsrc;
		} elsif ($targetdata->{method} =~ /^indir/) {
		    $target->{src} = $targetdata->{src};
		    $different++ if defined $targetdata->{src};
		} else {
		    # move target to best source
		    $target->{src} = $bestsrc;
		}
	    }

	    if ($different) {
		$heterogeneous++;
		push(@diff_sets, $set) if $separate;
	    } else {
		push(@same_sets, $set) if $separate;
	    }
	}
    }
    print STDERR "\n";
}

print "# unassignable targets: $n_unassignable\n";
print "# randomly assigned sets: $n_random\n";

if ($separate) {
    dumpsets(\@diff_sets, "targets.multi-src");
    dumpsets(\@same_sets, "targets.one-src");
} else {
    dumpsets(\@sets, "targets");
}

sub dumpsets($$) {
    my ($sets, $filename) = @_;

    if ($minoverlap_threads > 0) {
	## rearrange subsets to minimize overlap within sets
	# identify parent sets
	print STDERR "# identifying parent sets\n";
	my %parents;
	my $totalsize = 0;
	for my $set (@$sets) {
	    my $parentid = $set->{parentid};
	    push @{$parents{$parentid}{subsets}}, $set;
	    $parents{$parentid}{size} += $set->{size};
	    $totalsize += $set->{size};
	}
	my $threadsize = $totalsize / $minoverlap_threads;
	print STDERR "# distributing sets\n";
	my @singles; # subssets that are the only subset in their parent
	my @multis; # subsets that belong to parents with multiple subsets
	while (my ($parentid, $ref) = each %parents) {
	    if (scalar @{$ref->{subsets}} == 1) {
		my $subset = $ref->{subsets}[0];
		$subset->{earliest} = 0;
		$subset->{latest} = $threadsize - $subset->{size};
		push @singles, $subset;
	    } else {
		# Calculate number of threads needed by parent, and spread out the
		# subsets evenly among those threads.
		my $parent_threads = int(($ref->{size} + $threadsize - 1) / $threadsize);
		my $freespace = $parent_threads * $threadsize - $ref->{size};
		my $wiggle = $freespace / scalar @{$ref->{subsets}};
		# Calculate earliest and latest times each subset can start.
		my $t = 0;
		for my $subset (sort {$a->{vhi} <=> $b->{vhi}} @{$ref->{subsets}}) {
		    my $earliest = $t % $threadsize;
		    $subset->{earliest} = $earliest;
		    $subset->{latest} = $earliest + $wiggle;
		    $t += $earliest + $wiggle + $ref->{size};
		    push @multis, $subset;
		}
	    }
	}

	# reverse sort multis by earliest
	my @multis = sort {$b->{earliest} <=> $a->{earliest}} @multis;

	# reverse sort singles by latest
	my @singles = sort {$b->{latest} <=> $a->{latest}} @singles;

	print STDERR "# scheduling sets\n";
	my $nsets = @sets;
	@sets = ();
	my $t = 0; # clock
	my @active = (); # sets being actively probed
	my $tail = 0;
	print "\n";
	while (scalar @multis > 0 || scalar @singles > 0) {
	    print STDERR "\r# progress: ", scalar @sets, "/$nsets",
		"; singles: ", scalar @singles,
		"; multis: ", scalar @multis;
	    my $best_i = 0;
	    my $best_latest = 2147483647;

	    for (my $i = 1; $i <= @multis; $i++) {
		if ($multis[-$i]{earliest} > $t) {
		    $tail = $i-1;
		    print STDERR "; tail: $tail           \r";
		    last;
		}
		if ($multis[-$i]{latest} < $best_latest) {
		    $best_latest = $multis[-$i]{latest};
		    $best_i = $i;
		}
	    }

	    my $set;
	    if ($best_i > 0) {
		$set = splice @multis, -$best_i, 1; # remove set from @multis
	    } elsif (scalar @singles > 0) {
		$set = pop @singles; # remove set from @singles
	    } else {
		$set = pop @multis; # remove set from @multis
	    }

	    # add set to @active
	    push @active, $set;

	    push @sets, $set;
	    $set->{start} = $t;
	    $set->{end} = $t + $set->{size};
	    # if all "threads" are busy...
	    if (scalar @active >= $minoverlap_threads) {
		# find active set that will end first
		my $end_i = 0;
		for (my $i = 1; $i < @active; $i++) {
		    if ($active[$i]->{end} < $active[$end_i]->{end}) { $end_i = $i; }
		}
		my $ender = splice @active, $end_i, 1; # remove set from @active
		$t = $ender->{end}; # increment clock
	    }
	}
	print "\n";

    } elsif ($shuffle_flag) {
	print STDERR "# shuffling sets\n";
	@$sets = shuffle(@$sets);
    }

    if (0) {
	for (my $s = 0; $s < scalar @$sets; $s++) {
	    $sets->[$s]{setid} = $s;
	}
    }

    print STDERR "# writing sets\n";
    #my %load;
    open OUT, ">", $filename || die "$filename: $!\n";
    print STDERR "# writing $filename (", scalar @$sets, ")\n";
    print OUT "# command line: $cmdline\n";
    print OUT "# $cvsid\n";
    my $t = 0;

    for (my $s = 0; $s < scalar @$sets; $s++) {
	my $set = $sets->[$s];
	my $n = scalar @{$set->{targets}};
	#for my $src (@sources) {
	    #$load{$src}[$t+$n] = $load{$src}[$t];
	#}
	$t += $n;
	printf OUT "#%8s %3s %15s %15s %15s %3s %3s %5s %5s %11s\n",
	    "set", "idx", "target", "src", "dst", "hop", "trm", "cksum", "meth", "velocity";

	# Arrange the targets so that those with the same source are adjacent
	# (to optimize communication overhead in the prober), but use a
	# different source ordering for each set (to minimize contention in
	# the prober).  Also shuffle targets, so prober looks less like a
	# scanner.
	my %srcrand;
	for my $src (@sources) {
	    $srcrand{$src} = rand();
	}
	my @sortedtargets =
	    sort {$srcrand{$a->{src}} cmp $srcrand{$b->{src}}}
		shuffle(@{$set->{targets}});

	my $maxvelocity = 0;
	my $setsize = 0;
	my $idx = 0;
	for my $target (@sortedtargets) {
	    next unless defined $target->{src};
	    my $tgtdata = $targets{$target->{addr}};

	    $maxvelocity = $tgtdata->{velocity} if defined $tgtdata && $maxvelocity < $tgtdata->{velocity};
	    $setsize++;

	    #$load{$tgtdata->{src}}[$t]++;
	    printf OUT "%-9s %3d %15s %15s ",
		$set->{setid}, $idx, $target->{addr}, $target->{src};
	    $idx++;

	    if (!defined $tgtdata && defined $default_meth) {
		printf OUT "%15s %3d %3d %5d ", '0.0.0.0', 0, 0, 0;
		printf OUT "%5s %11.6f\n", $default_meth, 0;
	    } else {
		if ($target->{src} eq $tgtdata->{src}) {
		    printf OUT "%15s %3d %3d %5d ",
			$tgtdata->{dst}, $tgtdata->{hop},
			$tgtdata->{terminal_hop}, $tgtdata->{cksum};
		} else {
		    printf OUT "%15s %3d %3d %5d ", '0.0.0.0', 0, 0, 0;
		}
		printf OUT "%5s %11.6f\n",
		    $tgtdata->{method}, $tgtdata->{velocity};
	    }
	}

 	# We use win-target (i.e., estimation) velocity, because velocity
 	# during "break" runs may be inflated by induction
	$ids_per_round = $setsize * $maxvelocity * $delay;
	if ($max_ids_per_round < $ids_per_round) {
	    $max_ids_per_round = $ids_per_round if ($max_ids_per_round < $ids_per_round);
	}
	if ($ids_per_round > $allowed_ids_per_round) {
	    $n_overmax_ids_per_round++;
	    push @max_ids_per_round_offenders, "$ids_per_round IDs per round:  set " . $set->{setid} . ", size=$setsize, v=$maxvelocity\n";
	}
    }
    print OUT "# end\n";
    close OUT;
}

print STDERR "# ", scalar @sets, " sets\n";
if ($same_mon) {
    print STDERR "# $heterogeneous heterogeneous sets\n";
}

print STDERR "# max IDs per round: $max_ids_per_round (assuming a delay of $delay s)\n";
if ($max_ids_per_round > $allowed_ids_per_round) {
    print STDERR "# WARNING: $n_overmax_ids_per_round instances of ids per round > $allowed_ids_per_round\n";
    print STDERR "# offenders:\n", join('', sort {$b <=> $a} @max_ids_per_round_offenders);
}

#open GP, "|gnuplot";
#print GP "set terminal png\n";
#print GP "set output 'load.png'\n";
#print GP "set xlabel 'index'\n";
#print GP "set ylabel 'load'\n";
#my $prefix = "plot";
#for my $src (keys %sources) {
#    print GP "$prefix '-' using 1:2 title '$src' with lines";
#    $prefix = ",";
#}
#print GP "\n";
#for my $src (keys %sources) {
#    for (my $i = 0; $i < scalar @{$load{$src}}; $i++) {
#	if ($load{$src}[$i]) {
#	    print GP $i, ' ', $load{$src}[$i], "\n";
#	}
#    }
#    print GP "e\n";
#}
#close GP;
