/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Extract MIDAR targets from ark traceroute paths and generate target files.
 */

static const char *cvsID = "$Id: midar-targets.cc,v 1.75 2015/09/18 22:16:11 kkeys Exp $";

#include <float.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <cstdarg>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <map>
#include <algorithm>
// #include <ext/hash_set>
// #include <new>
// #include <string>
#include <exception>
#include <stdexcept>

#define HAVE_LIBZ 1

#include "../lib/config.h"
#include "../lib/infile.h"
#include "../lib/ip4addr.h"
#include "../lib/NetPrefix.h"
#include "MidarUtils.h"

using namespace std;

#define NO_DEBUG_MEMORY 0
#include "../lib/MemoryInfo.h"
MemoryInfo memoryInfo;

class myofstream : public ofstream {
public:
    string name;
    string tmpname;
    myofstream() { ofstream(); }
    void open(const string &suffix, bool tmp = false);
    void close() {
	ofstream::close();
	if (!tmpname.empty())
	    rename(tmpname.c_str(), name.c_str());
    }
};

myofstream out_log;

//#define debug_log out_log
#define debug_log sink

NetPrefixSet bogons;
time_t now;
string cmdline;
bool limitedTargets = false;
bool explicitMonitors = false;
int n_preferredMonitors = 0;
long n_forcedTargets = 0;

const class Sink : public ostream { public: Sink() {} } sink;
template<class T>
inline const Sink& operator<< (const Sink& out, const T&) { return out; }


struct Cfg {
    time_t start_time;
    bool extract;
    bool extract_all;
    bool movable_udp;
    bool point_to_point_mates;
    bool shuffle;
    size_t groupsize;
    double max_velocity;
    vector<const char*> bogonFiles;
    vector<const char*> traceFiles;
    vector<const char*> solutionFiles;
    vector<const char*> estimationFiles;
    vector<const char*> targetFiles;
    vector<const char*> includeFiles;
    vector<const char*> addrFiles;
    vector<const char*> monitorFiles;
    char filetype;
    char *output_basename;
    bool setFile(const char *filename);
private:
    void setOneFile(const char *filename);
} cfg;

void Cfg::setOneFile(const char *filename)
{
    switch (this->filetype) {
    case 'B':
	this->bogonFiles.push_back(filename);
	break;
    case 'P':
	this->traceFiles.push_back(filename);
	break;
    case 'S':
	this->solutionFiles.push_back(filename);
	break;
    case 'E':
	this->estimationFiles.push_back(filename);
	break;
    case 'T':
	this->targetFiles.push_back(filename);
	break;
    case 'I':
	this->includeFiles.push_back(filename);
	break;
    case 'A':
	this->addrFiles.push_back(filename);
	break;
    case 'M':
	this->monitorFiles.push_back(filename);
	break;
    default:
	break;
    }
}

bool Cfg::setFile(const char *filename)
{
    if (!this->filetype) return false;
    if (filename[0] == '@') {
	InFile filelist(filename+1);
	char buf[PATH_MAX];
	while (filelist.gets(buf, sizeof(buf))) {
	    strtok(buf, "\n");
	    setOneFile(strdup(buf));
	}
	filelist.close();
    } else {
	setOneFile(filename);
    }
    return true;
}

typedef uint32_t TraceID;

struct Source; // forward declaration;

struct Iface {
    ip4addr_t addr;
    Iface(ip4addr_t a) : addr(a) {}
};

struct partialTarget : public Iface { // hack to get around dependency loop
    double velocity;
    int n_solutions; // number of solutions
    bool isMate;
    int8_t meth;
    bool preferred; // meth is the preferred method for this target
    uint8_t probe_ttl; // probe TTL
    uint8_t ittl; // inferred initial TTL of target
    int8_t order; // ipid byte order of target
    partialTarget(ip4addr_t a) : Iface(a), velocity(-1), n_solutions(0),
	isMate(false), meth(-1), probe_ttl(0), ittl(0), order(-1) {}
};

struct IfaceLess {
    bool operator()(const Iface *a, const Iface *b) { return a->addr < b->addr; }
};

struct Solution {
    Source *source;
    partialTarget *target;
    ip4addr_t dest_addr;	// addr of dest in original probe
    uint16_t cksum;		// checksum of original probe
    int16_t target_hop;		// hop # of target in original probe
    int16_t path_len;		// path length to dest in original probe
    time_t age;			// how long ago was the path seen
    double value;		// how valuable is this solution
    Solution(Source *s, partialTarget *t) :
	source(s), target(t), dest_addr(ip4addr_t(0)),
	cksum(0), target_hop(-1), path_len(-1), age(0), value(0)
        {}
    static double calculate_value(uint16_t target_hop, uint16_t path_len, time_t age) {
	// Scale components so that 2 days of age is as bad as 1 hop.
	// Calculate a "Witch of Agnesi" curve for each component.
	// Sum the witches, with different weights for the importance of each
	// component.  (There's nothing special about this particular curve,
	// it just has the rough shape we want.)
	double scaled_age = age / (2 * 86400.0);
	double result =
	    0.5 / ((target_hop * target_hop) + 1.0) +
	    0.1 / ((path_len * path_len) + 1.0) +
	    0.4 / ((scaled_age * scaled_age) + 1.0);
	return result;
    }
};

struct Source : public Iface {
    mutable const string *name;
    long n_targets;
    long n_stuck;       // number of targets for which this source was the only choice
    vector<long> bucketed_targets;
    long n_mates;
    bool preferred;
    bool used;
    const char *getname() const {
	if (!name) { name = new string(addr); }
	return name->c_str();
    }
    Source(ip4addr_t a) : Iface(a), name(), n_targets(0), n_stuck(0),
	n_mates(0), preferred(), used() {};
};

typedef set<Source*, IfaceLess> SourceSet;

struct SolutionLess {
    bool operator()(const Solution *a, const Solution *b) {
	return a->target->addr != b->target->addr ?
	    a->target->addr < b->target->addr :
	    a->source->addr < b->source->addr;
    }
};

typedef multiset<Solution*, SolutionLess> SolutionSet; // multiset allows extract_all

struct Target : public partialTarget {
    SolutionSet::iterator firstSol; // first solution
    Solution *bestSol; // best solution
    Target(ip4addr_t a) : partialTarget(a), firstSol(), bestSol(0) {}
};

inline bool target_velocity_greater(const Target *a, const Target *b)
    { return a->velocity > b->velocity; }

inline bool target_source_less(const Target *a, const Target *b)
    { return a->bestSol->source->addr < b->bestSol->source->addr; }

inline bool target_source_less_velocity_greater(const Target *a, const Target *b) {
    return a->bestSol->source->addr != b->bestSol->source->addr ?
	a->bestSol->source->addr < b->bestSol->source->addr :
	a->velocity > b->velocity;
}

// true if a has a NONZERO number of solutions less than b
inline bool target_solutions_less(const Target *a, const Target *b)
    { return a->n_solutions && a->n_solutions < b->n_solutions; }

typedef set<Target*, IfaceLess> TargetSet;

typedef vector<Target*> TargetVec;

string *currentSourceName = 0;
SourceSet sources;
TargetSet targets;
TargetVec *vTargets = 0; // targets sorted by descending velocity
SolutionSet solutions;


static inline bool isAnon(ip4addr_t addr) { return (addr & 0xFF000000) == 0; }


static bool isBogus(const ip4addr_t addr)
{
    // assumes that bogons contains only the largest prefixes
    NetPrefix key(addr, 32);
    NetPrefixSet::const_iterator it = bogons.upper_bound(key);
    return (it != bogons.begin() && (*--it).contains(addr));
}

static Source *findOrInsertSource(ip4addr_t addr)
{
    Source key(addr);
    Source *source;
    SourceSet::const_iterator it = sources.find(&key);
    if (it == sources.end()) {
	source = new Source(addr);
	sources.insert(source);
    } else {
	source = (*it);
    }
    return source;
}

static inline Target *findTarget(const ip4addr_t &addr)
{
    Target key(addr);
    TargetSet::const_iterator it = targets.find(&key);
    return it != targets.end() ? *it : 0;
}

static inline bool observed(const ip4addr_t &addr)
{
    Target *target = findTarget(addr);
    return target && !target->isMate;
}

static Target *findOrInsertTarget(const ip4addr_t &addr)
{
    Target key(addr);
    Target *target;
    TargetSet::const_iterator it = targets.find(&key);
    if (it == targets.end()) {
	target = new Target(addr);
	targets.insert(target);
    } else {
	target = (*it);
    }
    return target;
}

static Solution *findOrInsertSolution(Source *source, Target *target)
{
    Solution key(source, target);
    Solution *solution;
    SolutionSet::const_iterator it = solutions.find(&key);
    if (it == solutions.end()) {
	solution = new Solution(source, target);
	solutions.insert(solution);
    } else {
	solution = (*it);
    }
    return solution;
}

#include "../lib/PathLoader.h"

PathLoader pathLoader;

class MyPathLoaderHandler : public PathLoaderHandler {
public:
    MyPathLoaderHandler() :
	PathLoaderHandler(out_log, debug_log != sink) {};

    int processHops(const ip4addr_t *hops, int n_hops, ip4addr_t src, ip4addr_t dst, void *strace)
    {
	if (src == ip4addr_t(0)) {
	    cerr << "illegal source address " << src << endl;
	    abort();
	}
	if (!strace) {
	    cerr << "not a scamper trace" << endl;
	    abort();
	}
#if HAVE_SCAMPER
	scamper_trace_t *scampertrace = static_cast<scamper_trace_t*>(strace);
	time_t age = now - scampertrace->start.tv_sec;
	if (age < 1) age = 1;

	// note:  include_ends == false
	Source *source = findOrInsertSource(src);
	if (!source->name)
	    source->name = currentSourceName;
	source->used = true;
	for (int i = (hops[0] == src); i < n_hops; i++) {
	    if (isBogus(hops[i])) continue;
	    Target *target;
	    if (limitedTargets) {
		target = findTarget(hops[i]);
		if (!target) continue;
	    } else {
		target = findOrInsertTarget(hops[i]);
	    }
	    Solution *solution;
	    if (cfg.extract_all) {
		solution = new Solution(source, target);
		solutions.insert(solution);
	    } else {
		solution = findOrInsertSolution(source, target);
	    }
	    double newvalue = Solution::calculate_value(i, n_hops, age);
	    // Replace old solution with new solution if it's as good or better
	    // than the old one.
	    if (newvalue >= solution->value) {
		debug_log << "# solution " << solution->source->addr << " -> " << solution->target->addr << ": dest " <<
		    dst << " " << i << " " << n_hops << " " << age << " (" << newvalue << ") beats " <<
		    solution->dest_addr << " " << solution->target_hop << " " << solution->path_len << " " << solution->age << " (" << solution->value << ")\n";
		solution->dest_addr = dst;
		solution->target_hop = i;
		solution->path_len = n_hops;
		solution->age = age;
		solution->value = newvalue;
		// Scamper stores the icmp checksum in the sport field.  We
		// need it to be able to reproduce the flow label for Paris
		// traceroute.
		solution->cksum = scampertrace->sport;
		{
		    // NB: Scamper stored the cksum in host order, not network
		    // order, and all of our ark boxes are little endian, so
		    // we must swap bytes here.
		    // TODO: We expect someday to deploy a newer version of
		    // scamper that stores the cksum in the correct order,
		    // along with a flag to indicate this fact.  This code
		    // must be updated to check the flag.
		    solution->cksum = (solution->cksum << 8) | (solution->cksum >> 8);
		}
	    }
	}

	++pathLoader.n_good_traces;
	return 1;
#else
	cerr << "scamper trace support is not enabled" << endl;
	abort();
#endif
    }
};

MyPathLoaderHandler pathLoaderHandler;

static void loadTraces(const string &filename)
{
    out_log << "# loadTraces: " << filename << endl;

    // if filename ends in ".xxx-xx.warts.gz", use "xxx-xx" as the source name.
    currentSourceName = 0;
    int idx = filename.length() - 16;
    if (idx >= 0 && filename.compare(idx+7, 9, ".warts.gz") == 0) {
	if (filename[idx+0] == '.' &&
	    isalpha(filename[idx+1]) &&
	    isalpha(filename[idx+2]) &&
	    isalpha(filename[idx+3]) &&
	    filename[idx+4] == '-' &&
	    isalpha(filename[idx+5]) &&
	    isalpha(filename[idx+6]))
	{
	    currentSourceName = new string(filename.substr(idx+1, 6));
	}
    }

    int n_traces = pathLoader.load(filename.c_str());
    currentSourceName = 0;

    out_log << "# traces=" << n_traces <<
	"/" << pathLoader.n_good_traces <<
	"/" << pathLoader.n_raw_traces <<
	" loops=" << pathLoader.n_loops <<
	" discarded=" << pathLoader.n_discarded_traces <<
	" solutions=" << solutions.size() <<
	endl;
    memoryInfo.print("load paths");
}

static void loadSolutions(const string &filename, bool fixed)
{
    out_log << "# loadSolutions: " << filename << endl;

    int offset = fixed ? 0 : 1;

    char buf[65536];
    InFile in(filename.c_str());
    while (in.gets(buf, sizeof(buf))) {
      try {
        if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty

        MidarUtils::FieldSplitter<7> fields(buf, " \t");

        ip4addr_t target_addr(fields[0 + offset]);
	Target *target;
	if (limitedTargets) {
	    target = findTarget(target_addr);
	    if (!target) continue;
	} else {
	    target = findOrInsertTarget(target_addr);
	}
	if (fixed) {
	    if (target->meth == MidarUtils::METH_INDIR ||
		(target->meth == MidarUtils::METH_UDP && !cfg.movable_udp))
	    {
		// target wants to use same solution as it did in Estimation
	    } else {
		continue;
	    }
	    if (target->bestSol) {
		throw std::runtime_error(string(target->addr) +
		    " already has solution");
	    }
	} else {
	    if (target->meth == MidarUtils::METH_INDIR && target->bestSol)
		continue;
	}

        ip4addr_t src_addr(fields[1 + offset]);
	Source *src = findOrInsertSource(src_addr);
	src->used = true;
	Solution *sol = findOrInsertSolution(src, target);

        sol->dest_addr = ip4addr_t(fields[2 + offset]);
        sol->path_len = atoi(fields[4 + offset]) - 1;
        sol->cksum = atoi(fields[5 + offset]);
	if (fixed) {
	    if (src->preferred || target->meth == MidarUtils::METH_INDIR) {
		sol->target_hop = target->probe_ttl - 1; // use value found in Estimation
		sol->age = 0;
		sol->value = 1;
		target->bestSol = sol;
		if (!src->preferred) {
		    n_forcedTargets++;
		}
	    }
	} else {
	    sol->target_hop = atoi(fields[4]) - 1;
	    sol->age = now - atoi(fields[0]);
	    sol->value = Solution::calculate_value(sol->target_hop, sol->path_len, sol->age);
	}
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    out_log << "# total targets: " << targets.size() << endl;
    memoryInfo.print("load solutions");
}

static void loadMonitors(const string &filename)
{
    out_log << "# loadMonitors: " << filename << endl;

    char buf[65536];
    InFile in(filename.c_str());
    while (in.gets(buf, sizeof(buf))) {
      try {
        if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty

	// format:  <name> <address>
        MidarUtils::FieldSplitter<2> fields(buf, " \t\n");

        ip4addr_t source_addr(fields[1]);
	Source *source = findOrInsertSource(source_addr);
	source->preferred = true;
	// source->name = new string(fields[0]); // monitor name
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    memoryInfo.print("load monitors");
}

static void loadAddrs(const string &filename, const string &label)
{
    out_log << "# loadAddrs: " << filename << ", " << label << endl;

    char buf[65536];
    InFile in(filename.c_str());
    while (in.gets(buf, sizeof(buf))) {
      try {
        if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty

        MidarUtils::FieldSplitter<1> fields(buf, " \t\n");

        ip4addr_t target_addr(fields[0]);
	if (isBogus(target_addr)) {
	    out_log << "# loadAddrs: skipping bogon " << target_addr << endl;
	    continue;
	}
	findOrInsertTarget(target_addr);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    memoryInfo.print("load addrs");
}

static void loadEstimation(const string &filename)
{
    out_log << "# loadEstimation: " << filename << endl;

    char buf[65536];
    InFile in(filename.c_str());
    while (in.gets(buf, sizeof(buf))) {
      try {
        if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty

        MidarUtils::FieldSplitter<14> fields(buf, " \t");
        const char *next_field = strtok(0, ";"); // classification

	int classification = 0;
	while (strcmp(next_field, MidarUtils::class_label[classification]) != 0) {
	    if (++classification == MidarUtils::N_CLASS) {
		throw std::runtime_error(string("bad classification \"") +
		    next_field + "\"");
	    }
	}
	if (classification != MidarUtils::CLASS_MONOTONIC) continue;

        ip4addr_t target_addr(fields[0]);
	Target *target;
	if (limitedTargets) {
	    target = findTarget(target_addr);
	    if (!target) continue;
	} else {
	    target = findOrInsertTarget(target_addr);
	}
	int ttl_expansion; // not used
	int meth;
	bool preferred = !!strchr(fields[1], '*');
	MidarUtils::parse_meth(fields[1], meth, ttl_expansion);

	// If this method is starred, use it.
	// (estimation-analyze marks the preferred method with a star.)
	// If none are starred (old versions of estimation-analyze),
	// we prefer the lowest numbered method.
	if (target->meth < 0 || preferred ||
	    (!target->preferred && meth < target->meth))
	{
	    target->meth = meth;
	    target->preferred = preferred;
	    target->velocity = atof(fields[10]);
	    target->probe_ttl = atoi(fields[11]);
	    target->order = atoi(fields[13]);
	}
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    memoryInfo.print("load estimation");
}

void myofstream::open(const string &suffix, bool tmp)
{
    name = string(cfg.output_basename ? cfg.output_basename :
	cfg.extract ? "extract" : "targets") + suffix;
    if (tmp) {
	tmpname = name;
	size_t i = tmpname.rfind("/");
	if (i == string::npos) i = 0; else i++;
	tmpname.replace(i, 0, ".tmp."); // leading "." prevents globbing
    }
    ofstream::open(tmp ? tmpname.c_str() : name.c_str());
    if (!*this) {
	cerr << "can't open " << name << ": " << strerror(errno) << endl;
	exit(1);
    }

    *this << "# version: " << cvsID << endl;
    *this << "# version: " << MidarUtils::cvsID << endl;
    *this << "# version: " << PathLoader::cvsID << endl;
    char timebuf[80];
    struct tm *tm = localtime(&cfg.start_time);
    strftime(timebuf, sizeof(timebuf), " (%F %T %Z)", tm);
    *this << "# start time: " << cfg.start_time << timebuf << endl;
    *this << "# command line: " << cmdline << endl;
    *this << endl << "#" << endl;
}

static void printSolutionHeader(ofstream &out, bool all = false)
{
    if (all)
	out << "# " << setw(8) << "time" << "   ";
    else
	out << "# ";
    out << setw(14) << "target";
    out << ' ' << setw(15) << "src";
    out << ' ' << setw(15) << "dst";
    out << ' ' << setw(3) << "hop";
    out << ' ' << setw(3) << "trm";
    out << ' ' << setw(5) << "cksum";
    if (!cfg.estimationFiles.empty()) {
	out << ' ' << setw(8) << "meth";
	out << ' ' << setw(11) << "velocity";
	out << ' ' << setw(4) << "ittl";
	out << ' ' << setw(2) << "LE";
    }
    out << endl;
}

static void printSolutionRow(ofstream &out, Solution *sol, bool all = false)
{
    if (all)
	out << setw(10) << (now - sol->age) << ' ';
    out << setw(16) << sol->target->addr;
    out << ' ' << setw(15) << sol->source->addr;
    out << ' ' << setw(15) << sol->dest_addr;
    out << ' ' << setw(3) << sol->target_hop + 1;
    out << ' ' << setw(3) << sol->path_len + 1;
    out << ' ' << setw(5) << sol->cksum;
    if (!cfg.estimationFiles.empty()) {
	out << ' ' << setw(8) << MidarUtils::methname[sol->target->meth];
	out << ' ' << setw(11) << MidarUtils::fmtfixed(sol->target->velocity, 6);
	out << ' ' << setw(4) << int(sol->target->ittl);
	out << ' ' << setw(2) << int(sol->target->order);
    }
    out << '\n';
}

static void printAllSolutions()
{
    SolutionSet::iterator si;
    SourceSet::iterator mi;
    TargetVec::iterator ti;

    myofstream out_solutions;
    out_solutions.open(".solutions", true);
    printSolutionHeader(out_solutions, true);
    for (si = solutions.begin(); si != solutions.end(); ++si) {
	printSolutionRow(out_solutions, (*si), true);
    }
    out_solutions.close();
    memoryInfo.print("printed solns");
}

struct GroupKey {
    int8_t meth;
    uint8_t ittl;
    uint8_t order;
    uint8_t prefix;
    GroupKey(int8_t _meth, uint8_t _ittl, uint8_t _order, uint8_t _prefix) :
	meth(_meth), ittl(_ittl), order(_order), prefix(_prefix) {}
};

bool operator<(const GroupKey &a, const GroupKey &b)
{
    return
	a.meth != b.meth ? a.meth < b.meth :
	a.ittl != b.ittl ? a.ittl < b.ittl :
	a.order != b.order ? a.order < b.order :
	a.prefix < b.prefix;
}

typedef map<GroupKey, TargetVec> GroupMap;
GroupMap groups;

static void selectSolutionsHelper(TargetVec &unsolvedTargets, bool byVelocity)
{
    TargetVec::iterator ti;
    double unconstrained_min_value = 0;
    double unconstrained_max_value = 0;
    double actual_value = 0;
    vector<long> histogram(sources.size() + 1);
    long n_targets = unsolvedTargets.size() - n_forcedTargets;
    double quota = n_targets / n_preferredMonitors;
    vector<double> bucket_velocity;
    vector<double> bucket_size;

    // We will walk targets ordered by number of solutions, so we can assign
    // the most constrained targets first.
    // TODO: assign all INDIRs first (by number of solutions), then all
    // non-INDIRs (by number of solutions).
    sort(unsolvedTargets.begin(), unsolvedTargets.end(), target_solutions_less);

    if (byVelocity && sources.size() > 1) {
	// Divide vTargets (targets sorted by velocity) into buckets whose
	// size depends on velocity.  Each source will be assigned its share
	// from each bucket, to ensure that sources have similar velocity
	// distributions.
	size_t i = 0, last_i = 0;
	double v = (*vTargets)[i]->velocity;
	while (true) {
	    i += (size_t)(n_preferredMonitors*30000/(v+50));
	    if (i >= vTargets->size()) break;
	    v = (*vTargets)[i-1]->velocity;
	    bucket_velocity.push_back(v);
	    bucket_size.push_back(i - last_i);
	    out_log << "# bucket " << (bucket_velocity.size() - 1) <<
		": size=" << bucket_size.back() <<
		", v=" << bucket_velocity.back() << endl;
	    last_i = i;
	}
	bucket_velocity.push_back(0); // last bucket
	bucket_size.push_back(vTargets->size() - last_i); // last bucket
	out_log << "# bucket " << (bucket_velocity.size() - 1) <<
	    ": size=" << bucket_size.back() <<
	    ", v=" << bucket_velocity.back() << endl;
    } else {
	// only one bucket
	bucket_velocity.push_back(-2);
	bucket_size.push_back(vTargets->size() / n_preferredMonitors);
    }

    for (SourceSet::iterator mi = sources.begin(); mi != sources.end(); ++mi) {
	(*mi)->n_targets = 0;
	(*mi)->bucketed_targets.insert((*mi)->bucketed_targets.end(),
	    bucket_velocity.size(), 0);
    }

    memoryInfo.print("helper setup");

    for (ti = unsolvedTargets.begin(); ti != unsolvedTargets.end(); ++ti) {
	double min_value = DBL_MAX;
	double max_value = 0;
	int i;
	SolutionSet::iterator si;
	for (i = 0, si = (*ti)->firstSol; i < (*ti)->n_solutions; ++i, ++si) {
	    if ((*si)->value < min_value)
		min_value = (*si)->value;
	    if ((*si)->value > max_value)
		max_value = (*si)->value;
	}
	unconstrained_min_value += min_value;
	unconstrained_max_value += max_value;
	histogram[(*ti)->n_solutions]++;
	Solution *best = 0;

	// find the bucket for this target
	size_t bucket;
	for (bucket = 0; bucket < bucket_velocity.size(); bucket++) {
	    if ((*ti)->velocity >= bucket_velocity[bucket])
		break;
	}

	// pick a preliminary solution for this target
	for (i = 0, si = (*ti)->firstSol; i < (*ti)->n_solutions; ++i, ++si) {
	    if (!best) {
		best = *si;
	    } else if (best->source->preferred != (*si)->source->preferred) {
		// one source is preferred, the other isn't.
		if ((*si)->source->preferred) {
		    best = *si;
		}
	    } else if (best->source->bucketed_targets[bucket] * n_preferredMonitors >= bucket_size[bucket]) {
		// best's src is over quota
		if ((*si)->source->bucketed_targets[bucket] < best->source->bucketed_targets[bucket]) {
		    // si's src is less full
		    best = *si;
		}
	    } else if ((*si)->source->bucketed_targets[bucket] * n_preferredMonitors < bucket_size[bucket]) {
		// best's src and si's src are both under quota.
#if 1
		if ((*si)->value * (bucket_size[bucket] - n_preferredMonitors * (*si)->source->bucketed_targets[bucket]) >
		    best->value * (bucket_size[bucket] - n_preferredMonitors * best->source->bucketed_targets[bucket]))
		{
		    // si is better (when their values are scaled by fraction of quota remaining).
		    best = *si;
		}
#else
		if ((*si)->source->bucketed_targets[bucket] < best->source->bucketed_targets[bucket]) {
		    // si's src is less full
		    best = *si;
		}
#endif
	    }
	}

	// move the target if necessary
	if ((*ti)->meth != MidarUtils::METH_INDIR) {
	    // Target isn't INDIR, so it can be assigned to any source.
	    if (!best || !best->source->preferred ||
		best->source->bucketed_targets[bucket] * n_preferredMonitors >= bucket_size[bucket])
	    {
		// There was no best source, or the best source is
		// non-preferred, or the best source is over quota.
		// Move the target to a source with the smallest bucket, even
		// if that source never saw the target.
		// TODO: don't do this here; do it after the ti loop, and for
		// the best-over-quota case, start with the lowest-valued
		// solutions.
		long bestTargets = best && best->source->preferred ?
		    best->source->bucketed_targets[bucket] : unsolvedTargets.size();
		Source *bestSource = 0;

		for (SourceSet::iterator mi = sources.begin(); mi != sources.end(); ++mi) {
		    if ((*mi)->preferred &&
			(*mi)->bucketed_targets[bucket] < bestTargets)
		    {
			bestTargets = (*mi)->bucketed_targets[bucket];
			bestSource = *mi;
		    }
		}

		if (bestSource) {
		    // generate a zero-value solution
		    best = new Solution(bestSource, (*ti));
		    best->value = 0;
		}
	    }
	}

	if (best) {
	    best->source->bucketed_targets[bucket]++;
	    best->source->n_targets++;
	    if ((*ti)->n_solutions == 1 && (*ti)->meth == MidarUtils::METH_INDIR)
		best->source->n_stuck++;
	    actual_value += best->value;
	    (*ti)->bestSol = best;
	}
    }

    memoryInfo.print("selected solns");

    out_log << "### bucket data:" << endl;
    for (size_t b = 0; b < bucket_velocity.size(); b++) {
	out_log << "# " << setw(3) << b << setw(9) << MidarUtils::fmtfixed(bucket_velocity[b], 3) << setw(6) << bucket_size[b];
	for (SourceSet::iterator mi = sources.begin(); mi != sources.end(); ++mi) {
	    out_log << ' ' << setw(5) << (*mi)->bucketed_targets[b];
	}
	out_log << endl;
    }
    out_log << "###" << endl;

    out_log << "# targets:                 " << setw(10) << unsolvedTargets.size() << endl;
    out_log << "# sources:                 " << setw(10) << sources.size() << endl;
    out_log << "# unconstrained_min_value: " << setw(10) << fixed << unconstrained_min_value << endl;
    out_log << "# unconstrained_max_value: " << setw(10) << fixed << unconstrained_max_value << endl;
    out_log << "# actual_value:            " << setw(10) << fixed << actual_value << endl;
    out_log << "# Num. of potential solutions vs. num. of targets with that many solutions:" << endl;
    for (unsigned i = 0; i < histogram.size(); i++) {
	out_log << "# " << setw(3) << i << " " << setw(8) << histogram[i] << endl;
    }

    out_log << "# quota: " << quota << endl;
    out_log << "# " << setw(15) << "Targets:" << " " << setw(7) << "stuck" << setw(7) << "actual";
    if (cfg.point_to_point_mates)
	out_log << setw(7) << "mates" << setw(7) << "total";
    out_log << endl;
    for (SourceSet::iterator mi = sources.begin(); mi != sources.end(); ++mi) {
	out_log << "# " << setw(15) << (*mi)->getname() << " " << setw(7) << (*mi)->n_stuck << setw(7) << ((*mi)->n_targets - (*mi)->n_mates);
	if (cfg.point_to_point_mates)
	    out_log << setw(7) << (*mi)->n_mates << setw(7) << (*mi)->n_targets;
	out_log << endl;
    }

    memoryInfo.print("printed stats");
}

static void selectSolutions()
{
    SolutionSet::iterator si;
    SourceSet::iterator mi;

    // eliminate unused sources
    for (mi = sources.begin(); mi != sources.end(); ) {
	SourceSet::iterator next = mi;
	++next;
	if (!(*mi)->used && !(*mi)->preferred)
	    sources.erase(mi);
	mi = next;
    }

    // eliminate targets that exceed max_velocity
    if (!cfg.estimationFiles.empty() && cfg.max_velocity > 0) {
	for (TargetSet::iterator tsi = targets.begin(); tsi != targets.end(); ) {
	    TargetSet::iterator next = tsi;
	    ++next;
	    if ((*tsi)->velocity > cfg.max_velocity) {
		targets.erase(tsi);
	    }
	    tsi = next;
	}
    }

    // match up solutions with targets
    {
	Target *target = 0;
	for (si = solutions.begin(); si != solutions.end(); ++si) {
	    if (target != (*si)->target) {
		target = static_cast<Target*>((*si)->target);
		target->firstSol = si;
	    }
	    target->n_solutions++;
	}
    }

    // eliminate INDIR targets without solutions
    if (!cfg.estimationFiles.empty()) {
	long targets_without_solutions = 0;
	for (TargetSet::iterator tsi = targets.begin(); tsi != targets.end(); ) {
	    TargetSet::iterator next = tsi;
	    ++next;
	    if (!(*tsi)->n_solutions && (*tsi)->meth == MidarUtils::METH_INDIR) {
		targets.erase(tsi);
		targets_without_solutions++;
	    }
	    tsi = next;
	}
	if (targets_without_solutions)
	    out_log << "# WARNING: discarded " << targets_without_solutions <<
		" targets without solutions" << endl;
    }
    memoryInfo.print("selection setup");

    if (cfg.groupsize) {
	for (TargetSet::iterator ti = targets.begin(); ti != targets.end(); ++ti) {
	    GroupKey key((*ti)->meth, (*ti)->ittl, (*ti)->order, (*ti)->addr >> 24);
	    groups[key].push_back(*ti);
	}
	// merge small groups into <groupsize> groups
	for (GroupMap::iterator gi = groups.begin(); gi != groups.end(); ++gi) {
	    GroupMap::iterator nextgroup = gi;
	    ++nextgroup;
	    while (nextgroup != groups.end() && ((*gi).second.size() + (*nextgroup).second.size() <= cfg.groupsize)) {
		out_log << "# merging groups sizes " << (*gi).second.size() << " and " << (*nextgroup).second.size();
		// append nextgroup's list to gi's list
		(*gi).second.insert((*gi).second.end(), (*nextgroup).second.begin(), (*nextgroup).second.end());
		out_log << ": " << (*gi).second.size() << endl;
		// erase nextgroup
		groups.erase(nextgroup);
		nextgroup = gi;
		++nextgroup;
	    }
	}
	memoryInfo.print("merged groups");
	long groupnum = 0;
	for (GroupMap::iterator gi = groups.begin(); gi != groups.end(); ++gi) {
	    out_log << "###### group " << groupnum << endl;
	    selectSolutionsHelper((*gi).second, false);
	    groupnum++;
	}

    } else {
	TargetVec unsolvedTargets(targets.begin(), targets.end());
	selectSolutionsHelper(unsolvedTargets, !cfg.estimationFiles.empty());
    }
}

static void generateMate(Target *target, ip4addr_t mateAddr)
{
    Target *mate = findOrInsertTarget(mateAddr);
    if (mate->n_solutions > 0) return; // already exists

    SolutionSet::iterator si;
    Source *bestSource = 0;
    int i;
    for (i = 0, si = target->firstSol; i < target->n_solutions; ++i, ++si) {
	if (!bestSource || bestSource->n_targets > (*si)->source->n_targets) {
	    bestSource = (*si)->source;
	}
    }
    Solution *solution = new Solution(bestSource, mate);
    mate->firstSol = solutions.insert(solution);
    mate->bestSol = *mate->firstSol;
    mate->n_solutions = 1;
    bestSource->n_targets++;
    bestSource->n_mates++;
    mate->isMate = true;
    debug_log << "# generated mate " << mate->addr << " for " << target->addr << "\n";
}

static void outOfMemory()
{
    out_log << "# OUT OF MEMORY" << endl;
    out_log << "# traces=" << pathLoader.n_good_traces << "/" << pathLoader.n_raw_traces <<
	" loops=" << pathLoader.n_loops <<
	" discarded=" << pathLoader.n_discarded_traces <<
	" solutions=" << solutions.size() <<
	endl;
    abort();
}

static void usageExit(const char *name, const char *badoption, int status) {
    if (badoption)
	cerr << "invalid option " << badoption << endl;
    cerr << endl;
    cerr << name << " [behavior-options] [file-options]" << endl;
    cerr << "Extract addreses from paths and assign to monitors for MIDAR." << endl;
    cerr << endl;
    cerr << "Behavior options:" << endl;
    cerr << "-x       do not assign targets to monitors; output best solution per monitor" << endl;
    cerr << "-X       do not assign targets to monitors; output ALL solutions" << endl;
//    cerr << "-m       also generate targets for point-to-point-mates" << endl;
    cerr << "-s       randomly shuffle output" << endl;
    cerr << "-g<size> group targets by similarity into groups of size <size> (requires -E)" << endl;
    cerr << "-v<vel>  ignore targets with velocity greater than <vel> (requires -E)" << endl;
    cerr << "-u       allow UDP targets to be moved to monitors that never saw them" << endl;
    cerr << "-l<arg>  how to handle loops in traces:" << endl;
    cerr << "    d    discard the entire trace" << endl;
    cerr << "    b    use only the part before the loop (default)" << endl;
    cerr << "    ba   use the parts before and after the loop (like APAR.c)" << endl;
    cerr << "-O <outfile>" << endl;
    cerr << "         The base name for result output files (default: \"extract\" if -x, " << endl;
    cerr << "         otherwise \"targets\")" << endl;
    cerr << endl;
    cerr << "File options:  each is an option followed by a list of filenames." << endl;
    cerr << "Files whose names end in \".gz\" or \".bz2\" will be automatically uncompressed." << endl;
    cerr << "Any filename preceeded by \"@\" (not part of the name) will be interpreted" << endl;
    cerr << "as a file containing a list of filenames, one per line." << endl;
    cerr << "-B <bogonfile>..." << endl;
    cerr << "   Text files containing bogon prefixes, one per line, in CIDR notation" << endl;
    cerr << "   (e.g., from http://www.cymru.com/Documents/bogon-bn-agg.txt)" << endl;
    cerr << "-P <pathfile>..." << endl;
    cerr << "   Files containing path traces: \"*.warts\" for warts files; \"trace.out.*\" for" << endl;
    cerr << "   iPlane files; otherwise text.  In text files, each trace starts with a \"#\"" << endl;
    cerr << "   line, followed by one hop address per line." << endl;
    cerr << "-S <solutionfile>..." << endl;
    cerr << "   Load files containing solutions previously extracted with -x -P" << endl;
    cerr << "-E <targetsummaryfile>" << endl;
    cerr << "   target-summary.txt from analysis of previous run.  If given, -P and -S will" << endl;
    cerr << "   ignore addresses not listed as monotonic in the file." << endl;
    cerr << "-T <targetfile>..." << endl;
    cerr << "   Target files used for estimation probe run." << endl;
    cerr << "-I <includefile>..." << endl;
    cerr << "   Include only addresses listed in <includefile>." << endl;
    cerr << "-A <addrfile>..." << endl;
    cerr << "   Load additional (non-traceroute) target addrs from <addrfile>" << endl;
    cerr << "-M <monitorfile>..." << endl;
    cerr << "   Assign targets only to monitors listed in <monitorfile>" << endl;
    cerr << "   (format: \"<name> <addr>\")." << endl;

    cerr << endl;
    cerr << "Normal usage:" << endl;
    cerr << "Generate intermediate solution files:" << endl;
    cerr << "  " << name << " [-B <bogons>] -x -P <paths> -O sol" << endl;
    cerr << "Generate targets for estimation run:" << endl;
    cerr << "  " << name << " -s [-m] [-M monitors] -S sol.solutions -O est" << endl;
    cerr << "Generate targets for sliding window resolution run:" << endl;
    cerr << "  " << name << " [-M monitors] -S sol.solutions -T estimation-targets/*.targets -E estimation-analysis/target-summary.txt -O win" << endl;
    cerr << "  " << "Then run sliding-window.pl scheduler." << endl;
    cerr << "Generate targets for grouped resolution run:" << endl;
    cerr << "  " << name << " [-v <max>] -g<groupsize> [-M monitors] -S sol.solutions -T estimation-targets/*.targets -E estimation-analysis/target-summary.txt -O grp" << endl;
    cerr << endl;
    exit(status);
}

int main(int argc, char *argv[])
{
    for (int i = 0; i < argc; i++) {
        cmdline += argv[i];
        cmdline += ' ';
    }

  try {
    set_new_handler(outOfMemory);
    time(&cfg.start_time);
    pathLoader.handler = &pathLoaderHandler;
    pathLoader.raw = false;
    pathLoader.include_src = false;
    pathLoader.include_dst = false;

    // default options
    cfg.filetype = 0;
    // -O targets
    cfg.output_basename = 0;

// allow "-xarg" or "-x arg"
#define get_optarg()  ( argv[optind][2] ? argv[optind] + 2 : \
    optind+1 < argc ? argv[++optind] : 0 )

    for (optind = 1; optind < argc; ++optind) {
	if (argv[optind][0] == '-') {
	    switch (argv[optind][1]) {
	    case 'O':
		optarg = get_optarg();
		cfg.output_basename = strdup(optarg);
		break;
	    case 'x':
		cfg.extract = true;
		break;
	    case 'X':
		cfg.extract_all = cfg.extract = true;
		break;
	    case 's':
		cfg.shuffle = true;
		break;
	    case 'u':
		cfg.movable_udp = true;
		break;
#if 0 // code rot
	    case 'm':
		cfg.point_to_point_mates = true;
		break;
#endif
	    case 'g':
		optarg = get_optarg();
		cfg.groupsize = atoi(optarg);
		if (cfg.groupsize == 0)
		    usageExit(argv[0], 0, 1);
		break;
	    case 'v':
		optarg = get_optarg();
		cfg.max_velocity = atof(optarg);
		if (cfg.max_velocity == 0)
		    usageExit(argv[0], 0, 1);
		break;
	    case 'l':
		optarg = get_optarg();
		if (strcmp(optarg, "d") == 0) {
		    pathLoader.loop_discard = true;
		} else if (strcmp(optarg, "b") == 0) {
		    pathLoader.loop_discard = false;
		    pathLoader.loop_after = false;
		} else if (strcmp(optarg, "ba") == 0) {
		    pathLoader.loop_discard = false;
		    pathLoader.loop_after = true;
		} else {
		    usageExit(argv[0], argv[optind], 1);
		}
		break;
	    case 'B': case 'P': case 'S': case 'E': case 'T': case 'I':
	    case 'A': case 'M':
		cfg.filetype = argv[optind][1];
		if (argv[optind][2]) // allow "-Xfile" without space
		    if (!cfg.setFile(argv[optind]+2))
			usageExit(argv[0], 0, 1);
		break;
	    default:
		usageExit(argv[0], argv[optind], 1);
	    }
	} else {
	    // allow "-X... file ..."
	    if (!cfg.setFile(argv[optind]))
		usageExit(argv[0], 0, 1);
	}
    }

    if (cfg.shuffle && (cfg.extract || cfg.groupsize)) {
	cerr << "-s option is not compatible with -g or -x\n";
	exit(1);
    }

    if ((cfg.groupsize || cfg.max_velocity > 0) && cfg.estimationFiles.empty()) {
	cerr << "-g and -v options require -E" << endl;
	exit(1);
    }

    if (cfg.targetFiles.empty() != cfg.estimationFiles.empty()) {
	cerr << "-T and -E must be both present or both absent" << endl;
	exit(1);
    }

    if (!cfg.includeFiles.empty() && !cfg.estimationFiles.empty()) {
	cerr << "-I and -E are not compatible" << endl; // ???
	exit(1);
    }

    cerr << "warning: assuming icmp cksum values are byte-swapped.\n";

    out_log.open(".log");

    memoryInfo.print("startup");

    // load bogons
    bogons.installStdBogons();
    out_log << "# loaded " << bogons.size() << " bogons" << endl;
    for (unsigned i = 0; i < cfg.bogonFiles.size(); ++i) {
	out_log << "# loadBogons: " << cfg.bogonFiles[i] << endl;
	bogons.load(cfg.bogonFiles[i]);
	out_log << "# loaded " << bogons.size() << " bogons" << endl;
    }
#if 0
    NetPrefixSet::const_iterator bit;
    out_log << "# Bogons:" << endl;
    for (bit = bogons.begin(); bit != bogons.end(); ++bit) {
	out_log << "# " << *bit << endl;
    }
#endif
    memoryInfo.print("load bogons");

    time(&now);

    // load monitor addresses
    if (!cfg.extract) {
	for (unsigned i = 0; i < cfg.monitorFiles.size(); ++i) {
	    loadMonitors(cfg.monitorFiles[i]);
	    explicitMonitors = true;
	    n_preferredMonitors = sources.size();
	}
    }

    // load include data
    for (unsigned i = 0; i < cfg.includeFiles.size(); ++i) {
	loadAddrs(cfg.includeFiles[i], "Include");
	limitedTargets = true;
    }

    // load estimation data
    for (unsigned i = 0; i < cfg.estimationFiles.size(); ++i) {
	loadEstimation(cfg.estimationFiles[i]);
	limitedTargets = true;
    }

    // load indir/udp targeting info
    for (unsigned i = 0; i < cfg.targetFiles.size(); ++i) {
	loadSolutions(cfg.targetFiles[i], true);
    }

    // load path traces
    for (unsigned i = 0; i < cfg.traceFiles.size(); ++i) {
	loadTraces(cfg.traceFiles[i]);
    }

    // load solutions
    for (unsigned i = 0; i < cfg.solutionFiles.size(); ++i) {
	loadSolutions(cfg.solutionFiles[i], false);
    }

    if (!explicitMonitors) {
	n_preferredMonitors = sources.size();
	if (n_preferredMonitors == 0) {
	    ip4addr_t source_addr("0.0.0.0");
	    findOrInsertSource(source_addr);
	    n_preferredMonitors = sources.size();
	}
	for (SourceSet::iterator mi = sources.begin(); mi != sources.end(); ++mi) {
	    (*mi)->preferred = true;
	}
    }

    // load additional addrs
    for (unsigned i = 0; i < cfg.addrFiles.size(); ++i) {
	loadAddrs(cfg.addrFiles[i], "Additional");
	out_log << "# total targets: " << targets.size() << endl;
    }

    if (cfg.extract) {
	printAllSolutions();
	memoryInfo.print("done");
	return 0;
    }

    // sort targets by descending velocity (if we have velocity data)
    vTargets = new TargetVec(targets.begin(), targets.end());
    if (!cfg.estimationFiles.empty())
	sort(vTargets->begin(), vTargets->end(), target_velocity_greater);
    memoryInfo.print("sort by velocity");

    cfg.groupsize *= sources.size();
    selectSolutions();

    if (cfg.point_to_point_mates) {
	for (TargetSet::iterator ti = targets.begin(); ti != targets.end(); ++ti) {
	    if ((*ti)->isMate) continue;
	    generateMate(*ti, ip4addr_t((*ti)->addr ^ 0x1)); // generate /31-mate
	    if (!observed(ip4addr_t(((*ti)->addr & ~0x3) | 0x0)) &&
		!observed(ip4addr_t(((*ti)->addr & ~0x3) | 0x3)))
	    {
		// neither of the /30's broadcast addresses were observed
		generateMate((*ti), ip4addr_t((*ti)->addr ^ 0x3)); // generate /30-mate
	    }
	}
	memoryInfo.print("generated mates");
    }

    // print targets

    if (cfg.groupsize) {
	for (SourceSet::iterator si = sources.begin(); si != sources.end(); ++si) {
	    int groupnum = 0;
	    for (GroupMap::iterator gi = groups.begin(); gi != groups.end(); ++gi) {
		ostringstream ost;
		ost << ".group-" << setw(4) << setfill('0') << groupnum << "." << (*si)->getname() << ".targets";

		myofstream out_targets;
		out_targets.open(ost.str(), true);
		printSolutionHeader(out_targets);
		sort((*gi).second.begin(), (*gi).second.end(), target_solutions_less);
		for (TargetVec::iterator ti = (*gi).second.begin(); ti != (*gi).second.end(); ++ti) {
		    if ((*si)->addr != (*ti)->bestSol->source->addr) continue;
		    printSolutionRow(out_targets, (*ti)->bestSol);
		}
		out_targets.close();

		groupnum++;
	    }
	}

    } else {
	if (cfg.shuffle) {
	    random_shuffle(vTargets->begin(), vTargets->end());
	    sort(vTargets->begin(), vTargets->end(), target_source_less);
	    memoryInfo.print("shuffle");
	} else {
	    sort(vTargets->begin(), vTargets->end(), target_source_less_velocity_greater);
	    memoryInfo.print("sort by source");
	}
	myofstream out_targets;
	Source *src = 0;
	for (TargetVec::iterator ti = vTargets->begin(); ti != vTargets->end(); ++ti) {
	    if (!src || (*ti)->bestSol->source->addr != src->addr) {
		if (src) out_targets.close(); // close old file
		src = (*ti)->bestSol->source;
		ostringstream ost;
		ost << "." << src->getname() << (src->preferred ? ".targets" : ".targets.omitted");
		out_targets.open(ost.str(), true); // open new file
		printSolutionHeader(out_targets);
	    }
	    printSolutionRow(out_targets, (*ti)->bestSol);
	}
	if (src) out_targets.close(); // close last file
    }

    memoryInfo.print("printed solns");

    memoryInfo.print("done");
    return 0;

  } catch (const std::exception &e) {
    cerr << e.what() << endl;
    exit(1);
  }
}
