/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../lib/config.h"
#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <assert.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <string>
#include <getopt.h>

using namespace std;

#include "../lib/infile.h"
#include "../lib/MemoryInfo.h"
#define NO_DEBUG_MEMORY 0
MemoryInfo memoryInfo;

#include "MidarUtils.h"
using namespace MidarUtils;

static const char *cvsID = "$Id: resolution-analyze.cc,v 1.138 2017/08/02 18:01:21 kkeys Exp $";
static string cmdline = "";
static const char *argv0;

static bool onlyknown = false;
static int n_threads = 1;
static const char *htmldir = 0;
static const char *targetsummaryfilename = 0;
// static const char *targetfilename = 0;
static const char *monitorfilename = 0;
static const char *includefilename = 0;
#ifdef ENABLE_FEATURES
static const char *featurefilename = 0;
#endif
static const char *outfilename = 0;
static const char *knownfilename = 0;
static const char *pairfilename = 0;
static const char *radargun_meth = 0;
static int alltest = 0;
static int res_flag = 1;
static const char *gptype = Gnuplotter::PNG;
static int gpscript = 1;
static int cmp_nonoverlap = 0;
static int write_option = RESTYPE_SHARED;
static BadclockSet badclocks;

static const double distdata_min = 0.01;
static const double distdata_offset = log(distdata_min);
static const double distdata_binsize = 0.02;
static const int distdata_nbins = int((log(32768) - distdata_offset) / distdata_binsize);
static int distdata_count = 0;
static OobDistribution oob_distribution;

static struct option longopts[] = {
    { "htmldir",       required_argument, 0,        'H' },
    { "targetsummary", required_argument, 0,        'S' },
// XXX    { "targets",  required_argument, 0,        'T' },
    { "output",        required_argument, 0,        'O' },
    { "include",       required_argument, 0,        'I' },
    { "monitors",      required_argument, 0,        'M' },
#ifdef ENABLE_FEATURES
    { "features",      required_argument, 0,        'F' },
#endif
    { "known",         required_argument, 0,        'K' },
    { "onlyknown",     required_argument, 0,        'k' },
    { "onlypairs",     required_argument, 0,        'p' },
    { "threads",       required_argument, 0,        'n' },
    { "overlap",       required_argument, 0,        'o' },
    { "oob-allowed",   required_argument, 0,        'a' },
    { "responsive",    required_argument, 0,        'r' },
    { "clockoffset",   required_argument, 0,        'c' },
    { "badclocks",     required_argument, 0,        'b' },
    { "radargun",      required_argument, 0,        'R' },
    { "alltest",       no_argument,       &alltest, 1 },
    { "resolution",    no_argument,       &res_flag, 1 },
    { "no-resolution", no_argument,       &res_flag, 0 },
    { "noresolution",  no_argument,       &res_flag, 0 },
    { "bits",          no_argument,       &bits_flag, 1 },
    { "no-bits",       no_argument,       &bits_flag, 0 },
    { "nobits",        no_argument,       &bits_flag, 0 },
    TTL_LONGOPT("no-ttl",     &ttl_flag,    0),
    TTL_LONGOPT("ttl",        &ttl_flag,    1),
    { "udp-diff",      no_argument,       &allow_udp_diff_addr, 1 },
    { "no-udp-diff",   no_argument,       &allow_udp_diff_addr, 0 },
    { "write-shared",  no_argument,       &write_option, RESTYPE_SHARED },
    { "write-maybe",   no_argument,       &write_option, RESTYPE_NOT_ENOUGH_INFO },
    { "write-all",     no_argument,       &write_option, RESTYPE_UNSHARED },
    VSIM_LONGOPT("vsim-none",     &vsim_option,    VSIM_NONE),
    VSIM_LONGOPT("vsim-nei",      &vsim_option,    VSIM_NEI),
    VSIM_LONGOPT("vsim-all",      &vsim_option,    VSIM_ALL),
    { "cmp-nonoverlap",no_argument,       &cmp_nonoverlap, 1 },
    DIST_LONGOPT("dist-none",   &dist_option, DIST_NONE),
    DIST_LONGOPT("dist-calc",   &dist_option, DIST_CALC),
    DIST_LONGOPT("dist-interp", &dist_option, DIST_INTERP),
    DIST_LONGOPT("dist-all",    &dist_option, DIST_ALL),
    { "gpscript",      no_argument,       &gpscript, 1 },
    { "no-gpscript",   no_argument,       &gpscript, 0 },
    { 0, 0, 0, 0 }
};

static ofstream outfile;
static Mutex outfilelock;
static ofstream html;
static int n_targets[N_METH];
static int n_probes[N_METH];
static int n_responses[N_METH];
#ifdef ENABLE_FEATURES
static int n_bad_feature_order[N_METH];
#endif
static uint64_t n_alias_pairs = 0; // known aliases
static uint64_t listed_pairs_not_compared = 0;
static vector<vector<Target*> > knownAliases;

class TargetPair {
    Target *data[2];
public:
    TargetPair() { data[0] = data[1] = 0; }
    Target*& operator[](size_t i) { return data[i]; }
};

static vector<TargetPair> pairs;

static set<ip4addr_t> includeaddrs;
static MultiMethTargetSet targets;

static MonitorSet monitors;
static vector<std::string> errormsgs;

void usageExit()
{
    cerr << "Usage: " << argv0 << " [<options>] <resolution_probe_result_file>...\n";
    cerr << "Options:\n";
    cerr << "--htmldir <dir>         write analysis graphs and webpage to directory <dir>\n";
    cerr << "--targetsummary <file>  write target summary to <file> (default: \n";
    cerr << "                        <htmldir>/target-summary.txt)\n";
    cerr << "--output <outfile>      write alias resolution results to <outfile>\n";
    cerr << "--write-shared          write only pairs that are SHARED (default)\n";
    cerr << "--write-maybe           write pairs that are SHARED or have not enough info\n";
    cerr << "--write-all             write all compared pairs\n";
#ifdef ENABLE_FEATURES
    cerr << "--features <file>       load estimation-analysis features from <file> \n";
#endif
    cerr << "--clockoffset <n>       maximum clock offset error, in ms (default " <<
	int(max_clock_offset * 1000) << ")\n";
    cerr << "--(no-)resolution       (do not) resolve aliases (default: on)\n";
    cerr << "--(no-)bits             (do not) apply bit precsion test (default: on)\n";
#if !ENABLE_TTL_OPTION
    cerr << "--ttl option requires recompiling with ENABLE_TTL_OPTION\n";
#endif
    cerr << "--(no-)ttl              (do not) use the DIFF_TTL test (default: off)\n";
    cerr << "--(no-)udp-diff         (do not) allow UDP response from different address\n";
    cerr << "                        (default: off)\n";
    cerr << "--badclocks             csv list of addrs of monitors with bad clocks\n";
    cerr << "--alltest               do not stop MBT after first failed point test\n";
    cerr << "                        (implies --write-all)\n";
#if !ENABLE_VSIM_OPTION
    cerr << "--vsim- options require recompiling with ENABLE_VSIM_OPTION\n";
#endif
    cerr << "--vsim-none             do not use VSIM test (default)\n";
    cerr << "--vsim-nei              use VSIM test if there's Not Enough Info for other tests\n";
    cerr << "--vsim-all              always use VSIM test, before other tests\n";
    cerr << "--cmp-nonoverlap        compare nonoverlapping targets (normally skipped for\n";
    cerr << "                        performance unless one of the --only* options is given)\n";
#if !ENABLE_DIST_OPTION
    cerr << "--dist- options require recompiling with ENABLE_DIST_OPTION\n";
#endif
    cerr << "--dist-none             do not use DIST test (default)\n";
    cerr << "--dist-calc             calculate dist, but do not use DIST test\n";
    cerr << "--dist-interp           use DIST test on interpolated points\n";
    cerr << "--dist-all              use DIST test on interpolated and extrapolated points\n";
    cerr << "--overlap <n>           require at least n overlapping points (default: 0)\n";
    cerr << "--oob-allowed <a>[,<b>[,<c>]]\n";
    cerr << "                        fraction of MBT point failures allowed for series to...\n";
    cerr << "                        <a>:  pass (default: 0.0)\n";
    cerr << "                        <b>:  pass, IF first & last points pass (default: <a>)\n";
    cerr << "                        <c>:  be inconclusive (default: <b>)\n";
    cerr << "--responsive <n>        treat >n responses as responsive, even if n<" <<
	(100 * mostly_thresh) << "% of tries\n";
    cerr << "                        (default: " << responsive_count << ")\n";
    cerr << "--no-gpscript           write gnuplot graphs instead of scripts\n";
    cerr << "--include <file>        include only target addresses listed in <file>\n";
    cerr << "--monitors <file>       load monitor names and addrs from <file>\n";
    cerr << "--known <file>          compare to known alias sets listed in <file>\n";
    cerr << "--onlyknown <file>      compare ONLY to known alias sets listed in <file>\n";
    cerr << "--onlypairs <file>      compare only alias pairs listed in <file>\n";
    cerr << "--radargun {tcp|udp|icmp}\n";
    cerr << "                        input is RadarGun probe data with given probe method\n";
#if HAVE_PTHREAD
    cerr << "--threads <n>           use <n> threads during alias resolution (1)\n";
#endif
    exit(1);
}

void loadIncludes(const char *filename)
{
    char buf[8192];
    InFile in(filename);
    cerr << "# loading include addresses from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	char *p = strchr(buf, '\n');
	if (p) *p = '\0';
	ip4addr_t addr(buf);
	includeaddrs.insert(addr);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << includeaddrs.size() << " addresses\n";
    memoryInfo.print("loaded includes");
}

#ifdef ENABLE_FEATURES
void loadFeatures(const char *filename)
{
    char buf[8192];
    long n = 0;
    InFile in(filename);
    cerr << "# loading feature data from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	const char *sep = " \t\n";
	const char *str;
	str = strtok(buf, sep); // address
	ip4addr_t addr(str);
	MultiMethTarget key(addr);
	MultiMethTargetSet::iterator it = targets.find(&key);
	if (it == targets.end()) {
	    continue;
	}
	(*it)->features = new Feature();
	int fid = 0;
	while ((str = strtok(0, sep))) {
	    if (fid >= N_FEATURES)
		throw std::runtime_error("too many feature values");
	    (*(*it)->features)[fid++] = atoi(str);
	}
	if (fid < N_FEATURES)
	    throw std::runtime_error("not enough feature values");
	n++;
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded features for " << n << " addresses\n";
    memoryInfo.print("loaded features");
}
#endif

void loadResponses(const char *filename)
{
    ResponseData rd;
    char buf[8192];
    bool finished = false;
    InFile in(filename);
    cerr << "# loading data from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (strncmp(buf, "## run finished ", 16) == 0) {
	    finished = true;
	}
	if (!buf[0] || buf[strlen(buf)-1] != '\n') {
	    cerr << "WARNING: input truncated at line " << in.linenum() << endl;
	    break;
	}
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	if (finished) {
	    std::cerr << "WARNING: ignoring text after \"run finished\", line "
		<< in.linenum() << endl;
	    break;
	}

	if (!radargun_meth) {
	    rd.parse(buf);
	} else {
	    rd.parseRadarGun(buf, radargun_meth);
	}
	if (rd.result == 'S') continue; // 'S' record is informational, not a real probe

	if (includefilename) {
	    if (includeaddrs.find(rd.target_addr) == includeaddrs.end())
		continue; // skip
	}

	int meth;
	parse_meth(rd.method, meth, rd.ttl_expansion);

	Monitor *monitor = monitors.countProbeFrom(rd.src_addr, meth);

	MultiMethTarget *mttarget = targets.findOrInsert(rd.target_addr);

	if (mttarget->meth < 0) {
	    mttarget->meth = meth;
	    mttarget->ptarget() = new Target(rd.target_addr);
	    mttarget->ptarget()->meth = meth;
	    n_targets[meth]++;
	} else if (mttarget->meth != meth) {
	    cerr << "WARNING: ignoring " << methname[meth] <<
		" response for " << mttarget->addr << " that already has " <<
		methname[mttarget->meth] << " response" << endl;
	    continue;
	}

	if (mttarget->ptarget()->probes == 0) {
	    mttarget->ptarget()->src_addr = rd.src_addr;
	} else if (mttarget->ptarget()->src_addr != rd.src_addr) {
	    // hack: if probed from multiple sources, set src_addr to 0.
	    mttarget->ptarget()->src_addr = ip4addr_t(0);
	}

	if (rd.ttl_expansion == 0)
	    mttarget->ptarget()->tries++;
	mttarget->ptarget()->probes++;
	n_probes[meth]++;
	if (rd.result != 'T' && rd.result != 'U' && rd.result != 'H')
	    continue;
	n_responses[meth]++;
	monitor->n_responses[meth]++;
	monitor->reply_ttl_histogram[meth][rd.reply_ttl]++;
	mttarget->ptarget()->processResponse(&rd);
	if (mttarget->ptarget()->probes == 1) {
	    monitor->time[meth].push_back((rd.resp_time + rd.probe_time) / 2);
	    monitor->id[meth].push_back(rd.reply_ipid);
	}
      } catch (const std::runtime_error &e) {
        if (!in.gets(buf, sizeof(buf))) { // error was in last line
            std::string msg(InFile::Error(in, e).what());
            msg += " (file appears to be truncated; ignoring error)";
            errormsgs.push_back(msg);
            cerr << msg << endl;
            break;
        } else {
            throw InFile::Error(in, e);
        }
      }
    }
    in.close();
    if (!finished)
	std::cerr << "WARNING: missing \"run finished\" line" << endl;
    memoryInfo.print("loaded responses");
}

void addAddrToSet(const char *str, long nodeid,
    long &n_observed_set_addrs, long &n_set_addrs)
{
    // cerr << "known address string: \"" << str << "\"\n";
    ip4addr_t addr(str);
    MultiMethTarget key(addr);
    MultiMethTargetSet::iterator it = targets.find(&key);
    if (it != targets.end() && (*it)->ptarget()->flags & MONOTONIC) {
	n_observed_set_addrs++;
	if ((*it)->ptarget()->known_nodeid) {
	    ostringstream oss;
	    oss << (*it)->addr << " already belongs to an alias set";
	    throw std::runtime_error(oss.str());
	}
	(*it)->ptarget()->known_nodeid = nodeid;
	if (onlyknown) {
	    knownAliases.back().push_back((*it)->ptarget());
	}
    }
    n_set_addrs++;
}

void loadKnownAliases(const char *filename)
{
    // Format can be either
    // "node 99: addr1 addr2 ... addrN\n", or
    // "# set 99: ...\n" followed by one or more address per line, or
    // "# name...\n" followed by one or more address per line
    enum { ADDR=1, NODE=2 };
    int fmt = 0;
    static long nodeid = 0;
    static long n_total_pairs = 0;
    static long n_observed_pairs = 0;
    long n_total_addrs = 0;
    long n_total_observed_addrs = 0;
    char buf[8192];
    InFile in(filename);
    cerr << "# loading known alias sets from " << filename << endl;
    long n_observed_set_addrs = 0;
    long n_set_addrs = 0;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (!fmt) {
	    if (strncmp(buf, "node ", 5) == 0) fmt = NODE;
	    else if (strncmp(buf, "# set", 5) == 0) fmt = ADDR;
	    else if (strncmp(buf, "# name", 6) == 0) fmt = ADDR;
	    else continue;
	}

	if (fmt == NODE) {
	    if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	    // start new set
	    nodeid++;
	    const char *sep = " \t\n";
	    const char *str;
	    char *p = buf;
	    if (strncmp(buf, "node ", 5) == 0) {
		p = strchr(buf+5, ':') + 1;
		while (isspace(*p)) p++;
	    }
	    n_observed_set_addrs = 0;
	    n_set_addrs = 0;
	    if (onlyknown) {
		if (knownAliases.empty() || !knownAliases.back().empty()) {
		    knownAliases.push_back(vector<Target*>());
		}
	    }
	    // add addrs to set
	    while ((str = strtok(p, sep))) {
		p = 0;
		addAddrToSet(str, nodeid, n_observed_set_addrs, n_set_addrs);
	    }
	    // close set
	    n_observed_pairs += choose2count(n_observed_set_addrs);
	    n_total_pairs += choose2count(n_set_addrs);
	    n_total_addrs += n_set_addrs;
	    n_total_observed_addrs += n_observed_set_addrs;
	    if (onlyknown && !knownAliases.empty() && knownAliases.back().size() < 2) {
		knownAliases.back().clear();
		// knownAliases.back() can be reused in the next set
	    }

	} else if (fmt == ADDR) {
	    if (strncmp(buf, "# set", 5) == 0 ||
		strncmp(buf, "# name", 6) == 0)
	    {
		// close previous set
		n_observed_pairs += choose2count(n_observed_set_addrs);
		n_total_pairs += choose2count(n_set_addrs);
		n_total_addrs += n_set_addrs;
		n_total_observed_addrs += n_observed_set_addrs;
		if (onlyknown && !knownAliases.empty() && knownAliases.back().size() < 2) {
		    knownAliases.back().clear();
		    // knownAliases.back() can be reused in the next set
		}
		// start new set
		nodeid++;
		n_observed_set_addrs = 0;
		n_set_addrs = 0;
		if (onlyknown) {
		    if (knownAliases.empty() || !knownAliases.back().empty()) {
			knownAliases.push_back(vector<Target*>());
		    }
		}
		continue;
	    }
	    if (buf[0] == '#') continue; // comment
	    const char *str;
	    char *p = buf;
	    while ((str = strtok(p, " \t\n"))) {
		p = 0;
		addAddrToSet(str, nodeid, n_observed_set_addrs, n_set_addrs);
	    }
	}
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    if (fmt == ADDR) {
	// close last set
	n_observed_pairs += choose2count(n_observed_set_addrs);
	n_total_pairs += choose2count(n_set_addrs);
	n_total_addrs += n_set_addrs;
	n_total_observed_addrs += n_observed_set_addrs;
    }
    if (onlyknown && knownAliases.back().size() < 2) {
	knownAliases.back().clear();
	knownAliases.pop_back();
    }
    cerr << "# loaded " << nodeid << " known alias sets\n";
    cerr << "#   " << n_total_pairs << " total pairs\n";
    cerr << "#   " << n_observed_pairs << " observed pairs\n";
    if (onlyknown)
	listed_pairs_not_compared = n_total_pairs - n_observed_pairs;
    n_alias_pairs = n_observed_pairs + listed_pairs_not_compared;
    memoryInfo.print("loaded aliases");
}

void loadPairs(const char *filename)
{
    static long n_total_pairs = 0;
    char buf[8192];
    InFile in(filename);
    cerr << "# loading pairs from " << filename << endl;
    TargetPair tp;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	const char *sep = " \t\n";
	const char *str;
	char *p = buf;
	tp[0] = tp[1] = 0;
	for (int i = 0; i < 2; i++) {
	    str = strtok(p, sep);
	    p = 0;
	    ip4addr_t addr(str);
	    MultiMethTarget key(addr);
	    MultiMethTargetSet::iterator it = targets.find(&key);
	    if (it != targets.end() && (*it)->ptarget()->flags & MONOTONIC) {
		tp[i] = (*it)->ptarget();
	    }
	}
	n_total_pairs++;
	if (tp[0] && tp[1]) {
	    pairs.push_back(tp);
	}
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << pairs.size() << " of " << n_total_pairs << " pairs\n";
    listed_pairs_not_compared = n_total_pairs - pairs.size();
    memoryInfo.print("loaded pairs");
}

void print_classification_row(ofstream &out, const char *label, int *counter)
{
    int total = 0;
    out << "<tr><td>" << label;
    for (int meth = 0; meth < N_METH; meth++) {
	if (!n_targets[meth]) continue;
	out << "<td colspan=2>" << htmlnum(counter[meth]) << " ";
	total += counter[meth];
    }
    out << "<td colspan=2>" << htmlnum(total) << "\n";
}

void plot_velocity(int flag)
{
    string label(
	(flag == MONOTONIC) ? "monotonic" :
	(flag == SHARED) ? "shared" :
	"?");
    // plot velocity
    if (htmldir) {
	html << "\n<h2>Velocity</h2>\n\n";
	vector<double> data[N_METH];
	gp.open("velocity-" + label,
	    "velocity distribution for " + label + " targets", gptype, gpscript);
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if (!((*it)->ptarget()->flags & flag)) continue;
	    data[(*it)->meth].push_back((*it)->ptarget()->series_velocity);
	}
	html << "<td><img src='velocity-" << label << ".png'>\n\n";
	fprintf(gp.f, "set logscale x\nset mxtics 10\n");
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'Velocity (ID/s)'\n");
	fprintf(gp.f, "set key top left\n");
	gp.plot_cdf(N_METH, data, methname);
	gp.close();
    }

    //  plot ids-per-probe
    if (htmldir) {
	html << "\n<h2>delta ID per probe</h2>\n\n";

	cerr << "# mean IDs per probe\n";
	gp.open("ids-per-probe-" + label,
	    "mean ids-per-probe distribution for " + label + " targets",
	    gptype, gpscript);
	vector<double> data[N_METH];
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if (!((*it)->ptarget()->flags & flag)) continue;
	    data[(*it)->meth].push_back(ids_per_probe((*it)->ptarget()->poi()->id));
	}
	html << "<br><img src='ids-per-probe-" << label << ".png'>\n\n";
	fprintf(gp.f, "set logscale x\nset mxtics 10\n");
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'IDs-per-probe'\n");
	fprintf(gp.f, "set key top left\n");
	gp.plot_cdf(N_METH, data, methname);

	cerr << "# median IDs per probe\n";
	gp.open("med-ids-per-probe-" + label,
	    "median ids-per-probe distribution for " + label + " targets",
	    gptype, gpscript);
	for (int meth = 0; meth < N_METH; meth++) data[meth].clear();
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if (!((*it)->ptarget()->flags & flag)) continue;
	    data[(*it)->meth].push_back(median_ids_per_probe((*it)->ptarget()->poi()->id));
	}
	html << "<br><img src='med-ids-per-probe-" << label << ".png'>\n\n";
	fprintf(gp.f, "set logscale x\nset mxtics 10\n");
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'IDs-per-probe'\n");
	fprintf(gp.f, "set key top left\n");
	gp.plot_cdf(N_METH, data, methname);
    }
}

void resolve();

int main(int argc, char *const *argv)
{
  try {
    // cerr.setprecision(6);
    // cerr.fixed();
    cerr << "# main" << endl;
    argv0 = argv[0];
    for (int i = 0; i < argc; i++) {
	cmdline += argv[i];
	cmdline += ' ';
    }

    InFile::fork = false;
    write_option = RESTYPE_SHARED;
    min_overlap_option = 0;

    int opt;
    int longidx;
    while ((opt = getopt_long(argc, argv, "", longopts, &longidx)) != -1) {
	switch (opt) {
	case 'H':
	    htmldir = strdup(optarg);
	    break;
	case 'S':
	    targetsummaryfilename = strdup(optarg);
	    break;
#if 0
	case 'T':
	    targetfilename = strdup(optarg);
	    break;
#endif
	case 'I':
	    includefilename = strdup(optarg);
	    break;
	case 'M':
	    monitorfilename = strdup(optarg);
	    break;
#ifdef ENABLE_FEATURES
	case 'F':
	    featurefilename = strdup(optarg);
	    break;
#endif
	case 'c':
	    max_clock_offset = atoi(optarg) / 1000.0;
	    break;
	case 'b':
	    badclocks.parse(optarg);
	    break;
	case 'O':
	    outfilename = strdup(optarg);
	    break;
	case 'k':
	    onlyknown = true;
	    // fall through
	case 'K':
	    knownfilename = strdup(optarg);
	    break;
	case 'p':
	    pairfilename = strdup(optarg);
	    break;
#if HAVE_PTHREAD
	case 'n':
	    n_threads = atoi(optarg);
	    break;
#endif
	case 'o':
	    min_overlap_option = atoi(optarg);
	    break;
	case 'a':
	    if (!Comparison::parse_oob_allowed(optarg))
		usageExit();
	    break;
	case 'r':
	    responsive_count = atoi(optarg);
	    break;
	case 'R':
	    radargun_meth = strdup(optarg);
	    break;
	case 0:
	    // flag
	    break;
	default:
	    usageExit();
	}
    }
    if (optind == argc) { usageExit(); }
    if (onlyknown && pairfilename) {
	cerr << "--onlyknown and --onlypairs are not compatible\n";
	exit(1);
    }
    if ((write_option <= RESTYPE_UNSHARED) && !onlyknown && !pairfilename) {
	cerr << "WARNING: --write-all without --onlyknown or --onlypairs "
	    "will write O(N^2) pairs\n";
    }

    if (dist_option) {
	distdata_count = knownfilename ? 6 : 4;
    }

    if (htmldir) {
	if (mkdir(htmldir, 0755) < 0 && errno != EEXIST) {
	    cerr << htmldir << ": " << strerror(errno) << endl;
	    exit(1);
	}
    }

    if (outfilename) {
	cerr << "# open outfile" << endl;
	outfile.open(outfilename);
	if (!outfile) {
	    cerr << "can't open " << outfilename << ": " << strerror(errno) << endl;
	    exit(1);
	}
	outfile << "# version: " << ::cvsID << endl;
	outfile << "# version: " << MidarUtils::cvsID << endl;
	outfile << "# command line: " << cmdline << endl << "#" << endl;
    }

    memoryInfo.print("startup");

    if (monitorfilename) monitors.load(monitorfilename);
    if (includefilename) loadIncludes(includefilename);

    // XXX TODO load target addrs

    for (int i = optind; i < argc; i++) {
	loadResponses(argv[i]);
    }

#ifdef ENABLE_FEATURES
    if (featurefilename) {
	loadFeatures(featurefilename);
    }
#endif

    cerr << "# analyzing" << endl;

    targetfilters.init();
    const int rows = targetfilters.size();
    const int cols = N_METH;

    Array2d<TargetStat> *targetstats = htmldir ? new Array2d<TargetStat>(rows, cols) : 0;

    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	classify((*it)->ptarget(), !alltest);
#ifdef ENABLE_FEATURES
	if ((*it)->ptarget()->order != ORDER_UNKNOWN && (*it)->ptarget()->features) {
	    int fid;
	    switch ((*it)->meth) {
		case METH_UDP: fid = UDP_IPID_ORDER; break;
		case METH_TCP: fid = TCP_IPID_ORDER; break;
		case METH_ICMP: fid = ICMP_IPID_ORDER; break;
		case METH_INDIR: fid = INDIR_IPID_ORDER; break;
		default: continue;
	    }
	    if ((*(*it)->ptarget()->features)[fid] != (*it)->ptarget()->order) {
		n_bad_feature_order[(*it)->meth]++;
		(*(*it)->ptarget()->features)[fid] = -1;
	    }
	}
#endif
	if (htmldir) {
	    for (int row = 0; row < rows; row++) {
		if ((*it)->ptarget() && targetfilters[row]((*it)->ptarget())) {
		    (*targetstats)(row, (*it)->meth).insert((*it)->ptarget());
		    (*targetstats)(row, (*it)->meth).count++;
		}
	    }
	}
    }
    memoryInfo.print("classified trgts");

    if (knownfilename) {
	// after classify() because we need to know which are monotonic
	loadKnownAliases(knownfilename);
    }

    if (pairfilename) {
	// after classify() because we need to know which are monotonic
	loadPairs(pairfilename);
	if (pairs.size() == 0) exit(1);
    }

    static ofstream summary;
    if (htmldir || targetsummaryfilename) {
	if (!targetsummaryfilename) {
	    targetsummaryfilename = strdup((string(htmldir) + "/target-summary.txt").c_str());
	}
	summary.open(targetsummaryfilename);
	if (!summary) {
	    cerr << "can't open " << targetsummaryfilename << ": " << strerror(errno) << endl;
	    exit(1);
	}
    }

    if (htmldir) {
	if (chdir(htmldir) < 0) {
	    cerr << htmldir << ": " << strerror(errno) << endl;
	    exit(1);
	}

	cerr << "# open index.html" << endl;
	html.open("index.html");
	if (!html) {
	    cerr << "can't open " << "index.html" << ": " << strerror(errno) << endl;
	    exit(1);
	}
	html << "<html>\n";
	html << "<head>\n";
	html << "<style>\n";
	html << ".bad { background:#FFBBBB; }\n";       // bad = red
	html << ".mediumbad { background:#FFDDBB; }\n"; // mediumbad = orange
	html << ".medium { background:#FFFFBB; }\n";    // medium = yellow
	html << ".good { background:#BBFFBB; }\n";      // good = green
	html << "td { text-align: right; }\n";
	html << "td.label { text-align: left; }\n";
	html << ".subgroup { font-size: small; }\n";
	html << ".subgroup .label { padding-left: 4ex; }\n";
	html << ".subgroup .label:before { content: \"-- \"; }\n";
	html << "td, th { border: 1px solid black; padding: 0.5ex; }\n";
	html << "td.percent { border-left-style: hidden; }\n";
	// Insert a thin space. This does not work in all browsers (notably IE),
	// but when it fails, it's harmless.
	html << ".numeric { white-space: nowrap; }\n";
	html << ".numeric span:after { content: \" \"; font-size: 50%; }\n";
	html << "</style>\n";
	html << "</head>\n";
	html << "<body>\n";
	html << "<h1>Resolution Analysis</h1>\n";

	monitors.sortByName();
	html << "\n<h2>The setup</h2>\n";
	html << "<p>Version: " << ::cvsID << "\n";
	html << "<p>Version: " << MidarUtils::cvsID << "\n";
	html << "<p>Command line: " << cmdline << "\n";
        if (!errormsgs.empty()) {
            html << "<p style='color: red;'>Errors:\n";
            html << "<ul style='color: red;'>\n";
            for (unsigned i = 0; i < errormsgs.size(); i++)
                html << "<li>" << errormsgs[i] << "\n";
            html << "</ul>\n";
        }
	html << "<p>Probed " << targets.size() << " router addresses from " <<
	    monitors.sizeByName() << " monitors.\n",
	// html << "<p>total time: %.2f s (%.2f hours)\n", $tot_time, $tot_time / 3600;
	html << "<p>Probe methods:\n";
	html << "<dl>\n";
	html << "<dt>tcp<dd>TCP SYN ACK sent to unopen port on target\n";
	html << "<dt>udp<dd>UDP packet sent to unopen port on target\n";
	html << "<dt>icmp<dd>ICMP Echo Request sent to target\n";
	html << "<dt>indir<dd>TTL-limited ICMP Echo Request sent to a destination <i>past</i> the target\n";
	html << "</dl>\n";
    }

    if (htmldir) {
	cerr << "# monitor stats\n";
	print_monitor_stats(html, monitors, n_targets);
	memoryInfo.print("per-monitor stats");
    }

    // print distribution of responses
    if (htmldir) {
	cerr << "# response distribution\n";
	html << "\n<h2>Response distribution</h2>\n";
	html << "Definitions:\n";
	html << "<ul>\n";
	html << "<li>An \"unresponsive\" target is one that responded to less " <<
	    "than " << (100*responsive_thresh) << "% of probes" << 
	    " and less than " << responsive_count << " probes total.\n";
	html << "<li>A \"nondegenerate\" response is one where the IP ID was not zero, " <<
	    "reflecting the probe, or a repeat of the previous response.\n";
	html << "<li>\"Mostly\" means more than " <<
	    (100*mostly_thresh) << "% of the responses for the target had the same classificiation.\n";
	html << "<li>An \"nonmonotonic\" target is one whose ID, on average, increases by more than " <<
	    (100*wraps_per_cycle_thresh) << "% of the ID space (i.e., " << int(default_wrapsize * wraps_per_cycle_thresh) <<
	    " for 16-bit counters) between probes.  " <<
	    "Such targets can not reliably be distinguished from targets with random ID counters.\n";
	html << "</ul>\n";

	html << "\n";
	html << "<p>\n";
	html << "<table style='border-collapse: collapse'>\n";
	html << "<tr><td>                 ";
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!n_targets[meth]) continue;
	    html << "<th colspan=2>" << methname[meth];
	}
	html << "<th colspan=2>" << "total";
	html << "</tr>\n";

	print_classification_row(html, "total targets", n_targets);
	print_classification_row(html, "probes",        n_probes);
	print_classification_row(html, "responses",     n_responses);

	for (int row = 0; row < rows; row++) {
	    // cerr << "# print_row: " << targetfilters[row].label << "\n";
	    html << "<tr " <<
		status_style[targetfilters[row].status] <<
		"><td class='label'>" << targetfilters[row].label << "\n";

	    uint64_t row_total = 0;
	    for (int col = 0; col < N_METH; col++) {
		if (!n_targets[col]) continue;
		string label(targetfilters[row].label);
		label += "-";
		label += methname[col];
		print_classification_cell(html, &(*targetstats)(row, col),
		    n_targets[col], label, gptype, gpscript);
		row_total += (*targetstats)(row, col).count;
	    }
	    html << "<td>" << htmlnum(row_total) << "<td class='percent'>" <<
		fmtpercent(row_total, targets.size()) << "\n";
	}
	delete targetstats; targetstats = 0; // no longer needed

#ifdef ENABLE_FEATURES
	if (featurefilename) {
	    html << "<tr><td>" << "bad feature order" << "     ";
	    int total = 0;
	    for (int meth = 0; meth < N_METH; meth++) {
		if (!n_targets[meth]) continue;
		html << "<td>" << htmlnum(n_bad_feature_order[meth]) <<
		    " <td class='percent'>" <<
		    fmtpercent(n_bad_feature_order[meth], n_targets[meth]);
		total += n_bad_feature_order[meth];
	    }
	    html << "<td>" << htmlnum(total) << " <td class='percent'>" <<
		fmtpercent(total, targets.size());
	    html << "</tr>\n";
	}
#endif

	html << "</table>\n";

	gp.open("responsiveness", "responsiveness", gptype, gpscript);
	vector<double> data[N_METH*2];
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    data[2 * (*it)->meth].push_back((*it)->ptarget()->expected / double((*it)->ptarget()->tries));
	    data[2 * (*it)->meth + 1].push_back((*it)->ptarget()->nondegenerate / double((*it)->ptarget()->tries));
	}

	html << "<br><img src='responsiveness.png'>\n\n";
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'Fraction of attempts'\n");
	fprintf(gp.f, "set key top left\n");

	string strlabels[N_METH*2];
	const char *labels[N_METH*2];
	for (int i = 0; i < N_METH; i++) {
	    strlabels[2*i] = string(methname[i]) + ": responses of expected type";
	    labels[2*i] = strlabels[2*i].c_str();
	    strlabels[2*i+1] = string(methname[i]) + ": nondegenerate";
	    labels[2*i+1] = strlabels[2*i+1].c_str();
	}
	gp.plot_cdf(N_METH * 2, data, labels);
	gp.close();
    }

    // initial ttl comparison
    if (htmldir) {
	init_ttl_comparison(html, targets, n_targets, gptype, gpscript);
    }

    // fwd/reverse path length comparison
    if (htmldir) {
	path_length_comparison(html, targets, gptype, gpscript);
    }

    //  print distribution of RTTs
    if (htmldir) {
	print_rtt_distribution(html, targets, gptype, gpscript);
    }

#if 0
    if (htmldir) {
	html << "\n<h2>Linearity</h2>\n\n";
	html << "<p>\n";

	gp.open("correlation", "correlation for nondegenerate targets", gptype, gpscript);
	vector<double> data[N_METH];
	for (int meth = 0; meth < N_METH; meth++) {
	    if (!n_targets[meth]) continue;
	    data[meth] = new vector<double>();
	}
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if (!((*it)->ptarget()->flags & GOOD_ID)) continue;
	    data[meth].push_back((*it)->ptarget()->correlation);
	}

	html << "<br><img src='correlation.png'>\n\n";
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'r^2\n");
	fprintf(gp.f, "set xrange [0:1]\n");
	fprintf(gp.f, "set key top left\n");
	gp.plot_cdf(data, methname);

	my $GP = open_gnuplot('correlation-monotonic', "correlation for nondegenerate monotonic targets", $gptype, $gpscript);
	my @data = ();
	while (my ($target_addr, $ref) = each %targets) {
	    if ($ref->[$i_monotonic_flag]) {
		push @{$data[$ref->[$i_meth]]}, $ref->[$i_correlation];
	    }
	}

	html << "<br><img src='correlation-monotonic.png'>\n\n";
	plot_cdf($GP, \@data, \@methname);
    }
#endif

#if 0
    // wrapping
    if (htmldir) {
	html << "\n<h2>Wrapping</h2>\n\n";
	html << "<p>A small number of negative ID deltas may indicate simple "
	    "wrapping that we can compensate for.  A large number of negative "
	    "deltas indicates either "
	    "random ID change, or wrapping too fast to distinguish from "
	    "random change.</p>\n\n";

	gp.open("wrapping", "wrapping", gptype, gpscript);
	vector<double> data[N_METH];
	for (TargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    // if (defined $ref->[$i_wrapfrac])
	    data[(*it)->meth].push_back((*it)->wrapfrac);
	}

	html << "<br><img src='wrapping.png'>\n\n";
	print $GP "set grid xtics ytics\n";
	print $GP "set xlabel 'fraction of negative deltas'\n";
	print $GP "set xrange [0:1]\n";
	print $GP "set key bottom right\n";
	plot_cdf($GP, \@data, \@methname);
    }
#endif

    plot_velocity(MONOTONIC);

    // print distribution of IDs
    if (htmldir) {
	const int bucketsize = 256;
	cerr << "# ID distribution\n";
	html << "\n<h2>ID distribution</h2>\n";
	html << "<p>bucket size: " << bucketsize << " (0x" << fmthex(bucketsize) << ")</p>\n";
	int data[N_METH][default_wrapsize/bucketsize];
	memset(data, 0, sizeof(data));
	gp.open("id-distribution", "ID distribution (1 response per address)", gptype, gpscript);
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if (!((*it)->ptarget()->flags & MONOTONIC) || (*it)->ptarget()->order != ORDER_BE) continue;
	    OrderInfo *oi = (*it)->ptarget()->poi();
	    ipid_t mask = 0xFFFF >> (16 - oi->bits);
	    // index_t n = oi->id->size(); // use all responses (biased towards addrs w/ more responses)
	    index_t n = 1; // use a single probe
	    for (index_t i = 0; i < n; i++) {
		ipid_t id = ((*oi->id)[i] & mask) | oi->prefix;
		data[(*it)->meth][id/bucketsize]++;
	    }
	}

	html << "<br><img src='id-distribution.png'>\n\n";
	// fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'ID (hex)'\n");
	fprintf(gp.f, "set xtics 8192\nset mxtics 2\nset format x '%%04x'\n");
	fprintf(gp.f, "set ylabel 'number of occurences'\n");
	fprintf(gp.f, "set xrange [0:65535]\n");
	fprintf(gp.f, "set yrange [0:]\n");
	fprintf(gp.f, "set key top right\n");

	const char *prefix = "plot";
	for (int i = 0; i < N_METH; i++) {
	    if (!n_targets[i]) continue;
	    fprintf(gp.f, "%s '-' using 1:2 title '%s' with points", prefix, methname[i]);
	    prefix = ",";
	}
	fprintf(gp.f, "\n");
	for (int i = 0; i < N_METH; i++) {
	    if (!n_targets[i]) continue;
	    for (unsigned j = 0; j < default_wrapsize/bucketsize; j++) {
		fprintf(gp.f, "%d %d\n", j*bucketsize, data[i][j]);
	    }
	    fprintf(gp.f, "e\n");
	}
    }

    // print distribution of bits in IDs
    if (htmldir) {
	cerr << "# ID bit distribution\n";
	html << "\n<h2>ID bit distribution</h2>\n";

	int data[N_METH][16];
	memset(data, 0, sizeof(data));
	int n[N_METH];
	memset(n, 0, sizeof(n));

	gp.open("id-bit-distribution", "ID bit distribution (all responses)", gptype, gpscript);
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if (!((*it)->ptarget()->flags & MONOTONIC) || (*it)->ptarget()->order != ORDER_BE) continue;
	    for (index_t i = 0; i < (*it)->ptarget()->id.size(); i++) {
		n[(*it)->meth]++;
		ipid_t id = (*it)->ptarget()->id[i];
		for (int j = 0; j < 16; j++) {
		    if (id & (1 << j))
			data[(*it)->meth][j]++;
		}
	    }
	}

	html << "<br><img src='id-bit-distribution.png'>\n\n";
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'bit'\n");
	fprintf(gp.f, "set ylabel 'fraction of occurences'\n");
	fprintf(gp.f, "set yrange [0:1]\n");
	fprintf(gp.f, "set ytics 0.1\n");
	fprintf(gp.f, "set xtics 1\n");
	fprintf(gp.f, "set key top right\n");

	const char *prefix = "plot";
	for (int i = 0; i < N_METH; i++) {
	    if (!n[i]) continue;
	    fprintf(gp.f, "%s '-' using 1:2 title '%s' with points", prefix, methname[i]);
	    prefix = ",";
	}
	fprintf(gp.f, "\n");
	for (int i = 0; i < N_METH; i++) {
	    if (!n[i]) continue;
	    for (int j = 0; j < 16; j++) {
		fprintf(gp.f, "%d %f\n", j, double(data[i][j]) / n[i]);
	    }
	    fprintf(gp.f, "e\n");
	}
	gp.close();
    }

    html << flush;
    memoryInfo.print("target plots");

    if (htmldir || targetsummaryfilename) {
	if (!targetsummaryfilename) {
	    // html << "<p><a href='target-summary.txt'>target summary</a></p>\n";
	}
	target_summary_header(summary, cmdline);
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    target_summary_row(summary, (*it)->ptarget());
	}
	summary << "# end\n";
	summary.close();
    }

    // free unneeded data
    for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	for (int meth = 0; meth < N_METH; meth++) {
	    if ((*it)->target[meth])
		(*it)->target[meth]->freeUnusedOrder();
	}
    }
    memoryInfo.print("freeUnusedOrder");

    // do alias resolution comparisons
    if (res_flag) resolve();

    if (htmldir) {
	html << "</body>\n";
	html << "</html>\n";
	html.close();
    }

    memoryInfo.print("done");
    cerr << "# done" << endl;

  } catch (const std::exception &e) {
    cerr << e.what() << endl;
    exit(1);
  }
}

CompFilterVector compSharedFilters;
CompFilterVector compFilters;

class AtomicIncrementer {
    unsigned i;
    Mutex lock;
public:
    AtomicIncrementer() : i(0) {}
    unsigned next() {
	lock.lock();
	unsigned val = i++;
	lock.unlock();
	return val;
    }
};

static inline int column(const bool &known_aliases) {
    return int(known_aliases);
}
static const unsigned N_COL = 2;
static const char *col_label_without_known[1] =
    { "pairs" };
static const char *col_label_with_known[2] =
    { "nonalias pairs", "alias pairs" };
const char **col_label;

struct ComparisonStatRow {
    ComparisonStat cs[N_COL];
    ComparisonStat &operator[](const unsigned &i) { return cs[i]; }
    const ComparisonStat &operator[](const unsigned &i) const { return cs[i]; }
    void operator+=(const ComparisonStatRow &that)
    {
	for (unsigned i = 0; i < N_COL; i++) {
	    cs[i].count += that.cs[i].count;
	    for (unsigned j = 0; j < that.cs[i].examples.size(); j++) {
		cs[i].insert(that.cs[i].examples[j]);
	    }
	}
    }
    bool empty() const {
	for (unsigned i = 0; i < N_COL; i++) {
	    if (cs[i].count) return false;
	}
	return true;
    }
};

static const int n_vsim_err = 8;
static const double vsim_err[n_vsim_err] =
    { 0, 0.125, 0.25, 0.5, 1, 2, 4, 8 };
static const char *vsim_label[n_vsim_err] =
    { "e = 0", "e = 0.125", "e = 0.25", "e = 0.5", "e = 1", "e = 2", "e = 4", "e = 8" };
static const int n_vsim_bins = 200;
static const double vsim_binsize = 1.001 / (n_vsim_bins);

class Resolver : public Callable
{
    AtomicIncrementer &incrementer;
    const vector<Target*> &stargets;
    const bool showprogress;
    vector<Comparison*> printableComps;
public:
    uint64_t n_pairs_not_compared;
    uint64_t n_alias_pairs_compared;
    uint64_t n_pairs[N_COL];

    uint64_t *vsim_data[RESTYPE_SHARED+1][n_vsim_err];
    uint64_t *truepositive_vsim_data[n_vsim_err];
    uint64_t *distdata[6];

    ComparisonStatRow *compResultStats;
    ComparisonStatRow *compFilterStats;
    ComparisonStatRow *compSharedStats;

    Resolver(AtomicIncrementer &_incrementer,
	const vector<Target*> &_stargets,
	bool _showprogress) :
	incrementer(_incrementer),
	stargets(_stargets),
	showprogress(_showprogress),
	printableComps(),
	n_pairs_not_compared(0),
	n_alias_pairs_compared(0),
	n_pairs()
    {
	compResultStats = new ComparisonStatRow[N_RESULT];
	compSharedStats = new ComparisonStatRow[compSharedFilters.size()];
	compFilterStats = new ComparisonStatRow[compFilters.size()];
	for (int i = 0; i < n_vsim_err; i++) {
	    vsim_data[RESTYPE_UNSHARED][i]        = newZeroedArray(uint64_t, n_vsim_bins);
	    vsim_data[RESTYPE_NOT_ENOUGH_INFO][i] = newZeroedArray(uint64_t, n_vsim_bins);
	    vsim_data[RESTYPE_SHARED][i]          = newZeroedArray(uint64_t, n_vsim_bins);
	    truepositive_vsim_data[i]             = newZeroedArray(uint64_t, n_vsim_bins);
	}
	for (int i = 0; i < distdata_count; i++) {
	    distdata[i] = newZeroedArray(uint64_t, distdata_nbins);
	}
    }

    ~Resolver() {
	delete[] compResultStats;
	delete[] compSharedStats;
	for (int i = 0; i < n_vsim_err; i++) {
	    delete[] vsim_data[RESTYPE_UNSHARED][i];
	    delete[] vsim_data[RESTYPE_NOT_ENOUGH_INFO][i];
	    delete[] vsim_data[RESTYPE_SHARED][i];
	    delete[] truepositive_vsim_data[i];
	}
	for (int i = 0; i < distdata_count; i++) {
	    delete[] distdata[i];
	}
    }

    bool comparePair(Comparison * &comp, Target * const &r1, Target * const &r2)
    {
	// assuption:  r1->series_velocity < r2->series_velocity
	double v1 = r1->series_velocity;
	double v2 = r2->series_velocity;

	if (comp->refcount > 0) {
	    // old comp is still in use, we need a new one
	    memlock.lock();
	    comp = new Comparison();
	    memlock.unlock();
	} else {
	    // we can recycle the old comp
	}
	bool has_badclock = !badclocks.empty() && r1->src_addr != r2->src_addr &&
	    (badclocks.contains(r1->src_addr) || badclocks.contains(r2->src_addr));
	if (has_badclock) {
	    comp->init(r1, r2);
	    comp->result = RESULT_BADCLOCK;
	} else {
	    comp->compare(r1, r2, !alltest);
	}

	if (alltest) {
	    oob_distribution.count(comp);
	}

	if (result_types[comp->result] >= write_option && outfilename) {
	    comp->refcount++;
	    printableComps.push_back(comp);
	}

	bool known_aliases = r1->known_nodeid &&
	    (r1->known_nodeid == r2->known_nodeid);
	if (known_aliases) {
	    n_alias_pairs_compared++;
	}

	if (htmldir) {
	    int col = column(known_aliases);

	    compResultStats[comp->result][col].insert(comp);
	    compResultStats[comp->result][col].count++;

	    for (unsigned k = 0; k < compFilters.size(); k++) {
		if (compFilters[k](comp)) {
		    compFilterStats[k][col].insert(comp);
		    compFilterStats[k][col].count++;
		}
	    }

	    if (comp->result == RESULT_SHARED) {
		// The condition above lets us skip this loop in most cases.
		for (unsigned k = 0; k < compSharedFilters.size(); k++) {
		    if (compSharedFilters[k](comp)) {
			compSharedStats[k][col].insert(comp);
			compSharedStats[k][col].count++;
		    }
		}
	    }
	    n_pairs[col]++;

#if 1
	    if (dist_option && !has_badclock) {
		int bin = comp->idist < distdata_min ? 0 :
		    int((log(comp->idist) - distdata_offset) / distdata_binsize);
		if (bin < 0) bin = 0;
		if (comp->result == RESULT_OUT_OF_BOUNDS)
		    distdata[0][bin]++;
		if (result_types[comp->result] == RESTYPE_UNSHARED)
		    distdata[1][bin]++;
		else if (result_types[comp->result] == RESTYPE_NOT_ENOUGH_INFO)
		    distdata[2][bin]++;
		else if (comp->result == RESULT_SHARED)
		    distdata[3][bin]++;
		if (knownfilename) {
		    distdata[4+int(known_aliases)][bin]++;
		}
	    }
#endif
	}

	if (v1 > 0) {
	    for (int ei = 0; ei < n_vsim_err; ei++) {
		double e = vsim_err[ei];
		double val = (v2 + e) / (v1 + e);
		int bin = int(val / vsim_binsize);
		assert(bin < n_vsim_bins);
		vsim_data[result_types[comp->result]][ei][bin]++;
		if (known_aliases && comp->result == RESULT_SHARED)
		    truepositive_vsim_data[ei][bin]++;
	    }
	}
	return true;
    }

    void dumpComparisons()
    {
	if (printableComps.empty()) return;
	char printbuf[1024];
	vector<Comparison*>::iterator it;
	outfilelock.lock(); // note: inet_ntoa needs lock
	for (it = printableComps.begin(); it != printableComps.end(); ++it) {
	    Comparison *comp = *it;
	    const Target *r1 = comp->r1;
	    const Target *r2 = comp->r2;
	    double v1 = r1->series_velocity;
	    double v2 = r2->series_velocity;
	    sprintf(printbuf, "%-15s %-15s %-6s %9.4f %-15s %-15s %-6s %9.4f %7.5f "
		"%-13s %3d %3d %3d %10.3f %3d",
		string(r1->addr).c_str(), string(r1->src_addr).c_str(),
		methname[r1->meth], v1,
		string(r2->addr).c_str(), string(r2->src_addr).c_str(),
		methname[r2->meth], v2,
		v2 / v1,
		result_label[comp->result],
		comp->n_bounded,
		comp->n_interp,
		comp->n_extrap,
		comp->overlap,
		comp->oob);
	    outfile << printbuf;
	    if (dist_option) {
		sprintf(printbuf, " %12.5f %12.5f", comp->idist, comp->edist);
		outfile << printbuf;
	    }
	    outfile << "\n";
	}
	outfilelock.unlock();
	memlock.lock();
	for (it = printableComps.begin(); it != printableComps.end(); ++it) {
	    if (--(*it)->refcount == 0) {
		delete (*it);
	    }
	}
	memlock.unlock();
	printableComps.clear();
    }


    void operator()()
    {
	memlock.lock();
	Comparison *comp = new Comparison();
	memlock.unlock();

	if (pairfilename) {
	    // compare only specified pairs
	    const int chunkSize = 100;
	    uint64_t n = pairs.size();
	    FmtNum<uint64_t> formatted_n(n);
	    while (true) {
		unsigned i = incrementer.next() * chunkSize;
		if (i >= pairs.size()) break; 
		unsigned chunkEnd = i + chunkSize;
		if (chunkEnd > pairs.size()) chunkEnd = pairs.size();
		for ( ; i < chunkEnd; i++) {
		    Target *r1 = pairs[i][0];
		    Target *r2 = pairs[i][1];
		    if (r1->series_velocity < r2->series_velocity)
			swap(r1, r2);
		    comparePair(comp, r1, r2);
		}
		++comp->refcount; // don't let dumpComparisons free it
		dumpComparisons();
		--comp->refcount;
		if (showprogress)
		    cerr << "\x1b[A# resolved " << fmtnum(i) << " of " << formatted_n << "   \n";
	    }

	} else if (onlyknown) {
	    // compare only pairs that belong to known alias sets
	    const int chunkSize = 1;
	    uint64_t n = knownAliases.size();
	    FmtNum<uint64_t> formatted_n(n);
	    while (true) {
		unsigned i = incrementer.next() * chunkSize;
		if (i >= knownAliases.size()) break; 
		unsigned chunkEnd = i + chunkSize;
		if (chunkEnd > knownAliases.size()) chunkEnd = knownAliases.size();
		for ( ; i < chunkEnd; i++) {
		    for (unsigned j = 0; j < knownAliases[i].size() - 1; j++) {
			for (unsigned k = j+1; k < knownAliases[i].size(); k++) {
			    Target *r1 = knownAliases[i][j];
			    Target *r2 = knownAliases[i][k];
			    if (r1->series_velocity < r2->series_velocity)
				swap(r1, r2);
			    comparePair(comp, r1, r2);
			}
		    }
		}
		++comp->refcount; // don't let dumpComparisons free it
		dumpComparisons();
		--comp->refcount;
		if (showprogress)
		    cerr << "\x1b[A# resolved " << fmtnum(i) << " of " <<
			formatted_n << "   \n";
	    }

	} else {
	    // compare all possible pairs
	    uint64_t n = choose2count(stargets.size());
	    FmtNum<uint64_t> formatted_n(n);
	    while (true) {
		size_t i = incrementer.next();
		if (i >= stargets.size()) break; 
		for (unsigned j = i; j > 0; ) {
		    j--;
		    if (!cmp_nonoverlap && stargets[j]->time.back().min < stargets[i]->time.front().max) {
			// target[j] does not overlap, and neither can any
			// target 0..j, because targets are sorted by
			// last timestamp
			n_pairs_not_compared += j + 1;
			break;
		    }
		    Target *r1 = stargets[i];
		    Target *r2 = stargets[j];
		    if (r1->series_velocity < r2->series_velocity)
			swap(r1, r2);
		    comparePair(comp, r1, r2);
		    if (j == 0) break;
		}
		++comp->refcount; // don't let dumpComparisons free it
		dumpComparisons();
		--comp->refcount;
		if (showprogress) {
		    uint64_t resolved = (i+1)*i/2;
		    cerr << "\x1b[A# resolved " <<
			fmtnum(resolved) << " of " << formatted_n << "   \n";
		}
	    }
	}

	if (comp && comp->refcount == 0) {
	    memlock.lock();
	    delete comp;
	    memlock.unlock();
	}
    }
    void print_n_pairs(const int &i, ostream &out)
    {
	out << "# resolver[" << i << "]: ";
	for (unsigned j = 0; j < N_COL; j++) {
	    out << " " << this->n_pairs[j];
	}
	out << "\n";
    }
};

void print_resolution_row(const string &rawlabel, ResultType restype,
    Resolver *resolver, const ComparisonStatRow &statrow)
{
    // cerr << "# print_row: " << rawlabel << "\n";
    html << "<tr " <<
	status_style[restype_status[restype]] <<
	"><td class='label'>" << rawlabel << "\n";

    for (int col = 0; col < (knownfilename ? 2 : 1); col++) {
	string label(rawlabel);
	label += "-";
	label += col_label[col];
	print_resolution_cell(html, &statrow[col],
	    resolver->n_pairs[col], label, gptype, gpscript);
    }
}

void resolve()
{
    vector<Resolver*> resolver(n_threads);

    using namespace ComparisonFilters;

    // set up general filters
    // These are expensive in a Discovery run, we're careful to enable only
    // the ones that are useful.
#ifdef ENABLE_FEATURES
    if (featurefilename) {
	compFilters("SHARED-featurediff", 0,
	    new ResultF(RESULT_SHARED), new FeaturediffF());
	compFilters("nonshared-featurediff", 0,
	    Not(ResultF(RESULT_SHARED)), new FeaturediffF());
    }
#endif
    if (dist_option) {
	compFilters("idist-le-500", 0,
	    new IDistF(0, 500));
	compFilters("idist-gt-500", 0,
	    new IDistF(500));
	compFilters("SHARED-idist-le-500", 0,
	    new ResultTypeF(RESTYPE_SHARED), new IDistF(0, 500));
	compFilters("SHARED-idist-gt-500", 0,
	    new ResultTypeF(RESTYPE_SHARED), new IDistF(500));
	compFilters("NOT_ENOUGH_INFO-idist-le-500", 0,
	    new ResultTypeF(RESTYPE_NOT_ENOUGH_INFO), new IDistF(0, 500));
	compFilters("NOT_ENOUGH_INFO-idist-gt-500", 0,
	    new ResultTypeF(RESTYPE_NOT_ENOUGH_INFO), new IDistF(500));
	compFilters("UNSHARED-idist-le-500", 0,
	    new ResultTypeF(RESTYPE_UNSHARED), new IDistF(0, 500));
	compFilters("UNSHARED-idist-gt-500", 0,
	    new ResultTypeF(RESTYPE_UNSHARED), new IDistF(500));
	compFilters("MBTpass-idist-le-500", 0,
	    new OOBcountF(0,0), new MBTcountF(min_overlap_option), new IDistF(0, 500));
	compFilters("MBTpass-idist-gt-500", 0,
	    new OOBcountF(0,0), new MBTcountF(min_overlap_option), new IDistF(500));
	compFilters("MBTfail-idist-le-500", 0,
	    new OOBcountF(1), new IDistF(0, 500));
	compFilters("MBTfail-idist-gt-500", 0,
	    new OOBcountF(1), new IDistF(500));
    
	compFilters("edist-le-500", 0,
	    new EDistF(0, 500));
	compFilters("edist-gt-500", 0,
	    new EDistF(500));
	compFilters("SHARED-edist-le-500", 0,
	    new ResultTypeF(RESTYPE_SHARED), new EDistF(0, 500));
	compFilters("SHARED-edist-gt-500", 0,
	    new ResultTypeF(RESTYPE_SHARED), new EDistF(500));
	compFilters("NOT_ENOUGH_INFO-edist-le-500", 0,
	    new ResultTypeF(RESTYPE_NOT_ENOUGH_INFO), new EDistF(0, 500));
	compFilters("NOT_ENOUGH_INFO-edist-gt-500", 0,
	    new ResultTypeF(RESTYPE_NOT_ENOUGH_INFO), new EDistF(500));
	compFilters("UNSHARED-edist-le-500", 0,
	    new ResultTypeF(RESTYPE_UNSHARED), new EDistF(0, 500));
	compFilters("UNSHARED-edist-gt-500", 0,
	    new ResultTypeF(RESTYPE_UNSHARED), new EDistF(500));
	compFilters("MBTpass-edist-le-500", 0,
	    new OOBcountF(0,0), new MBTcountF(min_overlap_option), new EDistF(0, 500));
	compFilters("MBTpass-edist-gt-500", 0,
	    new OOBcountF(0,0), new MBTcountF(min_overlap_option), new EDistF(500));
	compFilters("MBTfail-edist-le-500", 0,
	    new OOBcountF(1), new EDistF(0, 500));
	compFilters("MBTfail-edist-gt-500", 0,
	    new OOBcountF(1), new EDistF(500));
    }

#if 0
    compFilters("oob old=fail new=fail", 0,
	new FailOobF(true, true));
    compFilters("oob old=pass new=pass", 0,
	new FailOobF(false, false));
    compFilters("oob old=pass new=fail", 0,
	new FailOobF(false, true));
    compFilters("oob old=fail new=pass", 0,
	new FailOobF(true, false));
#endif

    if (alltest) {
	compFilters("OUT_OF_BOUNDS with < 20% OOB", 0,
	    new ResultF(RESULT_OUT_OF_BOUNDS), new OOBfracF(0,0.2));
    }

    // set up shared filters
    compSharedFilters("SHARED with xdisco_oob", 0,
	new XDiscoOOBcountF(1));
    if (Comparison::oob_frac_allowed || Comparison::oob_converge_allowed) {
	compSharedFilters("SHARED with OOB >= 1", 0,
	    new OOBcountF(1));
    }
    compSharedFilters("shared, high velocity", 0,
	new Velocity2F(100));
    compSharedFilters("shared, medium velocity", 0,
	new Velocity2F(10, 100));
    compSharedFilters("shared, low velocity", 0,
	new Velocity2F(0, 10));

    for (int t = 0; t < N_METH; t++) {
	if (!n_targets[t]) continue;
	compSharedFilters(string("at least one ") + methname[t] + " shared", 0,
	    new MethF(t));
    }

    for (int t = 0; t < N_METH; t++) {
	if (!n_targets[t]) continue;
	compSharedFilters(string("both ") + methname[t] + " shared", 0,
	    new Meth2F(t, t));
    }

    for (int t1 = 0; t1 < N_METH-1; t1++) {
	if (!n_targets[t1]) continue;
	for (int t2 = t1+1; t2 < N_METH; t2++) {
	    if (!n_targets[t2]) continue;
	    compSharedFilters(string(methname[t1]) + "-" + methname[t2] + " shared", 0,
		new Meth2F(t1, t2));
	}
    }

    if (monitors.size() > 1) {
	compSharedFilters("same source, shared", 0, new SameSourceF());
	compSharedFilters("different source, shared", 0, Not(SameSourceF()));
    }


    if (outfilename || htmldir) {
	memoryInfo.print("start resolving");
	cerr << "# alias resolution\n\n";

	// select monotonic addresses, and sort by descending velocity
	vector<Target*> stargets;
	for (MultiMethTargetSet::iterator it = targets.begin(); it != targets.end(); it++) {
	    if ((*it)->ptarget()->flags & MONOTONIC)
		stargets.push_back((*it)->ptarget());
	}
	sort(stargets.begin(), stargets.end(), target_last_timestamp_less);

	if (outfilename) {
	    if (write_option > RESTYPE_UNSHARED) {
		outfile << "# WARNING: this file does NOT contain every comparison\n";
	    }
	    char printbuf[1024];
	    sprintf(printbuf, "# %-13s %-15s %-6s %9s %-15s %-15s %-6s %9s %7s "
		"%-13s %3s %3s %3s %10s %3s",
		"addr1", "src1", "meth1", "v1", "addr2", "src2", "meth2", "v2", "v2/v1",
		"result", "bnd", "int", "ext", "overlap", "oob");
	    outfile << printbuf;
	    if (dist_option) {
		sprintf(printbuf, " %12s %12s", "idist", "edist");
		outfile << printbuf;
	    }
	    outfile << "\n";
	}

	AtomicIncrementer incrementer;
	resolver[0] = new Resolver(incrementer, stargets, true);
#if HAVE_PTHREAD
	vector<Thread*> thread(n_threads);
	for (int i = 1; i < n_threads; i++) {
	    resolver[i] = new Resolver(incrementer, stargets, false);
	    thread[i] = new Thread();
	    thread[i]->start(resolver[i]);
	}
#endif
	(*resolver[0])(); // resolver[0] runs in main thread, whether or not threads are enabled

#if HAVE_PTHREAD
	// resolver[0]->print_n_pairs(0, cerr);
	for (int i = 1; i < n_threads; i++) {
	    thread[i]->join();
	    // merge resolver[i] results into resolver[0] results
	    resolver[0]->n_pairs_not_compared += resolver[i]->n_pairs_not_compared;
	    resolver[0]->n_alias_pairs_compared += resolver[i]->n_alias_pairs_compared;
	    for (unsigned j = 0; j < N_COL; j++) {
		resolver[0]->n_pairs[j] += resolver[i]->n_pairs[j];
	    }
	    for (unsigned j = 0; j < N_RESULT; j++) {
		resolver[0]->compResultStats[j] += resolver[i]->compResultStats[j];
	    }
	    for (unsigned j = 0; j < compSharedFilters.size(); j++) {
		resolver[0]->compSharedStats[j] += resolver[i]->compSharedStats[j];
	    }
	    for (unsigned j = 0; j < compFilters.size(); j++) {
		resolver[0]->compFilterStats[j] += resolver[i]->compFilterStats[j];
	    }
	    for (int j = 0; j < n_vsim_err; j++) {
		for (int k = 0; k < n_vsim_bins; k++) {
		    for (int restype = RESTYPE_UNSHARED; restype <= RESTYPE_SHARED; restype++) {
			resolver[0]->vsim_data[restype][j][k] +=
			    resolver[i]->vsim_data[restype][j][k];
		    }
		    resolver[0]->truepositive_vsim_data[j][k] +=
			resolver[i]->truepositive_vsim_data[j][k];
		}
	    }
	    for (int j = 0; j < distdata_count; j++) {
		for (int k = 0; k < distdata_nbins; k++) {
		    resolver[0]->distdata[j][k] +=
			resolver[i]->distdata[j][k];
		}
	    }
	    // resolver[i]->print_n_pairs(i, cerr);
	    delete resolver[i]; resolver[i] = 0;
	    delete thread[i]; thread[i] = 0;
	}
#endif
	resolver[0]->n_pairs_not_compared += listed_pairs_not_compared;

	cerr << "\n";
	memoryInfo.print("done resolving");
    }

    if (htmldir) {
	// finish calculating totals
	int64_t n_alias_pairs_not_compared =
	    n_alias_pairs - resolver[0]->n_alias_pairs_compared;
	int64_t n_nonalias_pairs_not_compared =
	    resolver[0]->n_pairs_not_compared - n_alias_pairs_not_compared;
	resolver[0]->n_pairs[column(false)] += n_nonalias_pairs_not_compared;
	if (knownfilename) {
	    resolver[0]->n_pairs[column(true)] += n_alias_pairs_not_compared;
	}

	col_label = knownfilename ?
	    col_label_with_known : col_label_without_known;

	cerr << "# resolution comparisons\n";

	// print table
	html << "\n<h2>Resolution comparisons</h2>\n\n";
	html << "<p>Comparing IP ID time series collected from different " <<
	    "addresses identifies addresses that may share an ID counter " <<
	    "with each other.</p>\n";
	html << "<ul>\n";
	if (vsim_option) {
	    html << "<li>VSIM: (v<sub>2</sub> + e) / (v<sub>1</sub> + e) < r, where " <<
		"e = " << vsim_fudge << " and " <<
		"r = " << setprecision(6) << vsim_thresh << "\n";
	}
	if (min_overlap_option) {
	    html << "<li>POOR_OVERLAP: fewer than " << min_overlap_option <<
		" points overlapped\n";
	}
	if (Comparison::oob_frac_allowed) {
	    html << "<li>oob-allowed: " <<
		Comparison::oob_frac_allowed << "," <<
		Comparison::oob_converge_allowed << "," <<
		Comparison::oob_inconclusive_allowed << "\n";
	}
	html << "</ul>\n";

	html << "<table style='border-collapse: collapse'>\n";
	html << "<tr><th>category";
	for (unsigned col = 0; col < (knownfilename ? N_COL : N_COL/2); col++) {
	    html << "<th colspan=2>" << col_label[col];
	}
	html << "\n";

	html << "<tr><td>total pairs";
	html << "<td>" <<
	    htmlnum(resolver[0]->n_pairs[column(false)]) << "<td class='percent'>" <<
	    fmtpercent(resolver[0]->n_pairs[column(false)], resolver[0]->n_pairs[column(false)]) <<
	    "\n";
	if (knownfilename) {
	    html << "<td>" <<
		htmlnum(resolver[0]->n_pairs[column(true)]) << "<td class='percent'>" <<
		fmtpercent(resolver[0]->n_pairs[column(true)], resolver[0]->n_pairs[column(true)]) <<
		"\n";
	}

	{
	    html << "<tr " << status_style[STATUS_MEDIUM] << "><td>not compared";
	    html << "<td>" <<
		htmlnum(n_nonalias_pairs_not_compared) << "<td class='percent'>" <<
		fmtpercent(n_nonalias_pairs_not_compared, resolver[0]->n_pairs[column(false)]) <<
		"\n";
	    if (knownfilename) {
		html << "<td>" <<
		    htmlnum(n_alias_pairs_not_compared) << "<td class='percent'>" <<
		    fmtpercent(n_alias_pairs_not_compared, resolver[0]->n_pairs[column(true)]) <<
		    "\n";
	    }
	    html << "\n";
	}

	// calculate restype totals and grand total for comparison result counts
	ComparisonStatRow totalStats;
	ComparisonStatRow compRestypeStats[RESTYPE_SHARED+1];
	for (int result = 0; result < N_RESULT; result++) {
	    for (int col = 0; col < (knownfilename ? 2 : 1); col++) {
		compRestypeStats[result_types[result]][col].count += resolver[0]->compResultStats[result][col].count;
		totalStats[col].count += resolver[0]->compResultStats[result][col].count;
	    }
	}

	// print comparison result counts
	bool wantRes[N_RESULT];
	for (int result = 0; result < N_RESULT; result++) {
	    wantRes[result] = true;
	}
	wantRes[RESULT_UNKNOWN] = false;
#ifndef ENABLE_FEATURES
	wantRes[RESULT_FEATUREDIFF] = false;
#endif
	wantRes[RESULT_VSIM] = (vsim_option > VSIM_NEI);
	wantRes[RESULT_BADCLOCK] = !badclocks.empty();
	wantRes[RESULT_BIG_GAP_VSIM] = (vsim_option == VSIM_NEI);
	wantRes[RESULT_IDIST] = (dist_option >= DIST_INTERP);
	wantRes[RESULT_EDIST] = (dist_option >= DIST_ALL);
	wantRes[RESULT_NO_OVERLAP_VSIM] = (vsim_option == VSIM_NEI);
	wantRes[RESULT_POOR_OVERLAP] = (min_overlap_option > 0);
	wantRes[RESULT_POOR_OVERLAP_VSIM] = (min_overlap_option > 0) && (vsim_option == VSIM_NEI);
	wantRes[RESULT_DIFF_TTL] = ttl_flag;
	wantRes[RESULT_BITS] = bits_flag;

	for (int restype = RESTYPE_UNSHARED; restype <= RESTYPE_SHARED; restype++) {
	    for (int result = 0; result < N_RESULT; result++) {
		if (result_types[result] == restype &&
		    (wantRes[result] || !resolver[0]->compResultStats[result].empty()))
		{
		    print_resolution_row(string(result_label[result]),
			ResultType(restype), resolver[0], resolver[0]->compResultStats[result]);
		}
	    }
	    if (restype != RESTYPE_SHARED) {
		print_resolution_row(string("total ") + restype_label[restype],
		    ResultType(restype), resolver[0], compRestypeStats[restype]);
	    }
	}

	print_resolution_row("total pairs compared", RESTYPE_NONE, resolver[0], totalStats);

	// print comparison shared counts
	for (unsigned row = 0; row < compSharedFilters.size(); row++) {
	    print_resolution_row(compSharedFilters[row].label, RESTYPE_NONE, resolver[0], 
		resolver[0]->compSharedStats[row]);
	}
	// print comparison filtered counts
	for (unsigned row = 0; row < compFilters.size(); row++) {
	    print_resolution_row(compFilters[row].label, RESTYPE_NONE, resolver[0], 
		resolver[0]->compFilterStats[row]);
	}

	html << "</table>\n";
    }
    memoryInfo.print("post resolving");

    if (htmldir && alltest) {
	oob_distribution.plot(html, gptype, gpscript);
    }

    if (htmldir) {
	html << "\n<h2>Velocity similarity</h2>\n\n";

	const char *cfg =
	    "set grid xtics ytics\n"
	    "set xlabel '(v2 + e) / (v1 + e)'\n"
	    "set xtics 0.1\n"
	    "set mxtics 5\n"
	    "set key bottom right\n";

	html << "<table>\n";
	html << "<tr>\n";
	if (knownfilename) {
	    for (int logy = 0; logy <= 1; logy++) {
		string filename(string("truepositive-vsim-cdf") + (logy ? "-log" : ""));
		gp.open(filename,
		    "distribution of vsims for known aliases with shared counters",
		    gptype, gpscript);
		html << "<td><img src='" << filename << ".png'>\n\n";
		fprintf(gp.f, "%s", cfg);
		gp.plot_binned_cdf(false, n_vsim_err, n_vsim_bins,
		    resolver[0]->truepositive_vsim_data,
		    0, vsim_binsize, vsim_label, logy);
		gp.close();
	    }
	}

	for (int restype = RESTYPE_UNSHARED; restype <= RESTYPE_SHARED; restype++) {
	    html << "<tr>\n";
	    for (int logy = 0; logy <= 1; logy++) {
		string filename(restype_label[restype] + "-vsim-cdf" + (logy ? "-log" : ""));
		gp.open(filename,
		    "distribution of vsims for " + restype_label[restype] + " address pairs",
		    gptype, gpscript);
		html << "<td><img src='" << filename << ".png'>\n\n";
		fprintf(gp.f, "%s", cfg);
		gp.plot_binned_cdf(false, n_vsim_err, n_vsim_bins,
		    resolver[0]->vsim_data[restype],
		    0, vsim_binsize, vsim_label, logy);
		gp.close();
	    }
	}
	html << "</table>\n";
    }

    plot_velocity(SHARED);

    html << flush;
    // plot dist
    if (htmldir && dist_option) {
	html << "\n<h2>Average Distance</h2>\n\n";

	gp.open("dist-cdf", "distribution of average distances", gptype, gpscript);
	const char *labels[] = {"failed MBT", "unshared", "not enough info", "shared", "nonalias", "alias"};
	html << "<br><img src='dist-cdf.png'>\n\n";
	fprintf(gp.f, "set grid xtics ytics\n");
	fprintf(gp.f, "set xlabel 'Distance'\n");
	fprintf(gp.f, "set key top left\n");
	gp.plot_binned_cdf(true, distdata_count, distdata_nbins, resolver[0]->distdata,
	    distdata_offset, distdata_binsize, labels,
	    /*logy*/true, /*absolute*/true);
    }

    delete resolver[0]; resolver[0] = 0;
}

