/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../lib/config.h"
#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <assert.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <string>
#include <getopt.h>

using namespace std;

#include "../lib/infile.h"

#include "../lib/MemoryInfo.h"
#define NO_DEBUG_MEMORY 0
MemoryInfo memoryInfo;

#include "MidarUtils.h"
using namespace MidarUtils;

static const char *cvsID = "$Id: set-analysis.cc,v 1.38 2016/09/22 19:11:53 kkeys Exp $";
static string cmdline = "";
static const char *argv0;

static int n_threads = 1;
static const char *htmldir = 0;
static const char *graphfilename = 0;
static const char *outfilename = 0;
// static const char *gptype = Gnuplotter::PNG;
static int gpscript = 1;

static struct option longopts[] = {
    { "htmldir",       required_argument, 0,        'H' },
    { "output",        required_argument, 0,        'O' },
    { "cor-analysis",  required_argument, 0,        'A' },
    { "sets",          required_argument, 0,        'S' },
    { "pairs",         required_argument, 0,        'P' },
    { "threads",       required_argument, 0,        'n' },
    { "overlap",       required_argument, 0,        'o' },
    { "responsive",    required_argument, 0,        'r' },
    { "clockoffset",   required_argument, 0,        'c' },
    { "graph",         required_argument, 0,        'G' },
    { "bits",          no_argument,       &bits_flag, 1 },
    { "no-bits",       no_argument,       &bits_flag, 0 },
    { "nobits",        no_argument,       &bits_flag, 0 },
    TTL_LONGOPT("no-ttl",     &ttl_flag,    0),
    TTL_LONGOPT("ttl",        &ttl_flag,    1),
    { "udp-diff",      no_argument,       &allow_udp_diff_addr, 1 },
    { "no-udp-diff",   no_argument,       &allow_udp_diff_addr, 0 },
    VSIM_LONGOPT("vsim-none",     &vsim_option,    VSIM_NONE),
    VSIM_LONGOPT("vsim-nei",      &vsim_option,    VSIM_NEI),
    VSIM_LONGOPT("vsim-all",      &vsim_option,    VSIM_ALL),
    DIST_LONGOPT("dist-none",   &dist_option, DIST_NONE),
    DIST_LONGOPT("dist-calc",   &dist_option, DIST_CALC),
    DIST_LONGOPT("dist-interp", &dist_option, DIST_INTERP),
    DIST_LONGOPT("dist-all",    &dist_option, DIST_ALL),
    { "gpscript",      no_argument,       &gpscript, 1 },
    { "no-gpscript",   no_argument,       &gpscript, 0 },
    { 0, 0, 0, 0 }
};

static ostream *outfile = &cout;
static Mutex outfilelock;
static ofstream html;

static char *cor_analysis_filename;
static char *subjsets_filename;
static char *subjpairs_filename;
static AddrPairSet corPairs; // SHARED pairs in cor-analysis
static long n_classified = 0;
static long n_badclocks = 0;
static long n_preconfirmed = 0;
static long n_reconfirmed = 0;

const int MAX_DATASETS = 20; // XXX TODO: make this dynamic

struct AddrSet : public vector<ip4addr_t> {
    long n_badpairs[MAX_DATASETS+1];
    long n_sharedpairs[MAX_DATASETS+1];
    AddrSet() {
	memset(n_badpairs, 0, sizeof(n_badpairs));
	memset(n_sharedpairs, 0, sizeof(n_sharedpairs));
    }
};

struct AddrSetSet : public vector<AddrSet*> {
    void add(AddrSet* as) {
	sort(as->begin(), as->end()); // sort the addrs in as
	push_back(as);
    }
};

static AddrSetSet subjSets;
static AddrPairSet subjPairs;
static set<ip4addr_t> includeaddrs;

void usageExit()
{
    cerr << "Usage: " << argv0 << " [<options>] --sets <setfile> <dataset0> ; <dataset1> ; ... <datasetN>\n";
    cerr << "where <dataset> is:  [badclocks=<ip_list>] <probefile>...\n";
    cerr << "Each dataset is a set of output files from a single probe run.  Datasets\n";
    cerr << "must be separated by \" ; \".\n";
    cerr << "Badclocks specifies a list of monitors that should not be used in cross-monitor comparisons.\n";
    cerr << "Options:\n";
    cerr << "--sets <file>           load sets to be tested from <file> (required)\n";
    cerr << "--pairs <file>          consider only pairs listed in <file>\n";
//    cerr << "--htmldir <dir>         write optional analysis graphs and webpage to directory <dir>\n";
    cerr << "--output <file>         write results to <file>\n";
    cerr << "--graph <file>          write DOT graph to <file>\n";
    cerr << "--cor-analysis <file>   load corroboration analysis results from <file>\n";
    cerr << "--clockoffset <n>       maximum clock offset error, in milliseconds (default " <<
	int(max_clock_offset * 1000) << ")\n";
    cerr << "--(no-)bits             (do not) apply bit precsion test (default: on)\n";
#if !ENABLE_TTL_OPTION
    cerr << "--ttl option requires recompiling with ENABLE_TTL_OPTION\n";
#endif
    cerr << "--(no-)ttl              (do not) use the DIFF_TTL test (default: off)\n";
    cerr << "--(no-)udp-diff         (do not) allow UDP response from different address (default: off)\n";
#if !ENABLE_VSIM_OPTION
    cerr << "--vsim- options require recompiling with ENABLE_VSIM_OPTION\n";
#endif
    cerr << "--vsim-none             do not use VSIM test (default)\n";
    cerr << "--vsim-nei              use VSIM test on sets with Not Enough Info for other tests\n";
    cerr << "--vsim-all              always use VSIM test, before other tests\n";
#if !ENABLE_DIST_OPTION
    cerr << "--dist- options require recompiling with ENABLE_DIST_OPTION\n";
#endif
    cerr << "--dist-none             do not use DIST test (default)\n";
    cerr << "--dist-calc             calculate dist, but do not use DIST test\n";
    cerr << "--dist-interp           use DIST test on interpolated points\n";
    cerr << "--dist-all              use DIST test on interpolated and extrapolated points\n";
    cerr << "--overlap <n>           require at least <n> overlapping points (default: 0)\n";
    cerr << "--responsive <n>        treat >n responses as responsive, even if n<" <<
	(100 * mostly_thresh) << "% of tries (default: " << responsive_count << ")\n";
//    cerr << "--no-gpscript           write gnuplot graphs instead of scripts\n";
#if HAVE_PTHREAD
//    cerr << "--threads <n>           use <n> threads during alias resolution (1)\n";
#endif
    exit(1);
}

struct Dataset; // forward declaration

static vector<Dataset*> datasets;

struct Dataset {
    int id;
    MultiMethTargetSet targets;
    BadclockSet badclocks; // monitors with bad clocks that can't be compared to other monitors
    Dataset() : id(datasets.size()), targets(), badclocks() {
	datasets.push_back(this);
	cerr << "# starting new dataset " << id << endl;
    }
};

void loadResponses(const char *filename, Dataset *dataset)
{
    ResponseData rd;
    char buf[8192];
    bool finished = false;
    InFile in(filename);
    cerr << "# loading data from " << filename;
    cerr << " into dataset " << dataset->id << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (strncmp(buf, "## run finished ", 16) == 0) {
	    finished = true;
	}
	if (!buf[0] || buf[strlen(buf)-1] != '\n') {
	    cerr << "WARNING: input truncated at line " << in.linenum() << endl;
	    break;
	}
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	if (finished) {
	    std::cerr << "WARNING: ignoring text after \"run finished\", line "
		<< in.linenum() << endl;
	    break;
	}

	rd.parse(buf);
	if (rd.result == 'S') continue; // 'S' record is informational, not a real probe

#if 1
	if (includeaddrs.find(rd.target_addr) == includeaddrs.end())
	    continue;
#endif

	int meth;
	parse_meth(rd.method, meth, rd.ttl_expansion);

	// monitors.countProbeFrom(rd.src_addr);

	MultiMethTarget *mttarget = dataset->targets.findOrInsert(rd.target_addr);

	if (mttarget->meth < 0) {
	    mttarget->meth = meth;
	    mttarget->ptarget() = new Target(rd.target_addr);
	    mttarget->ptarget()->meth = meth;
	    // n_targets[meth]++;
	} else if (mttarget->meth != meth) {
	    cerr << "WARNING: ignoring " << methname[meth] <<
		" response for " << mttarget->addr << " that already has " <<
		methname[mttarget->meth] << " response" << endl;
	    continue;
	}

	if (mttarget->ptarget()->probes == 0) {
	    mttarget->ptarget()->src_addr = rd.src_addr;
	} else if (mttarget->ptarget()->src_addr != rd.src_addr) {
	    // hack: if probed from multiple sources, set src_addr to 0.
	    mttarget->ptarget()->src_addr = ip4addr_t(0);
	}

	if (rd.ttl_expansion == 0)
	    mttarget->ptarget()->tries++;
	mttarget->ptarget()->probes++;
	// n_probes[meth]++;
	if (rd.result != 'T' && rd.result != 'U' && rd.result != 'H')
	    continue;
	// n_responses[meth]++;
	mttarget->ptarget()->processResponse(&rd);
      } catch (const std::runtime_error &e) {
        if (!in.gets(buf, sizeof(buf))) { // error was in last line
            std::string msg(InFile::Error(in, e).what());
            msg += " (file appears to be truncated; ignoring error)";
            cerr << msg << endl;
            break;
        } else {
            throw InFile::Error(in, e);
        }
      }
    }
    in.close();
    if (!finished)
	std::cerr << "WARNING: missing \"run finished\" line" << endl;
    memoryInfo.print("loaded responses");
}

static void loadSharedPairs(const char *filename, AddrPairSet &pairs)
{   
    char buf[8192]; 
    InFile in(filename);
    cerr << "# loading shared pairs from " << filename << endl;
    AddrPair ap;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	FieldSplitter<10> fields(buf, " \t\n");
	if (strcmp(fields[9], "SHARED") != 0) continue;

	ap[0] = ip4addr_t(fields[0]);
	ap[1] = ip4addr_t(fields[4]);
	if (ap[0] != 0 && ap[1] != 0) {
	    if (ap[1] < ap[0]) swap(ap[0], ap[1]);
	    pairs.insert(ap);
	}
	// *outfile << "# loaded pair: " << setw(15) << ap[0] << " " << setw(15) << ap[1] << "\n"; // XXX
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << pairs.size() << " pairs\n";
    memoryInfo.print("loaded pairs");
}

static void loadAddrSets(const char *filename, AddrSetSet &sets)
{   
    char buf[8192]; 
    InFile in(filename);
    cerr << "# loading subject sets from " << filename << endl;
    AddrSet *as = 0;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (strncmp(buf, "# set ", 6) == 0 || strncmp(buf, "# name: ", 8) == 0) {
	    if (as) {
		if (as->empty()) continue;
		sets.add(as);
	    }
	    as = new AddrSet();
	    continue;
	}
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	char *p = strchr(buf, '\n');
	if (p) *p = '\0';
	ip4addr_t addr(buf);
	as->push_back(addr);
	includeaddrs.insert(addr);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    if (as && !as->empty()) {
	sets.add(as);
    } else if (as) {
	delete as;
    }
    cerr << "# loaded " << sets.size() << " sets\n";
    memoryInfo.print("loaded subj sets");
}

static void loadAddrPairs(const char *filename, AddrPairSet &pairs)
{   
    char buf[8192]; 
    InFile in(filename);
    cerr << "# loading subject pairs from " << filename << endl;
    AddrPair ap;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	FieldSplitter<2> fields(buf, " \t\n");
	ap[0] = ip4addr_t(fields[0]);
	ap[1] = ip4addr_t(fields[1]);
	if (ap[0] != 0 && ap[1] != 0) {
	    pairs.insert(ap);
	}
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << pairs.size() << " pairs\n";
    memoryInfo.print("loaded subj pairs");
}

void resolvePair(Comparison &comp, const AddrPair &ap, int dsi, unsigned &n_shared, unsigned &n_unshared, unsigned &n_nei)
{
    MultiMethTargetSet &targets = datasets[dsi]->targets;
    Target *target[2];
    MultiMethTarget key0(ap[0]);
    MultiMethTarget key1(ap[1]);
    MultiMethTargetSet::iterator mt0 = targets.find(&key0);
    MultiMethTargetSet::iterator mt1 = targets.find(&key1);
#if 1
    if (mt0 == targets.end() || mt1 == targets.end()) {
	// shouldn't happen
	n_nei++;
	return;
    }
#endif
    target[0] = (*mt0)->ptarget();
    target[1] = (*mt1)->ptarget();

    if (!datasets[dsi]->badclocks.empty() &&
	target[0]->src_addr != target[1]->src_addr &&
	(datasets[dsi]->badclocks.contains(target[0]->src_addr) ||
	datasets[dsi]->badclocks.contains(target[1]->src_addr)))
    {
	++n_badclocks;
	return;
    }

    if (!target[0]->classification) {
	classify(target[0], true);
	n_classified++;
	// cerr << "classify dataset " << dsi << " addr " << target[0]->addr << " = " << class_label[target[0]->classification] << endl;
    }
    if (!target[1]->classification) {
	classify(target[1], true);
	n_classified++;
	// cerr << "classify dataset " << dsi << " addr " << target[1]->addr << " = " << class_label[target[1]->classification] << endl;
    }
    if (target[0]->classification == CLASS_MONOTONIC &&
	target[1]->classification == CLASS_MONOTONIC)
    {
	comp.compare(target[0], target[1]);
	switch(result_types[comp.result]) {
	    case RESTYPE_SHARED:   n_shared++;   break;
	    case RESTYPE_UNSHARED: n_unshared++;
		*outfile << "# UNSHARED: dataset " << dsi << " addrs " << target[0]->addr << " " << target[1]->addr << " = " << result_label[comp.result] << endl;
		break;
	    default:               n_nei++;      break;
	}
    } else {
	// If either is not MONOTONIC, we can't say anything conclusive.  (It
	// might be UNRESPONSIVE due to rate limiting, or NONMONOTONIC because
	// its velocity was a little too high for us to measure, or MOSTLY_FOO
	// because it was just over the MOSTLY threshold....)
	n_nei++;
    }
}

static void plotPair(ofstream &graphfile, const AddrPair &ap, bool directional)
{
    unsigned n_shared = 0;
    unsigned n_unshared = 0;
    unsigned n_nei = 0;
    uint32_t color = 0;
    Comparison comp;

    if (corPairs.find(ap) != corPairs.end()) {
	// this pair is already confirmed
	color = 0x4040FF; // blue
    } else {
	for (size_t dsi = 0; dsi < datasets.size(); dsi++) {
	    resolvePair(comp, ap, dsi, n_shared, n_unshared, n_nei);
	}
	if (n_unshared > 0) {
	    double frac = n_unshared == datasets.size() ? 1.0 : (n_unshared-1.0) / (datasets.size()-1.0);
	    color = (255 << 16) | (int(255 * (1 - frac)) << 0); // magenta -> red
	} else if (n_shared > 0) {
	    double frac = n_shared == datasets.size() ? 1.0 : (n_shared-1.0) / (datasets.size()-1.0);
	    color = (255 << 8) | (int(255 * (1 - frac)) << 0); // cyan -> green
	} else {
	    color = 0x808080; // gray
	}
    }
    graphfile << "\"" << ap[0] << "\" " << (directional ? "->" : "--") << " \"" << ap[1] << "\" [ color=\"#" <<
	hex << setw(6) << setfill('0') << color << dec <<
	"\" ]; // " << n_unshared << " " << n_shared << "\n";
}

int main(int argc, char **argv)
{
  try {
    // cerr.setprecision(6);
    // cerr.fixed();
    cerr << "# main" << endl;
    argv0 = argv[0];
    for (int i = 0; i < argc; i++) {
	cmdline += argv[i];
	cmdline += ' ';
    }

    min_overlap_option = 0;

    int opt;
    int longidx;
    while ((opt = getopt_long(argc, argv, "", longopts, &longidx)) != -1) {
	switch (opt) {
	case 'H':
	    htmldir = strdup(optarg);
	    break;
	case 'O':
	    outfilename = strdup(optarg);
	    break;
	case 'A':
	    cor_analysis_filename = strdup(optarg);
	    break;
	case 'S':
	    subjsets_filename = strdup(optarg);
	    break;
	case 'P':
	    subjpairs_filename = strdup(optarg);
	    break;
#if HAVE_PTHREAD
	case 'n':
	    n_threads = atoi(optarg);
	    break;
#endif
	case 'o':
	    min_overlap_option = atoi(optarg);
	    break;
	case 'r':
	    responsive_count = atoi(optarg);
	    break;
	case 'c':
	    max_clock_offset = atoi(optarg) / 1000.0;
	    break;
	case 0:
	    // flag
	    break;
	case 'G':
	    graphfilename = strdup(optarg);
	    break;
	default:
	    usageExit();
	}
    }
    argc -= optind;
    argv += optind;

    if (!subjsets_filename) usageExit();

    if (htmldir) {
	if (mkdir(htmldir, 0755) < 0 && errno != EEXIST) {
	    cerr << htmldir << ": " << strerror(errno) << endl;
	    exit(1);
	}
    }

    cerr << "# open outfile" << endl;
    if (outfilename) {
	outfile = new ofstream(outfilename);
	if (!outfile) {
	    cerr << "can't open " << outfilename << ": " << strerror(errno) << endl;
	    exit(1);
	}
	*outfile << "# version: " << ::cvsID  << endl;
	*outfile << "# version: " << MidarUtils::cvsID << endl;
	*outfile << "# command line: " << cmdline << endl << "#" << endl;
    }

    memoryInfo.print("startup");

    if (cor_analysis_filename)
	loadSharedPairs(cor_analysis_filename, corPairs);
    if (subjsets_filename)
	loadAddrSets(subjsets_filename, subjSets);
    if (subjpairs_filename)
	loadAddrPairs(subjpairs_filename, subjPairs);

    Dataset *dataset = new Dataset();
    for (int i = 0; i < argc; i++) {
	if (strcmp(argv[i], ";") == 0) {
	    dataset = new Dataset();
	} else if (strncmp(argv[i], "badclocks=", 10) == 0) {
	    dataset->badclocks.parse(argv[i]+10);
	    // dump badclocks, to make sure it was parsed correctly
	    cerr << "# badclocks for dataset " << dataset->id << ": ";
	    bool first = true;
	    for (BadclockSet::iterator it = dataset->badclocks.begin();
		it != dataset->badclocks.end(); ++it)
	    {
		if (!first) cerr << ",";
		cerr << *it;
		first = false;
	    }
	    cerr << endl;
	} else {
	    loadResponses(argv[i], dataset);
	}
    }

    // delete empty final set (because argv ended with an unnecessary ";")
    if (datasets.back()->targets.empty()) {
	delete datasets.back();
	datasets.pop_back();
    }

#if 0
    for (int dsi = 0; dsi < datasets.size(); dsi++) {
	cerr << "### dataset " << dsi << " size: " << datasets[dsi]->targets.size() << endl;
	for (MultiMethTargetSet::iterator mti = datasets[dsi]->targets.begin();
	    mti != datasets[dsi]->targets.end(); ++mti)
	{
	    cerr << (*mti)->addr << endl;
	}
    }
#endif

    Comparison comp;
    long set_id = 0;
    *outfile << "#     ";
    *outfile << setw(6) << "set_id" << ": ";
    *outfile << setw(6) << "addrs" << " ";
    *outfile << setw(6) << "pairs";
    *outfile << " " << "unsh>0";
    for (size_t n_unshared = 1; n_unshared <= datasets.size(); n_unshared++) {
	*outfile << " " << "unsh=" << n_unshared;
    }
    *outfile << " ";
    for (size_t n_shared = 0; n_shared <= datasets.size(); n_shared++) {
	*outfile << " " << "  sh=" << n_shared;
    }
    *outfile << endl;
    for (AddrSetSet::iterator asi = subjSets.begin(); asi != subjSets.end(); ++asi) {
	bool conflict = false;
	for (size_t i = 0; i < (*asi)->size(); i++) {
	    for (size_t j = i+1; j < (*asi)->size(); j++) {
		AddrPair ap;
		ap[0] = (*(*asi))[i];
		ap[1] = (*(*asi))[j];
		if (corPairs.find(ap) != corPairs.end()) {
		    // this pair is already confirmed
		    (*asi)->n_sharedpairs[datasets.size()]++;
		    n_preconfirmed++;
		    continue;
		}
		if (!subjPairs.empty()) {
		    AddrPair ap_swapped;
		    ap_swapped[0] = (*(*asi))[j];
		    ap_swapped[1] = (*(*asi))[i];
		    if (subjPairs.find(ap) == subjPairs.end() &&
			subjPairs.find(ap_swapped) == subjPairs.end())
			    continue;
		}

		// number of datasets in which this pair was shared/unshared/nei
		unsigned n_shared = 0;
		unsigned n_unshared = 0;
		unsigned n_nei = 0;

		for (size_t dsi = 0; dsi < datasets.size(); dsi++) {
		    resolvePair(comp, ap, dsi, n_shared, n_unshared, n_nei);
		}

		if (n_unshared > 0) {
		    (*asi)->n_badpairs[n_unshared]++;
		    conflict = true;
		} else {
		    (*asi)->n_sharedpairs[n_shared]++;
		    if (n_shared == datasets.size())
			n_reconfirmed++;
		}
	    }
	}
	ostringstream header;
	header << "set " << setw(6) << set_id << ": ";
	header << setw(6) << (*asi)->size() << " ";
	header << setw(6) << choose2count((*asi)->size());
	size_t n_unshared_gt_0 = 0;
	for (size_t n_unshared = 1; n_unshared <= datasets.size(); n_unshared++) {
	    n_unshared_gt_0 += (*asi)->n_badpairs[n_unshared];
	}
	header << " " << setw(6) << n_unshared_gt_0;
	for (size_t n_unshared = 1; n_unshared <= datasets.size(); n_unshared++) {
	    header << " " << setw(6) << (*asi)->n_badpairs[n_unshared];
	}
	header << " ";
	for (size_t n_shared = 0; n_shared <= datasets.size(); n_shared++) {
	    header << " " << setw(6) << (*asi)->n_sharedpairs[n_shared];
	}

	*outfile << "# " << header.str() << endl;
	for (size_t i = 0; i < (*asi)->size(); i++) {
	    *outfile << (*(*asi))[i] << endl;
	}
	if (graphfilename && conflict) {
	    ostringstream name;
	    name << graphfilename << ".set" << set_id << ".dot";
	    ofstream graphfile(name.str().c_str());
	    graphfile << "// " << header.str() << "\n";
	    graphfile << (!subjPairs.empty() ? "digraph" : "graph") << " \"set" << set_id << "\" {\n";
	    graphfile << "edge [ color=black, style=solid, penwidth=.1, fontsize=28, arrowhead=\"halfopen\" ];\n";
	    graphfile << "node [ color=black, shape=point, label=\"\" ];\n";
	    graphfile << "subgraph {\n";
	    for (size_t i = 0; i < (*asi)->size(); i++) {
		for (size_t j = i+1; j < (*asi)->size(); j++) {
		    if (!subjPairs.empty()) {
			AddrPair ap;
			ap[0] = (*(*asi))[i];
			ap[1] = (*(*asi))[j];
			if (subjPairs.find(ap) != subjPairs.end())
			    plotPair(graphfile, ap, true);
			ap[0] = (*(*asi))[j];
			ap[1] = (*(*asi))[i];
			if (subjPairs.find(ap) != subjPairs.end())
			    plotPair(graphfile, ap, true);
		    } else {
			AddrPair ap;
			ap[0] = (*(*asi))[i];
			ap[1] = (*(*asi))[j];
			plotPair(graphfile, ap, false);
		    }
		}
	    }
	    graphfile << "}\n"; // end subgraph
	    graphfile << "}\n"; // end graph
	    graphfile.close();
	}
	set_id++;
    }
    *outfile << "# end" << endl;
    memoryInfo.print("analyzed");

    cerr << "# n_badclocks = " << n_badclocks << "\n";
    cerr << "# n_classified = " << n_classified << "\n";
    cerr << "# n_preconfirmed = " << n_preconfirmed << "\n";
    cerr << "# n_reconfirmed = " << n_reconfirmed << "\n";

#if 0
    if (htmldir) {
	html << "</body>\n";
	html << "</html>\n";
	html.close();
    }

#endif
    memoryInfo.print("done");
    cerr << "# done" << endl;

  } catch (const std::exception &e) {
    cerr << e.what() << endl;
    exit(1);
  }

}

