#!/usr/bin/env perl
## 
## Copyright (C) 2011-2018 The Regents of the University of California.
## 
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
## 

# Generate schedule of MIDAR probes done in a sliding window over targets sorted
# by descending estimated velocity.  The sort is descending so that high
# velocity targets are at the beginning of the list, when monitors are still
# well synced.

use strict;
use Getopt::Long;

my $cmdline = join(' ', $0, @ARGV);

#### parameters
my $htmldir = undef;
my $probes_per_sec = '100/1';
my $min_probes_per_target = 30;
my $max_period = 600;
my $min_period = 4;
my $per_round_overhead = 0.75; # delay between rounds (seconds)
my $min_probes_per_wrap = 4;
my $max_vsim = 2/3;
my $vsim_e = 2;
my $safe_delta_per_round = 655; # no need to even think about shortening round unless this is exceeded
my $gpscript = 0;

my $usage =
    "Usage: $0 [<options>] <target_file>...\n" .
    "Options:\n" .
    "--htmldir <output_directory>   write analysis graphs and webpage to <output_directory>\n" .
    "--rate <rate>                  use <rate> probes per second; can be written as a fraction (default: $probes_per_sec)\n" .
    "--delay <seconds>              delay between probing rounds (default: $per_round_overhead)\n" .
    "--gpscript                     write gnuplot scripts instead of plots\n";

sub usageExit() {
    print STDERR $usage;
    exit 1;
}

GetOptions(
    "gpscript" => \$gpscript,
    "htmldir=s" => \$htmldir,
    "rate=s" => \$probes_per_sec,
    "delay=f" => \$per_round_overhead
    ) or usageExit();

usageExit() if @ARGV < 1;

if ($probes_per_sec =~ m'^([\d.]+)/([\d.]+)$') {
    $probes_per_sec = $1 / $2;
} elsif ($probes_per_sec =~ m'^([\d.]+)$') {
    $probes_per_sec = $1;
} else {
    die "invalid rate \"$probes_per_sec\"\n";
}
usageExit() if $probes_per_sec <= 0;

#### state
my $actual_total_probes = 0;
my $actual_max_n_probes = 0;
my $actual_max_window = 0;
my $actual_min_window = 0xFFFFFFFF;
my $actual_max_period = 0;
my $actual_min_period = 0xFFFFFFFF;
my $actual_max_vsim = 0;
my $actual_min_probes_per_wrap = 0xFFFFFFFF;
my @round_events = ();
my @roundstats = ();

my $max_delta_per_round = 0x10000 / $min_probes_per_wrap;
my @sources = ();
my $n_unfinished_targets = 0;
my $max_targets_per_source = 0;
my $n_sources = 0;
my $max_targets_per_window = $max_period * $probes_per_sec;
my $min_targets_per_window = $min_period * $probes_per_sec;

sub vsim($$) {
    my ($vhi, $vlo) = @_;
    return ($vlo + $vsim_e) / ($vhi + $vsim_e);
}

sub load_targets() {
    my $line = 0;
    $n_sources = scalar @ARGV;
    for (my $s = 0; $s < $n_sources; $s++) {
	my $file = $ARGV[$s];
	print STDERR "# loading $file\n";
	$sources[$s]{file} = $ARGV[$s];
	$sources[$s]{label} =
	    ($sources[$s]{file} =~ /targets\.([-\w.]+)\.targets$/) ?
	    $1 : $sources[$s]{file};
	open DATA, "<", $ARGV[$s] || die "$ARGV[$s]: $!\n";
	while (<DATA>) {
	    $line++;
	    next if (/^#/ || /^$/);
	    my ($addr, $src, $dst, $hop, $terminal_hop, $cksum, $method, $velocity) = split;
	    if (!defined $velocity) {
		die "$file:$line: syntax error\n";
	    }
	    push @{$sources[$s]{targets}},
		{addr => $addr, velocity => 0.0 + $velocity, n_probes => 0};
	}
	next unless defined $sources[$s]{targets};
	my $n = scalar @{$sources[$s]{targets}};
	$max_targets_per_source = $n if ($n > $max_targets_per_source);
	$n_unfinished_targets += $n;
	#for my $target (@{$sources[$s]{targets}}) {
	#    print STDERR "target: ", $target->{addr}, " v=", $target->{velocity}, "\n";
	#}
    }
}

#### main
load_targets();

print STDERR "# checking for empty sources\n";
for (my $s = 0; $s < scalar @sources; $s++) {
    if (!defined $sources[$s]{targets}) {
	print STDERR "WARNING: no targets for ", $sources[$s]{label}, "\n";
	splice(@sources, $s, 1);
	$s--;
	$n_sources--;
    }
}
print STDERR "# done checking for empty sources\n";

if ($n_unfinished_targets == 0) {
    die "No targets.\n";
}

# verify that targets are sorted by estimated velocity, descending
print STDERR "# verifying sort\n";
my $max_v = 0;
my $min_v = 999999999;
for my $src (@sources) {
    my $targets = $src->{targets};
    my $n = scalar @{$src->{targets}};
    for (my $i = 1; $i < $n; $i++) {
	die "Targets must be sorted by descending velocity (" . $src->{file} . ":$i).\n" unless
	    $targets->[$i-1]{velocity} >= $targets->[$i]{velocity};
    }

    $max_v = $src->{targets}[0]{velocity} if ($src->{targets}[0]{velocity} > $max_v);
    $min_v = $src->{targets}[$n-1]{velocity} if ($src->{targets}[$n-1]{velocity} < $min_v);
}

print STDERR "# initializing\n";
for my $src (@sources) {
    $src->{begin} = 0; # index of target at beginning of window
    $src->{end} = 0;   # index of target after end of window
    $src->{max} = -1;  # maximum probed index
}
print STDERR "# calculating windows\n";
my $n_rounds = 0;
my $n_targets = $n_unfinished_targets;
my $time = 0;
while (1) {
    my $vhi = 0; # velocity of target at beginning of window
    for my $src (@sources) {
	my $v = $src->{targets}[$src->{begin}]{velocity};
	$vhi = $v if ($v > $vhi);
    }

    # Find the ideal end-of-window for each source.
    # print STDERR "# $n_rounds calculating window end\n";
    for my $src (@sources) {
	for ($src->{end} = $src->{max} + 1; $src->{end} < scalar @{$src->{targets}}; $src->{end}++) {
	    my $vlo = $src->{targets}[$src->{end}]{velocity};
	    # End round if we have exceeded max_period.
	    last if ($src->{end} - $src->{begin} + 1) > $max_targets_per_window;
	    # End round if the "badness" in probe frequency of the first
	    # (highest velocity) target would exceed the "badness" of the
	    # velocity similarity of first and last targets.
	    my $period = ($src->{end} - $src->{begin} + 1) / $probes_per_sec + $per_round_overhead;
	    if ($period >= $min_period && $vhi*$period > $safe_delta_per_round) {
		my $freq_badness = ($vhi * $period) / $max_delta_per_round;
		my $vsim_badness = (vsim($vhi, $vlo) / $max_vsim) **2;
		last if $freq_badness > $vsim_badness;
	    }
	}
	if ($src->{end} == scalar @{$src->{targets}} && $src->{end} - 1 > $src->{max}) {
	    push @round_events, $n_rounds;
	}
	$src->{max} = $src->{end} - 1;
    }

    # Count probes
    # print STDERR "# $n_rounds counting probes\n";
    my $max_window_size = 0;
    my $max_vlo = 0;
    for my $src (@sources) {
	for (my $i = $src->{begin}; $i < $src->{end}; $i++) {
	    my $target = $src->{targets}[$i];
	    $target->{n_probes}++;
	    if ($actual_max_n_probes < $target->{n_probes}) {
		$actual_max_n_probes = $target->{n_probes};
	    }

	    if ($target->{n_probes} == $min_probes_per_target) {
		$n_unfinished_targets--;
		if ($src->{begin} == 0) { push @round_events, $n_rounds; };
	    }
	}
	my $window_size = $src->{end} - $src->{begin};
	$actual_total_probes += $window_size;
	$max_window_size = $window_size if ($window_size > $max_window_size);
	$actual_max_window = $window_size if ($window_size > $actual_max_window);
	$actual_min_window = $window_size if ($window_size < $actual_min_window);
	if ($src->{targets}[$src->{max}]{velocity} > $max_vlo)
	    { $max_vlo = $src->{targets}[$src->{max}]{velocity}; }

#	print STDERR "# $n_rounds ", $src->{label}, " indexes: ",
#	    $src->{begin}, " ",
#	    $src->{end}-1, " (of ",
#	    scalar @{$src->{targets}}, ") ",
#	    " (end probes=", $src->{targets}[$src->{end}-1]{n_probes}, ")",
#	    " (total probes=$actual_total_probes)\n";
    }

    my $period = $max_window_size / $probes_per_sec + $per_round_overhead;

    if (0x10000 / ($vhi * $period) < $actual_min_probes_per_wrap) {
	$actual_min_probes_per_wrap = 0x10000 / ($vhi * $period);
    }
    if (vsim($vhi, $max_vlo) > $actual_max_vsim) {
	$actual_max_vsim = vsim($vhi, $max_vlo);
    }
    if ($period < $actual_min_period) {
	$actual_min_period = $period;
    }
    if ($period > $actual_max_period) {
	$actual_max_period = $period;
    }

    for my $src (@sources) {
	push @{$src->{roundstats}}, [
	    $time, $src->{begin}, $src->{max},
	    $src->{targets}[$src->{begin}]{velocity},
	    $src->{targets}[$src->{max}]{velocity},
	    $period,
	    $src->{targets}[$src->{begin}]{n_probes},
	    $src->{targets}[$src->{max}]{n_probes} ];
    }

    $period = $min_period if ($period < $min_period);
    $time += $period;
    $n_rounds++;

    print STDERR "# targets remaining: $n_unfinished_targets     \r";
    last if $n_unfinished_targets == 0;
    if ($n_rounds > 2000) { # XXX
	print STDERR "################################### ABORTING\n";
	exit 1;
    }

    # Find the maximum amount the beginning of each window can slide.
    my $max_vhi = 0;
    # print STDERR "# $n_rounds sliding\n";
    for my $src (@sources) {
	my $begin = $src->{begin};
	my $end = $src->{end};
	if ($end < scalar @{$src->{targets}}) {
	    # Calculate the ideal beginning of each window that isn't already
	    # up against the end of the list.  Set the maximum slide amount so
	    # that a target at the end of the window should stay in the window for
	    # $min_probes_per_target rounds, if sliding were to continue at that
	    # rate.  But slide less if a target that would fall out of the window
	    # hasn't been probed enough.
#	    my $max_begin =
#		$end >= scalar @{$src->{targets}} ? scalar @{$src->{targets}} :
#		$begin + int(($end - $begin) / ($min_probes_per_target + 1));
	    my $r = (scalar @{$src->{roundstats}} > $min_probes_per_target) ?
		(scalar @{$src->{roundstats}} - $min_probes_per_target) : 0;
	    my $max_begin = 
		$end >= scalar @{$src->{targets}} ? scalar @{$src->{targets}} :
		$begin + ($src->{roundstats}[$r][2] - $src->{roundstats}[$r][1] + 1) / ($min_probes_per_target);
	    while ($begin < $max_begin &&
		$src->{targets}[$begin]{n_probes} >= $min_probes_per_target)
	    {
		$begin++;
	    }
	    if ($begin < scalar @{$src->{targets}}) {
		my $vhi = $src->{targets}[$begin]{velocity};
		$max_vhi = $vhi if ($vhi > $max_vhi);
		# print STDERR "# $n_rounds ideal begin: $begin, $vhi\n";
	    }
	}
    }

    # Slide the beginning of each window to the same max_vhi velocity, unless
    # the window is already up against the end of the list.
    my $max_window_size = 0;
    for my $src (@sources) {
	my $begin = $src->{begin};
	if ($src->{end} < scalar @{$src->{targets}}) {
	    while ($src->{targets}[$begin]{velocity} > $max_vhi) {
		$begin++;
	    }
	    $src->{begin} = $begin;
	}
    }
}
print STDERR "\n";


my $actual_min_n_probes = 999999999;
for my $src (@sources) {
    for (my $i = 0; $i < scalar @{$src->{targets}}; $i++) {
	my $target = $src->{targets}[$i];
	if ($actual_min_n_probes > $target->{n_probes}) {
	    $actual_min_n_probes = $target->{n_probes};
	}
	$src->{probe_hist}[$target->{n_probes}]++;
    }
}

### html results
if ($htmldir) {
    -d $htmldir or mkdir $htmldir;
    open HTML, ">$htmldir/index.html";
    if ($gpscript) {
	open GP, ">$htmldir/script.gp";
    } else {
	open GP, "|gnuplot";
    }

    print HTML "<h2>MIDAR sliding window schedule</h2>\n";
    print HTML "Command line: $cmdline\n<br><br>\n";
    print HTML "Parameters:\n";
    print HTML "<ul>\n";
    print HTML "<li>targets: ", scalar $n_targets, "\n";
    print HTML "<li>Probes per second: $probes_per_sec\n";
    print HTML "<li>Min probes per target: $min_probes_per_target\n";
    printf HTML "<li>Max period: %.3f\n", $max_period;
    printf HTML "<li>Min period: %.3f\n", $min_period;
    print HTML "<li>Min probes per wrap: $min_probes_per_wrap\n";
    print HTML "<li>Max velocity similarity: $max_vsim\n";
    print HTML "</ul>\n";
    print HTML "Results:\n";
    print HTML "<ul>\n";
    print HTML "<li>Max window: $actual_max_window targets\n";
    print HTML "<li>Min window: $actual_min_window targets\n";
    printf HTML "<li>Max period: %.3f\n", $actual_max_period, " s";
    printf HTML "<li>Min period: %.3f\n", $actual_min_period, " s";
    printf HTML "<li>Min probes per wrap: %.3f\n", $actual_min_probes_per_wrap;
    printf HTML "<li>Max velocity similarity: %.3f\n", $actual_max_vsim;
    print HTML "<li>total probes: $actual_total_probes\n";
    printf HTML "<li>avg probes per target: %.3f\n", $actual_total_probes/$n_targets;
    print HTML "<li>min probes per target: ", $actual_min_n_probes, "\n";
    print HTML "<li>max probes per target: ", $actual_max_n_probes, "\n";
    printf HTML "<li>rounds: %d\n", $n_rounds;
    printf HTML "<li>time: %.2f s (%.2f hr)\n", $time, $time/3600;
    print HTML "</ul>\n";

    print HTML "<a href='data.txt'>raw data</a>\n";
}

for my $src (@sources) {
    my $outfile = $src->{file} . ".window";
    open OUT, ">", $outfile || die "$outfile: $!\n";
    printf OUT "# $cmdline\n";
    printf OUT "# probes per second: $probes_per_sec\n";
    printf OUT "# %7s %6s %6s %11s %11s %8s %7s %7s\n",
	'time', 'begin', 'end', 'v_begin', 'v_end', 'period', 'p_begin', 'p_end';
    for my $ref (@{$src->{roundstats}}) {
	printf OUT "%9.3f %6d %6d %11.6f %11.6f %8.3f %7d %7d\n", @$ref;
    }
    printf OUT "# end\n";
    close OUT;
    print STDERR "# output written to $outfile\n";
}

if ($htmldir) {
    print GP "set terminal png small\n";

    # probe count histogram
    print STDERR "# plotting probe-count-hist.png\n";
    print HTML "<br><img src='probe-count-hist.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/probe-count-hist.png'\n";
    print GP "set title 'probe count histogram'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'probe count'\n";
    print GP "set ylabel 'number of targets'\n";
    # print GP "set style data lines\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $n (@{$src->{probe_hist}}) {
	    print GP int($n), "\n";
	}
	print GP "e\n";
    }


    # probe counts
    print STDERR "# plotting probe-count.png\n";
    print HTML "<br><img src='probe-count.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/probe-count.png'\n";
    print GP "set title 'probe counts'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'index'\n";
    print GP "set ylabel 'number of probes'\n";
    print GP "set xrange [0:$max_targets_per_source]\n";
    print GP "set yrange [0:$actual_max_n_probes]\n";
    # print GP "set style data lines\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for (my $i = 0; $i < scalar @{$src->{targets}}; $i++) {
	    print GP int($src->{targets}[$i]{n_probes}), "\n";
	}
	print GP "e\n";
    }


    print STDERR "# plotting index-vs-round.png\n";
    print HTML "<br><img src='index-vs-round.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/index-vs-round.png'\n";
    print GP "set title 'index vs. round'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'round'\n";
    print GP "set ylabel 'index'\n";
    print GP "set xtics 20\n";
    print GP "set mxtics 4\n";
    #print GP "set grid xtics ytics\n";
    for my $round_event (@round_events) {
	print GP "set arrow from $round_event,0 to $round_event,$n_targets lt 0 nohead\n";
    }
    # print GP "set style data lines\n";
    print GP "set xrange [0:$n_rounds]\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	print GP ", '-' using 0:1 title '' w l ls ", $s+1;
	# print GP ", ", scalar @{$src->{targets}}, " title '' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $ref (@{$src->{roundstats}}) { print GP ${$ref}[1], "\n"; }
	print GP "e\n";
	for my $ref (@{$src->{roundstats}}) { print GP ${$ref}[2], "\n"; }
	print GP "e\n";
    }


    print STDERR "# plotting velocity-vs-time.png\n";
    print HTML "<br><img src='velocity-vs-time.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/velocity-vs-time.png'\n";
    print GP "set title 'velocity vs. time'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'time'\n";
    print GP "set ylabel 'velocity (id/s) [LOG SCALE]'\n";
    print GP "set logscale y\nset mytics 10\n";
    #print GP "set grid xtics ytics\n";
    # print GP "set style data lines\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 1:2 title '", $src->{label}, "' w l ls ", $s+1;
	print GP ", '-' using 1:2 title '' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $ref (@{$src->{roundstats}}) { print GP ${$ref}[0], " ", ${$ref}[3], "\n"; }
	print GP "e\n";
	for my $ref (@{$src->{roundstats}}) { print GP ${$ref}[0], " ", ${$ref}[4], "\n"; }
	print GP "e\n";
    }


    print STDERR "# plotting velocity-vs-round.png\n";
    print HTML "<br><img src='velocity-vs-round.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/velocity-vs-round.png'\n";
    print GP "set title 'velocity vs. round'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'round'\n";
    print GP "set ylabel 'velocity (id/s) [LOG SCALE]'\n";
    print GP "set logscale y\nset mytics 10\n";
    print GP "set xtics 20\n";
    print GP "set mxtics 4\n";
    #print GP "set grid xtics ytics\n";
    for my $round_event (@round_events) {
	print GP "set arrow from $round_event,$min_v to $round_event,$max_v lt 0 nohead\n";
    }
    # print GP "set style data lines\n";
    print GP "set xrange [0:$n_rounds]\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	print GP ", '-' using 0:1 title '' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $ref (@{$src->{roundstats}}) { print GP ${$ref}[3], "\n"; }
	print GP "e\n";
	for my $ref (@{$src->{roundstats}}) { print GP ${$ref}[4], "\n"; }
	print GP "e\n";
    }


    print STDERR "# plotting velocity-similarity.png\n";
    print HTML "<br><img src='velocity-similarity.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/velocity-similarity.png'\n";
    print GP "set title 'Velocity similarity vs. round'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'round'\n";
    print GP "set ylabel 'velocity similarity'\n";
    print GP "set xtics 20\n";
    print GP "set mxtics 4\n";
    #print GP "set grid xtics ytics y2tics\n";
    for my $round_event (@round_events) {
	print GP "set arrow from $round_event,0 to $round_event,1 lt 0 nohead\n";
    }
    print GP "set style data linespoints\n";
    print GP "set xrange [0:$n_rounds]\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $ref (@{$src->{roundstats}}) { print GP (vsim(${$ref}[3], ${$ref}[4]), "\n"); }
	print GP "e\n";
    }


    print STDERR "# plotting wrap.png\n";
    print HTML "<br><img src='wrap.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/wrap.png'\n";
    print GP "set title 'Wraps vs. round'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'round'\n";
    print GP "set ylabel 'wraps'\n";
    print GP "set xtics 20\n";
    print GP "set mxtics 4\n";
    #print GP "set grid xtics ytics y2tics\n";
    my $max_wraps_per_probe = 1 / $actual_min_probes_per_wrap;
    for my $round_event (@round_events) {
	print GP "set arrow from $round_event,0 to $round_event,$max_wraps_per_probe lt 0 nohead\n";
    }
    print GP "set style data linespoints\n";
    print GP "set xrange [0:$n_rounds]\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	print GP ", '-' using 0:1 title '' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $ref (@{$src->{roundstats}})
	    { print GP ((${$ref}[5] * ${$ref}[3] / 0x10000), "\n"); }
	print GP "e\n";
	for my $ref (@{$src->{roundstats}})
	    { print GP ((${$ref}[5] * ${$ref}[4] / 0x10000), "\n"); }
	print GP "e\n";
    }


    print STDERR "# plotting period.png\n";
    print HTML "<br><img src='period.png'>\n";
    print GP "reset\n";
    print GP "set output '$htmldir/period.png'\n";
    print GP "set title 'Period vs. round'\n";
    print GP "set key outside\n";
    print GP "set xlabel 'round'\n";
    print GP "set ylabel 'period'\n";
    print GP "set xtics 20\n";
    print GP "set mxtics 4\n";
    #print GP "set grid xtics ytics y2tics\n";
    for my $round_event (@round_events) {
	print GP "set arrow from $round_event,0 to $round_event,$actual_max_period lt 0 nohead\n";
    }
    print GP "set style data linespoints\n";
    print GP "set xrange [0:$n_rounds]\n";

    my $prefix = "plot";
    for (my $s = 0; $s < scalar @sources; $s++) {
	my $src = $sources[$s];
	print GP "$prefix '-' using 0:1 title '", $src->{label}, "' w l ls ", $s+1;
	$prefix = ",";
    }
    print GP "\n";

    for my $src (@sources) {
	for my $ref (@{$src->{roundstats}}) { print GP (${$ref}[5], "\n"); }
	print GP "e\n";
    }

    close GP;
    close HTML;
}

### stderr results
print STDERR "\n";
print STDERR "Max window: $actual_max_window targets\n";
print STDERR "Min window: $actual_min_window targets\n";
printf STDERR "Max period: %.3f\n", $actual_max_period, " s";
printf STDERR "Min period: %.3f\n", $actual_min_period, " s";
printf STDERR "Min probes per wrap: %.3f\n", $actual_min_probes_per_wrap;
printf STDERR "Max velocity similarity: %.3f\n", $actual_max_vsim;
print STDERR "total probes: $actual_total_probes\n";
printf STDERR "avg probes per target: %.3f\n", $actual_total_probes/$n_targets;
print STDERR "min probes per target: ", $actual_min_n_probes, "\n";
print STDERR "max probes per target: ", $actual_max_n_probes, "\n";
printf STDERR "rounds: %d\n", $n_rounds;
printf STDERR "time: %.2f s (%.2f hr)\n", $time, $time/3600;

