/* 
 * Copyright (C) 2011-2018 The Regents of the University of California.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <cstdio>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <assert.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <vector>
#include <set>
#include <map>
#include <string>
#include <getopt.h>

static const char *cvsID = "$Id: transitive-closure.cc,v 1.39 2015/09/16 00:39:52 kkeys Exp $";

using namespace std;

#include "../lib/config.h"
#include "../lib/infile.h"
#include "../lib/ip4addr.h"
#include "../lib/NetPrefix.h"

#include "../lib/MemoryInfo.h"
#define NO_DEBUG_MEMORY 0
MemoryInfo memoryInfo;

#include "MidarUtils.h"
using MidarUtils::AddrPair;
using MidarUtils::AddrPairSet;

static string cmdline = "";
static const char *argv0;

static const char *includeAddrsFile = 0;
static const char *includePairsFile = 0;
static const char *bogonFile = 0;
static NetPrefixSet bogons;
static set<ip4addr_t> includeAddrs;
static AddrPairSet includePairs;

static bool isBogus(const ip4addr_t addr)
{
    // assumes that bogons contains only the largest prefixes
    NetPrefix key(addr, 32);
    NetPrefixSet::const_iterator it = bogons.upper_bound(key);
    return (it != bogons.begin() && (*--it).contains(addr));
}

struct AddrSet {
    vector<ip4addr_t> addrs;
    vector<const AddrPair*> chainPairs;
    size_t n_found;
    double min_v, max_v;
    AddrSet() : addrs(), chainPairs(), n_found(0), min_v(-1), max_v(-1) {}
};

#if 0
static std::ostream& operator<< (std::ostream& out, const AddrSet & addrset) {
    bool first = true;
    for (vector<ip4addr_t>::const_iterator it = addrset.addrs.begin(); it != addrset.addrs.end(); ++it) {
	if (!first) out << ' ';
	out << *it;
	first = false;
    }
    return out;
}
#endif

struct AddrData {
    AddrSet *addrSet;
    double v;
    size_t n_pairs;
    AddrData() : addrSet(0), v(0), n_pairs(0) {}
};

typedef map<ip4addr_t, AddrData> AddrMap;
AddrMap addrs;

struct chainPair_velocity_less {
    bool operator()(const AddrPair* const &a, const AddrPair* const &b) {
	return (addrs[(*a)[0]].v < addrs[(*b)[0]].v);
    }
};

static AddrPairSet sharedPairs;
static AddrPairSet unsharedPairs;
bool opt_pairs = false;
bool opt_addrs = true;
bool opt_filter = false;
size_t opt_breaksize = 0;
size_t opt_size = 0;
string opt_outfile;

struct set_size_greater {
    bool operator()(const AddrSet * const &a, const AddrSet * const &b)
    {
	return a->addrs.size() != b->addrs.size() ? a->addrs.size() > b->addrs.size() :
	    a->addrs[0] < b->addrs[0];
    }
};

void usageExit()
{
    cerr << "Usage: " << argv0 << " [<options>] <file>...\n";
    cerr << "Finds the transitive closure of every pair that is SHARED in at least\n";
    cerr << "one file and never UNSHARED in any file.\n";
    cerr << "<file> may be in any of these formats:\n";
    cerr << "  MIDAR resolution-analyze, containing SHARED and UNSHARED pairs\n";
    cerr << "  iffinder\n";
    cerr << "  pairs\n";
    cerr << "Options:\n";
    cerr << "  -B <file>  load bogons from <file>\n";
    cerr << "  -I <file>  include only addresses listed in <file>\n";
    cerr << "  -P <file>  include only pairs listed in <file>\n";
    cerr << "  -a         print addresses in each set (default)\n";
    cerr << "  -p         print pairs in each set\n";
    cerr << "  -f         filter mode: print only SHARED pairs, not transitive closure\n";
    cerr << "             (useful for merging result files)\n";
    cerr << "  -n         print only a summary for each set\n";
    cerr << "  -s<n>      define <n> as max size of \"small\" sets\n";
    cerr << "  -o<name>   without -s, write results to <name>\n";
    cerr << "             with -s, write results to <name>-{small,large}\n";
    exit(1);
}

enum Formats { FMT_PAIRS, FMT_MIDAR, FMT_IFFINDER };

void loadIncludeAddrs(const char *filename)
{
    char buf[8192];
    InFile in(filename);
    cerr << "# loading include addresses from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	MidarUtils::FieldSplitter<1> fields(buf, " \t\n");
	includeAddrs.insert(ip4addr_t(fields[0]));
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << includeAddrs.size() << " addresses\n";
    memoryInfo.print("loaded inc addrs");
}

void loadIncludePairs(const char *filename)
{
    char buf[8192];
    InFile in(filename);
    cerr << "# loading include pairs from " << filename << endl;
    AddrPair ap;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	MidarUtils::FieldSplitter<10> fields(buf, " \t\n");
	ap[0] = ip4addr_t(fields[0]);
	ap[1] = ip4addr_t(fields[1]);
	if (ap[1] < ap[0]) swap(ap[0], ap[1]);
	includePairs.insert(ap);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();
    cerr << "# loaded " << includePairs.size() << " pairs\n";
    memoryInfo.print("loaded inc pairs");
}

void loadPairs(const char *filename)
{
    char buf[32768];
    InFile in(filename);
    AddrPair ap;
    AddrPairSet *pairs;
    int fmt = -1;
    long n_loaded = 0;
    long n_deleted = 0;
    cerr << "# loading pairs from " << filename << endl;
    while (in.gets(buf, sizeof(buf))) {
      try {
	if (buf[0] == '#') {
	    int oldfmt = fmt;
	    if (strstr(buf, "resolution-analy")) {
		fmt = FMT_MIDAR;
	    } else if (strstr(buf, "iffinder")) {
		fmt = FMT_IFFINDER;
	    }
	    if (oldfmt >= 0 && fmt != oldfmt) {
		cerr << "conflicting formats" << endl;
		exit(1);
	    }
	}
	if (buf[0] == '#' || buf[0] == '\n') continue; // comment or empty
	MidarUtils::FieldSplitter<10> fields(buf, " \t\n");
	ap[0] = ap[1] = ip4addr_t(0);
	if (fmt < 0) fmt = FMT_PAIRS;
	pairs = &sharedPairs;

	switch (fmt) {
	case FMT_MIDAR:
	    if (!fields[9])
		throw std::runtime_error("bad format");
	    if (strcmp(fields[9], "SHARED") != 0) {
		int result;
		for (result = 0; result < MidarUtils::N_RESULT; result++) {
		    if (strcmp(MidarUtils::result_label[result], fields[9]) == 0)
			break; // found result
		}
		if (result == MidarUtils::N_RESULT)
		    throw std::runtime_error(string("unknown result \"") + fields[9] + "\"");
		if (MidarUtils::result_types[result] != MidarUtils::RESTYPE_UNSHARED)
		    continue;
		pairs = &unsharedPairs;
	    }

	    ap[0] = ip4addr_t(fields[0]);
	    ap[1] = ip4addr_t(fields[4]);
	    addrs[ap[0]].v = atof(fields[3]);
	    addrs[ap[1]].v = atof(fields[7]);
	    break;

	case FMT_IFFINDER:
	    // # addr          alias           o-TTL-r        RTT code
	    // 127.0.0.41      127.0.0.2       244 241   0.141008 D       
	    // code is a letter, optionally followed by ':' and additional info
	    if (!fields[5] || !strchr(":", fields[5][1]))
		throw std::runtime_error("bad format");
	    if (strcmp(fields[5], "D") != 0) continue;
	    pairs = &sharedPairs;
	    ap[0] = ip4addr_t(fields[0]);
	    ap[1] = ip4addr_t(fields[1]);
	    break;

	case FMT_PAIRS:
	    if (!fields[1] || fields[2])
		throw std::runtime_error("bad format");
	    pairs = &sharedPairs;
	    ap[0] = ip4addr_t(fields[0]);
	    ap[1] = ip4addr_t(fields[1]);
	    break;
	}

	if (isBogus(ap[0]) || isBogus(ap[1])) continue;
	if (includeAddrsFile) {
	    if (includeAddrs.find(ap[0]) == includeAddrs.end() ||
		includeAddrs.find(ap[1]) == includeAddrs.end())
		    continue;
	}
	if (ap[1] < ap[0]) swap(ap[0], ap[1]);
	if (includePairsFile) {
	    if (includePairs.find(ap) == includePairs.end())
		continue;
	}

	if (pairs == &sharedPairs) {
	    if (unsharedPairs.find(ap) != unsharedPairs.end()) {
		// this pair is already known to be UNSHARED
		continue;
	    }
	    n_loaded++;
	} else {
	    // erase any previous occurances of this pair
	    n_deleted += sharedPairs.erase(ap);
	}
	pairs->insert(ap);
      } catch (const std::runtime_error &e) { throw InFile::Error(in, e); }
    }
    in.close();

    cerr << "# format: ";
    switch (fmt) {
	case FMT_PAIRS: cerr << "pairs\n"; break;
	case FMT_MIDAR: cerr << "midar\n"; break;
	case FMT_IFFINDER: cerr << "iffinder\n"; break;
	case -1: cerr << "empty\n"; break;
    }
    cerr << "# loaded " << n_loaded << " pairs\n";
    if (n_deleted) cerr << "# deleted " << n_deleted << " pairs\n";
    memoryInfo.print("loaded pairs");
}

struct outfile : public ofstream {
    void open(string name)
    {
	ofstream::open(name.c_str());
	if (!(*this)) {
	    cerr << "can't open " << name << ": " << strerror(errno) << endl;
	    exit(1);
	}
	*this << "# command line: " << cmdline << endl;
	*this << "# version: " << ::cvsID  << endl;
    }
};

int main(int argc, char *const *argv)
{
  try {
    argv0 = argv[0];
    int ch;

    for (int i = 0; i < argc; i++) {
	cmdline += argv[i];
	cmdline += ' ';
    }

    while ((ch = getopt(argc, argv, "B:I:P:apfns:b:o:")) != -1) {
	switch (ch) {
	    case 'B':
		bogonFile = strdup(optarg);
		break;
	    case 'I':
		includeAddrsFile = strdup(optarg);
		break;
	    case 'P':
		includePairsFile = strdup(optarg);
		break;
	    case 'a':
		opt_addrs = true;
		opt_pairs = false;
		opt_filter = false;
		break;
	    case 'p':
		opt_pairs = true;
		opt_addrs = false;
		opt_filter = false;
		break;
	    case 'f':
		opt_pairs = false;
		opt_addrs = false;
		opt_filter = true;
		break;
	    case 'n':
		opt_addrs = opt_pairs = false;
		break;
	    case 's':
		opt_size = atoi(optarg);
		break;
	    case 'b':
		opt_breaksize = atoi(optarg);
		break;
	    case 'o':
		opt_outfile = string(optarg);
		break;
	    default:
		usageExit();
	}
    }
    argc -= optind;
    argv += optind;

    if (argc < 1) usageExit();
    if ((opt_size || opt_breaksize) && opt_outfile.empty()) {
	cerr << "-s and -b require -o\n";
	exit(1);
    }

    memoryInfo.print("startup");

    bogons.installStdBogons();
    if (bogonFile) {
	bogons.load(bogonFile);
	cerr << "# loaded " << bogons.size() << " bogons\n";
    }
    if (includeAddrsFile) {
	loadIncludeAddrs(includeAddrsFile);
    }
    if (includePairsFile) {
	loadIncludePairs(includePairsFile);
    }

    for (int i = 0; i < argc; i++) {
	loadPairs(argv[i]);
    }
    if (sharedPairs.size() == 0) {
	cerr << "no pairs loaded.\n";
	exit(1);
    }
    unsharedPairs.clear(); // no longer needed
    includePairs.clear(); // no longer needed

    if (opt_filter) {
	outfile file;
	if (!opt_outfile.empty()) {
	    file.open(opt_outfile);
	}
	ostream &out = opt_outfile.empty() ? cout : file;

	for (AddrPairSet::iterator it = sharedPairs.begin(); it != sharedPairs.end(); ++it) {
	    out << (*it)[0] << ' ' << (*it)[1] << endl;
	}
	file.close();
	return 0;
    }

    cerr << "# merging " << sharedPairs.size() << " pairs...\n";
    long n_merged = 0;
    for (AddrPairSet::iterator it = sharedPairs.begin(); it != sharedPairs.end(); ++it) {
	AddrData &ad0 = addrs[(*it)[0]];
	AddrData &ad1 = addrs[(*it)[1]];
	ad0.n_pairs++;
	ad1.n_pairs++;
	if (ad0.addrSet && ad1.addrSet) {
	    if (ad0.addrSet == ad1.addrSet) {
		// already merged
		// cout << "# already merged: " << (*it)[0] << " " << (*it)[1] << "\n"; // XXX
		ad0.addrSet->n_found++;
	    } else {
		// merge existing sets (smaller set into bigger set)
		AddrSet *dst, *src;
		if (ad0.addrSet->addrs.size() > ad1.addrSet->addrs.size()) {
		    dst = ad0.addrSet;
		    src = ad1.addrSet;
		} else {
		    dst = ad1.addrSet;
		    src = ad0.addrSet;
		}
		dst->addrs.insert(dst->addrs.end(), src->addrs.begin(), src->addrs.end());
		for (vector<ip4addr_t>::const_iterator a = src->addrs.begin(); a != src->addrs.end(); ++a) {
		    addrs[*a].addrSet = dst;
		}
		if (src->max_v > dst->max_v) dst->max_v = src->max_v;
		if (src->min_v < dst->min_v) dst->min_v = src->min_v;
		dst->n_found += src->n_found + 1;
		delete src;
		// cout << "# merged: " << (*it)[0] << " " << (*it)[1] << "\n"; // XXX
	    }
	} else if (ad0.addrSet) {
	    // add 1 to 0's set
	    ad0.addrSet->addrs.push_back((*it)[1]);
	    ad1.addrSet = ad0.addrSet;
	    if (ad1.v > ad0.addrSet->max_v) ad0.addrSet->max_v = ad1.v;
	    if (ad1.v < ad0.addrSet->min_v) ad0.addrSet->min_v = ad1.v;
	    ad0.addrSet->n_found++;
	    // cout << "# added " << (*it)[1] << " to " << (*it)[0] << "\n"; // XXX
	} else if (ad1.addrSet) {
	    // add 0 to 1's set
	    ad1.addrSet->addrs.push_back((*it)[0]);
	    ad0.addrSet = ad1.addrSet;
	    if (ad0.v > ad1.addrSet->max_v) ad1.addrSet->max_v = ad0.v;
	    if (ad0.v < ad1.addrSet->min_v) ad1.addrSet->min_v = ad0.v;
	    ad1.addrSet->n_found++;
	    // cout << "# added " << (*it)[0] << " to " << (*it)[1] << "\n"; // XXX
	} else {
	    // create new set
	    AddrSet *addrset = new AddrSet();
	    addrset->addrs.push_back((*it)[0]);
	    ad0.addrSet = addrset;
	    addrset->addrs.push_back((*it)[1]);
	    ad1.addrSet = addrset;
	    if (ad0.v < ad1.v) {
		addrset->min_v = ad0.v;
		addrset->max_v = ad1.v;
	    } else {
		addrset->min_v = ad1.v;
		addrset->max_v = ad0.v;
	    }
	    addrset->n_found = 1;
	    // cout << "# created: " << (*it)[0] << " " << (*it)[1] << "\n"; // XXX
	}
	++n_merged;
	if (n_merged % 100 == 0)
	    cerr << "# merged  " << n_merged << " pairs\r";
    }
    cerr << "# merged " << n_merged << " pairs\n";
    memoryInfo.print("merged pairs");

    vector<AddrSet*> sets;

    long n_vectored = 0;
    cerr << "# vectoring " << addrs.size() << " addrs\n";
    for (AddrMap::iterator it = addrs.begin(); it != addrs.end(); ++it) {
	if (!(*it).second.addrSet) continue;
	if ((*it).second.addrSet->addrs[0] == (*it).first) {
	    sets.push_back((*it).second.addrSet);
	}
	++n_vectored;
	if (n_vectored % 100 == 0)
	    cerr << "# vectored " << n_vectored << " addrs\r";
    }
    cerr << "# vectored " << n_vectored << " addrs\n";

    cerr << "# sorting " << sets.size() << " sets\n";
    for (vector<AddrSet*>::iterator it = sets.begin(); it != sets.end(); ++it) {
	sort((*it)->addrs.begin(), (*it)->addrs.end());
    }
    sort(sets.begin(), sets.end(), set_size_greater());
    memoryInfo.print("sorted sets");

    cerr << "# dumping " << sets.size() << " sets\n";

    outfile smallfile;
    outfile largefile;

    if (!opt_outfile.empty()) {
	if (opt_size) {
	    smallfile.open(opt_outfile + "-small");
	    smallfile << "# total sets: " << sets.size() << endl;
	    largefile.open(opt_outfile + "-large");
	} else {
	    largefile.open(opt_outfile);
	}
	largefile << "# total sets: " << sets.size() << endl;
    } else {
	cout << "# total sets: " << sets.size() << endl;
    }

    size_t setnum = 0;
    long total_pairs = 0;
    long total_found = 0;
    for (vector<AddrSet*>::iterator it = sets.begin(); it != sets.end(); ++it) {
	size_t n_pairs = (*it)->addrs.size() * ((*it)->addrs.size() - 1) / 2;

	ostream &out = opt_outfile.empty() ? cout :
	    (*it)->addrs.size() <= opt_size ? smallfile : largefile;

	vector<AddrPair> mypairs;
	if (opt_pairs && n_pairs < sharedPairs.size()) {
	    // it's more efficient to test if every possible pair in the set was found
	    for (size_t i = 0; i < (*it)->addrs.size(); i++) {
		for (size_t j = i+1; j < (*it)->addrs.size(); j++) {
		    AddrPair ap;
		    ap[0] = (*it)->addrs[i];
		    ap[1] = (*it)->addrs[j];
		    if (ap[1] < ap[0]) swap(ap[0], ap[1]);
		    if (sharedPairs.find(ap) != sharedPairs.end()) {
			mypairs.push_back(ap);
		    }
		}
	    }
	}

	out << "# set " << setnum << ": " << (*it)->addrs.size() <<
	    " addrs; velocity " << (*it)->min_v << " - " << (*it)->max_v <<
	    "; " << n_pairs << " pairs, " << (n_pairs - (*it)->n_found) << " missing, " <<
	    (*it)->n_found << " found\n";
	total_pairs += n_pairs;
	total_found += (*it)->n_found;
	if (opt_pairs) {
	    if (!mypairs.empty()) {
		// it's more efficient to test if every possible pair in the set was found
		for (vector<AddrPair>::iterator pit = mypairs.begin();
		    pit != mypairs.end(); ++pit)
		{
		    out << (*pit)[0] << ' ' << (*pit)[1] << endl;
		}
		mypairs.clear();
	    } else {
		// it's more efficient to test if every found pair is in the set
		for (AddrPairSet::iterator pit = sharedPairs.begin();
		    pit != sharedPairs.end(); ++pit)
		{
		    if (addrs[(*pit)[0]].addrSet == (*it))
			out << (*pit)[0] << ' ' << (*pit)[1] << endl;
		}
	    }
	}
	if (opt_addrs) {
	    for (size_t i = 0; i < (*it)->addrs.size(); i++) {
		out << (*it)->addrs[i] << endl;
	    }
	}
	setnum++;
	cerr << "# dumped " << setnum << " sets\r";
    }
    memoryInfo.print("dumped sets");
    smallfile.close();
    largefile.close();
    cerr << "\n";
    cerr << "# total pairs:   " << setw(15) << total_pairs << "\n";
    cerr << "# total found:   " << setw(15) << total_found << "\n";
    cerr << "# total missing: " << setw(15) << (total_pairs - total_found) << "\n";

    // find chain links in large sets
    if (opt_breaksize) {
	for (AddrPairSet::iterator it = sharedPairs.begin(); it != sharedPairs.end(); ++it) {
	    AddrData &ad0 = addrs[(*it)[0]];
	    AddrData &ad1 = addrs[(*it)[1]];
	    if (ad0.addrSet->addrs.size() < 2 * opt_breaksize) continue;
	    if (ad0.n_pairs == 2 || ad1.n_pairs == 2) {
		ad0.addrSet->chainPairs.push_back(&(*it));
	    }
	}
	memoryInfo.print("found chain links");

	// in each large set, find chain breakpoints
	outfile bpfile;
	bpfile.open(opt_outfile + ".breakpoints");
	setnum = 0;
	for (vector<AddrSet*>::iterator it = sets.begin(); it != sets.end(); ++it) {
	    if (!(*it)->chainPairs.empty()) {
		sort((*it)->chainPairs.begin(), (*it)->chainPairs.end(), chainPair_velocity_less());
		size_t distance = 0; // distance since last breakpoint
		int subsetnum = 0;
		for (vector<const AddrPair*>::iterator cpit = (*it)->chainPairs.begin();
		    cpit != (*it)->chainPairs.end(); cpit++)
		{
		    if (++distance >= opt_breaksize && addrs[(**cpit)[0]].n_pairs == 2) {
			bpfile << "# set " << setnum << "." << subsetnum << endl;
			bpfile << (**cpit)[0] << endl;
			bpfile << (**cpit)[1] << endl;
			distance = 0;
			subsetnum++;
		    }
		}
	    }
	    setnum++;
	}
	bpfile.close();
	memoryInfo.print("found breakpoints");
    }

    memoryInfo.print("done");
    cerr << "# done" << endl;

  } catch (const std::exception &e) {
    cerr << e.what() << endl;
    exit(1);
  }
}

